// ===========================================================================
//	CGLContextFlags		 (c) 2013 Mark F. Coniglio	  All rights reserved.
// ===========================================================================

#ifndef _H_CGLContextFlags
#define _H_CGLContextFlags

#include "OpenGLIncludes.h"
#include "IsadoraTypes.h"
#include "CGLLightCommon.h"
#include "CGLColor.h"
#include "3DVectorDefs.h"
#include "IsadoraDebugging.h"
#include "ColorRefs.h"

#define SHARED_GL_CONTEXTS								TRUE

#define	DEBUG_GL_FLAGS									(TRUE && DEBUG)
#define	DEBUG_GL_FLAGS_EXTENSIVE						(TRUE && DEBUG_GL_FLAGS)

#if DEBUG_GL_FLAGS_EXTENSIVE
	#define DebugCheckGLFlagsForContext(ctx)	(ctx)->DebugCheckFlags()
#else
	#define DebugCheckGLFlagsForContext(...)	{ }
#endif

// ---------------------------------------------------------------------------
// LOCAL GL FLAG INDEXES
// ---------------------------------------------------------------------------

enum {
	
	kGLFlag_GL_COLOR_MATERIAL,
	kGLFlag_GL_RESCALE_NORMAL,
	kGLFlag_GL_NORMALIZE,
	
	kGLFlag_GL_LIGHTING,
	
	kGLFlag_GL_LIGHT0,
	kGLFlag_GL_LIGHT1,
	kGLFlag_GL_LIGHT2,
	kGLFlag_GL_LIGHT3,
	kGLFlag_GL_LIGHT4,
	kGLFlag_GL_LIGHT5,
	kGLFlag_GL_LIGHT6,
	kGLFlag_GL_LIGHT7,
	
	kGLFlag_GL_BLEND,
	kGLFlag_GL_DEPTH_TEST,
	kGLFlag_GL_CULL_FACE,
	
	kGLFlag_GL_MULTISAMPLE_ARB,
	kGLFlag_GL_LINE_SMOOTH,
	
	kGLFlag_TotalItems
};

// ---------------------------------------------------------------------------
// DEFINES
// ---------------------------------------------------------------------------

#ifdef __ISADORA_APP__
	#define CGLCF_DebugActive						DEBUG
#else
	#define CGLCF_DebugActive						FALSE
#endif

#if CGLCF_DebugActive
	#include "UDebugging.h"
	#include "COpenGLUtil.h"
	#define	CGLCF_CheckOpenGLError					CheckOpenGLError
	#define	CGLCF_Assert_							Assert_
	#define CGLCONTEXT_FLAGS_ENABLE_DEBUG_PRINT		(TRUE && DEBUG)
	#define cglcf_log_info							log_info
	#define cglcf_log_error							log_error
#else
	#define	CGLCF_CheckOpenGLError()				{ }
	#define	CGLCF_Assert_(...)						{ }
	#define CGLCONTEXT_FLAGS_ENABLE_DEBUG_PRINT		FALSE
	#define cglcf_log_info(...)						{ }
	#define cglcf_log_error(...)					{ }
#endif

// ---------------------------------------------------------------------------
// DEFINES
// ---------------------------------------------------------------------------

// allows set or clear of a flag... clients should use GLEnableFlag or GLDisableFlag
#define	GLSetEnableFlag(glFlag, glFlagValue) __GLSetEnableFlag(glFlag, kGLFlag_##glFlag, glFlagValue)

// replacement for glEnable
#define	GLEnableFlag(glFlag)	__GLSetEnableFlag(glFlag, kGLFlag_##glFlag, GL_TRUE)

// replacement for glDisable
#define	GLDisableFlag(glFlag)	__GLSetEnableFlag(glFlag, kGLFlag_##glFlag, GL_FALSE)

// replacement for glIsEnabled
#define	GLIsEnabled(glFlag)		__GLIsEnabled(glFlag, kGLFlag_##glFlag)

// maximum number of lights we can handle
#define	kGLCtxFlags_MaxLights	1

// ---------------------------------------------------------------------------
// CLightParams CLASS
// ---------------------------------------------------------------------------

class CLightParams
{
public:

	CLightParams() :
		mColorAmbient(),
		mColorDiffuse(),
		mColorSpecular(),
		mRotationAndDistance()
	{
	}
	
	CLightParams(
		const CGLColor&		inAmbientColor,
		const CGLColor&		inDiffuseColor,
		const CGLColor&		inSpecularColor,
		const CVector4&		inRotationAndDistance) :
			mColorAmbient(inAmbientColor),
			mColorDiffuse(inDiffuseColor),
			mColorSpecular(inSpecularColor),
			mRotationAndDistance(inRotationAndDistance)
	{
	}
	
	CGLColor		mColorAmbient;
	CGLColor		mColorDiffuse;
	CGLColor		mColorSpecular;
	CVector4		mRotationAndDistance;
};

// ---------------------------------------------------------------------------
// CAllLightParams CLASS
// ---------------------------------------------------------------------------

class CAllLightParams
{
public:
	CGLColor		mGlobalAmbientLightColor;
	CLightParams	mLightParams[IZZY_GL_MAX_LIGHTS];
};

// ---------------------------------------------------------------------------
// CGLContextFlagValues
// ---------------------------------------------------------------------------

#define GL_CACHED_FBO_BIND			TRUE

class CGLContextFlagValues
{
public:

	CGLContextFlagValues();

private:
	~CGLContextFlagValues();

public:	
	void		Retain();
	void		Release();
	GLuint		GetRetainCount() const { return mGLF_RetainCount; }
	
public:
	GLuint				mGLF_RetainCount;
	GLboolean			mGLF_Flags[kGLFlag_TotalItems];
#if GL_CACHED_FBO_BIND
	GLuint				mGLF_FrameBufferID;
#endif
	GLenum				mGLF_TextureTarget;
	GLboolean			mGLF_DepthWriteMask;
	GLint				mGLF_TexEnvMode;
	GLint				mGLF_LightModelColorControl;
	GLint				mGLF_ViewportCoords[4];
	GLfloat				mGLF_ClearColor[4];
	CAllLightParams		mGLF_AllLightParams;

	const GLenum*		mGLF_SupportedFlags;								// used only for debugging

	UInt32				mGLF_DebugCheckFlagsSuspendCount;					// used to suspend debug checks; meaningful in DEBUG version only

#ifdef __ISADORA_APP__
	static GLenum		mGLF_SupportedGLFlagsPrivate[kGLFlag_TotalItems+1];
#endif
};

// ---------------------------------------------------------------------------
// CGLContextFlags
// ---------------------------------------------------------------------------

class CGLContextFlags
{
public:

	// CONSTRUCTOR
					CGLContextFlags();
					
					~CGLContextFlags();
					
	// INITIALIZE
	void			InitializeFromContext();

	// DEFAULTS
	void			SetOpenGLDefaults(
						bool		inContextMultisamplingEnabled);
	
	bool			ContextFlagsValid() const
						{ return glfv != NULL; }
	
	// NOTE: releases the current CGLContextFlagValues
	// object pointed to by member variable 'glfv'. Then
	// stores inFlagValues into 'glfv' after retaining
	// it. inFlagValues must be a valid, instantiated
	// CGLContextFlagValues object!
	
	void			ReplaceCGLContextFlagValues(
						CGLContextFlagValues*		inFlagValues);
	
	CGLContextFlagValues*	GetCGLContextFlagValues() { return glfv; }
	
	// ENABLE FLAGS
	
	// enable getter
	GLboolean		__GLIsEnabled(
						GLenum	inGLFlagEnum,
						GLenum	inGLFlagIndex0) const
					{
						#if !CGLCF_DebugActive
						(void) inGLFlagEnum;	// to prevent unused warning
						#endif
						
						CGLCF_Assert_(inGLFlagIndex0 < kGLFlag_TotalItems);
						
						if (inGLFlagIndex0 < kGLFlag_TotalItems) {
							CGLCF_Assert_(glfv->mGLF_SupportedFlags[inGLFlagIndex0] == inGLFlagEnum);
							return glfv->mGLF_Flags[inGLFlagIndex0];
						}
						
						return false;
					}

	// enable setter
	void			__GLSetEnableFlag(
						GLenum		inGLFlagEnum,
						GLenum		inGLFlagIndex0,
						GLboolean	inFlag)
					{
						CGLCF_Assert_(inGLFlagIndex0 < kGLFlag_TotalItems);
						if (inGLFlagIndex0 < kGLFlag_TotalItems) {
							if (glfv->mGLF_Flags[inGLFlagIndex0] != inFlag) {
								glfv->mGLF_Flags[inGLFlagIndex0] = inFlag;
								if (inFlag) {
									glEnable(inGLFlagEnum);
								} else {
									glDisable(inGLFlagEnum);
								}
								CGLCF_CheckOpenGLError();
							}
						}
						
						#if DEBUG
							// SANITY CHECK: make sure the expected texture target is enabled
							GLboolean glEnabled = 0;
							glGetBooleanv(inGLFlagEnum, &glEnabled);
							CGLCF_Assert_(glEnabled == inFlag);						
						#endif
					}
	
	// CLEAR COLOR
	inline void			GetClearColor(
							GLfloat*	outClearColor)
						{
							outClearColor[0] = glfv->mGLF_ClearColor[0];
							outClearColor[1] = glfv->mGLF_ClearColor[1];
							outClearColor[2] = glfv->mGLF_ClearColor[2];
							outClearColor[3] = glfv->mGLF_ClearColor[3];
						}
					
	inline void			SetClearColor(
							ColorRef		inColoRef,
							bool			inForce = false)
						{
							float a = (float)(ColorRef_Alp(inColoRef) / 255.0);
							float r = (float)(ColorRef_Red(inColoRef) / 255.0);
							float g = (float)(ColorRef_Grn(inColoRef) / 255.0);
							float b = (float)(ColorRef_Blu(inColoRef) / 255.0);
							SetClearColor(r, g, b, a, inForce);
							
						}
						
	inline void			SetClearColor(
							GLfloat	r,
							GLfloat	g,
							GLfloat	b,
							GLfloat	a,
							bool	inForce = false)
						{
							if (inForce
							|| r != glfv->mGLF_ClearColor[0]
							|| g != glfv->mGLF_ClearColor[1]
							|| b != glfv->mGLF_ClearColor[2]
							|| a != glfv->mGLF_ClearColor[3]) {
							
								glClearColor(r, g, b, a);
								CGLCF_CheckOpenGLError();

								glfv->mGLF_ClearColor[0] = r;
								glfv->mGLF_ClearColor[1] = g;
								glfv->mGLF_ClearColor[2] = b;
								glfv->mGLF_ClearColor[3] = a;
								
							}
						}
						
	inline void			SetClearColor(
							const GLfloat*	inClearColor,
							bool			inForce = false)
						{
							if (inForce
							|| inClearColor[0] != glfv->mGLF_ClearColor[0]
							|| inClearColor[1] != glfv->mGLF_ClearColor[1]
							|| inClearColor[2] != glfv->mGLF_ClearColor[2]
							|| inClearColor[3] != glfv->mGLF_ClearColor[3]) {
							
								glClearColor(inClearColor[0], inClearColor[1], inClearColor[2], inClearColor[3]);
								CGLCF_CheckOpenGLError();

								glfv->mGLF_ClearColor[0] = inClearColor[0];
								glfv->mGLF_ClearColor[1] = inClearColor[1];
								glfv->mGLF_ClearColor[2] = inClearColor[2];
								glfv->mGLF_ClearColor[3] = inClearColor[3];
								
							}
						}
						

	// FRAME BUFFER OBJECTS
	#if GL_CACHED_FBO_BIND
	virtual void	BindFramebufferID(
						GLuint				inFrameBufferID);

	virtual GLuint	GetCurrentFramebufferID() const;
	#endif

	// TEXTURE TARGET
	
	GLenum			GetActiveTextureTarget() const { return glfv->mGLF_TextureTarget; }
								
	void			EnableTextureMapping(
						GLenum		inTextureTarget)
					{
						#if DEBUG_GL_FLAGS
						DebugCheckTextureMapping();
						DebugCheckFramebuffer();
						#endif

						if (inTextureTarget != glfv->mGLF_TextureTarget) {
						
							// disable the current texture mapping target if any
							if (glfv->mGLF_TextureTarget != 0) {
								CGLCF_CheckOpenGLError();
								glDisable(glfv->mGLF_TextureTarget);
								CGLCF_CheckOpenGLError();
								glfv->mGLF_TextureTarget = 0;
							}
							
							// enable new texture target
							if (inTextureTarget != 0) {
								glfv->mGLF_TextureTarget = inTextureTarget;
								CGLCF_Assert_(glfv->mGLF_TextureTarget == GL_TEXTURE_2D || glfv->mGLF_TextureTarget == GL_TEXTURE_RECTANGLE_ARB);
								CGLCF_CheckOpenGLError();
								glEnable(inTextureTarget);
								CGLCF_CheckOpenGLError();
							}
							
						}
						
						#if DEBUG_GL_FLAGS
						DebugCheckTextureMapping();
						DebugCheckFramebuffer();
						#endif
					}

	GLenum			DisableTextureMapping()
					{
						GLenum oldTextureTarget = glfv->mGLF_TextureTarget;
						
						if (glfv->mGLF_TextureTarget != 0) {
							glDisable(glfv->mGLF_TextureTarget);
							CGLCF_CheckOpenGLError();
							glfv->mGLF_TextureTarget = 0;
						}
						
						#if DEBUG_GL_FLAGS
						DebugCheckTextureMapping();
						DebugCheckFramebuffer();
						#endif
						
						DisableHAPQShader();

						return oldTextureTarget;
					}

	GLenum			GetTextureTarget() const
					{
						return glfv->mGLF_TextureTarget;
					}
	
	void			RestoreTextureTarget(
						GLenum			inTextureTarget) const
					{
						if (inTextureTarget == GL_TEXTURE_2D) {
							glEnable(GL_TEXTURE_2D);
							glDisable(GL_TEXTURE_RECTANGLE_ARB);
						} else if (inTextureTarget == GL_TEXTURE_RECTANGLE_ARB) {
							glEnable(GL_TEXTURE_RECTANGLE_ARB);
							glDisable(GL_TEXTURE_2D);
						} else {
							glDisable(GL_TEXTURE_2D);
							glDisable(GL_TEXTURE_RECTANGLE_ARB);
						}
					}
	
	void			ResetTextureTarget()
					{
						glDisable(GL_TEXTURE_2D);
						glDisable(GL_TEXTURE_RECTANGLE_ARB);
						glfv->mGLF_TextureTarget = 0;
						
						#if DEBUG
						GLboolean glEnable2d = 0;
						glGetBooleanv(GL_TEXTURE_2D, &glEnable2d);
						GLboolean glEnableRect = 0;
						glGetBooleanv(GL_TEXTURE_RECTANGLE_ARB, &glEnableRect);
						CGLCF_Assert_(glEnable2d == GL_FALSE && glEnableRect == GL_FALSE);
						#endif

				}

	// DEPTH WRITE MASK
	GLboolean		GLGetDepthWriteMask() const { return glfv->mGLF_DepthWriteMask; }
					
	void			GLSetDepthWriteMask(
						GLboolean	inMaskFlag)
					{
						if (inMaskFlag != glfv->mGLF_DepthWriteMask) {
							glfv->mGLF_DepthWriteMask = inMaskFlag;
							glDepthMask(inMaskFlag);
							CGLCF_CheckOpenGLError();
						}
					}
	
	// TEX ENV MODE
	GLint			GLGetTexEnvMode() const { return glfv->mGLF_TexEnvMode; }
					
	void			GLSetTexEnvMode(
						GLint				inTexEnvMode)
					{
						if (inTexEnvMode != glfv->mGLF_TexEnvMode) {
							glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, inTexEnvMode);
							CGLCF_CheckOpenGLError();
							glfv->mGLF_TexEnvMode = inTexEnvMode;
						}
					}

	
	// LIGHT - GLOBAL AMBIENT COLOR
	void			GLGetGlobalAmbientColor(
						CGLColor&			outGlobalAmbientColor)
							{ outGlobalAmbientColor = glfv->mGLF_AllLightParams.mGlobalAmbientLightColor; } 
	
	void			GLSetGlobalAmbientColor(
						const CGLColor&		inGlobalAmbientColor)
					{
						if (inGlobalAmbientColor != glfv->mGLF_AllLightParams.mGlobalAmbientLightColor) {
							glfv->mGLF_AllLightParams.mGlobalAmbientLightColor = inGlobalAmbientColor;
							glLightModelfv(GL_LIGHT_MODEL_AMBIENT, glfv->mGLF_AllLightParams.mGlobalAmbientLightColor);
							CGLCF_CheckOpenGLError();
						}
					}
				
	
	// LIGHT - AMBIENT COLOR
	void			GLGetLightAmbientColor(
						LightIndexT			inLightIndex0,
						CGLColor&			outAmbientColor)
					{
						if (inLightIndex0 < IZZY_GL_MAX_LIGHTS) {
							outAmbientColor = glfv->mGLF_AllLightParams.mLightParams[inLightIndex0].mColorAmbient;
						}
					}

	void			GLSetLightAmbientColor(
						LightIndexT			inLightIndex0,
						const CGLColor&		inAmbientColor)
					{
						if (inLightIndex0 < IZZY_GL_MAX_LIGHTS) {
							if (inAmbientColor != glfv->mGLF_AllLightParams.mLightParams[inLightIndex0].mColorAmbient) {
								glfv->mGLF_AllLightParams.mLightParams[inLightIndex0].mColorAmbient = inAmbientColor;
								glLightfv(GL_LIGHT0 + inLightIndex0, GL_AMBIENT, inAmbientColor);
								CGLCF_CheckOpenGLError();
							}
						}
					}
				
	
	// LIGHT - DIFFUSE COLOR
	void			GLGetLightDiffuseColor(
						LightIndexT			inLightIndex0,
						CGLColor&			outDiffuseColor)
					{
						if (inLightIndex0 < IZZY_GL_MAX_LIGHTS) {
							outDiffuseColor = glfv->mGLF_AllLightParams.mLightParams[inLightIndex0].mColorDiffuse;
						}
					}

	void			GLSetLightDiffuseColor(
						LightIndexT			inLightIndex0,
						const CGLColor&		inDiffuseColor)
					{
						if (inLightIndex0 < IZZY_GL_MAX_LIGHTS) {
							if (inDiffuseColor != glfv->mGLF_AllLightParams.mLightParams[inLightIndex0].mColorDiffuse) {
								glfv->mGLF_AllLightParams.mLightParams[inLightIndex0].mColorDiffuse = inDiffuseColor;
								glLightfv(GL_LIGHT0 + inLightIndex0, GL_DIFFUSE, inDiffuseColor);
								CGLCF_CheckOpenGLError();
							}
						}
					}
				
	
	// LIGHT - SPECULAR COLOR
	void			GLGetLightSpecularColor(
						LightIndexT			inLightIndex0,
						CGLColor&			outSpecularColor)
					{
						if (inLightIndex0 < IZZY_GL_MAX_LIGHTS) {
							outSpecularColor = glfv->mGLF_AllLightParams.mLightParams[inLightIndex0].mColorSpecular;
						}
					}

	void			GLSetLightSpecularColor(
						LightIndexT			inLightIndex0,
						const CGLColor&		inSpecularColor)
					{
						if (inLightIndex0 < IZZY_GL_MAX_LIGHTS) {
							if (inSpecularColor != glfv->mGLF_AllLightParams.mLightParams[inLightIndex0].mColorSpecular) {
								glfv->mGLF_AllLightParams.mLightParams[inLightIndex0].mColorSpecular = inSpecularColor;
								glLightfv(GL_LIGHT0 + inLightIndex0, GL_SPECULAR, inSpecularColor);
								CGLCF_CheckOpenGLError();
							}
						}
					}
				
	// LIGHT - ROTATION AND DISTANCE COLOR
	void			GLGetLightRotationAndDistance(
						LightIndexT			inLightIndex0,
						CVector4&			outLightRotatioAndDistance)
					{
						if (inLightIndex0 < IZZY_GL_MAX_LIGHTS) {
							outLightRotatioAndDistance = glfv->mGLF_AllLightParams.mLightParams[inLightIndex0].mRotationAndDistance;
						}
					}

	void			GLSetLightLightRotationAndDistance(
						LightIndexT			inLightIndex0,
						const CVector4&		inRotationAndDistance)
					{
						if (inLightIndex0 < IZZY_GL_MAX_LIGHTS) {
							if (inRotationAndDistance != glfv->mGLF_AllLightParams.mLightParams[inLightIndex0].mRotationAndDistance) {
				
								glfv->mGLF_AllLightParams.mLightParams[inLightIndex0].mRotationAndDistance = inRotationAndDistance;
									
								glPushMatrix();					// should be GL_MODELVIEW
								CGLCF_CheckOpenGLError();

								// rotate the world based on settings
								glRotatef(inRotationAndDistance.x, 1.0f, 0.0f, 0.0f);
								CGLCF_CheckOpenGLError();
								glRotatef(inRotationAndDistance.y, 0.0f, 1.0f, 0.0f);
								CGLCF_CheckOpenGLError();
								glRotatef(inRotationAndDistance.z, 0.0f, 0.0f, 1.0f);
								CGLCF_CheckOpenGLError();
								
								CVector4 pos(0.0f, 0.0f, +1.0f, inRotationAndDistance.w);
								glLightfv(GL_LIGHT0 + inLightIndex0, GL_POSITION, pos);
								CGLCF_CheckOpenGLError();

								glPopMatrix();					// should be GL_MODELVIEW
								CGLCF_CheckOpenGLError();

							}
						}
					}
				
	// LIGHT MODEL COLOR CONTROL
	GLint			GLGetLightModeColorControl() const { return glfv->mGLF_LightModelColorControl; }
					
	void			GLSetLightModeColorControl(
						GLint				inLightModeColorControl)
					{
						if (inLightModeColorControl != glfv->mGLF_LightModelColorControl) {
							glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL, inLightModeColorControl);
							CGLCF_CheckOpenGLError();
							glfv->mGLF_LightModelColorControl = inLightModeColorControl;
						}
					}
					

	// ALL LIGHT PARAMETERS
	CAllLightParams	GetAllLightParams() const
					{
						return glfv->mGLF_AllLightParams;
					}
					
	void			RestoreAllLightParams(
						const CAllLightParams&	inAllLightParams);
					
	// VIEWPORT
	bool			GLSetViewport(
						GLint		x,
						GLint		y,
						GLsizei		width,
						GLsizei		height,
						bool		inForceSetViewport = false)
					{
						if (inForceSetViewport || x != glfv->mGLF_ViewportCoords[0] || y != glfv->mGLF_ViewportCoords[1] || width != glfv->mGLF_ViewportCoords[2] || height != glfv->mGLF_ViewportCoords[3]) {
							CGLCF_CheckOpenGLError();
							glfv->mGLF_ViewportCoords[0] = x;
							glfv->mGLF_ViewportCoords[1] = y;
							glfv->mGLF_ViewportCoords[2] = width;
							glfv->mGLF_ViewportCoords[3] = height;
							glViewport(x, y, width, height);
							CGLCF_CheckOpenGLError();
							return true;
						}
						
						return false;
					}
				
	void			GLSetViewport(
						const GLint*	inViewportCoords)
					{
						CGLCF_CheckOpenGLError();
						glfv->mGLF_ViewportCoords[0] = inViewportCoords[0];
						glfv->mGLF_ViewportCoords[1] = inViewportCoords[1];
						glfv->mGLF_ViewportCoords[2] = inViewportCoords[2];
						glfv->mGLF_ViewportCoords[3] = inViewportCoords[3];
						glViewport(glfv->mGLF_ViewportCoords[0], glfv->mGLF_ViewportCoords[1], glfv->mGLF_ViewportCoords[2], glfv->mGLF_ViewportCoords[3]);
						CGLCF_CheckOpenGLError();
					}
					
	void			GLGetViewport(
						GLint*			outViewportCoords) const
					{
						outViewportCoords[0] = glfv->mGLF_ViewportCoords[0];
						outViewportCoords[1] = glfv->mGLF_ViewportCoords[1];
						outViewportCoords[2] = glfv->mGLF_ViewportCoords[2];
						outViewportCoords[3] = glfv->mGLF_ViewportCoords[3];
					}
	
	virtual void	EnableHAPQShader(
						bool				inPostDecodeUnpremultiply = false)
							{ EnableHAPQShader(CGLColor(), inPostDecodeUnpremultiply); }

	virtual void	EnableHAPQShader(
						const CGLColor&		/* inColorMod */,
						bool				/* inPostDecodeUnpremultiply */ = false)
							{ /* overridden by context implementation */ }

	virtual void	DisableHAPQShader()
							{ /* overridden by context implementation */ }
					
	virtual	bool	SetBlendFunction(
						RenderBlendMode		inBlendMode,
						bool				/* inIsPremultiplied */)
					{
						/* overridden by context implementation */
						return (inBlendMode != kBlendMode_Opaque);
					}
						
	#if CGLCONTEXT_FLAGS_ENABLE_DEBUG_PRINT
	void		DebugPrintContextFlags() const;
	#endif

	#if DEBUG_GL_FLAGS
	void		SuspendGLFlagsDebugChecks()		{ glfv->mGLF_DebugCheckFlagsSuspendCount += 1; }
	void		ResumeGLFlagsDebugChecks()		{ glfv->mGLF_DebugCheckFlagsSuspendCount -= 1; }
	void		DebugCheckFlags() const;
	void		DebugCheckTextureMapping() const;
	void		DebugCheckFramebuffer() const;
	#endif
				

protected:
	CGLContextFlagValues*	glfv;
};

#endif
