// ===========================================================================
//	IZDefinitions
// ===========================================================================
#ifndef _H_IZDefinitions
#define _H_IZDefinitions

#include "IZGraphics.h"
#include "IZTypes.h"

#if TARGET_OS_MAC
	#include <limits.h>
#elif TARGET_OS_WIN
	#include <intsafe.h>
#endif

// ---------------------------------------------------------------------------------
//	IZ_THREAD_LOCAL
// ---------------------------------------------------------------------------------
// A cross-platform definition of the storage-class modifier required to
// define Thread Local Storage (TLS) variables

#if TARGET_OS_MAC
#define IZ_THREAD_LOCAL	__thread
#elif TARGET_OS_WIN
#define IZ_THREAD_LOCAL	__declspec(thread)
#endif

// ---------------------------------------------------------------------------------
//	SIZE_T CONVERSION
// ---------------------------------------------------------------------------------

#ifdef __cplusplus

#ifdef __ISADORA_APP__
#define	IZ_RANGE_CHECK_SIZE_T	TRUE
#else
#define	IZ_RANGE_CHECK_SIZE_T	FALSE
#endif

#if IZ_RANGE_CHECK_SIZE_T
	SInt32		sizet_s32(size_t v);
	UInt32		sizet_u32(size_t v);
	SInt16		sizet_s16(size_t v);
	UInt16		sizet_u16(size_t v);
#else
	#define		sizet_s32(v) ((SInt32) (v))
	#define		sizet_u32(v) ((UInt32) (v))
	#define		sizet_s16(v) ((SInt16) (v))
	#define		sizet_u16(v) ((UInt16) (v))
#endif

#endif

// ---------------------------------------------------------------------------------
//	IzSel - SELECTION RANGE VALUE
// ---------------------------------------------------------------------------------

typedef	size_t			IzSel;

#define	IZ_SEL_MIN		0
#define IZ_SEL_MAX		SIZE_T_MAX

// ---------------------------------------------------------------------------------
//	DEFINES
// ---------------------------------------------------------------------------------

#define	IZ_INTENSIVE_CHECKS		TRUE && DEBUG

#ifdef __cplusplus
#define DEBUG_WITH_AUTO_RELEASE	(TRUE && DEBUG && TARGET_OS_MAC)

#if DEBUG_WITH_AUTO_RELEASE
	#define DebugAutoRelease()	CAutoReleasePool pool
#else
	#define DebugAutoRelease()	{ }
#endif

#endif

// ---------------------------------------------------------------------------------
//	TYPEDEFS
// ---------------------------------------------------------------------------------

typedef		SInt16		IzControlKind;	// **IMPORTANT** this must remain a 16 bit value

// ---------------------------------------------------------------------------------
//	RECT CONVERSIONS
// ---------------------------------------------------------------------------------

#if TARGET_OS_MAC && defined(_OBJC_OBJC_H_) && defined(__cplusplus)

inline NSRect
IzRectToNSRect(
	const IzRect&	inRect)
{
	NSRect outRect;
	outRect.origin.x = (CGFloat) inRect.left;
	outRect.origin.y = (CGFloat) inRect.top;
	outRect.size.width = (CGFloat) (inRect.right - inRect.left);
	outRect.size.height = (CGFloat) (inRect.bottom - inRect.top);
	return outRect;
}

inline IzRect
NSRectToIzRect(
	const NSRect&	inRect)
{
	IzRect outRect;
	outRect.left = (IzCoord)(floor(inRect.origin.x));
	outRect.top = (IzCoord)(floor(inRect.origin.y));
	outRect.right = (IzCoord)(ceil(inRect.origin.x + inRect.size.width));
	outRect.bottom = (IzCoord)(ceil(inRect.origin.y + inRect.size.height));
	return outRect;
}

inline NSPoint
IzPointToNSPoint(
	const IzPoint& inPoint)
{
	NSPoint outPoint;
	outPoint.x = inPoint.h;
	outPoint.y = inPoint.v;
	return outPoint;
}

inline IzPoint
NSPointToIzPoint(
	const NSPoint&	inPoint)
{
	Point outPoint;
	outPoint.h = inPoint.x;
	outPoint.v = inPoint.y;
	return outPoint;
}

inline NSRect
FlippedIzRectToNSRect(
	IzRect			inRect,
	CGFloat			inContainerHeight)
{
	NSRect outRect;
	outRect.origin.x = (CGFloat) inRect.left;
	outRect.origin.y = (CGFloat) inRect.top;
	outRect.size.width = (CGFloat) (inRect.right - inRect.left);
	outRect.size.height = (CGFloat) (inRect.bottom - inRect.top);
	outRect.origin.y = inContainerHeight - (outRect.origin.y + outRect.size.height);
	return outRect;
}

inline IzRect
FlippedNSRectToIzRect(
	NSRect			inRect,
	CGFloat			inContainerHeight)
{
	inRect.origin.y = inContainerHeight - (inRect.origin.y + inRect.size.height);
	
	IzRect outRect;
	outRect.left = (IzCoord)(floor(inRect.origin.x));
	outRect.top = (IzCoord)(floor(inRect.origin.y));
	outRect.right = (IzCoord)(ceil(inRect.origin.x + inRect.size.width));
	outRect.bottom = (IzCoord)(ceil(inRect.origin.y + inRect.size.height));
	return outRect;
}


inline NSRect
FlipNSRect(
	NSRect			inRect,
	CGFloat			inContainerHeight)
{
	inRect.origin.y = inContainerHeight - (inRect.origin.y + inRect.size.height);
	return inRect;
}

#endif

#if TARGET_OS_MAC
	typedef	UniChar	IzUniChar;
#elif TARGET_OS_WIN
	typedef WCHAR	IzUniChar;
#endif

// ---------------------------------------------------------------------------------
//	DEFINES
// ---------------------------------------------------------------------------------

// Note: the NativeObject type be one of the following
//
// -- NativeWindow
// -- NativeView
// -- NativeControl

#if TARGET_OS_MAC

	typedef struct OpaqueNativeObject*	NativeObject;
	typedef NativeObject				NativeActionObject;		// any object that can receive action callbacks

	typedef struct OpaqueNSMenu*		NativeMenu;

	typedef struct OpaqueNSDocument*	NativeDocument;

	typedef struct OpaqueNSWindow*		NativeWindow;
	typedef NativeWindow				NativeDialog;

	// on MacOS an alert is not a window; you must use
	// IZ_AlertGetWindow to get the NativeWindow
	// associated with the alert
	typedef struct OpaqueNSAlert*		NativeAlert;

	typedef struct OpaqueModalSession*	NativeModalSession;

	typedef struct OpaqueNSView*		NativeView;
	typedef NativeView					NativeScrollView;
	typedef NativeView					NativeTextView;			// subclass of NativeView

	typedef struct OpaqueNSControl*		NativeControl;
	typedef NativeControl				NativeTextField;		// subclass of NativeControl

	typedef CGContextRef				NativeContextRef;
	typedef CGContextRef				NativeBitmap;			// specifically, a CGBitmapContext
	typedef	CTFontRef					NativeFontRef;
	typedef HIMutableShapeRef			NativeRgn;

	typedef struct OpaqueNSCursor*		NativeCursor;
	typedef struct OpaqueNSTimer*		NativeTimer;

	#if defined(_OBJC_OBJC_H_) && defined(__cplusplus) && !defined(DONT_DEFINE_APPKIT_CONVERSIONS)

		inline NSObject*		IzNSObject(NativeObject ob)					{ return (NSObject*) ob; }
		inline NSObject*		IzNSActionObject(NativeActionObject ob)		{ return (NSObject*) ob; }

		inline NSMenu*			IzNSMenu(NativeMenu m)						{ return (NSMenu*) m; }

		inline NSDocument*		IzNSDocument(NativeDocument w)				{ return (NSDocument*) w; }

		inline NSWindow*		IzNSWindow(NativeWindow w)					{ return (NSWindow*) w; }
		inline NSWindow*		IzNSDialog(NativeDialog d)					{ return (NSWindow*) d; }
		inline NSAlert*			IzNSAlert(NativeAlert a)					{ return (NSAlert*) a; }

		inline NSModalSession*	IzNSModalSession(NativeModalSession ms)		{ return (NSModalSession*) ms; }

		inline NSView*			IzNSView(NativeView v)						{ return (NSView*) v; }
		inline NSTextView*		IzNSTextView(NativeTextView tv)				{ return (NSTextView*) tv; }

		inline NSScrollView*	IzNSScrollView(NativeScrollView sv)			{ return (NSScrollView*) sv; }

		inline NSControl*		IzNSControl(NativeControl c)				{ return (NSControl*) c; }
		inline NSTextField*		IzNSTextField(NativeTextField tf)			{ return (NSTextField*) tf; }

		inline NSTimer*			IzNSTimer(NativeTimer t)					{ return (NSTimer*) t; }
		inline NSCursor*		IzNsCursor(NativeCursor c)					{ return (NSCursor*) c; }

	#endif

#elif TARGET_OS_WIN

	typedef HWND						NativeObject;
	typedef NativeObject				NativeActionObject;		// any object that can receive action callbacks

	#if defined(__cplusplus)
		class CHMenu;
		typedef CHMenu*						NativeMenu;
	#else
		typedef void*						NativeMenu;
	#endif

	typedef void*						NativeDocument;
	
	typedef HWND						NativeWindow;
	typedef HWND						NativeDialog;
	typedef HWND						NativeAlert;

	typedef void*						NativeModalSession;

	typedef HWND						NativeView;
	typedef HWND						NativeScrollView;
	typedef HWND						NativeTextView;

	typedef HWND						NativeControl;
	typedef HWND						NativeTextField;

	typedef HDC							NativeContextRef;
	typedef HBITMAP						NativeBitmap;
	typedef	HFONT						NativeFontRef;
	typedef HRGN						NativeRgn;

	typedef HCURSOR						NativeCursor;

#endif

#endif	// #if _H_IZDefinitions
