// =================================================================================
//	IZFile						(c) 2018 Mark F. Coniglio -- All Rights Reserved
// =================================================================================

#ifndef _H_IZFile
#define _H_IZFile

#if TARGET_OS_WIN
#include <AEDataModel.h>
#endif

#ifdef __cplusplus

#define ALLOW_FS_SPEC	IZ_32BIT
#define ALLOW_FS_REF	TARGET_OS_MAC

#include <string>
#include <vector>
#include <stdio.h>

#include "IZDefinitions.h"

// ---------------------------------------------------------------------------
// IzNavTypeList
// ---------------------------------------------------------------------------

#if PRAGMA_STRUCT_ALIGN
	#pragma options align=mac68k
#elif PRAGMA_STRUCT_PACKPUSH
	#pragma pack(push, 2)
#elif PRAGMA_STRUCT_PACK
	#pragma pack(2)
#endif

struct IzNavTypeList {
	OSType			componentSignature;
	SInt16			reserved;
	SInt16			osTypeCount;
	OSType			osType[1];
};
typedef struct IzNavTypeList		IzNavTypeList;
typedef IzNavTypeList*				IzNavTypeListPtr;
typedef IzNavTypeListPtr*			IzNavTypeListHandle;

#if PRAGMA_STRUCT_ALIGN
	#pragma options align=reset
#elif PRAGMA_STRUCT_PACKPUSH
	#pragma pack(pop)
#elif PRAGMA_STRUCT_PACK
	#pragma pack()
#endif

// ---------------------------------------------------------------------------
//	ModTime STRUCT
// ---------------------------------------------------------------------------
// used for receiving file time modifications

class ModTime
{
public:

			ModTime() :
				mYear(0),
				mMonth(0),
				mDay(0),
				mHour(0),
				mMinute(0),
				mSecond(0)
			{
			}
	
	bool	IsValid()
			{
				return (mYear > 0);
			}
	
	UInt16	mYear;
	UInt16	mMonth;
	UInt16	mDay;
	UInt16	mHour;
	UInt16	mMinute;
	UInt16	mSecond;
};

// ---------------------------------------------------------------------------
// CONTSTANTS
// ---------------------------------------------------------------------------

const bool	SaveWhen_Closing	= false;
const bool	SaveWhen_Quitting	= true;

// ---------------------------------------------------------------------------
// TYPEDEFS
// ---------------------------------------------------------------------------

typedef OSType	IzFileType;

// ---------------------------------------------------------------------------
// CFileTypeDescriptor

class CFileTypeDescriptor
{
public:
			CFileTypeDescriptor();

			CFileTypeDescriptor(
				const std::string		inExtension,
				const std::string		inDescription = std::string());

	bool	IsValid() const
				{ return (mExtension.length() > 0 && mDescription.length() > 0); }
	
	std::string	mExtension;
	std::string	mDescription;
};


class CFileTypeDescriptorList : public std::vector<CFileTypeDescriptor>
{
public:
						CFileTypeDescriptorList()
							: std::vector<CFileTypeDescriptor>()
						{
						}

	void				AddAllFileTypeDescriptor();

	CFileTypeDescriptor	CreateAllFileTypeDescriptor() const;
};

// ---------------------------------------------------------------------------
// COMMON FILE TYPE DESCRIPTORS

inline CFileTypeDescriptor
GetQTMovieFileTypeDescriptor()
{
	return CFileTypeDescriptor("mov", "QuickTime Movie");
}

inline CFileTypeDescriptor
GetAVIMovieFileTypeDescriptor()
{
	return CFileTypeDescriptor("avi", "AVI Movie");
}

inline CFileTypeDescriptor
GetWMVMovieFileTypeDescriptor()
{
	return CFileTypeDescriptor("wmv", "WMV Movie");
}

inline CFileTypeDescriptor
GetMP4MovieFileTypeDescriptor()
{
	return CFileTypeDescriptor("mp4", "MP4 Movie");
}



// ---------------------------------------------------------------------------
// IZFileSpec

#if TARGET_OS_MAC
	#define kIzNativeFileSeparator	'/'
#elif TARGET_OS_WIN
	#define kIzNativeFileSeparator	'\\'
#endif

#define IZ_MAX_PATH		2048

void
IZ_EnsurePathEndsWithFileSeparator(
	std::string*		ioPath);
	
std::string
IZ_EnsurePathEndsWithFileSeparator(
	const std::string	inPath);
	

const size_t kIzFileRef_InvalidSeparatorOffset = (size_t) -1;

typedef FILE*	IzFileRef;
typedef int64_t	IzFileSize;

#define IZ_FILE_REF_INVALID	NULL

class IZFileSpec
{
public:

	static bool	FileSpecFromFullPath(
					std::string		inFullPath,
					IZFileSpec*		outFSSpec);
						
	static bool	FileSpecFromPathAndFileName(
					std::string		inPath,
					std::string		inFileName,
					IZFileSpec*		outFSSpec);
						
				IZFileSpec();
				
				IZFileSpec(
					std::string		inFullPath,
					char			inFileSeparatorOverride = 0);
				
				IZFileSpec(
					std::string		inPath,
					std::string		inFileName,
					char			inFileSeparatorOverride = 0);
	
	bool		operator==(
					const IZFileSpec&	inRhs) const;
	
	bool		operator!=(
					const IZFileSpec&	inRhs) const;
	
	bool		IsSpecified() const
					{ return strlen(name) > 0; }
					
	bool		HasValidPath() const;
	
	bool		HasValidFileWithExtension() const;
	
	void		Reset();
	
	void		SetFullPath(
					std::string		inPath,
					char			inFileSeparatorOverride = 0);
	
	std::string	GetFullPath() const	{ return std::string(name); }
	
	// return the path without the final component; if this is
	// already a directory path, then the returned path would
	// be the parent directory of the path.
	std::string	GetPath(
					bool			inIncludeTrailingSeparator = false) const;
	
	// return the last component in the path
	std::string	GetLastComponent() const;
	
	// return the last component in the path without any extension
	std::string GetPathWithoutExtension() const;
	
	// return the last component in the path without any extension
	std::string GetLastComponentWithoutExtension() const;
	
	bool		HasExtension() const;
	
	bool		HasExtensionMatching(
					std::string			inExtension) const;
	
	void		SetExtension(
					std::string			inExtension);
	
	std::string	GetExtension() const;
	
	// SEPARATE IMPLEMENTATIONS
	
	bool		Create(
					OSStatus*	outErr = NULL) const;
	
	bool		Create(
					OSType		inCreator,
					OSType		inFileType,
					OSStatus*	outErr = NULL) const;
	
	bool		SetCreatorAndFileType(
					OSType		inCreator,
					OSType		inFileType) const;
	
	bool		SetInvisibleFlag(
					bool		inInvisible) const;
	
	// returns true if the current path exists; this
	// function does not descrimiate between directory
	// paths and file paths. Optionally you may pass
	// a pointer to a bool that will indicate if the
	// path is a directory or file
	bool		PathExists(
					bool*	outIsDirectory = NULL) const;
	
	bool		DirectoryExists() const
				{
					bool isDir = false;
					return (PathExists(&isDir) && isDir);
				}
	
	bool		FileExists() const
				{
					bool isDir = false;
					return (PathExists(&isDir) && !isDir);
				}
	
	bool		CreateCopyFromExistingFile(
					const IZFileSpec&	inFileSpec,
					bool				inOverwriteExisting);

	bool		Delete() const;
	
	// note: this is a valid unix time in seconds
	// since Jan 1st 1970 at 00:00h GMT
	time_t		GetFileModificationTimeSeconds() const;
	
	ModTime		GetFileModificationTime() const;
	
	IzFileSize	GetFileSize() const;
	
#if ALLOW_FS_SPEC

				IZFileSpec(
					FSSpec			inFSSpec,
					bool			inAllowDirectory = false);
	
	bool		SetFileSpec(
					FSSpec			inFSSpec,
					bool			inAllowDirectory = false);
	
	OSStatus	GetFSSpec(
					FSSpec*			outFSSpec) const;
	
	operator	FSSpec() const;
	
#endif

#if ALLOW_FS_REF
				IZFileSpec(
					FSRef			inFSRef,
					bool			inAllowDirectory = false);
	
	bool		SetFileRef(
					FSRef			inFSRef,
					bool			inAllowDirectory = false);
	
	OSStatus	GetFSRef(
					FSRef*	outFSRef,
					bool*	isDirectory = NULL) const;
	
	operator	FSRef() const;
#endif

private:
	bool		StartsWithValidAbsolutePathChars() const;
	
	void		FindSeparators(
					char			inFileSeparatorOverride = 0);
	
	std::string	RemoveExtension(
					std::string		inString);
	
private:
	size_t	mPathComponentSeparatorOffset;
	size_t	mExtensionSeparatorOffset;

public:
	int16_t	vRefNum;
	int32_t	parID;
	char	name[512];
};

typedef IZFileSpec	IzSpec;

// ---------------------------------------------------------------------------
// IZFileRefList

class IZFileRefList : public std::vector<IZFileSpec>
{
public:
	IZFileRefList() :
		std::vector<IZFileSpec>()
	{
	}
	
	#if ALLOW_FS_SPEC
	bool	CreateAEDescFSSpecList(
				AEDescList*		outList) const;
	#endif

	bool	CreateAEDescFileURLList(
				AEDescList*		outList) const;
};

// ---------------------------------------------------------------------------
// FILE FUNCTIONS

#if ALLOW_FS_SPEC

bool
IZ_IsValidFSSpec(
	const FSSpec&	inFSSpec);

OSStatus
IZ_FileFSSpecToNativePath(
	const FSSpec&	inFSSpec,
	std::string&	outPath,
	bool*			isDirectory = NULL);		// CAN BE NULL

OSStatus
IZ_FileNativePathToFSSpec(
	std::string		inPath,
	FSSpec*			outFSSpec,
	bool*			outIsDirectory = NULL);		// CAN BE NULL

#endif

OSStatus
IZ_FileFSRefToNativePath(
	const FSRef&	inFSRef,
	std::string&	outPath,
	bool*			isDirectory = NULL);		// CAN BE NULL

#if ALLOW_FS_REF
OSStatus
IZ_FileNativePathToFSRef(
	std::string		inPath,
	FSRef*			outFSRef,
	bool*			outIsDirectory = NULL);		// CAN BE NULL
#endif

std::vector<IzSpec>
IZ_FileCreateFileSpecListFromAEDesc(
	const AEDesc&		inAEDesc);

std::vector<IzSpec>
IZ_FileCreateFileSpecListFromAppleEvent(
	const AppleEvent&	inAppleEvent,
	OSType				inKey);
	
// ---------------------------------------------------------------------------
// FILE DIALOGS + RELATED CLASSES
// ---------------------------------------------------------------------------

// ---------------------------------------------------------------------------
// IzFileDialogResponse + Related Constants

typedef OSType	IzFileDialogResponse;

enum {
	IzFileDialogResponse_Invalid	= FOUR_CHAR_CODE('invl'),
	IzFileDialogResponse_Running	= FOUR_CHAR_CODE('rnng'),
	IzFileDialogResponse_OK			= FOUR_CHAR_CODE('okok'),
	IzFileDialogResponse_Cancel		= FOUR_CHAR_CODE('cncl'),
	IzFileDialogResponse_Error		= FOUR_CHAR_CODE('erro')
};

enum IzFileDialogType {
	kIzFileDialogType_Invalid,
	kIzFileDialogType_Open,
	kIzFileDialogType_Save,
	kIzFileDialogType_SelectDirectory
};

// ---------------------------------------------------------------------------
// CFileDialogCallbackData
//
// Callers to the File Dialog routines below can pass this information to
// to allow window modal dialogs that return their responses via a callback.
//
// Classes that wish to do this should add the CIzFileDialogListener class
// as part of their interface and implement the FileDialogCompleteCallback
// method to respond to the completion of the dialog appropriately.

const OSType		kDefaultCallerRefID = FOUR_CHAR_CODE('fdcb');

typedef OSType		IzFDCallbackID;
typedef	uint64_t	IzFDCallbackData;

class CIzFileDialogListener;

class CFileDialogCallbackData
{
public:
	CFileDialogCallbackData();
	
	CFileDialogCallbackData(
		CIzFileDialogListener*		inFileDialogListener,					// CAN BE NULL
		IzFDCallbackID				inCallbackID = kDefaultCallerRefID,
		IzFDCallbackData			inCallbackData = 0);
	
	CFileDialogCallbackData(
		CFileDialogCallbackData*	inFileDialogUserRef);
	
	
	// listener that will be called with the file dialog data
	CIzFileDialogListener*		mFileDialogListener;
	
	// caller reference ID passed to the dialog
	// function; this allows the user to diferentiate
	// between different operations targeting the
	// same object
	IzFDCallbackID				mFileDialogCallbackID;
	
	// an additional field for user data. IzFDCallbackData
	// is currently defined as a uint64_t, which would
	// allow for a pointer if necessary.
	IzFDCallbackData			mFileDialogCallbackData;
};

// ---------------------------------------------------------------------------
// CIzFileDialogInfo
//
// This class is passed to the CIzFileDialogListener targeted in one
// of the file dialog methods below. It reports the outcome of the
// interaction with the user and the file dialog.

class CIzFileDialogInfo : public CFileDialogCallbackData
{
public:
							CIzFileDialogInfo() :
								CFileDialogCallbackData(),
								mFileDialogType(kIzFileDialogType_Invalid),
								mDialogResponse(IzFileDialogResponse_Invalid),
								mFileRefList(),
								mSaveReplaceFile(false)
							{
							}
							
							CIzFileDialogInfo(
								const CFileDialogCallbackData&	inFileDialogUserRef,
								IzFileDialogType				inDialogType,
								IzFileDialogResponse			inDialogResponse,
								const IZFileRefList&			inFileRefList = IZFileRefList(),
								bool							inSaveReplaceFile = false) :
									CFileDialogCallbackData(inFileDialogUserRef),
									mFileDialogType(inDialogType),
									mDialogResponse(inDialogResponse),
									mFileRefList(inFileRefList),
									mSaveReplaceFile(inSaveReplaceFile)
							{
							}
	
	size_t					GetFileCount() const
								{ return mFileRefList.size(); }
	
	const IZFileSpec&		GetFileRef(
								size_t		inIndex0) const
									{ return mFileRefList.at(inIndex0); }

	const IZFileRefList&	GetFileRefList() const
								{ return mFileRefList; }

	const IZFileSpec&		GetSaveFileName() const
								{ return mFileRefList.at(0); }

	// what kind of dialog sent this message
	IzFileDialogType			mFileDialogType;
	
	// the response code for the dialog, indicating
	// ok, cancel, or the value of a user button
	// if using IZ_OpenFileDialogWithUserButtons
	IzFileDialogResponse		mDialogResponse;
	
	// the list of chosen files for an open operation
	// or a select directory operation, or a single
	// item if a save operation
	IZFileRefList				mFileRefList;
	
	// if true, the save operation will replace the
	// file mFileRefList.at(0)
	bool						mSaveReplaceFile;
};

// ---------------------------------------------------------------------------
// CIzFileDialogListener
//
// Mixin class for callers who wish to support window modal dialog functions.
//
// The user should add this mixin and implement FileDialogCompleteCallback,
// keeping in mind that this callback will not be called until the dialog
// is dismissed.

class CIzFileDialogListener
{
public:
	virtual			~CIzFileDialogListener() { }
	
	virtual	void	FileDialogCompleteCallback(
						const CIzFileDialogInfo&	inFileDialogInfo) = 0;
};

// ---------------------------------------------------------------------------
// CIzFileDialogListener
//
// Static class that allows a user to easily call one of the file dialog
// functions without using the CIzFileDialogListener mixin. The results
// of the operation will be saved in the CIzFileDialogInfo member of this
// class
class StIzFileDialogListener : public CIzFileDialogListener, public CIzFileDialogInfo
{
public:
					StIzFileDialogListener() :
						CIzFileDialogListener(),
						CIzFileDialogInfo()
					{
						mFileDialogListener = this;
					}
	
	virtual	void	FileDialogCompleteCallback(
						const CIzFileDialogInfo&	inFileDialogInfo)
					{
						*((CIzFileDialogInfo*) this) = inFileDialogInfo;
					}
	
};

// ---------------------------------------------------------------------------
// IZFileButtonDesc

class CIZFileDialogButton
{
public:
	CIZFileDialogButton(
		std::string				inButtonText,
		IzFileDialogResponse	inDialogResponse) :
			mButtonText(inButtonText),
			mButtonResponse(inDialogResponse)
	{
	}
	
	std::string				mButtonText;
	IzFileDialogResponse	mButtonResponse;
};

class CIZFileDialogButtonList : public std::vector<CIZFileDialogButton>
{
public:
	CIZFileDialogButtonList(
		std::string		inButtonPrompt = std::string()) :
			std::vector<CIZFileDialogButton>(),
			mButtonPrompt(inButtonPrompt)
	{
	}
	
	std::string mButtonPrompt;
};

// ---------------------------------------------------------------------------
// LOW-LEVEL FILE FUNCTIONS

enum IzFilePerm {
	kIzFilePerm_CurPerm                     = 0x00,
	kIzFilePerm_RdPerm                      = 0x01,
	kIzFilePerm_WrPerm                      = 0x02,
	kIzFilePerm_RdWrPerm                    = 0x03,
	kIzFilePerm_RdWrShPerm                  = 0x04
};

enum IzSeekMode {
	kIzSeek_AtMark		= 0,    /* positioning modes in ioPosMode */
	kIzSeek_FromStart	= 1,
	kIzSeek_FromEOF		= 2,
	kIzSeek_FromMark	= 3
};

OSStatus
IZ_UnixErrorToMacOSError(
	int			inUnixErr);

OSStatus
IZ_FileOpen(
	const char*	inFullNativePath,
	IzFilePerm	inPermissions,
	IzFileRef&	outFileRef);

OSStatus
IZ_FileClose(
	IzFileRef	inFileRef);
	
OSStatus
IZ_FileRead(
	IzFileRef	inFileRef,
	IzFileSize*	ioDataLength,
	void*		inData);

OSStatus
IZ_FileWrite(
	IzFileRef	inFileRef,
	IzFileSize*	ioDataLength,
	const void*	inData);

OSStatus
IZ_FileSeek(
	IzFileRef	inFileRef,
	IzSeekMode	inSeekMode,
	IzFileSize	inSeekOffset);

OSStatus
IZ_FileGetPos(
	IzFileRef	inFileRef,
	IzFileSize*	outEOF);
	
OSStatus
IZ_FileSetPos(
	IzFileRef	inFileRef,
	IzFileSize	outEOF);

OSStatus
IZ_FileGetEOF(
	IzFileRef	inFileRef,
	IzFileSize*	outEOF);
	
OSStatus
IZ_FileSetEOF(
	IzFileRef	inFileRef,
	IzFileSize	outEOF);
	
// ---------------------------------------------------------------------------
// FILE DIALOG FUNCTIONS
// ---------------------------------------------------------------------------

// an application modal file open dialog
IZFileRefList
IZ_OpenFileDialog(
	std::string					inPrompt,
	bool						inSelectMultipleFiles,
	CFileTypeDescriptorList		inAllowedFileTypeDescriptors);

// an application modal file open dialog that
// uses the File Dialog callback mechanism
IzFileDialogResponse
IZ_OpenFileDialog(
	CFileDialogCallbackData		inFileDialogReceiver,
	std::string					inPrompt,
	bool						inSelectMultipleFiles,
	CFileTypeDescriptorList		inAllowedFileTypeDescriptors);

// a window modal file open dialog, optionally allowing
// an additional prompt and an optional list of buttons
// to be added to the dialog

IzFileDialogResponse
IZ_OpenFileDialog(
	CFileDialogCallbackData		inFileDialogReceiver,
	NativeWindow				inParentWindow,
	std::string					inPrompt,
	bool						inSelectMultipleFiles,
	CFileTypeDescriptorList		inAllowedFileTypeDescriptors,
	CIZFileDialogButtonList		inFileDialogButtonList = CIZFileDialogButtonList());

// will bring the most recently opened open file dialog to the front
// currently this is only impelemented for windows.
void
IZ_OpenFileDialogBringToFront();

// an application modal dialog that allows the user
// to select a directory
IZFileRefList
IZ_SelectDirectoryDialog(
	std::string					inPrompt);

// a window modal dialog that allows the user
// to select a directory
IzFileDialogResponse
IZ_SelectDirectoryDialog(
	CFileDialogCallbackData		inFileDialogReceiver,
	NativeWindow				inParentWindow,
	std::string					inPrompt);

// an application modal save dialog that allows the
// caller to specify a file.
IZFileRefList
IZ_SaveFileDialog(
	std::string					inPrompt,
	CFileTypeDescriptor			inFileDescriptor,
	const IZFileSpec*			ioExistingFileRef = NULL,
	bool*						outReplaceFile = NULL);

// a window modal save dialog that allows the
// caller to specify a file.
IzFileDialogResponse
IZ_SaveFileDialog(
	CFileDialogCallbackData		inFileDialogReceiver,
	NativeWindow				inParentWindow,
	std::string					inPrompt,
	CFileTypeDescriptor			inFileDescriptor,
	const IZFileSpec*			ioExistingFileRef = NULL);

#endif	// #ifdef __cplusplus

#endif	// #ifndef _H_IZAlert