// ===========================================================================
//	ImageBufferCommon			 (C)2000 Mark F. Coniglio. All rights reserved.
// ===========================================================================
//
// ImageBuffers define a bit mapped image.

#ifndef _H_ImageBufferCommon
#define _H_ImageBufferCommon

#include "IZQT.h"
#include "IZDefinitions.h"

#include "IsadoraTypes.h"
#include "ValueCommon.h"

#include "CoreImageCommon.h"

#if COMPILE_CORE_VIDEO
	#include "CoreImageBase.h"
#endif

// Magic Value to identify valid image buffers...
#define kImageBufferMagic		0x12345678

// ===========================================================================
// IBFormatInfo struct

struct IBFormatInfo
{
	IzInt32			mBitDepth;				// bit depth of this image
	IzInt32			mWidth;					// width of this image in pixels
	IzInt32			mHeight;				// height of this image in pixels
	OSType			mFormat;				// four char code for format
	IzUInt32		mUniqueID;				// image buffer's Unique ID
	IzInt32			mRowBytes;				// row size of this image in bytes
};

typedef struct IBFormatInfo	IBFormatInfo, *IBFormatInfoPtr;

// ===========================================================================
// IBMapFlags enum


enum {
	
	// when this flag is set, the row bytes of the incoming 
	// image buffers will be considered when performing
	// the mapping; this will force a conversion if the
	// row bytes of the input buffers do not match
	
	kIBMUpdateFlags_ConsiderRowBytes = (1 << 0),
	kIBMUpdateFlags_ConsiderUniqueID = (1 << 1)
};

typedef UInt32	IBMapFlags;

// ===========================================================================
// ImageBuffer struct

struct ImageBuffer
{
	DataValueInfo		mInfo;					// our data value info struct - required
	
	IzInt32				mBitDepth;				// bits per pixel of this image
	IzInt32				mWidth;					// width of this image in pixels
	IzInt32				mHeight;				// height of this image in pixels
	OSType				mFormat;				// four char code for format
	
	void*				mBaseAddress;			// Ptr to the raw bytes of the image,
												// if BitDepth is 16 each line should 
												// be seen ad gWidth unsigned shorts,
												// if BitDepth is 32 each line is 
												// gWidth unsigned ints.
									
	IzInt32				mRowBytes;				// byte offset between lines --
												// mBaseAddress + mRowBytes -> start of line 2;
												// mBaseAddress + 2 * mRowBytes -> start of third line, etc.

	NativeBitmap		mIzzyBitmap;			// Native Bitmap Object: v2.5: CGImageRef for Mac, HBITMAP for Windows; otherwise GWorldPtr
	
	void*				mUnused;				// was CIImage for CoreImage... but deprecated in v2.0

	IzUInt32			mUniqueID;				// this value will be unique for each ImageBuffer
												// created within a given Isadora session
	
	Boolean				mHasAlpha;				// true if there is alpha channel info in this image buffer
	
	IzUInt32			mMagic;					// magic values - always = kImageBufferMagic
	
	Boolean				mIsPremultiplied;		// v2.0: true if image buffer data is premutiplied
};

typedef struct ImageBuffer	ImageBuffer, *ImageBufferPtr;

// ===========================================================================
// GLOBALS

// empty core image buffer
extern ImageBuffer		sImageBuffer_CLEAR;

// ===========================================================================
// DEFINES

#define	DebugCheckImageBufferClear(a)	{ DebugAssert_((a)->mMagic == 0 && (a)->mIzzyBitmap == NULL); }

// ===========================================================================
// Image Buffer Functions

Boolean			ImageBuffer_IsValid(
					const ImageBuffer*	inImageBuffer);

void			ImageBuffer_InitImageBuffer(
					ImageBuffer*	ioImageBuffer);
						
void			ImageBuffer_Uninit(
					ImageBuffer*	ioImageBuffer);
						
void			ImageBuffer_Clear(
					ImageBuffer*	ioImageBuffer);
						
inline void		ImageBuffer_ClearBitmapInfo(
					ImageBuffer*	ioImageBuffer)
				{
					ioImageBuffer->mBaseAddress = NULL;
					ioImageBuffer->mRowBytes = 0;
					ioImageBuffer->mWidth = 0;
					ioImageBuffer->mHeight = 0;
					ioImageBuffer->mFormat = 0;
				}
						
ImageBufferPtr	ImageBuffer_MakeImageBuffer(
					SInt16				inWidth,
					SInt16				inHeight,
					SInt16				inPixelDepth,
					OSType				inFormat = 0);
					
void			ImageBuffer_DisposeImageBuffer(
					ImageBufferPtr		inImageBuffer);

NativeBitmap	ImageBuffer_StealAndReplaceImageBufferBitmap(
					ImageBufferPtr		ioImageBuffer);
#if TARGET_OS_MAC
bool			ImageBuffer_ReplaceBitmapWithStaticBuffer(
					ImageBuffer*	ioImageBuffer,
					OSType			inPixelFormat,
					const IzRect*	inBoundsRect,
					void*			inBaseAddress,
					long			inRowBytes,
					bool			inDisposeOldNativeBitmap);
#endif

bool			ImageBuffer_ReplaceBitmap(
					ImageBuffer*	ioImageBuffer,
					NativeBitmap	inNativeBitmap,
					bool			inDisposeOldNativeBitmap);

void			ImageBuffer_SetNativeBitmap(
					ImageBuffer*		mImageBuffer,
					NativeBitmap		inNativeBitmap);			// CAN BE NULL

#if IZ_ALLOW_QUICKTIME && !NATIVE_BITMAP_IS_GWORLD
void			ImageBuffer_SetGWorld(
					ImageBuffer*		mImageBuffer,
					GWorldPtr			inGWorld);					// CAN BE NULL
#endif

OSErr			ImageBuffer_CreateNativeBitmap(
					ImageBuffer*		inBuffer,
					SInt16				inWidth,
					SInt16				inHeight,
					SInt16				inPixelDepth,
					OSType				inFormat = 0);

void			ImageBuffer_DisposeNativeBitmap_DontClearParams(
					ImageBuffer*	inBuffer);

void			ImageBuffer_DisposeNativeBitmap(
					ImageBuffer*	inBuffer);

bool			ImageBuffer_IsBlack_Raw_ARGB(
					void*		inBuffer,
					IzInt32		inWidth,
					IzInt32		inHeight,
					IzInt32		inRowBytes);
					
bool			ImageBuffer_IsBlack(
					const ImageBuffer*	inBuffer);
					
void			ImageBuffer_EraseToColor(
					ImageBuffer*	inBuffer,
					UInt32			inARGBColor);

void			ImageBuffer_EraseToColorWithAlpha(
					ImageBuffer*	inBuffer,
					UInt32			inARGBColor);

void			ImageBuffer_IncrementFrameCount(
					ImageBuffer*	inBuffer);

void			ImageBuffer_Copy(
					ImageBufferPtr				inImageBufferSrc,
					ImageBufferPtr				inImageBufferDst,
					const IzRect*				inSrcRect,
					const IzRect*				inDstRect,
					SInt16						inXFerModeCode,
					RGBColor*					inRGBBlendColor = NULL,
					Boolean						inDontAccelerate = FALSE,
					Boolean						inFlipVertically = FALSE,
					Boolean						inInvertColors = FALSE);

// returns true if the specified image buffer contains a YUV bitmap 
Boolean			ImageBuffer_IsYUV(
					const ImageBufferPtr		inImageBuffer);

// ensures that the alpha byte for all pixels is set to 0xFF
void			ImageBuffer_FixAlpha(
					ImageBufferPtr		inImageBuffer);

void			ImageBuffer_AnalyzeBuffer(
					const ImageBufferPtr	inImageBuffer,
					bool*					outAllBlack,
					bool*					outAlphaAllZeroes,
					bool*					outAlphaAllOnes,
					UInt32*					outCenterPixel);
					
// if any alpha byte in the specified image buffer is not 0xFF
// then the mHasAlpha flag is set for the specified image buffer
void			ImageBuffer_CheckForAlpha(
					ImageBufferPtr		inImageBuffer);

// the caller is responsible for calling QTDisposeHandle on
// the returned result
#if IZ_ALLOW_QUICKTIME && NATIVE_BITMAP_IS_GWORLD

ImageDescriptionHandle	ImageBuffer_CreateImageDescription(
							ImageBufferPtr		inImageBuffer);
#endif

#if IZ_ALLOW_QUICKTIME && !NATIVE_BITMAP_IS_GWORLD

GWorldPtr				ImageBuffer_WrapGWorldAroundNativeImageBuffer(
							ImageBufferPtr		inImageBuffer);
#endif

#ifdef __cplusplus
	inline void		ImageBuffer_StoreFormatInfo(
						ImageBufferPtr		inImageBuffer,
						IBFormatInfoPtr		inFormatInfo)
					{
						inFormatInfo->mWidth = inImageBuffer->mWidth;
						inFormatInfo->mHeight = inImageBuffer->mHeight;
						inFormatInfo->mBitDepth = inImageBuffer->mBitDepth;
						inFormatInfo->mFormat = inImageBuffer->mFormat;
						inFormatInfo->mUniqueID = inImageBuffer->mUniqueID;
					}


	inline Boolean	ImageBuffer_AreFormatsDifferent(
						const ImageBufferPtr		inImageBuffer1,
						const ImageBufferPtr		inImageBuffer2)
					{
						return (inImageBuffer1->mWidth != inImageBuffer2->mWidth
							|| inImageBuffer1->mHeight != inImageBuffer2->mHeight
							|| inImageBuffer1->mBitDepth != inImageBuffer2->mBitDepth
							|| inImageBuffer1->mFormat != inImageBuffer2->mFormat);
					}
	
	inline Boolean	ImageBuffer_AreFormatsDifferent(
						const IBFormatInfoPtr		inFormatInfo1,
						const IBFormatInfoPtr		inFormatInfo2)
					{
						return (inFormatInfo1->mWidth != inFormatInfo2->mWidth
							|| inFormatInfo1->mHeight != inFormatInfo2->mHeight
							|| inFormatInfo1->mBitDepth != inFormatInfo2->mBitDepth
							|| inFormatInfo1->mFormat != inFormatInfo2->mFormat);
					}
	
	inline Boolean	ImageBuffer_HasFormatChanged(
						const ImageBufferPtr		inImageBuffer,
						const IBFormatInfoPtr		inFormatInfo)
					{
						return (inImageBuffer->mWidth != inFormatInfo->mWidth
							|| inImageBuffer->mHeight != inFormatInfo->mHeight
							|| inImageBuffer->mBitDepth != inFormatInfo->mBitDepth
							|| inImageBuffer->mFormat != inFormatInfo->mFormat);
					}
	
	inline Boolean	ImageBuffer_HasFormatChanged_WithFlags(
						const ImageBufferPtr		inImageBuffer,
						const IBFormatInfoPtr		inFormatInfo,
						const IBMapFlags			inIBMapFlags)
					{
						bool doesntMatch = (inImageBuffer->mWidth != inFormatInfo->mWidth
							|| inImageBuffer->mHeight != inFormatInfo->mHeight
							|| inImageBuffer->mBitDepth != inFormatInfo->mBitDepth
							|| inImageBuffer->mFormat != inFormatInfo->mFormat);
							
						if ((inIBMapFlags & kIBMUpdateFlags_ConsiderRowBytes) != 0) {
							if (inImageBuffer->mRowBytes != inFormatInfo->mRowBytes) {
								doesntMatch = true;
							}
						}
						if ((inIBMapFlags & kIBMUpdateFlags_ConsiderUniqueID) != 0) {
							if (inImageBuffer->mUniqueID != inFormatInfo->mUniqueID) {
								doesntMatch = true;
							}
						}
						
						return doesntMatch;
					}
	
	inline Boolean	ImageBuffer_HasFormatOrUniqueIDChanged(
						const ImageBufferPtr		inImageBuffer,
						const IBFormatInfoPtr		inFormatInfo)
					{
						return (inImageBuffer->mWidth != inFormatInfo->mWidth
							|| inImageBuffer->mHeight != inFormatInfo->mHeight
							|| inImageBuffer->mBitDepth != inFormatInfo->mBitDepth
							|| inImageBuffer->mFormat != inFormatInfo->mFormat
							|| inImageBuffer->mUniqueID != inFormatInfo->mUniqueID);
					}
	
	inline Boolean	ImageBuffer_HasFormatChanged_ForceFormat(
						const ImageBufferPtr		inImageBuffer,
						const IBFormatInfoPtr		inFormatInfo,
						const OSType				inForceFormat,
						const IzInt32				inForceBitDepth)
					{
						if (inForceFormat != 0) {
							return (inImageBuffer->mWidth != inFormatInfo->mWidth
								|| inImageBuffer->mHeight != inFormatInfo->mHeight
								|| inImageBuffer->mBitDepth != inForceBitDepth
								|| inImageBuffer->mFormat != inForceFormat);
						} else {
							return (inImageBuffer->mWidth != inFormatInfo->mWidth
								|| inImageBuffer->mHeight != inFormatInfo->mHeight
								|| inImageBuffer->mBitDepth != inFormatInfo->mBitDepth
								|| inImageBuffer->mFormat != inFormatInfo->mFormat);
						}
					}
	
	inline void		ImageBuffer_ClearFormatInfo(
						IBFormatInfoPtr				inFormatInfo)
					{
						inFormatInfo->mWidth = 0;
						inFormatInfo->mHeight = 0;
						inFormatInfo->mBitDepth = 0;
						inFormatInfo->mFormat = 0;
						inFormatInfo->mUniqueID = 0;
					}

#else
	void			ImageBuffer_StoreFormatInfo(
						ImageBufferPtr		inImageBuffer,
						IBFormatInfoPtr		inFormatInfo);
						
	Boolean			ImageBuffer_AreFormatsDifferent(
						const IBFormatInfoPtr		inFormatInfo1,
						const IBFormatInfoPtr		inFormatInfo2);
						
	void			ImageBuffer_ClearFormatInfo(
						IBFormatInfoPtr		inFormatInfo);

	Boolean			ImageBuffer_HasFormatChanged(
						const ImageBufferPtr		inImageBuffer,
						const IBFormatInfoPtr		inFormatInfo);

	Boolean			ImageBuffer_HasFormatOrUniqueIDChanged(
						const ImageBufferPtr		inImageBuffer,
						const IBFormatInfoPtr		inFormatInfo);

	Boolean			ImageBuffer_HasFormatChanged_ForceFormat(
						const ImageBufferPtr		inImageBuffer,
						const IBFormatInfoPtr		inFormatInfo,
						const OSType				inForceFormat,
						const IzInt32				inForceBitDepth);

	void			ImageBuffer_ClearFormatInfo(
						IBFormatInfoPtr				inFormatInfo);
#endif

#if NATIVE_BITMAP_IS_GWORLD

void			ImageBuffer_Copy_ARGB32_to_2yuv(
					ImageBufferPtr		inSrcBuffer,
					ImageBufferPtr		inDstBuffer,
					const IzRect*		inSrcRect = NULL,
					const IzRect*		inDstRect = NULL);
					
void			ImageBuffer_Copy_ARGB32_to_2yuv_2to1(
					ImageBufferPtr		inSrcBuffer,
					ImageBufferPtr		inDstBuffer,
					const IzRect*		inSrcRect = NULL,
					const IzRect*		inDstRect = NULL);

void			ImageBuffer_Copy_2yuv_to_ARGB32(
					ImageBufferPtr		inSrcBuffer,
					NativePixMapHandle	inDstMap,
					const IzRect*		inSrcRect = NULL,
					const IzRect*		inDstRect = NULL);

void			ImageBuffer_Copy_2yuv_to_ARGB32(
					ImageBufferPtr		inSrcBuffer,
					ImageBufferPtr		inDstBuffer,
					const IzRect*		inSrcRect = NULL,
					const IzRect*		inDstRect = NULL);
					
void			ImageBuffer_Copy_2yuv_to_ARGB32_1to2(
					ImageBufferPtr		inSrcBuffer,
					NativePixMapHandle	inDstMap,
					const IzRect*		inSrcRect /* = NULL */,
					const IzRect*		inDstRect /* = NULL */);
					
void			ImageBuffer_Copy_2yuv_to_ARGB32_1to2(
					ImageBufferPtr		inSrcBuffer,
					ImageBufferPtr		inDstBuffer,
					const IzRect*		inSrcRect /* = NULL */,
					const IzRect*		inDstRect /* = NULL */);

#endif

#if (DEBUG || CHECK_IMAGE_BUFFER_MAGIC)
	#define ImageBuffer_CheckMagic(ib)	Assert_(ib == NULL || (ib)->mMagic == kImageBufferMagic)
#else
	#define ImageBuffer_CheckMagic(ib)	{ }
#endif

#include "TextureBufferCommon.h"
						
#endif
