// ===========================================================================
//	StGLFlag					 (C)2004 Mark F. Coniglio	  All rights reserved.
// ===========================================================================
//
// A handy C++ class for setting and restoring OpenGL flags.
//
// Call the constructor to set an OpenGL flag. At the moment this object is
// instantiated, the current state of the constructor is stored, and the new
// state is set. When this object's destructor is called, this original
// state is automatically restored.
// 
// Example Usage: StGLFlag lights(GL_LIGHTING, TRUE); -- Turns on lighting,
// restoring GL_LIGHTING to its previous state upon destruction

#ifndef _H_StGLFlag
#define _H_StGLFlag

// do not allow StGLFlag inside the isadora app!
#ifndef __ISADORA_APP__

#include "3DPluginDrawUtil.h" // for gl includes, for now

#define kStGLReportError	1

#if kStGLReportError && defined(__ISADORA_APP__) && DEBUG
	#include "COpenGLUtil.h"
	#define	StGLErrReport()		COpenGLUtil::OpenGLReportError()
#else
	#define	StGLErrReport()		{ }
#endif

// --------------------------------------------------------------------------------

#if TARGET_OS_MAC && DEBUG
	#define NO_DEBUG_INLINE	__attribute__ ((noinline))
#elif TARGET_OS_WIN && DEBUG && !defined(__MWERKS__)
	#define NO_DEBUG_INLINE	__declspec(noinline) 
#else
	#define NO_DEBUG_INLINE
#endif

class StGLFlag
{
public:

	inline StGLFlag(
		GLenum		inGLFlag,
		GLboolean	inEnable)
	{
		// clear any pending open gl errors
		StGLFlag_CheckOpenGLError_DebugOnly();

		// remember GL flag being modified
		mGLFlag = inGLFlag;
		
		// get current flag setting
		mSavedEnableFlag = glIsEnabled(mGLFlag);
		
		// check for error after getting enable flag
		GLenum err = StGLFlag_CheckOpenGLError_DebugOnly();
		
		// assume we won't need restore
		mNeedsRestore = GL_FALSE;
		
		// if desired setting doesn't match current setting
		if (inEnable != mSavedEnableFlag || err != 0) {
			
			// mark us as needing to be restored
			mNeedsRestore = GL_TRUE;
			
			if (inEnable == GL_TRUE) {
				glEnable(mGLFlag);
				StGLFlag_CheckOpenGLError_DebugOnly();
			} else {
				glDisable(mGLFlag);
				StGLFlag_CheckOpenGLError_DebugOnly();
			}
			
			StGLErrReport();
		}
	}
	
	inline ~StGLFlag()
	{
		if (mNeedsRestore == GL_TRUE) {
			if (mSavedEnableFlag == GL_TRUE)
				glEnable(mGLFlag);
			else
				glDisable(mGLFlag);
			StGLErrReport();
		}
	}
	
	NO_DEBUG_INLINE GLenum StGLFlag_CheckOpenGLError()
	{
		GLenum lastError = GL_NO_ERROR;
		
		GLenum err;
		do {
			err = glGetError();
			if (err != GL_NO_ERROR) {
				lastError = err;
			}
		} while (err != 0);
		
		return lastError;
	}

	NO_DEBUG_INLINE GLenum StGLFlag_CheckOpenGLError_DebugOnly()
	{
		#if DEBUG
			return StGLFlag_CheckOpenGLError();
		#else
			return GL_NO_ERROR;
		#endif
	}

private:
	GLenum		mGLFlag;
	GLboolean	mNeedsRestore;
	GLboolean	mSavedEnableFlag;
};

// --------------------------------------------------------------------------------

class StDepthWriteMask
{
public:

	inline StDepthWriteMask(
		GLboolean	inEnable)
	{
		// get current depth write mask setting
		glGetBooleanv(GL_DEPTH_WRITEMASK, &mSavedDepthWriteMask);
		StGLErrReport();
		
		mNeedsRestore = GL_FALSE;
		
		// if desired setting doesn't match current setting
		if (inEnable != mSavedDepthWriteMask) {
			
			// indicate we'll need to be restored later
			mNeedsRestore = GL_TRUE;
			
			// change the depth mask setting
			if (inEnable == GL_TRUE) {
				glDepthMask(GL_TRUE);
			} else {
				glDepthMask(GL_FALSE);
			}
			StGLErrReport();
		}
	}
	
	inline ~StDepthWriteMask()
	{
		// restore if necessary
		if (mNeedsRestore == GL_TRUE) {
			glDepthMask(mSavedDepthWriteMask);
			StGLErrReport();
		}
	}

private:
	GLboolean	mNeedsRestore;
	GLboolean	mSavedDepthWriteMask;
};

// --------------------------------------------------------------------------------

class StTexEnvMode
{
public:

	inline StTexEnvMode(
		GLint	inTexEnvMode)
	{
		// get current tex env mode
		glGetTexEnviv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, &mSavedTexEnvMode);
		StGLErrReport();
		
		mNeedsRestore = GL_FALSE;
		
		// if desired setting doesn't match current setting
		if (inTexEnvMode != mSavedTexEnvMode) {
			// indicate we'll need to be restored later
			mNeedsRestore = GL_TRUE;
			// set tex env mode
			glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, inTexEnvMode);
			StGLErrReport();
		}
	}
	
	inline ~StTexEnvMode()
	{
		// restore if necessary
		if (mNeedsRestore == GL_TRUE) {
			glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, mSavedTexEnvMode);
			StGLErrReport();
		}
	}

private:
	GLboolean	mNeedsRestore;
	GLint		mSavedTexEnvMode;
};
#endif

#endif