// ===========================================================================
//	CValueCommon				(c) 2000 Mark F. Coniglio. All rights reserved
// ===========================================================================
//
// Structure that encapsulates a value, embodying its type and value.
//
// Isadora passes values as a structure that consists of the value's type
// and its actual value. There are four basic types, and a fifth type that
// allows plugins to define their own data types. The Value structure 
// represents this with two members: the 'type' member holds a constant
// that designates the value's type, and the 'value' member holds the
// actual value. The value is designated inside the struct as a signed
// 32 bit integer, but the meaning of the contents of this member is
// determined by the 'type' member.
// 
// The five basic types are as follows:
//
// kBoolean: a boolean value, non-zero being true and zero being false.
//
// kInteger: a 32-bit signed integer
//
// kFloat: a 32-bit floating point value
//
// kRange: a pair of signed 16 bit numbers designating a range of values
//
// kString: a pointer to a String structure
//
// The sixth type, kData allows artibrarily sized chunks of data to be
// passed through Isadora using the Value structure. In this case, the
// 'value' parameter points to a block of data that begins with a 
// DataValueInfo struct that gives some basic information about the
// data. Generally, your data struct will look something like this:
//
//	typedef struct {
//		DataValueInfo	info;
//		MyDataType		myData;
//	} MyDataBlock;
//
// In this case, the contents of the 'value' member would be a pointer
// to MyDataBlock. There is more information about the members of the
// DataValueInfo struct below.
//
// Note that one important, built-in use of the kData type is to pass
// ImageBuffers from actor to actor. See ImageBufferCommon.h for more
// information about ImageBuffers.

#ifndef _H_ValueCommon
#define _H_ValueCommon

#include "IZTypes.h"
#include <stdint.h>

// -----------------------------------------------------------------------------

typedef UInt16						ValueType;

// The ValueType constants
enum {
	
	kTypeInvalid,					// invalid value type
	
	// basic numeric types
	kBoolean,						// a true (non-zero) or false (zero) value
	kInteger,						// a 64-bit signed integer
	kFloat,							// a 64-bit floating point number
	kRange,							// a pair of 16-bit unsigned integers defining a range of values

	// data types
	kData,							// application defined data type - data member of Value
									// struct points data staring with DataValueType struct

	// string type
	kString,						// str member points to String type
	
	// timecode type
	kTimecode,						// bits 0-28, timecode frame number, bits 29-31 = timecode type (see TimecodeType in IsadoraTypes.h)
	
	kValueType_Min = kBoolean,		// minimum value type
	kValueType_Max = kTimecode		// maximum value type
};

#define	ValueTypeIsValid(a)		((a) >= kValueType_Min && (a) <= kValueType_Max)

// special marker used when broadcasting Values over a network.
// this value indicates that a string length and the string data
// will follow the Value struct in the packet. 

#define	kValueDataToFollowMarker64	0xFFFFFFFFFFFFFFFF
#define	kValueDataToFollowMarker32	0xFFFFFFFF

// --------------------
// IzValue_Int TYPE

typedef IzInt		IzValue_Int;

const IzValue_Int	IZ_VALUE_INT_MAX	= IZ_INT_MAX;
const IzValue_Int	IZ_VALUE_INT_MIN	= IZ_INT_MIN;

#define IZ_VALUE_INT_SIZE				 IZ_INT_SIZE

// --------------------
// IzValue_Int TYPE

typedef IzUInt		IzValue_UInt;

const IzValue_UInt	IZ_VALUE_UINT_MAX	= IZ_UINT_MAX;
const IzValue_UInt	IZ_VALUE_UINT_MIN	= IZ_UINT_MIN;

#define IZ_VALUE_UINT_SIZE				 IZ_UINT_SIZE

// --------------------
// IzValue_Flt TYPE

typedef IzReal		IzValue_Flt;

const IzReal		IZ_VALUE_FLT_MAX	= IZ_REAL_MAX;
const IzReal		IZ_VALUE_FLT_MIN	= IZ_REAL_MIN;

#define IZ_VALUE_FLT_SIZE				 IZ_REAL_SIZE

// --------------------
// IzValue_Range TYPE

typedef uint16_t	IzValue_Range;

const IzValue_Range	IZ_VALUE_RANGE_MAX = IZ_UINT16_MAX;
const IzValue_Range	IZ_VALUE_RANGE_MIN = IZ_UINT16_MIN;

#define IZ_VALUE_RANGE_SIZE				 IZ_UINT16_SIZE

// --------------------
// IzValue_Ptr TYPE

typedef IzPtr		IzValue_Ptr;

#define IZ_VALUE_PTR_SIZE				 IZ_PTR_SIZE

// --------------------
// IzValue_Bool TYPE

typedef uint8_t		IzValue_Bool;

const IzValue_Range	IZ_VALUE_BOOL_MAX = 1;
const IzValue_Range	IZ_VALUE_BOOL_MIN = 0;

#define IZ_VALUE_BOOL_SIZE				 1

// --------------------
// IzValue_Timecode TYPE

// timecode value structure
// bits  0-28 - timecode frame number
// bits 29-32 - timecode type (see TimecodeType in IsadoraTypes.h)

typedef int64_t	IzValue_Timecode;

typedef int32_t	TimecodeFrame;

typedef float	TimecodeSeconds;

const float	IZ_VALUE_TIMECODE_SECONDS_MIN = -FLT_MAX;	// min seconds @ 30fps = 24h * 60m * 60s
const float	IZ_VALUE_TIMECODE_SECONDS_MAX = +FLT_MAX;	// max seconds @ 30fps = 24h * 60m * 60s

// the following value can be used to check for timecode
// values that might be set to END. We don't support
// timecode values greater than 24 hours.

const float IZ_VALUE_TIMECODE_24HOURS = 86400.0;

#define IZ_VALUE_TIMECODE_SIZE		8

#define kIzValue_TimecodeTypeMaskHi	0x0000000F00000000	// bits 32-35 - 4 bits
#define kIzValue_TimecodeTypeMaskLo	0x000000000000000F	// bits 0-3 - 4 bits
#define kIzValue_TimecodeTypeShift	32					// type in high order 32 bits
#define kIzValue_TimecodeMask		0xFFFFFFFF			// low order 32 bits

#if TRUE && DEBUG && defined(__ISADORA_APP__)
	#define TC_INTENSIVE_CHECK 1
#else
	#define TC_INTENSIVE_CHECK 0
#endif

#if TC_INTENSIVE_CHECKS
#define	TimecodeIntensiveCheckAssert_		DebugAssert_
#else
#define	TimecodeIntensiveCheckAssert_(...)	 { }
#endif

inline uint8_t
GetIzTimecodeType(
	IzValue_Timecode	tc)
{
	return (uint8_t) ((tc >> kIzValue_TimecodeTypeShift) & kIzValue_TimecodeTypeMaskLo);
}

inline void
SetIzTimecode(
	IzValue_Timecode&	tc,
	uint8_t				t,
	TimecodeSeconds		s)
{
	TimecodeIntensiveCheckAssert_(GetIzTimecodeType(tc) != 9);
	uint32_t ss = *((uint32_t*) &s);
	tc = (((uint64_t(t) & kIzValue_TimecodeTypeMaskLo) << kIzValue_TimecodeTypeShift) | (ss & kIzValue_TimecodeMask) );
	TimecodeIntensiveCheckAssert_(GetIzTimecodeType(tc) != 9);
}

inline void
SetIzTimecodeType(
	IzValue_Timecode&	tc,
	uint8_t				t)
{
	TimecodeIntensiveCheckAssert_(t != 9);
	tc = (((uint64_t(t) & kIzValue_TimecodeTypeMaskLo) << kIzValue_TimecodeTypeShift) | (tc & kIzValue_TimecodeMask) );
	TimecodeIntensiveCheckAssert_(GetIzTimecodeType(tc) != 9);
}

inline TimecodeSeconds
GetIzTimecodeSeconds(
	IzValue_Timecode	tc)
{
	TimecodeIntensiveCheckAssert_(GetIzTimecodeType(tc) != 9);
	uint32_t ss = (tc & kIzValue_TimecodeMask);
	return *((TimecodeSeconds*) & ss);
}

inline void
SetIzTimecodeSeconds(
	IzValue_Timecode&	tc,
	TimecodeSeconds		s)
{
	uint32_t ss = *((uint32_t*) &s);
	tc = ((tc & kIzValue_TimecodeTypeMaskHi) | (ss & kIzValue_TimecodeMask));
	TimecodeIntensiveCheckAssert_(GetIzTimecodeType(tc) != 9);
}

// -----------------------------------------------------------------------------
// The Range structure, used inside of the Value struct
//
// the hi and lo members were made unsigned in v2.6. this
// should not have any negative impact since no Isadora
// plugins allowed a negative value for these members

struct Range {
	IzValue_Range	hi;		// v2.6
	IzValue_Range	lo;		// v2.6
};

typedef struct	Range 	Range;

const char		kPropRangeDelimiter			= '-';

// -----------------------------------------------------------------------------
// The String structure, used inside of the Value struct

#define kStringMagic	0x12345678

struct String {
	IzInt32	useCount;					// use count for this string
	#ifdef kStringMagic
	UInt32	magic;						// magic value for checking - only if kStringMagic defined
	#endif
	char	strData[1];					// the actual string data
};

typedef struct String 	String;

#ifdef kStringMagic
	#define	AssertStringMagicValid_(value)			Assert_( ((value).type == kString) && ( (value).u.str == nil || (value).u.str->magic == kStringMagic ) )
	#define	PluginAssertStringMagicValid_(value)	PluginAssert_( ((value).type == kString) && ( (value).u.str == nil || (value).u.str->magic == kStringMagic ) )
#else
	#define	AssertStringMagicValid_(valuePtr)		0
	#define	PluginAssertStringMagicValid_(valuePtr)	0
#endif

// -----------------------------------------------------------------------------
// The Value structure

struct Value {
	ValueType	type;						// type - use this to choose which member of the union 'u' to access
	
	union {
		IzValue_Int			ivalue;			// type == kInteger or kBoolean
		IzValue_Flt			fvalue;			// type == kFloat
		Range				range;			// type == kRange
		IzValue_Ptr			data;			// type == kData
		String*				str;			// type == kString
		IzValue_Timecode	tcvalue;		// type == kTimecode
	} u;
};

typedef struct Value			Value;
typedef Value*					ValuePtr;

// -----------------------------------------------------------------------------
// The MultiValue structure

struct MultiValue {
	UInt32	mValueCount;
	Value	mValues[1];			// array of values -- actual len determined by mValueCount
};

typedef struct MultiValue		MultiValue;
typedef MultiValue*				MultiValuePtr;

// -----------------------------------------------------------------------------
// Isadora assumes that a Value struct with a type of kData
// points to a block of memory that begins with the following
// information

typedef	OSType	DataType;

struct DataValueInfo {
	DataType	mDataType;			// set to a unique four character id for this data type

	Boolean		mDescNeverChanges;	// prevents the description in the thing from being redrawn when
									// the new data is sent out.

	UInt32		mDataChangeCount;	// should be incremented whenever the internal data changes
};
typedef struct DataValueInfo	DataValueInfo;

// -----------------------------------------------------------------------------
// the CustomDataValueInfo struct is a variant of the DataValueInfo
// struct, with name and length fields attached to the end -- this allows
// custom data types to be added by third parties

struct CustomDataValueInfo {
	DataValueInfo	mDataValueInfo;		// embedded DataValueInfo struct
	char			mDataName[16];		// name of the data - to be displayed by Isadora
	SInt32			mDataLen;			// optional -- number of bytes of data to follow
};
typedef struct CustomDataValueInfo	CustomDataValueInfo;

// macro to return the data type of a from the DataValueInfo struct
// pointed to by a Value struct with a type of kData
#define	GetDataValueDataType(inValuePtr)  ((((ValuePtr)(inValuePtr))->type == kData \
	&& ((ValuePtr)(inValuePtr))->u.data != nil) \
		? ((DataValueInfo*)((ValuePtr)(inValuePtr))->u.data)->mDataType : 0)

// macro returns true if the data type of the Value pointed to by
// inValuePtr is kData, and if the data value is not null, and if
// the data type matches inDataType
#define	ValueUsesDataType(inValuePtr, inDataType) (((ValuePtr)(inValuePtr))->type == kData \
	&& ((ValuePtr)(inValuePtr))->u.data != nil \
	&& ((DataValueInfo*)((ValuePtr)(inValuePtr))->u.data)->mDataType == (inDataType))

// macro returns the data value pointed if the data value matches the
// one specified, nil otherwise.
#define GetDataValueOfType(inValuePtr, inDataType, inDesiredType) ((inDesiredType)(ValueUsesDataType(inValuePtr, inDataType) ? (((ValuePtr)(inValuePtr))->u.data) : nil))

#define GetCustomDataName(inValuePtr) (((CustomDataValueInfo*) ((ValuePtr)(inValuePtr))->u.data)->mDataName )

// -----------------------------------------------------------------------------
// Useful Constants for the Value struct

#define	kFloatMin				((IzValue_Real) IZ_VALUE_FLT_MIN)
#define	kFloatMax				((IzValue_Real) IZ_VALUE_FLT_MAX)
#define	kIntMin					((IzValue_Int) IZ_VALUE_INT_MIN)
#define	kIntMax					((IzValue_Int) IZ_VALUE_INT_MAX)
#define	kTimecodeMin			((IzValue_Timecode) IZ_VALUE_TIMECODE_FRAME_MIN)
#define	kTimecodeMax			((IzValue_Timecode) IZ_VALUE_TIMECODE_FRAME_MAX)
#define	kRangeMin				((IzValue_Range) IZ_VALUE_RANGE_MIN)		// v2.6 - changed limits for unsigned range; min = 0
#define kRangeMax				((IzValue_Range) IZ_VALUE_RANGE_MAX)		// v2.6 - changed limits for unsigned range; max = 65535

#endif	// ifdef _H_ValueCommon
