// ===========================================================================
//	Isadora Demo Plugin			(c) 2021 Mark F. Coniglio. All rights reserved
// ===========================================================================

/*

# ISADORA DEMO PLUGIN - READ ME

## DESCRIPTION

This plugin demonstrations a simple video effect implemented both as an
OpenGL shader. The shader provided simply modulates the red, green, and blue
components of the source video stream as found in the "Colorizer" actor that
comes built-in with Isdaora. To perform a different effect, you simply need
to replace the shader under the heading "COLORIZER SHADER"

## ABOUT VIDEO TYPES

In Isadora, video data comes in three "flavors":

- **vid-gpu**: Image is stored as an OpenGL texture on your graphics card.
Image data is processed using OpenGL shaders

- **vid-ci**: Image is stored as
a CIImage on your graphics card (macOS only). Image data is processed using
OpenGL shaders

- **vid-cpu**: Image is stored in the main memory of your
computer. image data is processed using standard C++.  format is only used by
legacy Isadora plugins, usually marked by the word classic" in their name.

This plugin only handles 'vid-gpu' streams only, as this is by far the most
common video format used by Isadora. To see a plugin that offers a mutable
input that can receive all three flavors, please take a look at the Analyze
Image Demo project within the Isadora SDK.

## CUSTOMIZING THIS SOURCE CODE

To customize this file, search for the text ###. All of the places where you
will need to customize the file are marked with this pattern of characters.

## LICENSE

IMPORTANT: This source code ("the software") is supplied to you in
consideration of your agreement to the following terms. If you do not agree
to the terms, do not install, use, modify or redistribute the software.

Mark Coniglio (aka TroikaTronix) grants you a personal, non exclusive license
to use, reproduce, modify this software with and to redistribute it, with or
without modifications, in source and/or binary form. Except as expressly
stated in this license, no other rights are granted, express or implied, to
you by TroikaTronix.

This software is provided on an "AS IS" basis. TroikaTronix makes no
warranties, express or implied, including without limitation the implied
warranties of non-infringement, merchantability, and fitness for a
particular purpurse, regarding this software or its use and operation alone
or in combination with your products.

In no event shall TroikaTronix be liable for any special, indirect,
incidental, or consequential damages arising in any way out of the use,
reproduction, modification and/or distribution of this software.

*/

// ---------------------------------------------------------------------------------
// INCLUDES
// ---------------------------------------------------------------------------------

// ISADORA INCLUDES
// These three includes should be present at the top of every Isadora plugin.

#define _STDINT_H

#include "IsadoraPluginPrefix.h"
#include "IsadoraTypes.h"
#include "IsadoraCallbacks.h"

// PLUGIN SPECIFIC INCLUDES
// required for plugins that process GPU and CoreImage based image data
#include "MultiVideoInOutUtil.h"
#include "CGLShader.h"

// GPU: useful for plugins that process CPU based image data
#include "PluginDrawUtil.h"

// STANDARD INCLUDES
#include <string.h>
#include <stdio.h>

// ---------------------------------------------------------------------------------
// MacOS Specific
// ---------------------------------------------------------------------------------
#if TARGET_OS_MAC
#define EXPORT_
#endif

#pragma mark ----- COLORIZER SHADER -----

// ---------------------------------------------------------------------------------
// Win32  Specific
// ---------------------------------------------------------------------------------
#if TARGET_OS_WIN	
	#define EXPORT_ __declspec(dllexport)
#endif

// ---------------------------------------------------------------------------------
//	Exported Function Definitions
// ---------------------------------------------------------------------------------

#ifdef __cplusplus
extern "C" {
#endif

EXPORT_ void GetActorInfo(void* inParam, ActorInfo* outActorParams);

#ifdef __cplusplus
}
#endif

// ---------------------------------------------------------------------------------
//	FORWARD DECLARTIONS
// ---------------------------------------------------------------------------------
static void
ReceiveMessage(
	IsadoraParameters*	ip,
	MessageMask			inMessageMask,
	PortIndex			inPortIndex1,
	const MsgData*		inData,
	UInt32				inLen,
	IzRefCon			inRefCon);

// ---------------------------------------------------------------------------------
// GLOBAL VARIABLES
// ---------------------------------------------------------------------------------
// ### Declare global variables, common to all instantiations of this plugin here

// Example: static int gMyGlobalVariable = 5;

// ---------------------------------------------------------------------------------
// PluginInfo struct
// ---------------------------------------------------------------------------------
// ### This structure neeeds to contain all variables used by your plugin. Memory for
// this struct is allocated during the CreateActor function, and disposed during
// the DisposeActor function, and is private to each copy of the plugin.
//
// If your plugin needs global data, declare them as static variables within this
// file. Any static variable will be global to all instantiations of the plugin.

typedef struct {

	ActorInfo*				mActorInfoPtr;		// our ActorInfo Pointer - set during create actor function
	MessageReceiverRef		mMessageReceiver;	// pointer to our message receiver reference
	
	PixelSize				mSize;				// ### Demo Plugin - size of the bitmap and the texture output
	bool					mSizeChanged;		// ### Demo Plugin - true when the size of the bitmap has changed

	float					mRedAmount;			// ### Demo Plugin - value = 0 to 1.0
	float					mGreenAmount;		// ### Demo Plugin - value = 0 to 1.0
	float					mBlueAmount;		// ### Demo Plugin - value = 0 to 1.0
	bool					mColorChanged;		// ### Demo Plugin - true when the color of the bitmap has changed

	Value					mBitmapValue;		// ### Demo Plugin - we enclose our bitmap in this value to feed to CMultiVideoInput
	ImageBufferPtr			mBitmap;			// ### Demo Plugin - an Isadora Bitmap Buffer
	
// GPU: MULTI VIDEO INPUT OUTPUT
	CMultiVideoInOut*		mv;

} PluginInfo;

// A handy macro for casting the mActorDataPtr to PluginInfo*
#if __cplusplus
#define	GetPluginInfo_(actorDataPtr)		static_cast<PluginInfo*>((actorDataPtr)->mActorDataPtr);
#else
#define	GetPluginInfo_(actorDataPtr)		(PluginInfo*)((actorDataPtr)->mActorDataPtr);
#endif

// ---------------------------------------------------------------------------------
//	Constants
// ---------------------------------------------------------------------------------
//	Defines various constants used throughout the plugin.

// CONSTANTS FOR GPU VERSION

// ### Set the maximum number of video inputs here
static const size_t	kMaxVideoInputs = 1;		// we have one video input, from our self generated bitmap

// ### Set the maximum number of FBOs you require here.
static const size_t	kMaxFBO			= 1;		// GPU: Max FBOs

// ### GROUP ID
// Define the group under which this plugin will be displayed in the Isadora interface.
// These are defined under "Actor Types" in IsadoraTypes.h

static const OSType	kActorClass 	= kGroupVideo;

// ### PLUGIN IN
// Define the plugin's unique four character identifier. Contact TroikaTronix to
// obtain a unique four character code if you want to ensure that someone else
// has not developed a plugin with the same code. Note that TroikaTronix reserves
// all plugin codes that begin with an underline, an at-sign, and a pound sign
// (e.g., '_', '@', and '#'.)

static const OSType	kActorID		= FOUR_CHAR_CODE('_BTT');

// ### ACTOR NAME
// The name of the actor. This is the name that will be shown in the User Interface.

static const char* kActorName		= "Bitmap to Texture Demo";

// ### PROPERTY DEFINITION STRING
// The property string. This string determines the inputs and outputs for your plugin.
// See the IsadoraCallbacks.h under the heading "PROPERTY DEFINITION STRING" for the
// meaning ofthese codes. (The IsadoraCallbacks.h header can be seen by opening up
// the IzzySDK Framework while in the Files view.)
//
// IMPORTANT: You cannot use spaces in the property name. Instead, use underscores (_)
// where you want to have a space.
//
// Note that each line ends with a carriage return (\r), and that only the last line of
// the bunch ends with a semicolon. This means that what you see below is one long
// null-terminated c-string, with the individual lines separated by carriage returns.

static const char* sPropertyDefinitionString =

// INPUT PROPERTY DEFINITIONS
//	TYPE 	PROPERTY NAME	ID		DATATYPE	DISPLAY FMT			MIN		MAX		INIT VALUE
	"INPROP width			widt	int			number				1		4096	1280\r"
	"INPROP height			hegt	int			number				1		4096	720\r"
	"INPROP red				red		float		number				0		100		100\r"
	"INPROP green			grn		float		number				0		100		0\r"
	"INPROP blue			blu		float		number				0		100		0\r"

// OUTPUT PROPERTY DEFINITIONS
//	TYPE 	 PROPERTY NAME	ID		DATATYPE	DISPLAY FMT			MIN		MAX		INIT VALUE
	"OUTPROP video_out		vout	data		texture				*		*		0\r";	// GPU;

// ### Property Index Constants
// Properties are referenced by a one-based index. The first input property will
// be 1, the second 2, etc. Similarly, the first output property starts at 1.
// You whould have one constant for each input and output property defined in the 
// property definition string.

enum
{
	kInputWidth = 1,
	kInputHeight,
	kInputRed,
	kInputGreen,
	kInputBlue,
	
	kOutputVideo = 1
};


// ---------------------
//	Help String
// ---------------------
// ### Help Strings
//
// The first help string is for the actor in general. This followed by help strings
// for all of the inputs, and then by the help strings for all of the outputs. These
// should be given in the order that they are defined in the Property Definition
// String above.
//
// In all, the total number of help strings should be (num inputs + num outputs + 1)
//
// Note that each string is followed by a comma -- it is a common mistake to forget the
// comma which results in the two strings being concatenated into one.

const char* sHelpStrings[] =
{
	// ACTOR HELP
	"Demonstration of how to convert a ARGB bitmap to a texture that can be sent to a 'vid-gpu' output.",

	// INPUT HELP
	"The width of the bitmap.",
	
	"The height of the bitmap.",
	
	"The amount by which the red in the bitmap.",

	"The amount of green in the bitmap.",

	"The amount of blue in the bitmap.",
		
	// OUTPUT HELP
	"The colorized video output."
};

// ---------------------------------------------------------------------------------
//		 DisposeBitmap
// ---------------------------------------------------------------------------------
// Called when we need to dispose the ARGB bitmap

static void
DisposeBitmap(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo)
{
	PluginInfo* info = GetPluginInfo_(inActorInfo);
	if (info->mBitmap) {
		ReleaseColorImageBuffer_(ip, info->mBitmap);
		info->mBitmap = NULL;
	}
}


// ---------------------------------------------------------------------------------
//		 CreateBitmap
// ---------------------------------------------------------------------------------
// Called when the bitmap needs to be reallocated

static void
CreateBitmap(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo)
{
	PluginInfo* info = GetPluginInfo_(inActorInfo);
	
	DisposeBitmap(ip, inActorInfo);
	
	info->mBitmap = CreateAndRetainColorImageBuffer_(ip, GetActorDefaultPixelFormat_(ip, inActorInfo), info->mSize.width, info->mSize.height, 0xFF000000);
}

// ---------------------------------------------------------------------------------
//		 FillBitmap
// ---------------------------------------------------------------------------------
// Called when the bitmap changes color

static void
FillBitmap(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo)
{
	PluginInfo* info = GetPluginInfo_(inActorInfo);
	
	if (info->mBitmap != NULL) {
	
		uint32_t red = (uint32_t)(info->mRedAmount * 255.0 + 0.5);
		uint32_t green = (uint32_t)(info->mGreenAmount * 255.0 + 0.5);
		uint32_t blue = (uint32_t)(info->mBlueAmount * 255.0 + 0.5);
	
		uint32_t colorRef = 0xFF000000 | (red << 16) | (green << 8) | (blue << 0);
		
		// we'll draw a white cross in the center of the image that is
		// always there, regardless of the RGB colors calculated above.
		// we do this in a very "down and dirty" way by calculating the
		// minimum and maximum horizontal pixel and the minimum and
		// maximum vertical pixel that will be white. We then simply
		// do a cmparison of the horizontal and vertical pixel below
		// do decide where the white should be drawn
		SInt32 white_line_size = info->mBitmap->mWidth / 100;
		if (white_line_size % 2 != 0) {
			white_line_size += 1;
		}
		if (white_line_size < 2) {
			white_line_size = 2;
		}
		SInt32 white_left = (info->mBitmap->mWidth - white_line_size) / 2;
		SInt32 white_right = white_left + white_line_size;
		SInt32 white_top = (info->mBitmap->mHeight - white_line_size) / 2;
		SInt32 white_bottom = white_top + white_line_size;

		// info->mBitmap->mRowBytes will always be a multiple of 4
		// for an ARGB bitmap; to hit every pixel in the bitmap
		// (including those beyond the actual width) we divide
		// the row bytes by four and use that as our horizontal
		// limit
		
		const SInt32 x_max = info->mBitmap->mRowBytes/4;
		const SInt32 y_max = info->mBitmap->mHeight;
		
		uint32_t* pixelPtr = (uint32_t*) info->mBitmap->mBaseAddress;
		for (SInt32 y=0; y<y_max; y++) {

			for (SInt32 x=0; x<x_max; x++) {
				if (x >= white_left && x < white_right) {
					*(pixelPtr++) = 0xFFFFFFFF;
				} else if (y >= white_top && y < white_bottom) {
					*(pixelPtr++) = 0xFFFFFFFF;
				} else {
					*(pixelPtr++) = colorRef;
				}
			}
		}
		
		// whenever we change the bitmap, we need to increment
		// its data change count to ensure those referencing
		// the bitmap know it has changed
		info->mBitmap->mInfo.mDataChangeCount += 1;
	}
}

// ---------------------------------------------------------------------------------
//		 CreateActor
// ---------------------------------------------------------------------------------
// Called once, prior to the first activation of an actor in its Scene. The
// corresponding DisposeActor actor function will not be called until the file
// owning this actor is closed, or the actor is destroyed as a result of being
// cut or deleted.

static void
CreateActor(
	IsadoraParameters*	ip,	
	ActorInfo*			ioActorInfo)		// pointer to this actor's ActorInfo struct - unique to each instance of an actor
{
	// creat the PluginInfo struct - initializing it to all zeroes
	PluginInfo* info = (PluginInfo*) IzzyMallocClear_(ip, sizeof(PluginInfo));
	PluginAssert_(ip, info != NULL);
	
	ioActorInfo->mActorDataPtr = info;
	info->mActorInfoPtr = ioActorInfo;

	// ### allocation and initialization of private member variables
	
}

// ---------------------------------------------------------------------------------
//		 DisposeActor
// ---------------------------------------------------------------------------------
// Called when the file owning this actor is closed, or when the actor is destroyed
// as a result of its being cut or deleted.
//
static void
DisposeActor(
	IsadoraParameters*	ip,
	ActorInfo*			ioActorInfo)		// pointer to this actor's ActorInfo struct - unique to each instance of an actor
{
	PluginInfo* info = GetPluginInfo_(ioActorInfo);
	PluginAssert_(ip, info != NULL);
	
	// ### destruction of private member variables
	
	// destroy the PluginInfo struct allocated with IzzyMallocClear_ the CreateActor function
	PluginAssert_(ip, ioActorInfo->mActorDataPtr != NULL);
	IzzyFree_(ip, ioActorInfo->mActorDataPtr);
}

// ---------------------------------------------------------------------------------
//		 ActivateActor
// ---------------------------------------------------------------------------------
//	Called when the scene that owns this actor is activated or deactivated. The
//	inActivate flag will be true when the scene is activated, false when deactivated.
//
static void
ActivateActor(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,		// pointer to this actor's ActorInfo struct - unique to each instance of an actor
	Boolean				inActivate)			// true when actor is becoming active, false otherwise.
{
	PluginInfo* info = GetPluginInfo_(inActorInfo);
	
	// ------------------------
	// ACTIVATE
	// ------------------------
	
	if (inActivate) {
	
		// Isadora passes various messages to plugins that request them.
		// These include Mouse Moved messages, Key Down/Key Up messages,
		// Video Frame Clock messages, etc. The complete list can be found
		// in the enumeration in MessageReceiverCommon.h
		
		// You ask Isadora for these messages by calling CreateMessageReceiver_
		// with a pointer to your function, and the message types you would
		// like to receive. (These are bitmapped flags, so you can combine as
		// many as you like: kWantKeyDown | kWantKeyDown for instance.)
		
		// Here we request that our ReceiveMessage function is called
		// whenever the Isadora New Video Frame message is sent,
		// which happens periodically, 30 times per second. We set the ref
		// con to our ActorInfo ptr so that we can access that information
		// from ReceiveMessage callback.
		
		// we should not already have a message receiver
		PluginAssert_(ip, info->mMessageReceiver == NULL);
		
		// create a message receiver that will be notified of
		// video frame ticks
		info->mMessageReceiver = CreateActorMessageReceiver_(
			ip,
			inActorInfo,
			ReceiveMessage,
			0,
			kWantVideoFrameTick | kWantStageAboutToDispose,		// GPU: add kWantStageAboutToDispose
			PtrToIzRefCon(inActorInfo));
		
		// GPU: setup Multi Video Input object
		//
		// The CMultiVideoInOut object handles all kinds of necessary chores
		// including managing one or more FBOs, compiling shaders, routing the
		// output video to the actor's output and more. Here, we give three
		// important parameters:
		//
		// inInputCount = the number of video inputs; in this example 1
		// inVideoOutputPropertyIndex1 = the property index for the video output
		//
		// inFBOCount = the number of FBO's you will require. The FBOs still need
		// be be created/updated using the UpdateIzzyFrameBuffer_ command. But
		// you need to tell CMultiVideoInOut the maximum you will require whe
		// the object is created.
		
		info->mv = new CMultiVideoInOut(ip, inActorInfo, kMaxVideoInputs, kOutputVideo, kMaxFBO);
		IzzyDebugAssert_(info->mv != NULL);
		
		// GPU: it is important to call this function to ensure
		// that the CMultiVideoInOut knows what video format to
		// expect.
		info->mv->SetVideoOutputTypeBasedOnProperty(kOutputProperty, kOutputVideo);
		
		// ensure the bitmap is set to NULL when we're activated
		info->mBitmap = NULL;

	
	// ------------------------
	// DEACTIVATE
	// ------------------------
	
	} else {
	
		// GPU: makes sure to clear the video output
		info->mv->InvalidateVideoOutput();
		
		// dispose our message receiver when we are deactivated.
		if (info->mMessageReceiver != NULL) {
			DisposeMessageReceiver_(ip, info->mMessageReceiver);
			info->mMessageReceiver = NULL;
		}
		
		// ### dispose any data that you don't need when
		// you are not active.
	
		// GPU: Multi Video Input
		IzzyDebugAssert_(info->mv != NULL);
		delete info->mv;
		info->mv = NULL;
		
		DisposeBitmap(ip, inActorInfo);
	}
}

// ---------------------------------------------------------------------------------
//		 GetParameterString
// ---------------------------------------------------------------------------------
//	Returns the property definition string. Called when an instance of the actor
//	needs to be instantiated.

static const char*
GetParameterString(
	IsadoraParameters*	/* ip */,
	ActorInfo*			/* inActorInfo */)
{
	return sPropertyDefinitionString;
}

// ---------------------------------------------------------------------------------
//		 GetHelpString
// ---------------------------------------------------------------------------------
//	Returns the help string for a particular property. If you have a fixed number of
//	input and output properties, it is best to use the PropertyTypeAndIndexToHelpIndex_
//	function to determine the correct help string to return.

static void
GetHelpString(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	PropertyType		inPropertyType,			// kPropertyTypeInvalid when requesting help for the actor
												// or kInputProperty or kOutputProperty when requesting help for a specific property
	PropertyIndex		inPropertyIndex1,		// the one-based index of the property (when inPropertyType is not kPropertyTypeInvalid)
	char*				outParamaterString,		// receives the help string
	UInt32				inMaxCharacters)		// size of the outParamaterString buffer
{
	const char* helpstr = NULL;
	
	// The PropertyTypeAndIndexToHelpIndex_ converts the inPropertyType and
	// inPropertyIndex1 parameters to determine the zero-based index into
	// your list of help strings.
	UInt32 index1 = PropertyTypeAndIndexToHelpIndex_(ip, inActorInfo, inPropertyType, inPropertyIndex1);
	
	// get the help string
	helpstr = sHelpStrings[index1];
	
	// copy it to the output string
	strncpy(outParamaterString, helpstr, inMaxCharacters);
}
	
// ---------------------------------------------------------------------------------
//		 HandlePropertyChangeValue	[INTERRUPT SAFE]
// ---------------------------------------------------------------------------------
//	### This function is called whenever one of the input values of an actor changes.
//	The one-based property index of the input is given by inPropertyIndex1.
//	The new value is given by inNewValue, the previous value by inOldValue.
//
static void
HandlePropertyChangeValue(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	PropertyIndex		inPropertyIndex1,			// the one-based index of the property than changed values
	ValuePtr			/* inOldValue */,			// the property's old value
	ValuePtr			inNewValue,					// the property's new value
	Boolean				inInitializing)				// true if the value is being set when an actor is first initalized
{
	PluginInfo* info = GetPluginInfo_(inActorInfo);

	// ### When you add/change/remove properties, you will need to add cases
	// to this switch statement, to process the messages for your
	// input properties
	
	// The value comes to you encapsulated in a Value structure. See 
	// ValueCommon.h for details about the contents of this structure.
	
	switch (inPropertyIndex1) {
	
	// ### the pending size (width & height) of the bitmap; when
	// this differs from mCurrentHeight, a new bitmap will be
	// generated
	case kInputWidth:
		{
			info->mSize.width = inNewValue->u.ivalue;
			info->mSizeChanged = true;
		}
		break;
	
	case kInputHeight:
		{
			info->mSize.height = inNewValue->u.ivalue;
			info->mSizeChanged = true;
		}
		break;
	
	// ### The red, green and blue inputs are all processed in the same way:
	// the floating point value is in the u.fvalue member of the Value struct
	// pointed to by inNewValue. Because the min and max values specfieid in
	// the Property Definition string was 0 to 100, the resulting value here
	// is 0.0 to 1.0.
	
	case kInputRed:
		{
			// set new red amount, as value between 0.0 and 1.0
			info->mRedAmount = inNewValue->u.fvalue / 100.0;
			info->mColorChanged = true;
		}
		break;
	
	case kInputGreen:
		{
			// set new green amount, as value between 0.0 and 1.0
			info->mGreenAmount = inNewValue->u.fvalue / 100.0;
			info->mColorChanged = true;
		}
		break;
	
	case kInputBlue:
		{
			// set new blue amount, as value between 0.0 and 1.0
			info->mBlueAmount = inNewValue->u.fvalue / 100.0;
			info->mColorChanged = true;
		}
		break;
	}
}

// ---------------------------------------------------------------------------------
//		 GetActorDefinedArea
// ---------------------------------------------------------------------------------
//	If the mGetActorDefinedAreaProc in the ActorInfo struct points to this function,
//	it indicates to Isadora that the object would like to draw either an icon or else
//	an graphic representation of its function.
//
//	### This function uses the 'PICT' 0 resource stored with the plugin to draw an icon.
//  You should replace this picture (located in the Plugin Resources.rsrc file) with
//  the icon for your actor.
// 
static ActorPictInfo	gPictInfo = { false, NULL, NULL, 0, 0 };

static Boolean
GetActorDefinedArea(
	IsadoraParameters*			ip,			
	ActorInfo*					inActorInfo,
	SInt16*						outTopAreaWidth,			// returns the width to reserve for the top Actor Defined Area
	SInt16*						outTopAreaMinHeight,		// returns the minimum height of the top area
	SInt16*						outBotAreaHeight,			// returns the width to reserve for the bottom Actor Defined Area
	SInt16*						outBotAreaMinWidth)			// returns the minimum width of the bottom area
{
	if (gPictInfo.mInitialized == false) {
		PrepareActorDefinedAreaPict_(ip, inActorInfo, 0, &gPictInfo);
	}
	
	// place picture in top area
	*outTopAreaWidth = gPictInfo.mWidth;
	*outTopAreaMinHeight = gPictInfo.mHeight;
	
	// don't draw anything in bottom area
	*outBotAreaHeight = 0;
	*outBotAreaMinWidth = 0;
	
	return true;
}

// ---------------------------------------------------------------------------------
//		 DrawActorDefinedArea
// ---------------------------------------------------------------------------------
//	If GetActorDefinedArea is defined, then this function will be called whenever
//	your ActorDefinedArea needs to be drawn.
//
//	Beacuse we are using the PICT 0 resource stored with this plugin, we can use
//	the DrawActorDefinedAreaPict_ supplied by the Isadora callbacks.
//
//  DrawActorDefinedAreaPict_ is Alpha Channel aware, so you can have nice
//	shading if you like.

static void
DrawActorDefinedArea(
	IsadoraParameters*			ip,
	ActorInfo*					inActorInfo,
	void*						/* inDrawingContext */,		// unused at present
	ActorDefinedAreaPart		inActorDefinedAreaPart,		// the part of the actor that needs to be drawn
	ActorAreaDrawFlagsT			/* inAreaDrawFlags */,		// actor draw flags
	Rect*						inADAArea,					// rect enclosing the entire Actor Defined Area
	Rect*						/* inUpdateArea */,			// subset of inADAArea that needs updating
	Boolean						inSelected)					// TRUE if actor is currently selected
{
	if (inActorDefinedAreaPart == kActorDefinedAreaTop
	&& gPictInfo.mInitialized == true) {
		DrawActorDefinedAreaPict_(ip, inActorInfo, inSelected, inADAArea, &gPictInfo);
	}
}
	
// ---------------------------------------------------------------------------------
//		 GetActorInfo
// ---------------------------------------------------------------------------------
//	This is function is called by to get the actor's class and ID, and to get
//	pointers to the all of the plugin functions declared locally.
//
//	All members of the ActorInfo struct pointed to by outActorParams have been
//	set to 0 on entry. You only need set functions defined by your plugin
//	
EXPORT_ void
GetActorInfo(
	void*				/* inParam */,
	ActorInfo*			outActorParams)
{
	// REQUIRED information
	outActorParams->mActorName							= kActorName;
	outActorParams->mClass								= kActorClass;
	outActorParams->mID									= kActorID;
	outActorParams->mCompatibleWithVersion				= kActorCompatilibilityVersion_v1;
	
	// REQUIRED functions
	outActorParams->mGetActorParameterStringProc		= GetParameterString;
	outActorParams->mGetActorHelpStringProc				= GetHelpString;
	outActorParams->mCreateActorProc					= CreateActor;
	outActorParams->mDisposeActorProc					= DisposeActor;
	outActorParams->mActivateActorProc					= ActivateActor;
	outActorParams->mHandlePropertyChangeValueProc		= HandlePropertyChangeValue;
	
	// OPTIONAL FUNCTIONS
	outActorParams->mHandlePropertyChangeTypeProc		= NULL;
	outActorParams->mHandlePropertyConnectProc			= NULL;
	outActorParams->mPropertyValueToStringProc			= NULL;
	outActorParams->mPropertyStringToValueProc			= NULL;
	outActorParams->mGetActorDefinedAreaProc			= GetActorDefinedArea;
	outActorParams->mDrawActorDefinedAreaProc			= DrawActorDefinedArea;
	outActorParams->mMouseTrackInActorDefinedAreaProc	= NULL;
}

// ---------------------------------------------------------------------------------
//		 ReceiveMessage
// ---------------------------------------------------------------------------------
//	Handles incoming video frame changed messages

static void
ReceiveMessage(
	IsadoraParameters*	ip,
	MessageMask			inMessageMask,
	PortIndex			/* inIndex1 */,
	const MsgData*		inData,
	UInt32				/* inLen */,
	IzRefCon			inRefCon)
{
	ActorInfo* inActorInfo = reinterpret_cast<ActorInfo*>(inRefCon);

	PluginInfo* info = GetPluginInfo_(inActorInfo);

	if ((inMessageMask & kWantStageAboutToDispose) != 0) {
		
		// GPU: important! if the stage is being disposed, we
		// must deallocate all of the openGL resources associated
		// with that stage. It is very important that you call
		// HandleAboutToDisposeStageMessage to ensure those
		// resources are properly disposed!
		
		if (info->mv->HandleAboutToDisposeStageMessage(inData)) {
		}

		// we can't process right now; there's no stage so we return
		return;
	}

	IzzyDebugAssert_(info != NULL && info->mv != NULL);

	// GPU: GetOutputResolutionForImageProcessingMode wil return the required
	// resolution in the CFrameFormatInfo object. The result is calculated
	// after examining the resolution of each video input, and then consulting
	// the the "When Processing Video..." settings in the video tab of the
	// Isadora preferences.
	
	CFrameFormatInfo ffi;
	bool imageInputWasRemapped = false;
	info->mv->GetOutputResolutionForImageProcessingMode(ip, ffi, &imageInputWasRemapped);
	if (imageInputWasRemapped) {
		info->mv->SetNeedsDraw();
	}

	// call EnterVideoProcessing_ so that Isadora will accumulate the
	// amount of time spent processing the video data - this is not
	// requried by highly recommended so that the VPO value in the
	// Status Window stays accurate.
	UInt64 vpStart = EnterVideoProcessing_(ip);
	
	// if the size of the bitmap has changed, we need to
	// reallocate the bitmap; trigger that by disposing
	// the old bitmap which sets info->mBitmap to NULL
	if (info->mSizeChanged) {
		DisposeBitmap(ip, inActorInfo);
		info->mColorChanged = true;
		info->mSizeChanged = false;
	}
	
	// if we don't have a bitmap, then allocate it
	if (info->mBitmap == NULL) {
		CreateBitmap(ip, inActorInfo);
	}
	
	if (info->mBitmap) {
	
		// if the color has changed, then upadte the bitmap's color
		// and call SetNeedsDraw() to ensure the texture output is
		// generated again.
		if (info->mColorChanged) {
			FillBitmap(ip, inActorInfo);
			info->mv->SetNeedsDraw();
			info->mColorChanged = false;
		}
		
		if (info->mv->GetNeedsDraw()) {
		
			info->mv->UpdateTextureOutput(0, info->mSize);
			
			IzzyFrameBuffer fboOut = info->mv->GetOutputFrameBuffer();

			if (fboOut != NULL) {
				
				Value v;
				v.type = kData;
				v.u.data = NULL;
				
				// UpdateIzzyFrameBuffer_ is used to both create and update the the
				// OpenGL Framebuffer into which you'll render or, this case, into
				// which we'll load the ARGB bitmap data/.
				//
				// The first time this function is called, the FBO will be created at
				// the resolution specified by info->mSize. After that, the FBO will
				// remain intact unless info->mSize stays the same. When the size
				// changes, the FBO will be destoryed and rebuilt at the new size.
				//
				// Note that if the flags variable were to change, this could also lead
				// to the FBO being rebuilt, but in this example this parameter never
				// changes.
				//
				// Note also that don't have to use actual power of two sizes; Isadora
				// assumes GL_TEXTURE_2D textures can be allocated even at non-POT sizes.
				//
				// You can optionally specify the clear color used to erase the FBO by
				// providing a pointer to an GLfloat array with the rgba colors. In this
				// example that last paraemter to UpdateIzzyFrameBuffer_() is set to NULL
				
				CreateFBOOptions flags = kCreateFBOOptions_PowerOfTwo;
				if (UpdateIzzyFrameBuffer_(ip, inActorInfo, fboOut, 0, info->mSize.width, info->mSize.height, flags, NULL)) {
					
					// BindIzzyFrameBuffer_ binds the frame buffer so you can draw into it.
					// In this example, we won't actually draw into the FBO using a shader
					// but instead will simply upload the ARGB bitmap into the texdture that
					// backs the FBO. Nevertheless, we need to ensure the OpenGL context is
					// current and BindIzzyFrameBuffer_ does that for us.

					if (BindIzzyFrameBuffer_(ip, inActorInfo, fboOut, TRUE)) {
						
						// get a pointer to the output FBO
						IzzyTexture tx = fboOut->FBO_IzzyTexture();
						
						// bind the backing texture of the FBO
						glBindTexture(tx->GetTextureTarget(), tx->GetTextureName());
						
						// set the row length and pack alignment for the bitmap upload
						glPixelStorei(GL_PACK_ROW_LENGTH, info->mBitmap->mRowBytes / 4);
						glPixelStorei(GL_PACK_ALIGNMENT, 4);
						
						// update the texture to the FBO's backing texture
						#if TARGET_OS_MAC
							glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, info->mSize.width, info->mSize.height, 0, GL_BGRA, GL_UNSIGNED_INT_8_8_8_8_REV, info->mBitmap->mBaseAddress);
						#elif TARGET_OS_WIN
							glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, info->mSize.width, info->mSize.height, 0, GL_BGRA, GL_UNSIGNED_BYTE, info->mBitmap->mBaseAddress);
						#endif
					
						// unbind the texture
						glBindTexture(tx->GetTextureTarget(), 0);
						
						// unbind the FBO which returns a TextureBufferPtr to use -- the
						// TextureBufferPtr contains the vid-gpu texture we want to send
						// to the output
						TextureBufferPtr outputBuffer = UnbindIzzyFrameBuffer_(ip, inActorInfo, fboOut);
						v.u.data = outputBuffer;
					}
				}
				
				// if v contains a valid texture, then send it to the output
				if (v.u.data != NULL) {
					info->mv->SetVideoOutput(&v);
				}

				// clear the needs draw flag until it is set again
				info->mv->ClearNeedsDraw();

			}
			
		}
	
	}
	
	// make sure to compliment EnterVideoProcessing_ with 
	// and ExitVideoProcessing_ call
	ExitVideoProcessing_(ip, vpStart);
}
