// ===========================================================================
//	Isadora Demo Plugin			(c) 2021 Mark F. Coniglio. All rights reserved
// ===========================================================================

/*

# ISADORA DEMO PLUGIN - READ ME

## DESCRIPTION

This plugin demonstrations a simple video effect implemented both as an
OpenGL shader. The shader provided simply modulates the red, green, and blue
components of the source video stream as found in the "Colorizer" actor that
comes built-in with Isdaora. To perform a different effect, you simply need
to replace the shader under the heading "COLORIZER SHADER"

## ABOUT VIDEO TYPES

In Isadora, video data comes in three "flavors":

- **vid-gpu**: Image is stored as an OpenGL texture on your graphics card.
Image data is processed using OpenGL shaders

- **vid-ci**: Image is stored as
a CIImage on your graphics card (macOS only). Image data is processed using
OpenGL shaders

- **vid-cpu**: Image is stored in the main memory of your
computer. image data is processed using standard C++.  format is only used by
legacy Isadora plugins, usually marked by the word classic" in their name.

This plugin only handles 'vid-gpu' streams only, as this is by far the most
common video format used by Isadora. To see a plugin that offers a mutable
input that can receive all three flavors, please take a look at the Analyze
Image Demo project within the Isadora SDK.

## CUSTOMIZING THIS SOURCE CODE

To customize this file, search for the text ###. All of the places where you
will need to customize the file are marked with this pattern of characters.

## LICENSE

IMPORTANT: This source code ("the software") is supplied to you in
consideration of your agreement to the following terms. If you do not agree
to the terms, do not install, use, modify or redistribute the software.

Mark Coniglio (aka TroikaTronix) grants you a personal, non exclusive license
to use, reproduce, modify this software with and to redistribute it, with or
without modifications, in source and/or binary form. Except as expressly
stated in this license, no other rights are granted, express or implied, to
you by TroikaTronix.

This software is provided on an "AS IS" basis. TroikaTronix makes no
warranties, express or implied, including without limitation the implied
warranties of non-infringement, merchantability, and fitness for a
particular purpurse, regarding this software or its use and operation alone
or in combination with your products.

In no event shall TroikaTronix be liable for any special, indirect,
incidental, or consequential damages arising in any way out of the use,
reproduction, modification and/or distribution of this software.

*/

// ---------------------------------------------------------------------------------
// INCLUDES
// ---------------------------------------------------------------------------------

// ISADORA INCLUDES
// These three includes should be present at the top of every Isadora plugin.

#include "IsadoraPluginPrefix.h"
#include "IsadoraTypes.h"
#include "IsadoraCallbacks.h"

// PLUGIN SPECIFIC INCLUDES
// required for plugins that process GPU and CoreImage based image data
#include "MultiVideoInOutUtil.h"
#include "CGLShader.h"

// GPU: useful for plugins that process CPU based image data
#include "PluginDrawUtil.h"

// STANDARD INCLUDES
#include <string.h>
#include <stdio.h>

// ---------------------------------------------------------------------------------
// MacOS Specific
// ---------------------------------------------------------------------------------
#if TARGET_OS_MAC
#define EXPORT_
#endif

#pragma mark ----- COLORIZER SHADER -----

// ---------------------------------------------------------------------------
// COLORIZER SHADER
// ---------------------------------------------------------------------------

// vertex shader
static const GLcharARB* sOneToOne_VertexShader =
"void main(void) {"
    "gl_Position = ftransform();"
    "gl_TexCoord[0] = gl_TextureMatrix[0] * gl_MultiTexCoord0;"
"}";

#define QUOTE(...) #__VA_ARGS__

// fragment shader
static const GLcharARB* sColorizer_FragmentShader = QUOTE(

uniform sampler2D inputTexture;
uniform float m_RedAmount;
uniform float m_GreenAmount;
uniform float m_BlueAmount;
void main()
{
	vec4 color = texture2D(inputTexture, gl_TexCoord[0].st);
	color.r *= m_RedAmount;
	color.g *= m_GreenAmount;
	color.b *= m_BlueAmount;
	color = clamp(color, 0.0, 1.0);
	gl_FragColor = color;
}

);

// ---------------------------------------------------------------------------------
// Win32  Specific
// ---------------------------------------------------------------------------------
#if TARGET_OS_WIN	
	#define EXPORT_ __declspec(dllexport)
#endif

// ---------------------------------------------------------------------------------
//	Exported Function Definitions
// ---------------------------------------------------------------------------------

#ifdef __cplusplus
extern "C" {
#endif

EXPORT_ void GetActorInfo(void* inParam, ActorInfo* outActorParams);

#ifdef __cplusplus
}
#endif

// ---------------------------------------------------------------------------------
//	FORWARD DECLARTIONS
// ---------------------------------------------------------------------------------
static void
ReceiveMessage(
	IsadoraParameters*	ip,
	MessageMask			inMessageMask,
	PortIndex			inPortIndex1,
	const MsgData*		inData,
	UInt32				inLen,
	IzRefCon			inRefCon);

// ---------------------------------------------------------------------------------
// GLOBAL VARIABLES
// ---------------------------------------------------------------------------------
// ### Declare global variables, common to all instantiations of this plugin here

// Example: static int gMyGlobalVariable = 5;

// ---------------------------------------------------------------------------------
// PluginInfo struct
// ---------------------------------------------------------------------------------
// ### This structure neeeds to contain all variables used by your plugin. Memory for
// this struct is allocated during the CreateActor function, and disposed during
// the DisposeActor function, and is private to each copy of the plugin.
//
// If your plugin needs global data, declare them as static variables within this
// file. Any static variable will be global to all instantiations of the plugin.

typedef struct {

	ActorInfo*				mActorInfoPtr;		// our ActorInfo Pointer - set during create actor function
	MessageReceiverRef		mMessageReceiver;	// pointer to our message receiver reference
	
	float					mRedAmount;			// ### Demo Plugin - value = -1.0 to +1.0
	float					mGreenAmount;		// ### Demo Plugin - value = -1.0 to +1.0
	float					mBlueAmount;		// ### Demo Plugin - value = -1.0 to +1.0
	Boolean					mBypass;

// SHADER UNIFORIM IDs
	size_t					mRedIndex0;
	size_t					mGreenIndex0;
	size_t					mBlueIndex0;

// GPU: MULTI VIDEO INPUT OUTPUT
	CMultiVideoInOut*		mv;

} PluginInfo;

// A handy macro for casting the mActorDataPtr to PluginInfo*
#if __cplusplus
#define	GetPluginInfo_(actorDataPtr)		static_cast<PluginInfo*>((actorDataPtr)->mActorDataPtr);
#else
#define	GetPluginInfo_(actorDataPtr)		(PluginInfo*)((actorDataPtr)->mActorDataPtr);
#endif

// ---------------------------------------------------------------------------------
//	Constants
// ---------------------------------------------------------------------------------
//	Defines various constants used throughout the plugin.

// CONSTANTS FOR GPU VERSION

// ### Set the maximum number of video inputs here
static const size_t	kMaxVideoInputs = 1;		// GPU: Max video inputs

// ### Set the maximum number of FBOs you require here.
static const size_t	kMaxFBO			= 1;		// GPU: Max FBOs

// ### GROUP ID
// Define the group under which this plugin will be displayed in the Isadora interface.
// These are defined under "Actor Types" in IsadoraTypes.h

static const OSType	kActorClass 	= kGroupVideo;

// ### PLUGIN IN
// Define the plugin's unique four character identifier. Contact TroikaTronix to
// obtain a unique four character code if you want to ensure that someone else
// has not developed a plugin with the same code. Note that TroikaTronix reserves
// all plugin codes that begin with an underline, an at-sign, and a pound sign
// (e.g., '_', '@', and '#'.)

static const OSType	kActorID		= FOUR_CHAR_CODE('_DGP');

// ### ACTOR NAME
// The name of the actor. This is the name that will be shown in the User Interface.

static const char* kActorName		= "IsadoraDemoPlugin";

// ### PROPERTY DEFINITION STRING
// The property string. This string determines the inputs and outputs for your plugin.
// See the IsadoraCallbacks.h under the heading "PROPERTY DEFINITION STRING" for the
// meaning ofthese codes. (The IsadoraCallbacks.h header can be seen by opening up
// the IzzySDK Framework while in the Files view.)
//
// IMPORTANT: You cannot use spaces in the property name. Instead, use underscores (_)
// where you want to have a space.
//
// Note that each line ends with a carriage return (\r), and that only the last line of
// the bunch ends with a semicolon. This means that what you see below is one long
// null-terminated c-string, with the individual lines separated by carriage returns.

static const char* sPropertyDefinitionString =

// INPUT PROPERTY DEFINITIONS
//	TYPE 	PROPERTY NAME	ID		DATATYPE	DISPLAY FMT			MIN		MAX		INIT VALUE
	"INPROP video_in		vin		data		texture				*		*		0\r"	// GPU
	"INPROP red				red		float		number				-100	100		0\r"
	"INPROP green			grn		float		number				-100	100		0\r"
	"INPROP blue			blu		float		number				-100	100		0\r"
	"INPROP	bypass			byps	bool		onoff				0		1		0\r"

// OUTPUT PROPERTY DEFINITIONS
//	TYPE 	 PROPERTY NAME	ID		DATATYPE	DISPLAY FMT			MIN		MAX		INIT VALUE
	"OUTPROP video_out		vout	data		texture				*		*		0\r";	// GPU;

// ### Property Index Constants
// Properties are referenced by a one-based index. The first input property will
// be 1, the second 2, etc. Similarly, the first output property starts at 1.
// You whould have one constant for each input and output property defined in the 
// property definition string.

enum
{
	kInputVideoIn = 1,
	kInputRed,
	kInputGreen,
	kInputBlue,
	kInputBypass,
	
	kOutputVideo = 1
};


// ---------------------
//	Help String
// ---------------------
// ### Help Strings
//
// The first help string is for the actor in general. This followed by help strings
// for all of the inputs, and then by the help strings for all of the outputs. These
// should be given in the order that they are defined in the Property Definition
// String above.
//
// In all, the total number of help strings should be (num inputs + num outputs + 1)
//
// Note that each string is followed by a comma -- it is a common mistake to forget the
// comma which results in the two strings being concatenated into one.

const char* sHelpStrings[] =
{
	// ACTOR HELP
	"Changes the level of the red, green and blue components of a video image.",

	// INPUT HELP
	"The video source that will be colorized.",
	
	"The amount by which the red component of the image will be changed. Values greater than zero "
		"increase the amount of red, values less than zero decrease it.",

	"The amount by which the green component of the image will be changed. Values greater than zero "
		"increase the amount of green, values less than zero decrease it.",

	"The amount by which the blue component of the image will be changed. Values greater than zero "
		"increase the amount of blue, values less than zero decrease it.",
		
	"Bypasses this effect when set to 'on'. When 'off' the effect functions normally.",
	
	// OUTPUT HELP
	"The colorized video output."
};

// ---------------------------------------------------------------------------------
//		 CreateActor
// ---------------------------------------------------------------------------------
// Called once, prior to the first activation of an actor in its Scene. The
// corresponding DisposeActor actor function will not be called until the file
// owning this actor is closed, or the actor is destroyed as a result of being
// cut or deleted.

static void
CreateActor(
	IsadoraParameters*	ip,	
	ActorInfo*			ioActorInfo)		// pointer to this actor's ActorInfo struct - unique to each instance of an actor
{
	// creat the PluginInfo struct - initializing it to all zeroes
	PluginInfo* info = (PluginInfo*) IzzyMallocClear_(ip, sizeof(PluginInfo));
	PluginAssert_(ip, info != NULL);
	
	ioActorInfo->mActorDataPtr = info;
	info->mActorInfoPtr = ioActorInfo;

	// ### allocation and initialization of private member variables
	
}

// ---------------------------------------------------------------------------------
//		 DisposeActor
// ---------------------------------------------------------------------------------
// Called when the file owning this actor is closed, or when the actor is destroyed
// as a result of its being cut or deleted.
//
static void
DisposeActor(
	IsadoraParameters*	ip,
	ActorInfo*			ioActorInfo)		// pointer to this actor's ActorInfo struct - unique to each instance of an actor
{
	PluginInfo* info = GetPluginInfo_(ioActorInfo);
	PluginAssert_(ip, info != NULL);
	
	// ### destruction of private member variables
	
	// destroy the PluginInfo struct allocated with IzzyMallocClear_ the CreateActor function
	PluginAssert_(ip, ioActorInfo->mActorDataPtr != NULL);
	IzzyFree_(ip, ioActorInfo->mActorDataPtr);
}

// ---------------------------------------------------------------------------------
//		 ActivateActor
// ---------------------------------------------------------------------------------
//	Called when the scene that owns this actor is activated or deactivated. The
//	inActivate flag will be true when the scene is activated, false when deactivated.
//
static void
ActivateActor(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,		// pointer to this actor's ActorInfo struct - unique to each instance of an actor
	Boolean				inActivate)			// true when actor is becoming active, false otherwise.
{
	PluginInfo* info = GetPluginInfo_(inActorInfo);
	
	// ------------------------
	// ACTIVATE
	// ------------------------
	
	if (inActivate) {
	
		// Isadora passes various messages to plugins that request them.
		// These include Mouse Moved messages, Key Down/Key Up messages,
		// Video Frame Clock messages, etc. The complete list can be found
		// in the enumeration in MessageReceiverCommon.h
		
		// You ask Isadora for these messages by calling CreateMessageReceiver_
		// with a pointer to your function, and the message types you would
		// like to receive. (These are bitmapped flags, so you can combine as
		// many as you like: kWantKeyDown | kWantKeyDown for instance.)
		
		// Here we request that our ReceiveMessage function is called
		// whenever the Isadora New Video Frame message is sent,
		// which happens periodically, 30 times per second. We set the ref
		// con to our ActorInfo ptr so that we can access that information
		// from ReceiveMessage callback.
		
		// we should not already have a message receiver
		PluginAssert_(ip, info->mMessageReceiver == NULL);
		
		// create a message receiver that will be notified of
		// video frame ticks
		info->mMessageReceiver = CreateActorMessageReceiver_(
			ip,
			inActorInfo,
			ReceiveMessage,
			0,
			kWantVideoFrameTick | kWantStageAboutToDispose,		// GPU: add kWantStageAboutToDispose
			PtrToIzRefCon(inActorInfo));
		
		// GPU: setup Multi Video Input object
		//
		// The CMultiVideoInOut object handles all kinds of necessary chores
		// including managing one or more FBOs, compiling shaders, routing the
		// output video to the actor's output and more. Here, we give three
		// important parameters:
		//
		// inInputCount = the number of video inputs; in this example 1
		// inVideoOutputPropertyIndex1 = the property index for the video output
		//
		// inFBOCount = the number of FBO's you will require. The FBOs still need
		// be be created/updated using the UpdateIzzyFrameBuffer_ command. But
		// you need to tell CMultiVideoInOut the maximum you will require whe
		// the object is created.
		
		info->mv = new CMultiVideoInOut(ip, inActorInfo, kMaxVideoInputs, kOutputVideo, kMaxFBO);
		IzzyDebugAssert_(info->mv != NULL);
		
		// GPU: it is important to call this function to ensure
		// that the CMultiVideoInOut knows what video format to
		// expect.
		info->mv->SetVideoOutputTypeBasedOnProperty(kOutputProperty, kOutputVideo);

	
	// ------------------------
	// DEACTIVATE
	// ------------------------
	
	} else {
	
		// GPU: makes sure to clear the video output
		info->mv->InvalidateVideoOutput();
		
		// dispose our message receiver when we are deactivated.
		if (info->mMessageReceiver != NULL) {
			DisposeMessageReceiver_(ip, info->mMessageReceiver);
			info->mMessageReceiver = NULL;
		}
		
		// ### dispose any data that you don't need when
		// you are not active.
	
		// GPU: Multi Video Input
		IzzyDebugAssert_(info->mv != NULL);
		delete info->mv;
		info->mv = NULL;
	}
}

// ---------------------------------------------------------------------------------
//		 GetParameterString
// ---------------------------------------------------------------------------------
//	Returns the property definition string. Called when an instance of the actor
//	needs to be instantiated.

static const char*
GetParameterString(
	IsadoraParameters*	/* ip */,
	ActorInfo*			/* inActorInfo */)
{
	return sPropertyDefinitionString;
}

// ---------------------------------------------------------------------------------
//		 GetHelpString
// ---------------------------------------------------------------------------------
//	Returns the help string for a particular property. If you have a fixed number of
//	input and output properties, it is best to use the PropertyTypeAndIndexToHelpIndex_
//	function to determine the correct help string to return.

static void
GetHelpString(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	PropertyType		inPropertyType,			// kPropertyTypeInvalid when requesting help for the actor
												// or kInputProperty or kOutputProperty when requesting help for a specific property
	PropertyIndex		inPropertyIndex1,		// the one-based index of the property (when inPropertyType is not kPropertyTypeInvalid)
	char*				outParamaterString,		// receives the help string
	UInt32				inMaxCharacters)		// size of the outParamaterString buffer
{
	const char* helpstr = NULL;
	
	// The PropertyTypeAndIndexToHelpIndex_ converts the inPropertyType and
	// inPropertyIndex1 parameters to determine the zero-based index into
	// your list of help strings.
	UInt32 index1 = PropertyTypeAndIndexToHelpIndex_(ip, inActorInfo, inPropertyType, inPropertyIndex1);
	
	// get the help string
	helpstr = sHelpStrings[index1];
	
	// copy it to the output string
	strncpy(outParamaterString, helpstr, inMaxCharacters);
}
	
// ---------------------------------------------------------------------------------
//		 HandlePropertyChangeValue	[INTERRUPT SAFE]
// ---------------------------------------------------------------------------------
//	### This function is called whenever one of the input values of an actor changes.
//	The one-based property index of the input is given by inPropertyIndex1.
//	The new value is given by inNewValue, the previous value by inOldValue.
//
static void
HandlePropertyChangeValue(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	PropertyIndex		inPropertyIndex1,			// the one-based index of the property than changed values
	ValuePtr			/* inOldValue */,			// the property's old value
	ValuePtr			inNewValue,					// the property's new value
	Boolean				inInitializing)				// true if the value is being set when an actor is first initalized
{
	PluginInfo* info = GetPluginInfo_(inActorInfo);

	// ### When you add/change/remove properties, you will need to add cases
	// to this switch statement, to process the messages for your
	// input properties
	
	// The value comes to you encapsulated in a Value structure. See 
	// ValueCommon.h for details about the contents of this structure.
	
	switch (inPropertyIndex1) {
	case kInputVideoIn:
		{
			// if bypass is off, then we store the incoming video frame reference
			// into our image buffer -- it will be processed when our ReceiveMessage
			// function next receives a kWantVideoFrameTick message
			if (!info->mBypass) {									// GPU
				info->mv->SetVideoInputFromValue(0, inNewValue);	// GPU
				info->mv->SetNeedsDraw();							// GPU
			
			// if bypass is on, we simply send the incoming video frame reference
			// on to the output -- bypassing all processing entirely.
			} else {
				info->mv->SetVideoOutput(inNewValue);
			}
		}
		break;
	
	// ### The red, green and blue inputs are all processed in the same way:
	// the floating point value is in the u.fvalue member of the Value struct
	// pointed to by inNewValue. Because the min and max values specfieid in
	// the Property Definition string was -100 to +100, the resulting value
	// here is -1.0 to +1.0.
	
	case kInputRed:
		{
			// set new red amount, as value between 0.0 and 1.0
			info->mRedAmount = inNewValue->u.fvalue / 100.0;
			// set mNeedsDraw flag to ensure that new video image is drawn
			info->mv->SetNeedsDraw();
		}
		break;
	
	case kInputGreen:
		{
			// set new green amount, as value between 0.0 and 1.0
			info->mGreenAmount = inNewValue->u.fvalue / 100.0;
			// set mNeedsDraw flag to ensure that new video image is drawn
			info->mv->SetNeedsDraw();
		}
		break;
	
	case kInputBlue:
		{
			// set new blue amount, as value between 0.0 and 1.0
			info->mBlueAmount = inNewValue->u.fvalue / 100.0;
			// set mNeedsDraw flag to ensure that new video image is drawn
			info->mv->SetNeedsDraw();
		}
		break;
	
	// When the bypass input property is set to 0 (off = default) this
	// actor processes the video normally. When it is set to 1 (on) all
	// processing is bypassed and the input video is passed directly
	// to the output.
	
	case kInputBypass:
		{
			// store member variable for on/off
			info->mBypass = (inNewValue->u.ivalue != 0);
			
			// if "bypass" is going from on to off, we need to
			// update our message receiver so that we will
			// start receiving video frame tick messages again
			if (info->mBypass == false) {
				UpdateActorMessageReceiver_(
					ip,
					info->mMessageReceiver,
					inActorInfo,
					ReceiveMessage,
					0,
					kWantVideoFrameTick | kWantStageAboutToDispose,
					PtrToIzRefCon(inActorInfo));
				info->mv->SetNeedsDraw();
			
			} else {
				// if "bypass" is going from off to on, then we want to
				// stop processing video. We update our message receiver
				// to stop receiving video ticks here to save processing
				// power -- when bypass is "on" the incoming video is
				// sent directly to the output -- see the ase kInputVideoIn
				// above.
				UpdateActorMessageReceiver_(
					ip,
					info->mMessageReceiver,
					inActorInfo,
					ReceiveMessage,
					0,
					kWantStageAboutToDispose,
					PtrToIzRefCon(inActorInfo));
				
				if (inInitializing == false) {
					ValuePtr v = GetInputPropertyValue_(ip, inActorInfo, kInputVideoIn);
					info->mv->SetVideoOutput(v);
				}
			}
		}
		break;
	}
}

// ---------------------------------------------------------------------------------
//		 GetActorDefinedArea
// ---------------------------------------------------------------------------------
//	If the mGetActorDefinedAreaProc in the ActorInfo struct points to this function,
//	it indicates to Isadora that the object would like to draw either an icon or else
//	an graphic representation of its function.
//
//	### This function uses the 'PICT' 0 resource stored with the plugin to draw an icon.
//  You should replace this picture (located in the Plugin Resources.rsrc file) with
//  the icon for your actor.
// 
static ActorPictInfo	gPictInfo = { false, NULL, NULL, 0, 0 };

static Boolean
GetActorDefinedArea(
	IsadoraParameters*			ip,			
	ActorInfo*					inActorInfo,
	SInt16*						outTopAreaWidth,			// returns the width to reserve for the top Actor Defined Area
	SInt16*						outTopAreaMinHeight,		// returns the minimum height of the top area
	SInt16*						outBotAreaHeight,			// returns the width to reserve for the bottom Actor Defined Area
	SInt16*						outBotAreaMinWidth)			// returns the minimum width of the bottom area
{
	if (gPictInfo.mInitialized == false) {
		PrepareActorDefinedAreaPict_(ip, inActorInfo, 0, &gPictInfo);
	}
	
	// place picture in top area
	*outTopAreaWidth = gPictInfo.mWidth;
	*outTopAreaMinHeight = gPictInfo.mHeight;
	
	// don't draw anything in bottom area
	*outBotAreaHeight = 0;
	*outBotAreaMinWidth = 0;
	
	return true;
}

// ---------------------------------------------------------------------------------
//		 DrawActorDefinedArea
// ---------------------------------------------------------------------------------
//	If GetActorDefinedArea is defined, then this function will be called whenever
//	your ActorDefinedArea needs to be drawn.
//
//	Beacuse we are using the PICT 0 resource stored with this plugin, we can use
//	the DrawActorDefinedAreaPict_ supplied by the Isadora callbacks.
//
//  DrawActorDefinedAreaPict_ is Alpha Channel aware, so you can have nice
//	shading if you like.

static void
DrawActorDefinedArea(
	IsadoraParameters*			ip,
	ActorInfo*					inActorInfo,
	void*						/* inDrawingContext */,		// unused at present
	ActorDefinedAreaPart		inActorDefinedAreaPart,		// the part of the actor that needs to be drawn
	ActorAreaDrawFlagsT			/* inAreaDrawFlags */,		// actor draw flags
	Rect*						inADAArea,					// rect enclosing the entire Actor Defined Area
	Rect*						/* inUpdateArea */,			// subset of inADAArea that needs updating
	Boolean						inSelected)					// TRUE if actor is currently selected
{
	if (inActorDefinedAreaPart == kActorDefinedAreaTop
	&& gPictInfo.mInitialized == true) {
		DrawActorDefinedAreaPict_(ip, inActorInfo, inSelected, inADAArea, &gPictInfo);
	}
}
	
// ---------------------------------------------------------------------------------
//		 GetActorInfo
// ---------------------------------------------------------------------------------
//	This is function is called by to get the actor's class and ID, and to get
//	pointers to the all of the plugin functions declared locally.
//
//	All members of the ActorInfo struct pointed to by outActorParams have been
//	set to 0 on entry. You only need set functions defined by your plugin
//	
EXPORT_ void
GetActorInfo(
	void*				/* inParam */,
	ActorInfo*			outActorParams)
{
	// REQUIRED information
	outActorParams->mActorName							= kActorName;
	outActorParams->mClass								= kActorClass;
	outActorParams->mID									= kActorID;
	outActorParams->mCompatibleWithVersion				= kActorCompatilibilityVersion_v1;
	
	// REQUIRED functions
	outActorParams->mGetActorParameterStringProc		= GetParameterString;
	outActorParams->mGetActorHelpStringProc				= GetHelpString;
	outActorParams->mCreateActorProc					= CreateActor;
	outActorParams->mDisposeActorProc					= DisposeActor;
	outActorParams->mActivateActorProc					= ActivateActor;
	outActorParams->mHandlePropertyChangeValueProc		= HandlePropertyChangeValue;
	
	// OPTIONAL FUNCTIONS
	outActorParams->mHandlePropertyChangeTypeProc		= NULL;
	outActorParams->mHandlePropertyConnectProc			= NULL;
	outActorParams->mPropertyValueToStringProc			= NULL;
	outActorParams->mPropertyStringToValueProc			= NULL;
	outActorParams->mGetActorDefinedAreaProc			= GetActorDefinedArea;
	outActorParams->mDrawActorDefinedAreaProc			= DrawActorDefinedArea;
	outActorParams->mMouseTrackInActorDefinedAreaProc	= NULL;
}

// ---------------------------------------------------------------------------------
//		DrawUnityQuad DEFINES
// ---------------------------------------------------------------------------------
/*
#define		TL			0
#define		TR			1
#define		BL			2
#define		BR			3

#define		VX_X		0
#define		VX_Y		1
#define		TX_X		2
#define		TX_Y		3

void
DrawUnityQuad_Textured(
	const float			inTxCoordRight,
	const float			inTxCoordBottom)
{
	static GLfloat sCoordinates[4][4] = {
		// vertex x/y, texture x/y
		{ -1.0f, -1.0, +0.0f, +0.0f },	// upper left
		{ +1.0f, -1.0, +1.0f, +0.0f },	// upper right
		{ -1.0f, +1.0, +0.0f, +1.0f },	// lower left
		{ +1.0f, +1.0, +1.0f, +1.0f }	// lower right
	};

	sCoordinates[TR][TX_X] = sCoordinates[BR][TX_X] = inTxCoordRight;
	sCoordinates[BL][TX_Y] = sCoordinates[BR][TX_Y] = inTxCoordBottom;
	
	glEnableClientState(GL_VERTEX_ARRAY);
	glVertexPointer(2, GL_FLOAT, 4 * sizeof(GLfloat), &sCoordinates[0][0]);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glTexCoordPointer(2, GL_FLOAT, 4 * sizeof(GLfloat), &sCoordinates[0][2]);
	
	glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);

	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glDisableClientState(GL_VERTEX_ARRAY);
}
*/

// ---------------------------------------------------------------------------------
//		ProcessTextureImages
// ---------------------------------------------------------------------------------
static void
ProcessTextureImages(
	IsadoraParameters*		ip,
	ActorInfo*				inActorInfo,
	const CFrameFormatInfo&	inFFI)
{
	PluginInfo*	info = (PluginInfo*) inActorInfo->mActorDataPtr;

	TextureBufferPtr tb = info->mv->mValidInputArray[0]->GetTextureImageInput();
	
	if (tb != NULL && tb->mTexture != NULL && tb->mTargetStageIndex0 == inFFI.mTargetStageIndex0) {
		
		IzzyTexture tx = tb->mTexture;
					
		info->mv->UpdateTextureOutput(inFFI.mTargetStageIndex0, inFFI.mFrameSize);

		IzzyFrameBuffer fboOut = info->mv->GetOutputFrameBuffer();
		
		#define __ctx__accessor	(fboOut->mFBOGLFunctions)
		
		if (fboOut != NULL) {
			
			Value v;
			v.type = kData;
			v.u.data = NULL;
			
			// UpdateIzzyFrameBuffer_ is used to both create and update the
			// the OpenGL Framebuffer into which you'll render.
			//
			// The first time this function is called, the FBO will be created
			// at the resolution specified by inFFI.mFrameSize. After that, as
			// long as the size and the Force Power of Two (POT) flag do not change.
			// (You don't have to use actual power of two sizes; Isadora assumes
			// that GL_TEXTURE_2D textures can be allocated even at non-POT sizes.
			//
			// After that, the FBO will only be disposed and recreated if the size
			// or the POT flag change.
			//
			// You can optionally specified the clear color used to erase the FBO by
			// providing a pointer to an GLfloat array with the rgba colors.
			
			if (UpdateIzzyFrameBuffer_(ip, inActorInfo, fboOut, inFFI.mTargetStageIndex0, inFFI.mFrameSize.width, inFFI.mFrameSize.height, TRUE, NULL)) {
				
				// BindIzzyFrameBuffer_ binds the frame buffer so you can draw into it.
				// It does all the necessary preparations, including making the FBO's
				// context current, setting up the viewport as a standard -1 to +1 sized
				// area, etc. From that point on, all you need to do is start drawing.

				if (BindIzzyFrameBuffer_(ip, inActorInfo, fboOut, TRUE)) {
				
					// if the CMultiVideoInOut does not yet have a shader, we will want
					// to initialize it with our shader code here.
					
					if (!info->mv->HasShader()) {
						// create and compile the the shader using the specified vertex and fragment shader
						info->mv->CreateShader(sOneToOne_VertexShader, sColorizer_FragmentShader);
						// in debug mode, we show an error if the shader compilation failes
						IzzyDebugAssert_(info->mv->HasShader() == true);
						// now get the indexes to the uniform inputs
						info->mRedIndex0 = info->mv->GetShader()->AttachUniform("m_RedAmount", 1);
						info->mGreenIndex0 = info->mv->GetShader()->AttachUniform("m_GreenAmount", 1);
						info->mBlueIndex0 = info->mv->GetShader()->AttachUniform("m_BlueAmount", 1);
					}

					// even though the texture is visible size is almost always
					// the same as the GPU size, we make the proper calculation
					// here for tx_right and tx_bottom. the top and left are
					// both assumed to be 0.0.
					
					const float tx_right = float(tx->GetTextureVisibleWidth()) / float(tx->GetTextureGPUWidth());
					const float tx_bottom = float(tx->GetTextureVisibleHeight()) / float(tx->GetTextureGPUHeight());
					
					// setup to draw
					
					glDisable(GL_BLEND);
					glBlendFunc(GL_ONE, GL_ZERO);
					
					glEnable(GL_TEXTURE_2D);
					glBindTexture(GL_TEXTURE_2D, tx->GetTextureName());

					// enable the shader
					info->mv->EnableShader();
					
					// set the uniform values
					float amt[3];
					amt[0] = info->mRedAmount + 1.0f;
					amt[1] = info->mGreenAmount + 1.0f;
					amt[2] = info->mBlueAmount + 1.0f;
					
					info->mv->GetShader()->SetUniformValue(info->mRedIndex0, &amt[0]);
					info->mv->GetShader()->SetUniformValue(info->mGreenIndex0, &amt[1]);
					info->mv->GetShader()->SetUniformValue(info->mBlueIndex0, &amt[2]);
					
					// use the built in drawing function to render the full quad...
					// or do your own drawing here...
					
					// (FYI: the source for DrawUnityQuad_Textured can be seen above)
					
					CMultiVideoInOut::DrawUnityQuad_Textured(tx_right, tx_bottom);
					
					info->mv->DisableShader();
					
					// unbind and disable the texture map
					glBindTexture(GL_TEXTURE_2D, 0);
					glDisable(GL_TEXTURE_2D);
					
					// UnbindIzzyFrameBuffer_ unbinds the framebuffer and returns
					// an Isadora TextureBufferPtr, which contains a reference
					// to the texture. As shown below, you can simply pass this on
					// to the output
					TextureBufferPtr outputBuffer = UnbindIzzyFrameBuffer_(ip, inActorInfo, fboOut);
					v.u.data = outputBuffer;
				}
			}
			
			if (v.u.data != NULL) {
				info->mv->SetVideoOutput(&v);
			}
			
		} else {
			info->mv->InvalidateVideoOutput();
		}
		
	} else {
		info->mv->InvalidateVideoOutput();
	}
}

// ---------------------------------------------------------------------------------
//		 ProcessVideoFrame
// ---------------------------------------------------------------------------------
//	### This is the code that does the actual processing of a video frame. Modify
// this code to create your own filter.
//

static void
ProcessVideoFrame(
	IsadoraParameters*	/* ip */,	// not used in this function, but needed to call PluginAssert_
	PluginInfo*			info,
	ImageBufferPtr		srcBuf,
	ImageBufferPtr		outBuf)
{
	UInt32* srcData = static_cast<UInt32*>(srcBuf->mBaseAddress);
	UInt32 srcStride = srcBuf->mRowBytes - srcBuf->mWidth * sizeof(UInt32);

	UInt32* outData = static_cast<UInt32*>(outBuf->mBaseAddress);
	UInt32 outStride = outBuf->mRowBytes - outBuf->mWidth * sizeof(UInt32);
	
	// ### this is where your video processing code would go
	// here, we are increasing or decreasing the red, green,
	// and blue components of the video image.
	SInt16 redFactor = static_cast<SInt16>(256.0 * info->mRedAmount);
	SInt16 greenFactor = static_cast<SInt16>(256.0 * info->mGreenAmount);
	SInt16 blueFactor = static_cast<SInt16>(256.0 * info->mBlueAmount);
	
	// for each row
	SInt16 row = 0;
	while (row < outBuf->mHeight) {
		
		// and for each column in that row
		SInt16 col = 0;
		while (col < outBuf->mWidth) {
		
			// IMRORTANT: For Mac/Windows cross-platform compatbility,
			// make sure to use the RED_, GREEN_ and BLUE_ macros to
			// extract pixels from a the raw data, and use RGB_ to 
			// combine them back.
			//
			// On the Mac, the pixels are arranged 00RRGGBB with
			// the blue in the low 8 bits.
			// Under Quicktime for Windows, the data is arranged
			// BBGGRR00 - in reverse order. Using the macros is
			// a easy way to ensure your code will operate on 
			// both platforms.
			//
			
			SInt16 red = RED_(*srcData);
			SInt16 green = GREEN_(*srcData);
			SInt16 blue = BLUE_(*srcData);
			
			// adjust the colors
			red += (red * redFactor) / 256;
			if (red < 0)
				red = 0;
			else if (red > 255)
				red = 255;
			
			green += (green * greenFactor) / 256;
			if (green < 0)
				green = 0;
			else if (green > 255)
				green = 255;
			
			blue += (blue * blueFactor) / 256;
			if (blue < 0)
				blue = 0;
			else if (blue > 255)
				blue = 255;

			*(outData) = ARGB_(255, red, green, blue);
			
			// increment src and out pixel
			srcData++;
			outData++;
			
			// increment column count
			col++;
		}

		// skip to the next line of video, using the
		// stride values computed above
		srcData = (UInt32*)((char*) srcData + srcStride);
		outData = (UInt32*)((char*) outData + outStride);
		
		// increment row count
		row++;
	}
}

// ---------------------------------------------------------------------------------
//		 ReceiveMessage
// ---------------------------------------------------------------------------------
//	Handles incoming video frame changed messages

static void
ReceiveMessage(
	IsadoraParameters*	ip,
	MessageMask			inMessageMask,
	PortIndex			/* inIndex1 */,
	const MsgData*		inData,
	UInt32				/* inLen */,
	IzRefCon			inRefCon)
{
	ActorInfo* inActorInfo = reinterpret_cast<ActorInfo*>(inRefCon);

	PluginInfo* info = GetPluginInfo_(inActorInfo);

	if ((inMessageMask & kWantStageAboutToDispose) != 0) {
		
		// GPU: important! if the stage is being disposed, we
		// must deallocate all of the openGL resources associated
		// with that stage. It is very important that you call
		// HandleAboutToDisposeStageMessage to ensure those
		// resources are properly disposed!
		
		if (info->mv->HandleAboutToDisposeStageMessage(inData)) {
		}

		// we can't process right now; there's no stage so we return
		return;
	}

	IzzyDebugAssert_(info != NULL && info->mv != NULL);

	// GPU: GetOutputResolutionForImageProcessingMode wil return the required
	// resolution in the CFrameFormatInfo object. The result is calculated
	// after examining the resolution of each video input, and then consulting
	// the the "When Processing Video..." settings in the video tab of the
	// Isadora preferences.
	
	CFrameFormatInfo ffi;
	bool imageInputWasRemapped = false;
	info->mv->GetOutputResolutionForImageProcessingMode(ip, ffi, &imageInputWasRemapped);
	if (imageInputWasRemapped) {
		info->mv->SetNeedsDraw();
	}

	// call EnterVideoProcessing_ so that Isadora will accumulate the
	// amount of time spent processing the video data - this is not
	// requried by highly recommended so that the VPO value in the
	// Status Window stays accurate.
	UInt64 vpStart = EnterVideoProcessing_(ip);
		
	if (info->mv->GetNeedsDraw()) {
	
		if (info->mv->GetValidVideoInputCount() > 0 && ffi.mFrameSize.width > 0 && ffi.mFrameSize.height > 0) {
		
			switch (info->mv->GetOutputVideoBufferType()) {
			
			case kVideoBuffer_Texture:
				if (ffi.mTargetStageIndex0 != kInvalidStageIndex) {
					ProcessTextureImages(ip, inActorInfo, ffi);
				} else {
					info->mv->InvalidateVideoOutput();
				}
				break;
			}
		} else {
			info->mv->InvalidateVideoOutput();
		}
		
		info->mv->ClearNeedsDraw();
		
	}
	
	// make sure to compliment EnterVideoProcessing_ with 
	// and ExitVideoProcessing_ call
	ExitVideoProcessing_(ip, vpStart);
}
