/*
Author : Benot Vogel -- BASED ON A 'BLANK' PLUGIN FROM THE ISADORA 1.3 SDK PROJECT WIZARD
Original idea : dbengali
Date : 20/02/2014
*/

// ===========================================================================
//	Isadora Demo Plugin			   2003 Mark F. Coniglio. All rights reserved.
// ===========================================================================
//
//	IMPORTANT: This source code ("the software") is supplied to you in
//	consideration of your agreement to the following terms. If you do not
//	agree to the terms, do not install, use, modify or redistribute the
//	software.
//
//	Mark Coniglio (dba TroikaTronix) grants you a personal, non exclusive
//	license to use, reproduce, modify this software with and to redistribute it,
//	with or without modifications, in source and/or binary form. Except as
//	expressly stated in this license, no other rights are granted, express
//	or implied, to you by TroikaTronix.
//
//	This software is provided on an "AS IS" basis. TroikaTronix makes no
//	warranties, express or implied, including without limitation the implied
//	warranties of non-infringement, merchantability, and fitness for a 
//	particular purpurse, regarding this software or its use and operation
//	alone or in combination with your products.
//
//	In no event shall TroikaTronix be liable for any special, indirect, incidental,
//	or consequential damages arising in any way out of the use, reproduction,
//	modification and/or distribution of this software.
//
// ===========================================================================
//
// CUSTOMIZING THIS SOURCE CODE
// To customize this file, search for the text ###. All of the places where
// you will need to customize the file are marked with this pattern of 
// characters.
//
// ABOUT IMAGE BUFFER MAPS:
//
// The ImageBufferMap structure, and its accompanying functions,
// exists as a convenience to those writing video processing plugins.
//
// Basically, an image buffer contains an arbitrary number of input and
// output buffers (in the form of ImageBuffers). The ImageBufferMap code
// will automatically create intermediary buffers if needed, so that the
// size and depth of the source image buffers sent to your callback are
// the same for all buffers.
// 
// Typically, the ImageBufferMap is created in your CreateActor function,
// and dispose in the DiposeActor function.

// ---------------------------------------------------------------------------------
// INCLUDES
// ---------------------------------------------------------------------------------

#include "IsadoraPluginPrefix.h"

#include "IsadoraTypes.h"
#include "IsadoraCallbacks.h"

// STANDARD INCLUDES
#include <string.h>
#include <stdio.h>

#include "MyIncludeFile.h"

// ---------------------------------------------------------------------------------
// MacOS Specific
// ---------------------------------------------------------------------------------
#if TARGET_OS_MAC
#define EXPORT_
#endif

// ---------------------------------------------------------------------------------
// Win32  Specific
// ---------------------------------------------------------------------------------
#if TARGET_OS_WIN

	#include <windows.h>
	
	#define EXPORT_ __declspec(dllexport)
	
	#ifdef __cplusplus
	extern "C" {
	#endif

	BOOL WINAPI DllMain ( HINSTANCE hInst, DWORD wDataSeg, LPVOID lpvReserved );

	#ifdef __cplusplus
	}
	#endif

	BOOL WINAPI DllMain (
		HINSTANCE	/* hInst */,
		DWORD		wDataSeg,
		LPVOID		/* lpvReserved */)
	{
	switch(wDataSeg) {
	
	case DLL_PROCESS_ATTACH:
		return 1;
		break;
	case DLL_PROCESS_DETACH:
		break;
		
	default:
		return 1;
		break;
	}
	return 0;
	}

#endif

// ---------------------------------------------------------------------------------
//	Exported Function Definitions
// ---------------------------------------------------------------------------------

#ifdef __cplusplus
extern "C" {
#endif

EXPORT_ void GetActorInfo(void* inParam, ActorInfo* outActorParams);

#ifdef __cplusplus
}
#endif

// ---------------------------------------------------------------------------------
//	FORWARD DECLARTIONS
// ---------------------------------------------------------------------------------
static void
ReceiveMessage(
	IsadoraParameters*	ip,
	MessageMask			inMessageMask,
	PortIndex			inPortIndex1,
	const MsgData*		inData,
	UInt32				inLen,
	long				inRefCon);

// ---------------------------------------------------------------------------------
// GLOBAL VARIABLES
// ---------------------------------------------------------------------------------
// ### Declare global variables, common to all instantiations of this plugin here

// Example: static int gMyGlobalVariable = 5;

// ---------------------------------------------------------------------------------
// PluginInfo struct
// ---------------------------------------------------------------------------------
// ### This structure neeeds to contain all variables used by your plugin. Memory for
// this struct is allocated during the CreateActor function, and disposed during
// the DisposeActor function, and is private to each copy of the plugin.
//
// If your plugin needs global data, declare them as static variables within this
// file. Any static variable will be global to all instantiations of the plugin.

typedef struct {

	ActorInfo*				mActorInfoPtr;		// our ActorInfo Pointer - set during create actor function
	
	String*					mName;	// the string to look for in scene names
	IzValue_Int				mFadeOut; // fade out value in milliseconds
	IzValue_Int				mFadeIn; // fade in value in milliseconds
	bool					mDelimited; // use a delimiter char? true/false
	char					mDelimiter;	// delimiter char value

} PluginInfo;

// A handy macro for casting the mActorDataPtr to PluginInfo*
#if __cplusplus
#define	GetPluginInfo_(actorDataPtr)		static_cast<PluginInfo*>((actorDataPtr)->mActorDataPtr);
#else
#define	GetPluginInfo_(actorDataPtr)		(PluginInfo*)((actorDataPtr)->mActorDataPtr);
#endif

// ---------------------------------------------------------------------------------
//	Constants
// ---------------------------------------------------------------------------------
//	Defines various constants used throughout the plugin.

// ### GROUP ID
// Define the group under which this plugin will be displayed in the Isadora interface.
// These are defined under "Actor Types" in IsadoraTypes.h

static const OSType	kActorClass 	= kGroupControl;

// ### PLUGIN IN
// Define the plugin's unique four character identifier. Contact TroikaTronix to
// obtain a unique four character code if you want to ensure that someone else
// has not developed a plugin with the same code. Note that TroikaTronix reserves
// all plugin codes that begin with an underline, an at-sign, and a pound sign
// (e.g., '_', '@', and '#'.)

static const OSType	kActorID		= FOUR_CHAR_CODE('JkPs');

// ### ACTOR NAME
// The name of the actor. This is the name that will be shown in the User Interface.

static const char* kActorName		= "Jump By Name";

// ### PROPERTY DEFINITION STRING
// The property string. This string determines the inputs and outputs for your plugin.
// See the IsadoraCallbacks.h under the heading "PROPERTY DEFINITION STRING" for the
// meaning ofthese codes. (The IsadoraCallbacks.h header can be seen by opening up
// the IzzySDK Framework while in the Files view.)
//
// IMPORTANT: You cannot use spaces in the property name. Instead, use underscores (_)
// where you want to have a space.
//
// Note that each line ends with a carriage return (\r), and that only the last line of
// the bunch ends with a semicolon. This means that what you see below is one long
// null-terminated c-string, with the individual lines separated by carriage returns.

static const char* sPropertyDefinitionString =

// INPUT PROPERTY DEFINITIONS
//	TYPE 	PROPERTY NAME	ID		DATATYPE	DISPLAY FMT			MIN		MAX		INIT VALUE
	"INPROP name			nam		string		text				*		*		\r"
	"INPROP fade_out		fado	int			seconds				0		*		0\r"
	"INPROP fade_in			fadi	int			seconds				0		*		0\r"
	"INPROP delimited		delf	bool		onoff				0		1		1\r"
	"INPROP delimiter		del 	int			char				32		126		58\r" // char(58) = ':'
	"INPROP	trigger			trg		bool		trig				0		1		0\r";

// OUTPUT PROPERTY DEFINITIONS
//	TYPE 	 PROPERTY NAME	ID		DATATYPE	DISPLAY FMT			MIN		MAX		INIT VALUE
	//"OUTPROP video_out		vout	data		video				*		*		0\r";

// ### Property Index Constants
// Properties are referenced by a one-based index. The first input property will
// be 1, the second 2, etc. Similarly, the first output property starts at 1.
// You whould have one constant for each input and output property defined in the 
// property definition string.

enum
{
	kInputName = 1,
	kInputFadeOut,
	kInputFadeIn,
	kInputDelimited,
	kInputDelimiter,
	kInputTrigger//,
	
	//kOutputVideo = 1
};


// ---------------------
//	Help String
// ---------------------
// ### Help Strings
//
// The first help string is for the actor in general. This followed by help strings
// for all of the inputs, and then by the help strings for all of the outputs. These
// should be given in the order that they are defined in the Property Definition
// String above.
//
// In all, the total number of help strings should be (num inputs + num outputs + 1)
//
// Note that each string is followed by a comma -- it is a common mistake to forget the
// comma which results in the two strings being concatenated into one.

const char* sHelpStrings[] =
{
	"Jumps to the first scene whose name begins with the specified text. This actor searches "
		"from the beginning of the scene list and jumps to the first matching scene, "
		"or does not jump if no matching scene is found. A match is defined as the "
		"search text matching the first characters of a scene name.\nStrange behaviours "
		"will occur if this property contains the delimiter character in it.",

	"The text to match when locating a scene to jump to. A match is defined as the "
		"search text matching the first characters of a scene name. If this is true, subsequent characters are ignored.",
	
	"The fade out time of the current scene, just like in the Jump++ actor.",

	"The fade in time of the new scene, just like in the Jump++ actor.",

	"If ON, a match requires the NAME to be found at the beginning of a scene name, "
		"followed by the delimiter character. Subsequent characters are ignored.\r"
		"This is useful when receiving MIDI Show Control, so that, for example, "
		"Cue 10 does not erroneously match a scene labeled 100.",
		
	"The delimiter character to be used in scene names.\r"
	"List of accepted characters:\r"
		"space + !\"#$%&'()*+,-./\r"
		":;<=>?@[\\]^_`{|}~.\r"
		"0123456789\r"
		"ABCDEFGHIJKLMNOPQRSTUVWXYZ\r"
		"abcdefghijklmnopqrstuvwxyz",
	
	"Triggers the jump."
};

// ---------------------------------------------------------------------------------
//		 CreateActor
// ---------------------------------------------------------------------------------
// Called once, prior to the first activation of an actor in its Scene. The
// corresponding DisposeActor actor function will not be called until the file
// owning this actor is closed, or the actor is destroyed as a result of being
// cut or deleted.

static void
CreateActor(
	IsadoraParameters*	ip,	
	ActorInfo*			ioActorInfo)		// pointer to this actor's ActorInfo struct - unique to each instance of an actor
{
	// creat the PluginInfo struct - initializing it to all zeroes
	PluginInfo* info = (PluginInfo*) IzzyMallocClear_(ip, sizeof(PluginInfo));
	PluginAssert_(ip, info != nil);
	
	ioActorInfo->mActorDataPtr = info;
	info->mActorInfoPtr = ioActorInfo;

	// ### allocation and initialization of private member variables
	


}

// ---------------------------------------------------------------------------------
//		 DisposeActor
// ---------------------------------------------------------------------------------
// Called when the file owning this actor is closed, or when the actor is destroyed
// as a result of its being cut or deleted.
//
static void
DisposeActor(
	IsadoraParameters*	ip,
	ActorInfo*			ioActorInfo)		// pointer to this actor's ActorInfo struct - unique to each instance of an actor
{
	PluginInfo* info = GetPluginInfo_(ioActorInfo);
	PluginAssert_(ip, info != nil);
	
	// ### destruction of private member variables
	
	// destroy the PluginInfo struct allocated with IzzyMallocClear_ the CreateActor function
	PluginAssert_(ip, ioActorInfo->mActorDataPtr != nil);
	IzzyFree_(ip, ioActorInfo->mActorDataPtr);
}



// ---------------------------------------------------------------------------------
//		 GetParameterString
// ---------------------------------------------------------------------------------
//	Returns the property definition string. Called when an instance of the actor
//	needs to be instantiated.

static const char*
GetParameterString(
	IsadoraParameters*	/* ip */,
	ActorInfo*			/* inActorInfo */)
{
	return sPropertyDefinitionString;
}

// ---------------------------------------------------------------------------------
//		 GetHelpString
// ---------------------------------------------------------------------------------
//	Returns the help string for a particular property. If you have a fixed number of
//	input and output properties, it is best to use the PropertyTypeAndIndexToHelpIndex_
//	function to determine the correct help string to return.

static void
GetHelpString(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	PropertyType		inPropertyType,			// kPropertyTypeInvalid when requesting help for the actor
												// or kInputProperty or kOutputProperty when requesting help for a specific property
	PropertyIndex		inPropertyIndex1,		// the one-based index of the property (when inPropertyType is not kPropertyTypeInvalid)
	char*				outParamaterString,		// receives the help string
	UInt32				inMaxCharacters)		// size of the outParamaterString buffer
{
	const char* helpstr = nil;
	
	// The PropertyTypeAndIndexToHelpIndex_ converts the inPropertyType and
	// inPropertyIndex1 parameters to determine the zero-based index into
	// your list of help strings.
	UInt32 index1 = PropertyTypeAndIndexToHelpIndex_(ip, inActorInfo, inPropertyType, inPropertyIndex1);
	
	// get the help string
	helpstr = sHelpStrings[index1];
	
	// copy it to the output string
	strncpy(outParamaterString, helpstr, inMaxCharacters);
}







// ---------------------------------------------------------------------------------
//		REMOVECHARINSTRING FUNCTION
// ---------------------------------------------------------------------------------
//	This function removes all occurrences of a char parameter in a char* parameter.
//

static void removeCharInString (char c, char* s)
{
	char* pRead = s;
	char* pWrite = s;

	while(*pRead != '\0') // while not end of string
	{
		if (*pRead != c)
		{
			*pWrite = *pRead;
			pWrite++; // advance one step further only if read char differs from c

		}

		pRead++; // always read next char

		

	}
	*pWrite = '\0'; // finally, null-terminate the string

}



// ---------------------------------------------------------------------------------
//		EXTRACTFIRSTPART FUNCTION
// ---------------------------------------------------------------------------------
//	Returns a string containing every character of a string (str parameter) until some char (delimiter parameter) is found.
//	If no delimiter is found, the function returns str as is.
//	The returned string must be freed afterwards.
//

static char* ExtractFirstPart(char *str, char delim)
{
   size_t len;
   char *new_str;
   char *delim_pos = strchr(str, delim);

   /* new string is the length from the start of the old string to the
    * delimiter, or if it doesn't exist, a copy of the whole string */
   if (delim_pos == NULL)
      return strdup(str);

   len = delim_pos - str;
   new_str = (char*)malloc(len + 1);
   memcpy(new_str, str, len);
   new_str[len] = '\0'; /* null-terminate the new string */

   return new_str;
}


// ---------------------------------------------------------------------------------
//		FINDNAMEPARAMETERINSCENENAME FUNCTION
// ---------------------------------------------------------------------------------
//	This function returns true if its inputName parameter and the beginning of the sceneName parameter are the same, else it returns false.
//	- if delimited == false, the beginning of inputName must match the beginning of sceneName
//	- if delimited == true, the beginning of inputName must match all characters of sceneName until the delimiter char is found (delimiter excluded)


static bool
FindNameParameterInSceneName(
	char*	inputName,
	char*	sceneName,
	bool	delimited,
	char	delimiter
	)
{	
	

	if (delimited == false) // If delimited property is off
	{
		// Don't use the delimiter, the scene name only has to begin with the input name property
		return (strncmp(sceneName, inputName, strlen(inputName)) == 0); // if beginning of scene name matches the input name property: return true, else return false

	}else{ // If delimited property is on

		
		// If the delimiter exists in scene name, extract the beginning of the scene name (if not, use entire scene name as a base for comparison)
		char* strFirstPart = ExtractFirstPart(sceneName, delimiter);

		if (strcmp(inputName, strFirstPart) == 0) // if input name parameter and extracted first part of scene name (or entire scene name, if it contains no delimiter) are the same
		{
			free(strFirstPart); // free memory allocated by ExtractFirstPart()
			return true;
		}else{
			free(strFirstPart); // free memory allocated by ExtractFirstPart()
			return false;
		}
	}
}

// ---------------------------------------------------------------------------------
//		LAUNCHSCENE FUNCTION
// ---------------------------------------------------------------------------------
//	This function looks for a scene name containing the 'name' input parameter string in it, according to the actor's variables. Those must be already set.
//	When a scene is found, make Isadora jump to it.
//

static void
LaunchScene(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	PluginInfo*			info)
{
	JumpMode mode = kJumpAbsolute; // always use absolute jump mode
	
	
	UInt32 fadeInDuration_mS = (UInt32)info->mFadeIn;
	UInt32 fadeOutDuration_mS = (UInt32)info->mFadeOut;

	// BROWSE ALL SCENES: IF A SCENE SATISFIES THE REQUIREMENTS BELOW, JUMP TO IT
	//
	// Requirements:
	// - if delimited == false
	//		scene name == 'name' input parameter
	// - if delimited == true AND delimiter exists in scene name
	//		1st part of scene name == 'name' input property
	// - if delimited == true AND doesn't exist in scene name
	//		scene name == 'name' input property
	//
	// Note: the Scene Index is one-based

	char loopCurrentSceneName[SCENE_NAME_MAX_LENGTH];

	int i = 1; // scene index (one-based)

	do
	{
		
		GetSceneName_(ip, inActorInfo, mode, i, SCENE_NAME_MAX_LENGTH-1, loopCurrentSceneName); // loopCurrentSceneName pointer now points to the scene name at current index

		if (FindNameParameterInSceneName(info->mName->strData, loopCurrentSceneName, info->mDelimited, info->mDelimiter))
		{
			JumpToSceneWithFadeInFadeOut_(ip, inActorInfo, mode, (SInt32)i, fadeOutDuration_mS, fadeInDuration_mS); // fadeInDuration_mS and fadeOutDuration_mS have to be switched??? bug in the API???
		}
		
		i++;

	}while (strcmp(loopCurrentSceneName, "") != 0); // Loop while a scene name is found in scene list


}





	
// ---------------------------------------------------------------------------------
//		 HandlePropertyChangeValue	[INTERRUPT SAFE]
// ---------------------------------------------------------------------------------
//	### This function is called whenever one of the input values of an actor changes.
//	The one-based property index of the input is given by inPropertyIndex1.
//	The new value is given by inNewValue, the previous value by inOldValue.
//
static void
HandlePropertyChangeValue(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	PropertyIndex		inPropertyIndex1,			// the one-based index of the property than changed values
	ValuePtr			/* inOldValue */,			// the property's old value
	ValuePtr			inNewValue,					// the property's new value
	Boolean				/* inInitializing */)		// true if the value is being set when an actor is first initalized
{
	PluginInfo* info = GetPluginInfo_(inActorInfo);

	// ### When you add/change/remove properties, you will need to add cases
	// to this switch statement, to process the messages for your
	// input properties
	
	// The value comes to you encapsulated in a Value structure. See 
	// ValueCommon.h for details about the contents of this structure.
	
	switch (inPropertyIndex1) {

	case kInputName:
		{
			// name property has changed
			info->mName = inNewValue->u.str;

			// remove all forbidden characters ('\r', '\n', '\t')
			removeCharInString ('\r', info->mName->strData);
			removeCharInString ('\n', info->mName->strData);
			removeCharInString ('\t', info->mName->strData);

		}
		break;

	case kInputFadeOut:
		{
			// fade out property has changed
			info->mFadeOut = inNewValue->u.ivalue;
		}
		break;

	case kInputFadeIn:
		{
			// fade in property has changed
			info->mFadeIn = inNewValue->u.ivalue;
		}
		break;

	case kInputDelimited:
		{
			// delimited property has changed
		if (inNewValue->u.ivalue == 0)
			info->mDelimited = false;
		else
			info->mDelimited = true;
		}
		break;

	case kInputDelimiter:
		{
			// delimiter property has changed
			// MFC: limit this to valid ASCII characters
			// TODO: accept UTF-16 characters
			if (inNewValue->u.ivalue >= 0x20 && inNewValue->u.ivalue <= 0x7F) {
				info->mDelimiter = (char) inNewValue->u.ivalue;
			}
		}
		break;

	case kInputTrigger:
		{
			// trigger has been fired
			LaunchScene(ip, inActorInfo, info);
		}
		break;

	}
}


// ---------------------------------------------------------------------------------
//		 GetActorDefinedArea
// ---------------------------------------------------------------------------------
//	If the mGetActorDefinedAreaProc in the ActorInfo struct points to this function,
//	it indicates to Isadora that the object would like to draw either an icon or else
//	an graphic representation of its function.
//
//	### This function uses the 'PICT' 0 resource stored with the plugin to draw an icon.
//  You should replace this picture (located in the Plugin Resources.rsrc file) with
//  the icon for your actor.
// 
static ActorPictInfo	gPictInfo = { false, nil, nil, 0, 0 };

static Boolean
GetActorDefinedArea(
	IsadoraParameters*			ip,			
	ActorInfo*					inActorInfo,
	SInt16*						outTopAreaWidth,			// returns the width to reserve for the top Actor Defined Area
	SInt16*						outTopAreaMinHeight,		// returns the minimum height of the top area
	SInt16*						outBotAreaHeight,			// returns the width to reserve for the bottom Actor Defined Area
	SInt16*						outBotAreaMinWidth)			// returns the minimum width of the bottom area
{
	if (!gPictInfo.mInitialized) {
		PrepareActorDefinedAreaPict_(ip, inActorInfo, 0, &gPictInfo);
	}
	
	// place picture in top area
	*outTopAreaWidth = gPictInfo.mWidth;
	*outTopAreaMinHeight = gPictInfo.mHeight;
	
	// don't draw anything in bottom area
	*outBotAreaHeight = 0;
	*outBotAreaMinWidth = 0;
	
	return true;
}

// ---------------------------------------------------------------------------------
//		 DrawActorDefinedArea
// ---------------------------------------------------------------------------------
//	If GetActorDefinedArea is defined, then this function will be called whenever
//	your ActorDefinedArea needs to be drawn.
//
//	Beacuse we are using the PICT 0 resource stored with this plugin, we can use
//	the DrawActorDefinedAreaPict_ supplied by the Isadora callbacks.
//
//  DrawActorDefinedAreaPict_ is Alpha Channel aware, so you can have nice
//	shading if you like.

static void
DrawActorDefinedArea(
	IsadoraParameters*			ip,
	ActorInfo*					inActorInfo,
	void*						/* inDrawingContext */,		// unused at present
	ActorDefinedAreaPart		inActorDefinedAreaPart,		// the part of the actor that needs to be drawn
	ActorAreaDrawFlagsT			/* inAreaDrawFlags */,		// actor draw flags
	Rect*						inADAArea,					// rect enclosing the entire Actor Defined Area
	Rect*						/* inUpdateArea */,			// subset of inADAArea that needs updating
	Boolean						inSelected)					// TRUE if actor is currently selected
{
	if (inActorDefinedAreaPart == kActorDefinedAreaTop && gPictInfo.mInitialized) {
		DrawActorDefinedAreaPict_(ip, inActorInfo, inSelected, inADAArea, &gPictInfo);
	}
}
	
// ---------------------------------------------------------------------------------
//		 GetActorInfo
// ---------------------------------------------------------------------------------
//	This is function is called by to get the actor's class and ID, and to get
//	pointers to the all of the plugin functions declared locally.
//
//	All members of the ActorInfo struct pointed to by outActorParams have been
//	set to 0 on entry. You only need set functions defined by your plugin
//	
EXPORT_ void
GetActorInfo(
	void*				/* inParam */,
	ActorInfo*			outActorParams)
{
	// REQUIRED information
	outActorParams->mActorName							= kActorName;
	outActorParams->mClass								= kActorClass;
	outActorParams->mID									= kActorID;
	outActorParams->mCompatibleWithVersion				= kActorCompatilibilityVersion_v1;
	outActorParams->mActorFlags							= kActorFlags_Plugin_CheckForUpdates;
	
	// REQUIRED functions
	outActorParams->mGetActorParameterStringProc		= GetParameterString;
	outActorParams->mGetActorHelpStringProc				= GetHelpString;
	outActorParams->mCreateActorProc					= CreateActor;
	outActorParams->mDisposeActorProc					= DisposeActor;
	outActorParams->mActivateActorProc					= NULL; // I removed this one, though it is marked as required, hope it's OK
	outActorParams->mHandlePropertyChangeValueProc		= HandlePropertyChangeValue;
	
	// OPTIONAL FUNCTIONS
	outActorParams->mHandlePropertyChangeTypeProc		= NULL;
	outActorParams->mHandlePropertyConnectProc			= NULL;
	outActorParams->mPropertyValueToStringProc			= NULL;
	outActorParams->mPropertyStringToValueProc			= NULL;
	outActorParams->mGetActorDefinedAreaProc			= GetActorDefinedArea;
	outActorParams->mDrawActorDefinedAreaProc			= DrawActorDefinedArea;
	outActorParams->mMouseTrackInActorDefinedAreaProc	= NULL;
}

