/*
Author : Benot Vogel -- BASED ON A 'BLANK' PLUGIN FROM THE ISADORA 1.3 SDK PROJECT WIZARD
Date : 24/05/2013
*/


// ===========================================================================
//	Isadora Demo Plugin			   2003 Mark F. Coniglio. All rights reserved.
// ===========================================================================
//
//	IMPORTANT: This source code ("the software") is supplied to you in
//	consideration of your agreement to the following terms. If you do not
//	agree to the terms, do not install, use, modify or redistribute the
//	software.
//
//	Mark Coniglio (dba TroikaTronix) grants you a personal, non exclusive
//	license to use, reproduce, modify this software with and to redistribute it,
//	with or without modifications, in source and/or binary form. Except as
//	expressly stated in this license, no other rights are granted, express
//	or implied, to you by TroikaTronix.
//
//	This software is provided on an "AS IS" basis. TroikaTronix makes no
//	warranties, express or implied, including without limitation the implied
//	warranties of non-infringement, merchantability, and fitness for a 
//	particular purpurse, regarding this software or its use and operation
//	alone or in combination with your products.
//
//	In no event shall TroikaTronix be liable for any special, indirect, incidental,
//	or consequential damages arising in any way out of the use, reproduction,
//	modification and/or distribution of this software.
//
// ===========================================================================
//
// CUSTOMIZING THIS SOURCE CODE
// To customize this file, search for the text ###. All of the places where
// you will need to customize the file are marked with this pattern of 
// characters.
//
// ABOUT IMAGE BUFFER MAPS:
//
// The ImageBufferMap structure, and its accompanying functions,
// exists as a convenience to those writing video processing plugins.
//
// Basically, an image buffer contains an arbitrary number of input and
// output buffers (in the form of ImageBuffers). The ImageBufferMap code
// will automatically create intermediary buffers if needed, so that the
// size and depth of the source image buffers sent to your callback are
// the same for all buffers.
// 
// Typically, the ImageBufferMap is created in your CreateActor function,
// and dispose in the DiposeActor function.

// ---------------------------------------------------------------------------------
// INCLUDES
// ---------------------------------------------------------------------------------

#include "IsadoraPluginPrefix.h"

#include "IsadoraTypes.h"
#include "IsadoraCallbacks.h"
#include "ImageBufferUtil.h"
#include "PluginDrawUtil.h"

#include <string>

// STANDARD INCLUDES
#include <string.h>
#include <stdio.h>

#include <sys/types.h>
#include <sys/stat.h>

// PLUGIN SPECIFIC INCLUDES
#include "ReadTextFromFile.h"

// ---------------------------------------------------------------------------------
// MacOS Specific
// ---------------------------------------------------------------------------------
#if TARGET_OS_MAC
#define EXPORT_
#endif

// ---------------------------------------------------------------------------------
// Win32  Specific
// ---------------------------------------------------------------------------------
#if TARGET_OS_WIN

	#include <windows.h>
	
	#define EXPORT_ __declspec(dllexport)
	
	#ifdef __cplusplus
	extern "C" {
	#endif

	BOOL WINAPI DllMain ( HINSTANCE hInst, DWORD wDataSeg, LPVOID lpvReserved );

	#ifdef __cplusplus
	}
	#endif

	BOOL WINAPI DllMain (
		HINSTANCE	/* hInst */,
		DWORD		wDataSeg,
		LPVOID		/* lpvReserved */)
	{
	switch(wDataSeg) {
	
	case DLL_PROCESS_ATTACH:
		return 1;
		break;
	case DLL_PROCESS_DETACH:
		break;
		
	default:
		return 1;
		break;
	}
	return 0;
	}

#endif

// ---------------------------------------------------------------------------------
//	Exported Function Definitions
// ---------------------------------------------------------------------------------

#ifdef __cplusplus
extern "C" {
#endif

EXPORT_ void GetActorInfo(void* inParam, ActorInfo* outActorParams);

#ifdef __cplusplus
}
#endif

// ---------------------------------------------------------------------------------
//	FORWARD DECLARTIONS
// ---------------------------------------------------------------------------------


// ---------------------------------------------------------------------------------
// GLOBAL VARIABLES
// ---------------------------------------------------------------------------------
// ### Declare global variables, common to all instantiations of this plugin here

// Example: static int gMyGlobalVariable = 5;

// Possible strings for 'read mode' and 'new line mode' input properties

static const char* sReadModeStrings[5] = { "all", "char", "line", "paragraph", "undefined" };

static const char* sNewLineModeStrings[4] = { "windows", "unix", "mac", "undefined" };

static const char* sFileLineEndingType[5] = { "unknown", "windows", "unix", "mac", "mixed" };


// ---------------------------------------------------------------------------------
// PluginInfo struct
// ---------------------------------------------------------------------------------
// ### This structure neeeds to contain all variables used by your plugin. Memory for
// this struct is allocated during the CreateActor function, and disposed during
// the DisposeActor function, and is private to each copy of the plugin.
//
// If your plugin needs global data, declare them as static variables within this
// file. Any static variable will be global to all instantiations of the plugin.

#if TARGET_OS_MAC
typedef IzUniChar*		RTFFString;
#elif TARGET_OS_WIN
typedef IzzyUnicodeStr	RTFFString;
#endif

typedef struct {

	ActorInfo*				mActorInfoPtr;		// our ActorInfo Pointer - set during create actor function
	MessageReceiverRef		mMessageReceiver;	// pointer to our message receiver reference
	
	
	std::string*			mPath;				// retained full or partial path string (UTF-8)

	RTTFReadMode			mReadMode;			// F_ALL = all (default), F_CHR = character, F_LINE = line, F_PARAGRAPH = paragraph
	int						mStart;				// where the plugin should start reading (nth char/nth char/nth line/nth paragraph, respectively).

	bool					mAutoRead;			// true if auto-read is on. When false, the user has to user the read trigger manually when he wants to read the file and output its text.

	bool					mFileOK;			// true if chosen file exists and can be opened with 'read' attribute
	RTTFLineEndingType		mLineEndingType;	// line ending type discovered in the file after reading it
	RTFFString				mAllocatedString;	// string used to store the content of the entire text file
	RTFFString				mFinalString;		// final string for output text. Only contains text filtered according to readmode and start parameters.
	bool					mBeginningFound;	// flag set to true if beginning of final selection has been found
	bool					mEndFound;			// flag set to true if end of final selection has been found

	RTTFNewLineMode			mNewLineMode;		// Output String Line Endings: F_WINDOWS for CR+LF text files, F_UNIX for LF, F_MAC for CR
	
} PluginInfo;


// A handy macro for casting the mActorDataPtr to PluginInfo*
#if __cplusplus
#define	GetPluginInfo_(actorDataPtr)		static_cast<PluginInfo*>((actorDataPtr)->mActorDataPtr);
#else
#define	GetPluginInfo_(actorDataPtr)		(PluginInfo*)((actorDataPtr)->mActorDataPtr);
#endif

// ---------------------------------------------------------------------------------
//	Constants
// ---------------------------------------------------------------------------------
//	Defines various constants used throughout the plugin.

// ### GROUP ID
// Define the group under which this plugin will be displayed in the Isadora interface.
// These are defined under "Actor Types" in IsadoraTypes.h

static const OSType	kActorClass 	= kGroupControl;

// ### PLUGIN IN
// Define the plugin's unique four character identifier. Contact TroikaTronix to
// obtain a unique four character code if you want to ensure that someone else
// has not developed a plugin with the same code. Note that TroikaTronix reserves
// all plugin codes that begin with an underline, an at-sign, and a pound sign
// (e.g., '_', '@', and '#'.)

static const OSType	kActorID		= FOUR_CHAR_CODE('PsIO');

// ### ACTOR NAME
// The name of the actor. This is the name that will be shown in the User Interface.

static const char* kActorName		= "Read Text From File";

// ### PROPERTY DEFINITION STRING
// The property string. This string determines the inputs and outputs for your plugin.
// See the IsadoraCallbacks.h under the heading "PROPERTY DEFINITION STRING" for the
// meaning ofthese codes. (The IsadoraCallbacks.h header can be seen by opening up
// the IzzySDK Framework while in the Files view.)
//
// IMPORTANT: You cannot use spaces in the property name. Instead, use underscores (_)
// where you want to have a space.
//
// Note that each line ends with a carriage return (\r), and that only the last line of
// the bunch ends with a semicolon. This means that what you see below is one long
// null-terminated c-string, with the individual lines separated by carriage returns.

static const char* sPropertyDefinitionString =

// INPUT PROPERTY DEFINITIONS
//	TYPE 	PROPERTY NAME	ID		DATATYPE	DISPLAY FMT			MIN		MAX		INIT VALUE
	"INPROP file_path		path	string		text				-100	100		file.txt\r"

#if TARGET_OS_MAC
	"INPROP line_break_out	lbmd	int			propdef				0		2		2\r"
#elif TARGET_OS_WIN
	"INPROP line_break_out	lbmd	int			propdef				0		2		0\r"
#endif

	"INPROP read_mode		rdmd	int			propdef				0		3		0\r"
	"INPROP start			strt	int			number				0		*		0\r"
	"INPROP auto_read		autr	bool		onoff				0		1		0\r"
	"INPROP reset_output	rsto	bool		trig				0		1		0\r"
	"INPROP	read			trgr	bool		trig				0		1		0\r"

// OUTPUT PROPERTY DEFINITIONS
//	TYPE 	 PROPERTY NAME	ID		DATATYPE	DISPLAY FMT			MIN		MAX		INIT VALUE
	"OUTPROP text_out		outt	string		text				*		*		\r"
	"OUTPROP file_available	fexs	bool		onoff				*		*		0\r"
	"OUTPROP has_been_read	fred	bool		trig				*		*		0\r"
	"OUTPROP line_endings	lend	int			propdef				0		4		0\r";

// ### Property Index Constants
// Properties are referenced by a one-based index. The first input property will
// be 1, the second 2, etc. Similarly, the first output property starts at 1.
// You whould have one constant for each input and output property defined in the 
// property definition string.

enum
{
	kInputFilePath = 1,
	kInputLinebreakOut,
	kInputReadMode,
	kInputStart,
	kInputAutoRead,
	kInputResetOutput,
	kInputTriggerRead,
	
	kOutputText = 1,
	kOutputFileWasFound,
	kOutputFileWasRead,
	kOutputFileLineEndingType
};

// ---------------------
//	Help String
// ---------------------
// ### Help Strings
//
// The first help string is for the actor in general. This followed by help strings
// for all of the inputs, and then by the help strings for all of the outputs. These
// should be given in the order that they are defined in the Property Definition
// String above.
//
// In all, the total number of help strings should be (num inputs + num outputs + 1)
//
// Note that each string is followed by a comma -- it is a common mistake to forget the
// comma which results in the two strings being concatenated into one.

const char* sHelpStrings[] =
{
	// ACTOR HELP
	"This actor reads a text file in the ASCII or UTF-8 format and outputs its text after optional filtering. Attempt to read files that "
		"are not saved as ASCII or UTF-8 (e.g., MacOS Roman or ISO-8859-1) will not be correctly interpreted.\r"
		
		"This actor correctly interprets all three line ending types (CR for macOS, CR/LF for Windows, and LF for Unix). You can choose "
		"the line ending format in the output text by setting the 'line break out' input, though Isaodra should respond correctly to all three.\r"
		
		"This plugin was written and contributed to the community by Beno\xc3\xaet Vogel.",

	// INPUT HELP
	"The file name, or the full or partial path name, of the text file you want to read.\r"
	
		"Line breaks in the path will be ignored, so feel free to cut long paths into several lines.\r"

		"The best practice is to place the file you wish to read in the same folder as your Isadora (.izz) file. If you "
		"do this, then you need only to enter the file name and extension for this property, e.g., file-name.txt.\r"
	
		"Otherwise, you can specify a full or partial path name. Full path names start with a '/' on macOS, or 'X:\' "
		"on Windows, where X is a drive letter. On macOS, if you wish to read from a drive other than the main hard drive, "
		"the full path name will start with '/Volumes/NAME' where NAME is the name of the drive.\r"
	
		"Partial path names are specified relative to the Isadora document, and can use the special code '..' to "
		"designate the folder that is one above the current directory. For example, if your file is in a folder called "
		"'myfolder' that is in the same folder as your Isadora document, you could use the partial path "
		"'myfolder/file-name.txt. If the text file is in the directory above your Isadora document, then you could "
		"enter '../file-name.txt.",
	
	"Specifies the line break type used for the output text. Set this input to 'windows' to output CR+LF line breaks, "
		"'unix' for LFs, or 'mac' for CRs.",

	"'read mode', in coordination with the 'start'  input property, determines which part of the file will be read:\r"
		"'all' = read until end of file is reached.\r"
		"'char' = read only 1 character.\r"
		"'line' = read only 1 line of text (a blank line also counts as a line).\r"
		"'paragraph' = read only 1 paragraph (paragraphs are separated by 2 blank lines or more).",

	"How this parameter is interpreted depends on the setting of the 'read mode' input. When 'read mode' is set to:\r"
		"'all': read from the position specified by the 'start' input property until the end.\r"
		"'char': read only the character at the position specified by the 'start'  input property.\r"
		"'line': read the line specified by the 'start'  input property.\r"
		"'paragraph': read the paragraph specified by the 'start'  input property.\r"
		"This parameter starts from 0 : first position is 0, not 1.",

	"Auto-read mode:\r"
		"When set to 'on', the input file will be read and the output text will be updated whenever the 'file path', 'read mode', or 'start' input properties change.\r"
		"When set to 'off', you will have to use the 'read' input trigger each time you want to read the file.",

	"Clears the output string and resets the 'file available' property to off.",

	"When triggered, read the file and send its text to the 'text out' output after applying any filtering required by the 'read mode' and 'start' input properties. "
		"The 'file available' output properties will indicate if the file exists and could be read. The 'has been read' output will send a trigger after the 'text out' "
		"output property has been updated.",

	// OUTPUT HELP
	"The output text using the line break type specified by the 'linebreak out' input property.",

	"'on' if input file can be opened and read. 'off' if not (which usually means that the input path is wrong).",

	"Sends a trigger once input file has been read and output text has changed accordingly.",

	"Displays the type of line endings found in the file once it has been read. If there are no line endings in the file, "
		"'windows' for CR/LF, 'mac' for CR, 'unix' for LF, and 'mixed' if multiple line ending types are found."
};

// ---------------------------------------------------------------------------------
//		 CFStringToRTFFString
// ---------------------------------------------------------------------------------
#if TARGET_OS_MAC
RTFFString
CFStringToRTFFString(
	IzzyUnicodeStr	inIzzyUnicodeStr)
{
	CFIndex len = ::CFStringGetLength(inIzzyUnicodeStr);
	RTFFString outUniChar = (RTFFString) malloc((len+1) * sizeof(UniChar));
	if (outUniChar != NULL) {
		::CFStringGetCharacters(inIzzyUnicodeStr,  CFRangeMake(0, len), outUniChar);
		outUniChar[len] = 0;
	}
	return outUniChar;
}
#endif

// ---------------------------------------------------------------------------------
//		 RTFFStringToUTF8String
// ---------------------------------------------------------------------------------
#if TARGET_OS_MAC
char*
RTFFStringToUTF8String(
	RTFFString	inUnicodeString)
{
	char* outUTF8String = NULL;
	
	// get string length
	CFIndex len = 0;
	RTFFString c = inUnicodeString;
	while (*c != 0) {
		len++;
		c++;
	}
	
	CFStringRef cfStr = ::CFStringCreateWithCharacters(kCFAllocatorDefault, inUnicodeString, len);
	if (cfStr != NULL) {
		CFIndex length = ::CFStringGetMaximumSizeForEncoding(CFStringGetLength(cfStr), kCFStringEncodingUTF8);
		outUTF8String = (char*) malloc(length+1);
		if (outUTF8String != NULL) {
			if (!::CFStringGetCString(cfStr, outUTF8String, length+1, kCFStringEncodingUTF8)) {
				free(outUTF8String);
				outUTF8String = NULL;
			}
		}
		::CFRelease(cfStr);
	}

	return outUTF8String;
}
#endif

// ---------------------------------------------------------------------------------
//		 CreateRTFFStringFromUTF8
// ---------------------------------------------------------------------------------
RTFFString
CreateRTFFStringFromUTF8(
	IsadoraParameters*	ip,
	const char*			inUTFString,
	size_t				inUTFStringLen)
{
	RTFFString outString = NULL;
	IzzyUnicodeStr iustr = CreateIzzyUnicodeStringWithUTF8Text_(ip, inUTFString, inUTFStringLen);
	#if TARGET_OS_MAC
		outString = ::CFStringToRTFFString(iustr);
		DisposeIzzyUnicodeString_(ip, iustr);
	#elif TARGET_OS_WIN
		outString = iustr;
	#endif
	return outString;
}

// ---------------------------------------------------------------------------------
//		 DisposeAllocatedString
// ---------------------------------------------------------------------------------
void
DisposeAllocatedString(
	IsadoraParameters*	ip,
	PluginInfo*			info)
{
	if (info->mAllocatedString) {
		#if TARGET_OS_MAC
			free(info->mAllocatedString);
		#elif TARGET_OS_WIN
			DisposeIzzyUnicodeString_(ip, info->mAllocatedString);
		#endif
		info->mAllocatedString = NULL;
	}
}

// ---------------------------------------------------------------------------------
//		 DisposeFinalString
// ---------------------------------------------------------------------------------
void
DisposeFinalString(
	IsadoraParameters*	ip,
	PluginInfo*			info)
{
	if (info->mFinalString) {
		free(info->mFinalString);
		info->mFinalString = NULL;
	}
}

// ---------------------------------------------------------------------------------
//		 CreateActor
// ---------------------------------------------------------------------------------
// Called once, prior to the first activation of an actor in its Scene. The
// corresponding DisposeActor actor function will not be called until the file
// owning this actor is closed, or the actor is destroyed as a result of being
// cut or deleted.

static void
CreateActor(
	IsadoraParameters*	ip,	
	ActorInfo*			ioActorInfo)		// pointer to this actor's ActorInfo struct - unique to each instance of an actor
{
	// creat the PluginInfo struct - initializing it to all zeroes
	PluginInfo* info = (PluginInfo*) IzzyMallocClear_(ip, sizeof(PluginInfo));
	PluginAssert_(ip, info != nil);

	if (info != NULL) {
	
		ioActorInfo->mActorDataPtr = info;
		info->mActorInfoPtr = ioActorInfo;

		// ### allocation and initialization of private member variables
	
		info->mPath = NULL;
		info->mAllocatedString = NULL;
		info->mFinalString = NULL;

		// use default line ending for current platform
		#if TARGET_OS_MAC
		info->mNewLineMode = F_MAC;
		#elif TARGET_OS_WIN
		info->mNewLineMode = F_WINDOWS;
		#endif

		// By default, read the entire file
		info->mReadMode = F_ALL;
	
		// Initialize the starting line with the very 1st line of the file
		info->mStart = 0;

		// Initialize auto-read mode to off
		info->mAutoRead = false;

	}
}

// ---------------------------------------------------------------------------------
//		 DisposeActor
// ---------------------------------------------------------------------------------
// Called when the file owning this actor is closed, or when the actor is destroyed
// as a result of its being cut or deleted.
//
static void
DisposeActor(
	IsadoraParameters*	ip,
	ActorInfo*			ioActorInfo)		// pointer to this actor's ActorInfo struct - unique to each instance of an actor
{
	PluginInfo* info = GetPluginInfo_(ioActorInfo);
	PluginAssert_(ip, info != nil);
	
	// ### destruction of private member variables
	
	// destroy the PluginInfo struct allocated with IzzyMallocClear_ the CreateActor function
	PluginAssert_(ip, ioActorInfo->mActorDataPtr != nil);
	IzzyFree_(ip, ioActorInfo->mActorDataPtr);
}

// ---------------------------------------------------------------------------------
//		 ActivateActor
// ---------------------------------------------------------------------------------
//	Called when the scene that owns this actor is activated or deactivated. The
//	inActivate flag will be true when the scene is activated, false when deactivated.
//

static void
ActivateActor(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	Boolean				inActivate)
{
	PluginInfo* info = GetPluginInfo_(inActorInfo);
	PluginAssert_(ip, info != nil);

	if (inActivate) {

		if (ip->mCallbacks->mCallbackVersionNumber >= kIsadoraCallbacksVersion_308b39) {
			SetPropertyFilePathModeProc_(ip, inActorInfo, kInputProperty, kInputFilePath, kPropertyFilePath_PathRead);
		}
		
		// init members that will be allocated and may need disposal later
		info->mPath = NULL;
		info->mAllocatedString = NULL;
		info->mFinalString = NULL;

	} else {
		
		if (info->mPath != NULL) {
			delete info->mPath;
			info->mPath = NULL;
		}
		DisposeAllocatedString(ip, info);
		DisposeFinalString(ip, info);
	}
}


// ---------------------------------------------------------------------------------
//		 GetParameterString
// ---------------------------------------------------------------------------------
//	Returns the property definition string. Called when an instance of the actor
//	needs to be instantiated.

static const char*
GetParameterString(
	IsadoraParameters*	/* ip */,
	ActorInfo*			/* inActorInfo */)
{
	return sPropertyDefinitionString;
}

// ---------------------------------------------------------------------------------
//		 GetHelpString
// ---------------------------------------------------------------------------------
//	Returns the help string for a particular property. If you have a fixed number of
//	input and output properties, it is best to use the PropertyTypeAndIndexToHelpIndex_
//	function to determine the correct help string to return.

static void
GetHelpString(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	PropertyType		inPropertyType,			// kPropertyTypeInvalid when requesting help for the actor
												// or kInputProperty or kOutputProperty when requesting help for a specific property
	PropertyIndex		inPropertyIndex1,		// the one-based index of the property (when inPropertyType is not kPropertyTypeInvalid)
	char*				outParamaterString,		// receives the help string
	UInt32				inMaxCharacters)		// size of the outParamaterString buffer
{
	const char* helpstr = nil;
	
	// The PropertyTypeAndIndexToHelpIndex_ converts the inPropertyType and
	// inPropertyIndex1 parameters to determine the zero-based index into
	// your list of help strings.
	UInt32 index1 = PropertyTypeAndIndexToHelpIndex_(ip, inActorInfo, inPropertyType, inPropertyIndex1);
	
	// get the help string
	helpstr = sHelpStrings[index1];
	
	// copy it to the output string
	strncpy(outParamaterString, helpstr, inMaxCharacters);
}
	


// ---------------------------------------------------------------------------------
//	###	 PropertyValueToString
// ---------------------------------------------------------------------------------
//	### This function defines custom display strings for "read mode" input parameter
//	### since we declared this parameter as "PROPDEF" type

static Boolean
PropertyValueToString(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	PropertyType		inPropertyType,				// property type
	PropertyIndex		inPropertyIndex1,			// the one-based index of the property than changed values
	ValuePtr			inValue,					// the property's current value
	char*				outString)					// string to display (max length of 256 characters)
{
	Boolean outStringChanged = false; // I don't know if it's necessary, let's assume it is (function declared with Boolean return)

	if (inPropertyType == kInputProperty) // input parameters only
	{

		switch (inPropertyIndex1) {

		case kInputReadMode : // 'read mode' input property
			
			switch (inValue->u.ivalue) {
				
			case F_ALL :
				strcpy(outString, sReadModeStrings[F_ALL]);
				outStringChanged = true;
				break;
			case F_CHAR :
				strcpy(outString, sReadModeStrings[F_CHAR]);
				outStringChanged = true;
				break;
			case F_LINE :
				strcpy(outString, sReadModeStrings[F_LINE]);
				outStringChanged = true;
				break;
			case F_PARAGRAPH :
				strcpy(outString, sReadModeStrings[F_PARAGRAPH]);
				outStringChanged = true;
				break;
			default :
				strcpy(outString, sReadModeStrings[F_UNDEFINED]);
				outStringChanged = true;
				break;

			}
			break;

		case kInputLinebreakOut: // 'line break mode' input property

			switch (inValue->u.ivalue) {

			case F_WINDOWS:
				strcpy(outString, sNewLineModeStrings[F_WINDOWS]);
				outStringChanged = true;
				break;
			case F_UNIX:
				strcpy(outString, sNewLineModeStrings[F_UNIX]);
				outStringChanged = true;
				break;
			case F_MAC:
				strcpy(outString, sNewLineModeStrings[F_MAC]);
				outStringChanged = true;
				break;
			default:
				strcpy(outString, sNewLineModeStrings[F_NLUNDEFINED]);
				outStringChanged = true;
				break;

			}
			break;

		default :
			break;
		}
	}
	if (inPropertyType == kOutputProperty) // output parameters only
	{

		switch (inPropertyIndex1) {

		case kOutputFileLineEndingType: // 'read mode' input property

			switch (inValue->u.ivalue) {

			case kRTTFLineEndingType_Unknown:
				strcpy(outString, sFileLineEndingType[kRTTFLineEndingType_Unknown]);
				outStringChanged = true;
				break;
			case kRTTFLineEndingType_Windows:
				strcpy(outString, sFileLineEndingType[kRTTFLineEndingType_Windows]);
				outStringChanged = true;
				break;
			case kRTTFLineEndingType_Unix:
				strcpy(outString, sFileLineEndingType[kRTTFLineEndingType_Unix]);
				outStringChanged = true;
				break;
			case kRTTFLineEndingType_MacOS:
				strcpy(outString, sFileLineEndingType[kRTTFLineEndingType_MacOS]);
				outStringChanged = true;
				break;
			case kRTTFLineEndingType_Mixed:
				strcpy(outString, sFileLineEndingType[kRTTFLineEndingType_Mixed]);
				outStringChanged = true;
				break;
			}
			break;
		}
	}

	if (outStringChanged)
		return true;
	else
		return false;

}

// ---------------------------------------------------------------------------------
//		 PropertyStringToValue
// ---------------------------------------------------------------------------------

static Boolean
PropertyStringToValue(
	IsadoraParameters*	ip,
	ActorInfo*			/* inActorInfo */,
	PropertyType		inPropertyType,
	PropertyIndex		inPropertyIndex1,
	const char*			inString,
	ValuePtr			outValue)
{
	Boolean encodeSuccessful = false;
	
	if (inPropertyType == kInputProperty && inPropertyIndex1 == kInputLinebreakOut) {
		if (strlen(inString) == 1 && inString[0] >= '0' && inString[0] <= '2') {
			outValue->type = kInteger;
			outValue->u.ivalue = (SInt32) inString[0] - '0';
			encodeSuccessful = true;
		} else {
			SInt32 matchIndex = LookupPartialStringInList_(ip, 3, sNewLineModeStrings, inString);
			if (matchIndex >= 0) {
				outValue->type = kInteger;
				outValue->u.ivalue = matchIndex;
				encodeSuccessful = true;
			}
		}
	}
	
	if (inPropertyType == kInputProperty && inPropertyIndex1 == kInputReadMode) {
		if (strlen(inString) == 1 && inString[0] >= '0' && inString[0] <= '3') {
			outValue->type = kInteger;
			outValue->u.ivalue = (SInt32)inString[0] - '0';
			encodeSuccessful = true;
		} else {
			SInt32 matchIndex = LookupPartialStringInList_(ip, 4, sReadModeStrings, inString);
			if (matchIndex >= 0) {
				outValue->type = kInteger;
				outValue->u.ivalue = matchIndex;
				encodeSuccessful = true;
			}
		}
	}

	if (inPropertyType == kOutputProperty && inPropertyIndex1 == kOutputFileLineEndingType) {
		if (strlen(inString) == 1 && inString[0] >= '0' && inString[0] <= '4') {
			outValue->type = kInteger;
			outValue->u.ivalue = (SInt32)inString[0] - '0';
			encodeSuccessful = true;
		} else {
			SInt32 matchIndex = LookupPartialStringInList_(ip, 5, sFileLineEndingType, inString);
			if (matchIndex >= 0) {
				outValue->type = kInteger;
				outValue->u.ivalue = matchIndex;
				encodeSuccessful = true;
			}
		}
	}

	return encodeSuccessful;
}

// ---------------------------------------------------------------------------------
//		IsRelativePathName
// ---------------------------------------------------------------------------------
// IS-1009: Read Text From File v1.1.1 Does Not Allow for Partial Path Names

#if TARGET_OS_MAC
const char kSeparatorChar = '/';
#elif TARGET_OS_WIN
const char kSeparatorChar = '\\';
#endif

static bool
IsRelativePathName(
	const char*		c)
{
	bool isRelativePath = true;
	
	#if TARGET_OS_MAC
	if (strlen(c) >= 2) {
		if (*c == kSeparatorChar) {
			isRelativePath = false;
		}
	}
	#elif TARGET_OS_WIN
	if (strlen(c) >= 3) {
		if ( ((*c >= 'a' && *c <= 'z') || (*c >= 'A' && *c <= 'Z')) && *(c+1) == ':' && *(c+2) == kSeparatorChar) {
			isRelativePath = false;
		}
	}
	#endif
	
	return isRelativePath;
}

// ---------------------------------------------------------------------------------
//	CountBackslashes
// ---------------------------------------------------------------------------------

#if TARGET_OS_WIN
int
CountBackslashes(
	const char* c)
{
	int backslashCount = 0;
	while (*c != 0) {
		if (*c == '\\') {
			backslashCount += 1;
		}
		c++;
	}
	return backslashCount;
}
#endif

// ---------------------------------------------------------------------------------
//		READTEXTFILE FUNCTION
// ---------------------------------------------------------------------------------
// This function reads the text file according to the actor's variables.
// Those must be already set.
//

static void
ReadTextFile(
	IsadoraParameters*	ip,
	ActorInfo*			actorInfo,
	PluginInfo*			info)
{
	// trigger file reading and save it into the 'selected text' string
			
	// READ FILE - VARIABLES

	// pointer to the FILE structure used when reading the file
	// assume NULL in case one the next section fails
	FILE* fileToRead = NULL;

	// stat structure and stat result; used when determining the length
	// of the file.
	int stat_result = 0;
	#if TARGET_OS_MAC
	struct stat fileStat = { 0 };
	#elif TARGET_OS_WIN
	struct _stat64i32 fileStat = { 0 };
	#endif

	size_t fileSize = -1;

	// --------------------------------------------------------------------------------
	// GET FULL PATH NAME, HANDLING RELATIVE PATH IF NEEDED
	// --------------------------------------------------------------------------------
	// Format the file path by removing any carriage returns or line-feeds and also
	// by doubling the backslashes if we're running on Windows

	// get the initial path from the parameter passed by the user
	std::string utf8Path;
	if (info != NULL && info->mPath != NULL && info->mPath->length() > 0) {
		utf8Path = *info->mPath;
		// if the path length is greater than zero and if it is
		// a relative path, then we'll attempt to get the path
		// to the document and append it to the beginning of
		// the path supplied by the user
		if (utf8Path.length() > 0 && IsRelativePathName(utf8Path.c_str())) {
			// IS-1009: Read Text From File v1.1.1 Does Not Allow for Partial Path Names
			char documentPath[4096] = { 0 };
			// if we are able to get the document path (file is saved) then the
			// final path will be the document path + the supplied relative path
			if (GetUTF8PathToCurrentDocument_(ip, actorInfo->mThing, sizeof(documentPath), documentPath)) {
				utf8Path = std::string(documentPath) + utf8Path;
			}
		}
	}

	// N.B. It is possible that the UTF-8 path at this point will be empty...
	
	// --------------------------------------------------------------------------------
	// FORMAT FULL FILE PATH
	// --------------------------------------------------------------------------------
	// Format the file path by removing any carriage returns or line-feeds and also
	// by doubling the backslashes if we're running on Windows

	// get pointer to start of source UTF-8 path
	const char* fpSrc = utf8Path.c_str();

	// calculate the size of the destination UTF-8 path
	// allow one extra character for null termination
	int fpDstLen = (int) utf8Path.length() + 1;

	// if we're running on windows, increase fpDstLen by the number
	// of backslashes in the path because we will double them.
	#if TARGET_OS_WIN
		fpDstLen += CountBackslashes(fpSrc);
	#endif

	// build the final path string, with doubled slashes for the windows version of fopen
	char* fullUTF8FilePath = (char*) malloc(fpDstLen * sizeof(char));

	if (fullUTF8FilePath != NULL) {

		// get pointer to buffer that will hold the final file path
		char* fpDst = fullUTF8FilePath;

		while (*fpSrc != 0) {
		
			// if we're running on windows and the current character
			// is a backslash, we write an extra one here; since it
			// will again be copied below, this has the effect of 
			// doubling the backslash
			#if TARGET_OS_WIN
			if (*fpSrc == '\\') {
				*(fpDst++) = '\\';
			}
			#endif

			// if the current
			if (*fpSrc != '\r' && *fpSrc != '\n') {
				*(fpDst++) = *fpSrc;
			}

			// skip over the character we just examined
			fpSrc++;
	
		}
		// null terminate the updated string
		*fpDst = 0;
	}

	// --------------------------------------------------------------------------------
	// OPEN THE FILE
	// --------------------------------------------------------------------------------

	// if we were able to allocate the full file path
	if (fullUTF8FilePath != NULL) {

		#if TARGET_OS_MAC
	
			stat_result = stat(fullUTF8FilePath, &fileStat);

			// if stat succeeds we can continue
			if (stat_result >= 0 && (fileStat.st_mode & S_IFREG) != 0) {
				// save the file size from the stat structure
				fileSize = fileStat.st_size;
				// if file size is actually greater than zero
				if (fileSize > 0) {
					// on macOS, we can use the UTF8 path
					fileToRead = fopen(fullUTF8FilePath, "r"); // try to open the file using the read attribute
				}
			}
	
		#elif TARGET_OS_WIN
	
			// on Windows, we must convert the path to a wide character string
	
			int fullUTF8FilePathLen = (int) strlen(fullUTF8FilePath);

			// create wide char string from UTF-8 string
			RTFFString fullWidePath = CreateRTFFStringFromUTF8(ip, fullUTF8FilePath, fullUTF8FilePathLen);
			if (fullWidePath != NULL) {
				// use stat to get the file's information
				stat_result = _wstat(fullWidePath, &fileStat);
				if (stat_result >= 0 && (fileStat.st_mode & _S_IFREG) != 0) {
					// save the file size from the stat structure
					fileSize = fileStat.st_size;
					// if file size is actually greater than zero
					if (fileSize > 0) {
						// attempt to open the file for reading
						fileToRead = _wfopen(fullWidePath, L"rb");
					}
				}
				DisposeIzzyUnicodeString_(ip, fullWidePath);
			}

		#endif
	}

	// if the full file path was created, then free it here
	if (fullUTF8FilePath != NULL) {
		free(fullUTF8FilePath);
		fullUTF8FilePath = NULL;
	}


	/*
	- SET VALUE FOR "file available" OUTPUT PARAMETER
	- IF FILE IS OPEN, READ IT AND STORE ITS CONTENTS TO A STRING
	- FIND BEGINNING AND END OF SELECTION ACCORDING TO READMODE INPUT PROPERTY
	- CLOSE THE FILE
	*/

	// test if file has been found and can be read

	if (fileToRead == NULL) {
		
		// prepare "file available" output param --> can't be opened in read mode
		info->mFileOK = false;

		// but do nothing else

	} else {
	
		// prepare "file available" output param --> has been opened in read mode
		info->mFileOK = true;

		// read every character in the file into a temporary buffer
		// and convert the result from UTF-8 to Unicode
		char* buffer = (char*)malloc((fileSize + 1) * sizeof(char));
		// if we succeeded in allocating the buffer

		if (buffer != NULL) {
		
			// read entire file into buffer
			int cc;
			char* dst = buffer;
			while ((cc = fgetc(fileToRead)) != EOF) {
				*(dst++) = cc;
			}
			// add null terminator
			*dst = 0;
			
			// here we both sense the line ending type found in the file
			// and normalize all line endings to '\n', whether the original
			// was CR/LF, LF, or CR.
			bool macLineEndingFound = false;
			bool winLineEndingFound = false;
			bool unixLineEndingFound = false;
			
			char* src = buffer;
			dst = buffer;
			while (*src != 0) {
				// if
				if (*(src+0) == '\r' && *(src+1) == '\n') {
					winLineEndingFound = true;
					src += 2;
					*(dst++) = '\n';
				} else if (*src == '\n') {
					unixLineEndingFound = true;
					*(dst++) = '\n';
					src += 1;
				} else if (*src == '\r') {
					macLineEndingFound = true;
					*(dst++) = '\n';
					src += 1;
				} else {
					*(dst++) = *src;
					src += 1;
				}
			}
			
			info->mLineEndingType = kRTTFLineEndingType_Unknown;
			if (macLineEndingFound && winLineEndingFound && unixLineEndingFound) {
				info->mLineEndingType = kRTTFLineEndingType_Mixed;
			} else if (macLineEndingFound && winLineEndingFound) {
				info->mLineEndingType = kRTTFLineEndingType_Mixed;
			} else if (winLineEndingFound && unixLineEndingFound) {
				info->mLineEndingType = kRTTFLineEndingType_Mixed;
			} else if (unixLineEndingFound && macLineEndingFound) {
				info->mLineEndingType = kRTTFLineEndingType_Mixed;
			} else if (macLineEndingFound) {
				info->mLineEndingType = kRTTFLineEndingType_MacOS;
			} else if (winLineEndingFound) {
				info->mLineEndingType = kRTTFLineEndingType_Windows;
			} else if (unixLineEndingFound) {
				info->mLineEndingType = kRTTFLineEndingType_Unix;
			}

			// dispose any existing string
			DisposeAllocatedString(ip, info);
			
			// create a new unicode string from the UTF8 string
			info->mAllocatedString = CreateRTFFStringFromUTF8(ip, buffer, (dst - buffer));

			free(buffer);
		}
				
		const IzUniChar* rmStart = nil;		// 1st character of the text we want to keep
		const IzUniChar* rmEnd = nil;		// last character of the text we want to keep (should point to '\0' at the end of the process)

		// if the data from teh file was read and the conversion to unicode succeeded
		if (info->mAllocatedString != NULL) {

			// FILTER STRING ACCORDING TO "READMODE" AND "START" PARAMETERS

			// find beginning and end according to what we want to keep in the string //
					
			// initialize beginning/end found flags
			info->mBeginningFound = false;
			info->mEndFound = false;

			// if there are blank lines at the beginning, ignore them first (read on
			// until 1st non-linebreak char is reached)

			IzUniChar *c = info->mAllocatedString;
			while ((*c == '\r') || (*c == '\n')) {
				c++;
			}

			// if end of string has already been reached (=info->mAllocatedString
			// is empty or only contains new line characters) [this shouldn't happen
			// normally, but who knows if the file is empty or not?] an empty string
			// can already be set for selection
			
			if (*c == '\0') {

				// Set beginning of selection
				rmStart = c;
							
				// Set end of selection
				rmEnd = c;

				// Set mBeginningFound & mEndFound flag in order to stop searching
				info->mBeginningFound = true;
				info->mEndFound = true;
							
			}

			int curLine;				// line counter
			int curParagraph;			// paragraph counter
			bool newLineFound = false;	// this flag will be set to true if a new line has been detected while parsing info->mAllocatedString

			switch (info->mReadMode) { // from now on the algorithm differs according to 'read mode' input property

			case F_UNDEFINED:
				break;
				
			case F_ALL:
						
				// set the bounds if they are not set yet //

				c = info->mAllocatedString;
						
				if (info->mBeginningFound == false) {

					int i = 0;

					// read until start character is reached
					while (i < info->mStart) {
								
						// note: i = info->mStart && *c == '\0' at the same
						// time can't happen here because this case has already been tested

						// if 'start' input property is larger than number of characters in the file
						if (*c == '\0') {

							// set end of selection (beginning will be set after while loop) and associated flag
							rmEnd = c;
							info->mEndFound = true;

							// force end of while loop (a bit nasty but it works)
							i = info->mStart;

						} else {
						
							// end of string has not been reached yet, just read on
							c++;
							i++;

						}					
					}
							
					rmStart = c;		// set beginning of selection
					info->mBeginningFound = true;	// and corresponding flag
				}

				// if '\0' char has not been reached yet, read on until we find it
				if (info->mEndFound == false) {

					while(*c != '\0')
						c++;

					// and set end of selection & corresponding flag
					rmEnd = c;

					info->mEndFound = true;							
				}
						
				break;

			case F_CHAR:
						
				// extract only 1 char
						
				// set the bounds if they are not set yet

				c = info->mAllocatedString;
						
				if (info->mBeginningFound == false) {

					int i = 0;

					while (i < info->mStart) { // read until start character is reached
								
						// note : i = info->mStart && *c == '\0' at the same
						// time can't happen here because this case has already been tested

						if (*c == '\0') { // if 'start' input property is larger than number of characters in the file

							// set end of selection (beginning will be set after while loop) and associated flag
									
							rmEnd = c;
							info->mEndFound = true;

							// force end of while loop (a bit nasty but it works)
							i = info->mStart;

						} else {

							c++;
							i++;

						}					
					}
							
					rmStart = c; // set beginning of selection
					info->mBeginningFound = true; // and corresponding flag
				}

				// if '\0' char has not been reached yet, read only 1 char
				if (info->mEndFound == false) {
							
					// set end of selection & corresponding flag
					rmEnd = rmStart + 1;

					info->mEndFound = true;							
				}
						
				break;

			case F_LINE:

				// find beginning according to the line number ('start' input property)
				c = info->mAllocatedString;

				curLine = 0;
				newLineFound = false;
						
				if (info->mBeginningFound == false) {

					// Initialize beginning of selection in case the 'start' property is 0
					rmStart = c;

					// if 'start' property is greater than 0
					while (curLine < info->mStart) {


						if (*c == '\0') {

							// set end of selection & corresponding flag
							rmEnd = c;
							info->mEndFound = true;
							
							// force exit of while loop (a bit nasty but it works...)
							curLine = info->mStart;

						} else {

							// detect linebreaks
							if (*c == '\n')
								newLineFound = true;				

							// read on
							c++;

						}

						if (newLineFound) {

							// Increase current line no
							curLine++;

							// reset flag to be able to find new new-line characters from now on
							newLineFound = false;

						}
								
						// Set start of selection
						rmStart = c;
					}

					info->mBeginningFound = true;
				}


				// if '\0' char has not been reached yet, read only 1 line
				if (info->mEndFound == false) {

					// find line end from here

					while ((*c != '\r') && (*c != '\n') && (*c != '\0')) {
						c++;
					}

					// Set end of selection & corresponding flag
					rmEnd = c;

					info->mEndFound = true;

				}

				break;

			case F_PARAGRAPH :

				// find beginning according to the paragraph number ('start' input property)
				c = info->mAllocatedString;

				IzUniChar previousChar = NULL; // initialize previous character
				curParagraph = 0;
						
				if (info->mBeginningFound == false) {

					// initialize beginning of selection in case the 'start' property is 0
					rmStart = c;

					// if 'start' property is greater than 0
					while (curParagraph < info->mStart) {

						if (*c == '\0') {

							// Set end of selection & corresponding flag
							rmEnd = c;

							info->mEndFound = true;

							curParagraph = info->mStart; // Force exit of while loop (a bit nasty but it works)


						} else {

							// detect single new lines
							if (*c == '\n') {

								// If previous char was already a new line char
								if (previousChar == '\n')
								{
									// ... the end of a paragraph has been reached

									// Increase current paragraph #
									curParagraph++;

									// Read on until next non-new-line char is found (can be the beginning of a new paragraph or just '\0')
									while (*c == '\r' || *c == '\n') {
										c++;
									}

									c--; // just because we move on to the next char once again using *c++ below, we should move one step backwards first
								}


							}

							// READ ON
							c++;
						}

						// Set start of selection
						rmStart = c;

						previousChar = *(c-1); // Set previous char in order to be able to detect new paragraphs, not only single line endings

					}

					info->mBeginningFound = true;
				}

				if (info->mEndFound == false) { // if '\0' char has not been read yet, read only 1 paragraph (or until 'end of string' char is reached)

					// find paragraph end from here

					bool paragraphEndReached = false;

					while (*c != '\0' && paragraphEndReached == false) {
							
						if (*c == '\n') { // If current char is a new line char

							// test next chars to detect end of paragraph: if next char != '\0' and next char == new line char
							if (*(c+1) != '\0') {

								if (*(c+1) == '\n') {
									paragraphEndReached = true;
								}

							}
						}

						c++;

					}

					// do not step back to last char if no paragraph should be read (this is to
					// prevent a bug when the file end is composed of several linebreaks)
					
					if (*rmStart != '\0') {

						// remove every '\r's and '\n's at the end of the selected paragraph

						while ( *(c-1) == '\r' || *(c-1) == '\n') {
							c--;
						}

					}

					// Set end of selection & corresponding flag
					rmEnd = c;

					info->mEndFound = true;

				}
				break;

			}
		}

		// CLOSE THE FILE
		fclose(fileToRead);
				
		// BUILD NEW STRING info->mFinalString USING BEGINNING (rmStart) AND END (rmEnd) POINTERS FROM PREVIOUS FILTER
			
		if (rmStart != NULL && rmEnd != NULL) {

			// if windows CRLF mode, count how many linebreaks there are, as we need to allocate space for additional '\r's
			int windowsNewLinesCounter = 0;
			if (info->mNewLineMode == F_WINDOWS) {
				// count new line characters in output
				if (rmStart != NULL) {
					const IzUniChar* lc = rmStart;
					while (*lc != '\0' && lc < rmEnd) {
						if (*lc == '\n')
							windowsNewLinesCounter++;
						lc++;
					}
				}
			}

			// if the final string exists already, then dispose it
			DisposeFinalString(ip, info);
			
			// allocate final unicode string
			size_t charCount = (rmEnd - rmStart) + windowsNewLinesCounter + 1;
			info->mFinalString = (IzUniChar*) malloc(charCount * sizeof(IzUniChar));

			if (info->mFinalString != NULL) {

				const IzUniChar* src = rmStart;
				IzUniChar* dst = info->mFinalString;
	
				while (src < rmEnd) {

					if (*src == '\n') {
						if (info->mNewLineMode == F_WINDOWS) {
							*(dst++) = '\r';
							*(dst++) = '\n';
						} else if (info->mNewLineMode == F_MAC) {
							*(dst++) = '\r';
						} else if (info->mNewLineMode == F_UNIX) {
							*(dst++) = '\n';
						}
					} else {
						// copy current character, incrementing the dst and src pointers
						*(dst++) = *src;
					}

					src += 1;

				}

				*dst = '\0'; // add string ending char (as it is not always the case)


				// SET ALL OUTPUT PARAMETERS FOR FEEDBACK (except 'file available' as it should
				// always be updated, not only when the file could be read)

				// convert the unicode string back to UTF-8 so we can output it
				#if TARGET_OS_MAC
				char* utf8Output = RTFFStringToUTF8String(info->mFinalString);
				#else
				char* utf8Output = CreateUTF8StringWithIzzyUnicodeString_(ip, info->mFinalString);
				#endif
				
				// if that succeeds, then send the output value
				if (utf8Output != NULL) {
					Value vstr = { kString, 0 };			
					AllocateValueString_(ip, utf8Output, &vstr);
					SetOutputPropertyValue_(ip, info->mActorInfoPtr, kOutputText, &vstr);
					ReleaseValueString_(ip, &vstr);
					// dispose the allocated UTF-8 string
					IzzyFree_(ip, utf8Output);
				}

				// if we were able to open the file, then send the 
				if (info->mFileOK) {
					Value vtrig = { kBoolean, 0 };
					vtrig.u.ivalue = 1;
					SetOutputPropertyValue_(ip, info->mActorInfoPtr, kOutputFileWasRead, &vtrig);
				}

			}
		}


	} // end if fopen is successful
	
	// UPDATE 'file found' output
	Value vbool = { kBoolean, 0 };
	vbool.u.ivalue = info->mFileOK;
	SetOutputPropertyValue_(ip, info->mActorInfoPtr, kOutputFileWasFound, &vbool);

	// UPDATE 'line ending type' output
	Value vint = { kInteger, 0 };
	vint.u.ivalue = (int) info->mLineEndingType;
	SetOutputPropertyValue_(ip, info->mActorInfoPtr, kOutputFileLineEndingType, &vint);

}




// ---------------------------------------------------------------------------------
//		 HandlePropertyChangeValue	[INTERRUPT SAFE]
// ---------------------------------------------------------------------------------
//	### This function is called whenever one of the input values of an actor changes.
//	The one-based property index of the input is given by inPropertyIndex1.
//	The new value is given by inNewValue, the previous value by inOldValue.
//
static void
HandlePropertyChangeValue(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	PropertyIndex		inPropertyIndex1,			// the one-based index of the property than changed values
	ValuePtr			/* inOldValue */,			// the property's old value
	ValuePtr			inNewValue,					// the property's new value
	Boolean				/* inInitializing */)		// true if the value is being set when an actor is first initalized
{
	PluginInfo* info = GetPluginInfo_(inActorInfo);

	// ### When you add/change/remove properties, you will need to add cases
	// to this switch statement, to process the messages for your
	// input properties
	
	// The value comes to you encapsulated in a Value structure. See 
	// ValueCommon.h for details about the contents of this structure.
	
	Value kOutTextValue = { kString, 0 }; // Output text as a Value
	Value kFileIsAccessibleValue = { kBoolean, 0 }; // 'File is accessible' boolean Value
	
	switch (inPropertyIndex1) {

	case kInputFilePath: // Input path changed
		
		/*
		Paths beginning with white spaces and/or linebreaks will bugs later on while trying to read the file.
		Therefore we should clean this path by trimming all white spaces and linebreaks at the beginning of the input path.

		(totally empty input path string will be tested and ignored).

		This operation should be done here so that the user gets feedback directly into his input property box.
		*/

		// Find the real beginning of the string (skip white spaces and linebreaks)

		if (inNewValue->type == kString && inNewValue->u.str != NULL) {

			if (info->mPath != NULL) {
				delete info->mPath;
				info->mPath = NULL;
			}

			// skip leading whitespace
			const char* firstPathChar = inNewValue->u.str->strData;
			while (*firstPathChar != 0 && std::isspace(*firstPathChar)) {
				firstPathChar++;
			}

			// allocate new std::string with the path
			info->mPath = new std::string(firstPathChar);
		}

		// if auto-read is on, launch read function
		if (info->mAutoRead)
			ReadTextFile(ip, inActorInfo, info);	// IS-1009

		break;


	// linebreak mode changed
	case kInputLinebreakOut:

		// set new 'new line' mode (should not be necessary, but just in 'case' ^^)
		// and corresponding linebreak char (this is necessary)
		switch (inNewValue->u.ivalue) {
			
		case F_WINDOWS:
			info->mNewLineMode = F_WINDOWS;
			break;
		case F_UNIX:
			info->mNewLineMode = F_UNIX;
				break;
		case F_MAC:
			info->mNewLineMode = F_MAC;
			break;
		}
		break;


	case kInputReadMode: // Read mode changed

		// set new read mode (should not be necessary, but just in "case" ^^)
		switch (inNewValue->u.ivalue) {
		case F_ALL: info->mReadMode = F_ALL; break;
		case F_CHAR: info->mReadMode = F_CHAR; break;
		case F_LINE: info->mReadMode = F_LINE; break;
		case F_PARAGRAPH: info->mReadMode = F_PARAGRAPH; break;
		}

		// If auto-read is on, launch read function
		if (info->mAutoRead)
			ReadTextFile(ip, inActorInfo, info);	// IS-1009

		break;

	// starting char/line/paragraph changed
	case kInputStart:

		// set new section start for reading
		info->mStart = (int) inNewValue->u.ivalue;

		// If auto-read is on, launch read function
		if (info->mAutoRead)
			ReadTextFile(ip, inActorInfo, info);	// IS-1009

		break;

	// auto-read status changed
	case kInputAutoRead:

		// set new section start for reading
		if (inNewValue->u.ivalue == 0)
			info->mAutoRead = false;
		else
			info->mAutoRead = true;

		// Launch read function it auto-read is now on
		if (info->mAutoRead)
			ReadTextFile(ip, inActorInfo, info);	// IS-1009

		break;


	// reset output is triggered
	case kInputResetOutput:

		// update 'file available' output property (set to off)
		kFileIsAccessibleValue.u.ivalue = false;
		SetOutputPropertyValue_(ip, info->mActorInfoPtr, kOutputFileWasFound, &kFileIsAccessibleValue);

		// clean output text property
		
		AllocateValueString_(ip, "", &kOutTextValue);
		SetOutputPropertyValue_(ip, info->mActorInfoPtr, kOutputText, &kOutTextValue);
		ReleaseValueString_(ip, &kOutTextValue);

		break;

	// reading is initiated
	case kInputTriggerRead:

		// launch read function
		ReadTextFile(ip, inActorInfo, info);	// IS-1009

		break;
	}
}


// ---------------------------------------------------------------------------------
//		 GetActorDefinedArea
// ---------------------------------------------------------------------------------
//	If the mGetActorDefinedAreaProc in the ActorInfo struct points to this function,
//	it indicates to Isadora that the object would like to draw either an icon or else
//	an graphic representation of its function.
//
//	### This function uses the 'PICT' 0 resource stored with the plugin to draw an icon.
//  You should replace this picture (located in the Plugin Resources.rsrc file) with
//  the icon for your actor.
// 
static ActorPictInfo	gPictInfo = { false, nil, nil, 0, 0 };

static Boolean
GetActorDefinedArea(
	IsadoraParameters*			ip,			
	ActorInfo*					inActorInfo,
	SInt16*						outTopAreaWidth,			// returns the width to reserve for the top Actor Defined Area
	SInt16*						outTopAreaMinHeight,		// returns the minimum height of the top area
	SInt16*						outBotAreaHeight,			// returns the width to reserve for the bottom Actor Defined Area
	SInt16*						outBotAreaMinWidth)			// returns the minimum width of the bottom area
{
	if (!gPictInfo.mInitialized) {
		// PrepareActorDefinedAreaPict_(ip, inActorInfo, 0, &gPictInfo);
	}
	
	// place picture in top area
	*outTopAreaWidth = gPictInfo.mWidth;
	*outTopAreaMinHeight = gPictInfo.mHeight;
	
	// don't draw anything in bottom area
	*outBotAreaHeight = 0;
	*outBotAreaMinWidth = 0;
	
	return true;
}

// ---------------------------------------------------------------------------------
//		 DrawActorDefinedArea
// ---------------------------------------------------------------------------------
//	If GetActorDefinedArea is defined, then this function will be called whenever
//	your ActorDefinedArea needs to be drawn.
//
//	Beacuse we are using the PICT 0 resource stored with this plugin, we can use
//	the DrawActorDefinedAreaPict_ supplied by the Isadora callbacks.
//
//  DrawActorDefinedAreaPict_ is Alpha Channel aware, so you can have nice
//	shading if you like.

static void
DrawActorDefinedArea(
	IsadoraParameters*			ip,
	ActorInfo*					inActorInfo,
	void*						/* inDrawingContext */,		// unused at present
	ActorDefinedAreaPart		inActorDefinedAreaPart,		// the part of the actor that needs to be drawn
	ActorAreaDrawFlagsT			/* inAreaDrawFlags */,		// actor draw flags
	Rect*						inADAArea,					// rect enclosing the entire Actor Defined Area
	Rect*						/* inUpdateArea */,			// subset of inADAArea that needs updating
	Boolean						inSelected)					// TRUE if actor is currently selected
{
	if (inActorDefinedAreaPart == kActorDefinedAreaTop && gPictInfo.mInitialized) {
		DrawActorDefinedAreaPict_(ip, inActorInfo, inSelected, inADAArea, &gPictInfo);
	}
}
	
// ---------------------------------------------------------------------------------
//		 GetActorInfo
// ---------------------------------------------------------------------------------
//	This is function is called by to get the actor's class and ID, and to get
//	pointers to the all of the plugin functions declared locally.
//
//	All members of the ActorInfo struct pointed to by outActorParams have been
//	set to 0 on entry. You only need set functions defined by your plugin
//	
EXPORT_ void
GetActorInfo(
	void*				/* inParam */,
	ActorInfo*			outActorParams)
{
	// REQUIRED information
	outActorParams->mActorName							= kActorName;
	outActorParams->mClass								= kActorClass;
	outActorParams->mID									= kActorID;
	outActorParams->mCompatibleWithVersion				= kActorCompatilibilityVersion_v1;
	outActorParams->mActorFlags							= kActorFlags_Plugin_CheckForUpdates;
	
	// REQUIRED functions
	outActorParams->mGetActorParameterStringProc		= GetParameterString;
	outActorParams->mGetActorHelpStringProc				= GetHelpString;
	outActorParams->mCreateActorProc					= CreateActor;
	outActorParams->mDisposeActorProc					= DisposeActor;
	outActorParams->mActivateActorProc					= ActivateActor;
	outActorParams->mHandlePropertyChangeValueProc		= HandlePropertyChangeValue;
	
	// OPTIONAL FUNCTIONS
	outActorParams->mHandlePropertyChangeTypeProc		= NULL;
	outActorParams->mHandlePropertyConnectProc			= NULL;
	outActorParams->mPropertyValueToStringProc			= PropertyValueToString;	// For read mode and linebreak mode input properties
	outActorParams->mPropertyStringToValueProc			= PropertyStringToValue;
	outActorParams->mGetActorDefinedAreaProc			= GetActorDefinedArea;
	outActorParams->mDrawActorDefinedAreaProc			= DrawActorDefinedArea;
	outActorParams->mMouseTrackInActorDefinedAreaProc	= NULL;
}

// ---------------------------------------------------------------------------------
//		 ProcessVideoFrame
// ---------------------------------------------------------------------------------
//	### This is the code that does the actual processing of a video frame. Modify
// this code to create your own filter.
//


// ---------------------------------------------------------------------------------
//		 ReceiveMessage
// ---------------------------------------------------------------------------------
//	Isadora broadcasts messages to its Message Receives depending on what message
//	they are listening to. In this case, we are listening for kWantVideoFrameTick,
//	which is broadcast periodically (30 times per second.) When we receive the
//	message, we check to see if our video frame needs to be updated. If so, we
//	process the incoming video and pass the newly generated frame to the output.

