// ===========================================================================
//	Isadora Demo Plugin			(c) 2021 Mark F. Coniglio. All rights reserved
// ===========================================================================

/*

# ISADORA DEMO PLUGIN - READ ME

## DESCRIPTION

This plugin demonstrations a simple video effect implemented both as an
OpenGL shader. The shader provided simply modulates the red, green, and blue
components of the source video stream as found in the "Colorizer" actor that
comes built-in with Isdaora. To perform a different effect, you simply need
to replace the shader under the heading "COLORIZER SHADER"

## ABOUT VIDEO TYPES

In Isadora, video data comes in three "flavors":

- **vid-gpu**: Image is stored as an OpenGL texture on your graphics card.
Image data is processed using OpenGL shaders

- **vid-ci**: Image is stored as
a CIImage on your graphics card (macOS only). Image data is processed using
OpenGL shaders

- **vid-cpu**: Image is stored in the main memory of your
computer. image data is processed using standard C++.  format is only used by
legacy Isadora plugins, usually marked by the word classic" in their name.

This plugin only handles 'vid-gpu' streams only, as this is by far the most
common video format used by Isadora. To see a plugin that offers a mutable
input that can receive all three flavors, please take a look at the Analyze
Image Demo project within the Isadora SDK.

## CUSTOMIZING THIS SOURCE CODE

To customize this file, search for the text ###. All of the places where you
will need to customize the file are marked with this pattern of characters.

## LICENSE

IMPORTANT: This source code ("the software") is supplied to you in
consideration of your agreement to the following terms. If you do not agree
to the terms, do not install, use, modify or redistribute the software.

Mark Coniglio (aka TroikaTronix) grants you a personal, non exclusive license
to use, reproduce, modify this software with and to redistribute it, with or
without modifications, in source and/or binary form. Except as expressly
stated in this license, no other rights are granted, express or implied, to
you by TroikaTronix.

This software is provided on an "AS IS" basis. TroikaTronix makes no
warranties, express or implied, including without limitation the implied
warranties of non-infringement, merchantability, and fitness for a
particular purpurse, regarding this software or its use and operation alone
or in combination with your products.

In no event shall TroikaTronix be liable for any special, indirect,
incidental, or consequential damages arising in any way out of the use,
reproduction, modification and/or distribution of this software.

*/

// ---------------------------------------------------------------------------------
// INCLUDES
// ---------------------------------------------------------------------------------

// ISADORA INCLUDES
// These three includes should be present at the top of every Isadora plugin.

#include "IsadoraPluginPrefix.h"
#include "IsadoraTypes.h"
#include "IsadoraCallbacks.h"

// PLUGIN SPECIFIC INCLUDES
// required for plugins that process GPU and CoreImage based image data
#include "MultiVideoInOutUtil.h"
#include "CGLShader.h"

// GPU: useful for plugins that process CPU based image data
#include "PluginDrawUtil.h"

// STANDARD INCLUDES
#include <string.h>
#include <stdio.h>

// ---------------------------------------------------------------------------------
// MacOS Specific
// ---------------------------------------------------------------------------------
#if TARGET_OS_MAC
#define EXPORT_
#endif

// ---------------------------------------------------------------------------------
// Win32  Specific
// ---------------------------------------------------------------------------------
#if TARGET_OS_WIN	
	#define EXPORT_ __declspec(dllexport)
#endif

// ---------------------------------------------------------------------------------
//	Exported Function Definitions
// ---------------------------------------------------------------------------------

#ifdef __cplusplus
extern "C" {
#endif

EXPORT_ void GetActorInfo(void* inParam, ActorInfo* outActorParams);

#ifdef __cplusplus
}
#endif

#include "IsadoraCallbacks.h"

#include <string.h>
#include <stdio.h>

using namespace std;

// ---------------------------------------------------------------------------------
// CONSTANTS
// ---------------------------------------------------------------------------------
const SInt16	kVolPanMatrixIndicatorWidthPixels	= 41;
const SInt16	kVolPanIndicatorHeightPixels		= 41;
const IzCoord	kVolPanMatrixMargin					= 4;
const SInt16	kMatrixIndicatorHeightPixels		= 53;

const float		kMinVolume							= 0.0;
const float		kMaxVolume							= 200.0;
const float		kMinPan								= 0.0;
const float		kMaxPan								= 100.0;

enum {
	// kDrawLoopPosition = 1,
	// kDrawPlayPosition = 2,
	kDrawVolumePan = 4,
	kDrawVolumeMatrix = 8
};

// ---------------------------------------------------------------------------------
//	FORWARD DECLARTIONS
// ---------------------------------------------------------------------------------
static void
ReceiveMessage(
	IsadoraParameters*	ip,
	MessageMask			inMessageMask,
	PortIndex			inPortIndex1,
	const MsgData*		inData,
	UInt32				inLen,
	IzRefCon			inRefCon);

// ---------------------------------------------------------------------------------
// GLOBAL VARIABLES
// ---------------------------------------------------------------------------------
// ### Declare global variables, common to all instantiations of this plugin here

// Example: static int gMyGlobalVariable = 5;

// ---------------------------------------------------------------------------------
// PluginInfo struct
// ---------------------------------------------------------------------------------
// ### This structure neeeds to contain all variables used by your plugin. Memory for
// this struct is allocated during the CreateActor function, and disposed during
// the DisposeActor function, and is private to each copy of the plugin.
//
// If your plugin needs global data, declare them as static variables within this
// file. Any static variable will be global to all instantiations of the plugin.

typedef struct {

	ActorInfo*				mActorInfoPtr;
	MessageReceiverRef		mMessageReceiver;
	UInt32					mNextSampleRateCheck_mS;

	ADAAudioControlsInfo	mAudioCtlInfo;

	IzValue_Int				mReqSampleRate;
	IzValue_Int				mReqChannels;
	IzValue_Int				mActualSampleRate;
	IzValue_Int				mActualChannels;
	IzValue_Flt				mFrequency;
	IzValue_Flt				mVolume;
	IzValue_Flt				mPan;
	std::string*			mAudioRouting;
	double					mPhase;
	bool					mEnabled;
	
	IzzyAudioStreamRef		mAudioStreamRef;
	
} PluginInfo;

// A handy macro for casting the mActorDataPtr to PluginInfo*
#if __cplusplus
#define	GetPluginInfo_(actorDataPtr)		static_cast<PluginInfo*>((actorDataPtr)->mActorDataPtr);
#else
#define	GetPluginInfo_(actorDataPtr)		(PluginInfo*)((actorDataPtr)->mActorDataPtr);
#endif

// ---------------------------------------------------------------------------------
//	Constants
// ---------------------------------------------------------------------------------
//	Defines various constants used throughout the plugin.

// CONSTANTS FOR GPU VERSION

// ### GROUP ID
// Define the group under which this plugin will be displayed in the Isadora interface.
// These are defined under "Actor Types" in IsadoraTypes.h

static const OSType	kActorClass 	= kGroupSound;

// ### PLUGIN IN
// Define the plugin's unique four character identifier. Contact TroikaTronix to
// obtain a unique four character code if you want to ensure that someone else
// has not developed a plugin with the same code. Note that TroikaTronix reserves
// all plugin codes that begin with an underline, an at-sign, and a pound sign
// (e.g., '_', '@', and '#'.)

static const OSType	kActorID		= FOUR_CHAR_CODE('_TGD');

// ### ACTOR NAME
// The name of the actor. This is the name that will be shown in the User Interface.

static const char* kActorName		= "Tone Geneator Demo";

// ### PROPERTY DEFINITION STRING
// The property string. This string determines the inputs and outputs for your plugin.
// See the IsadoraCallbacks.h under the heading "PROPERTY DEFINITION STRING" for the
// meaning ofthese codes. (The IsadoraCallbacks.h header can be seen by opening up
// the IzzySDK Framework while in the Files view.)
//
// IMPORTANT: You cannot use spaces in the property name. Instead, use underscores (_)
// where you want to have a space.
//
// Note that each line ends with a carriage return (\r), and that only the last line of
// the bunch ends with a semicolon. This means that what you see below is one long
// null-terminated c-string, with the individual lines separated by carriage returns.

static const char* sPropertyDefinitionString =

// INPUT PROPERTY DEFINITIONS
//	TYPE 	PROPERTY NAME	ID		DATATYPE	DISPLAY FMT			MIN		MAX		INIT VALUE
		"INPROP		sample_rate		samp	int		number				0		192000	48000\r"
		"INPROP		channels		chan	int		number				0		16		1\r"
		"INPROP		frequency		freq	float	number				30		88100	1000\r"
		"INPROP		volume			volu	float	number				0		100		25\r"
		"INPROP		pan				pann	float	number				0		100		50\r"
		"INPROP		audio_routing	rout	string	text				*		*		\r"
		"INPROP		enable			enab	int		onoff				0		1		0\r";

// OUTPUT PROPERTY DEFINITIONS
//	TYPE 	 PROPERTY NAME	ID		DATATYPE	DISPLAY FMT			MIN		MAX		INIT VALUE

	// NO OUTPUTS FOR THIS PLUGIN!

// ### Property Index Constants
// Properties are referenced by a one-based index. The first input property will
// be 1, the second 2, etc. Similarly, the first output property starts at 1.
// You whould have one constant for each input and output property defined in the 
// property definition string.

enum
{
	kInputSampleRate = 1,
	kInputChannels,
	kInputFrequency,
	kInputVolume,
	kInputPan,
	kInputAudioRouting,
	kInputEnable
};


// ---------------------
//	Help String
// ---------------------
// ### Help Strings
//
// The first help string is for the actor in general. This followed by help strings
// for all of the inputs, and then by the help strings for all of the outputs. These
// should be given in the order that they are defined in the Property Definition
// String above.
//
// In all, the total number of help strings should be (num inputs + num outputs + 1)
//
// Note that each string is followed by a comma -- it is a common mistake to forget the
// comma which results in the two strings being concatenated into one.

const char* sHelpStrings[] =
{
	"Sends a test tone to the audio output device currently selected in the Audio View of this Isadora document.",
		
	"The sample rate used when generating the tone,specified as the number of samples per second. Set this to 0 to use "
		"the default sample rate for the audio output device currently selected in the Audio View of this Isadora "
		"document. Otherise, rates less than 11025 Hz will be limited to a minimum of 11025.",
	
	"The number of channels used when generating the tone. Set this to 0 to use the default number of channels "
		"offered by the audio output device currently selected in the Audio View of this Isadora document.",
	
	"The frequency of the tone, specified in Hertz (cycles per second).",
	
	"The volume of the tone, from 0 to 100%.",
	
	"A string specifying the audio_routing from the source channel(s) in the received audio to the output channels on the "
		"audio output device currently selected in the Audio View of this Isadora document.\r"
		"You can edit this routing by clicking on this input, or by clicking in the matrix routing display shown in the "
		"center of this actor.\r"
		"When specifying the routing using a string, the general form is 's:d' where 's' is the source channel in the "
		"sound file and 'd' is the destination channel on the audio output. To add volume control, use use 's:d@n dB' "
		"to specify a volume increase or decrease in decibels (where n is is a positive or negative number), or 's:d@n%' "
		"to specfy a volume percentage from 0 to 100%. You can specify several routings by separating them by commas, "
		"e.g., 's1:d1, s2:d2@-3db', etc.",

	"The panning of the tone when routed to more than one output channel, specified as a percentage between "
		"0 and 100 percent, where 50% is centered, 0% is fully left, and 100% is fully right.",
		
	"When turned on, generates a tone. When turned off this actor is silent"
};

// ---------------------------------------------------------------------------------
//		GetAudioSamples
// ---------------------------------------------------------------------------------
// Callback function used to generate a sine wave at the frequency and volume
// specified by the mFrequency and mActualSampleRate members of the plugin info
// struct. The sample rate is controlled by the mActualSampleRate member.

void
GetAudioSamples(
	ActorInfo*				inActorInfo,
	void*					/* inRefCon */,
	IzzyAudioNotification*	inNotif)
{
	PluginInfo* info = GetPluginInfo_(inActorInfo);

	if (inNotif->mNotificationVersion <= IzzyAudioNotificationVersion_Current) {
	
		if (inNotif->mNotificationMsg == kIzzyAudio_GetSamples) {
		
			// put phase in local variable.
			double phase = info->mPhase;
			float* samples = inNotif->mGetSamples.mSampleBuf;

			for (unsigned int i=0; i<inNotif->mGetSamples.mSamplesRequested; i++) {
				
				// get the value for the sine wave
				double value = info->mVolume * sin(phase);
				
				// increment the phase based on the frequency and sample rate
				phase += (2.0 * M_PI) * (info->mFrequency / info->mActualSampleRate);
				
				// use this loop to write the current sample value
				// for all channels in the non-interleaved format
				float* dstSample = samples;
				for (uint32_t i=0; i<info->mActualChannels; i++) {
					*dstSample = value;
					dstSample += inNotif->mGetSamples.mChannelOffset;
				}
				// increment the sample pointer by one
				samples++;
			}
			
			inNotif->mGetSamples.mSamplesWritten = inNotif->mGetSamples.mSamplesRequested;
			
			// store phase for next callback
			info->mPhase = phase;
		}
	}
}

// ---------------------------------------------------------------------------------
//	CADAEditUndo CLASS
// ---------------------------------------------------------------------------------
// The CADAEditUndo class contains undo and redo information for the volume and
// pan inputs of this actor. The class is created by the DACreateUndoRedoData
// callback function which is passed to the ADAAudioControls_AttemptToHandleClick_
// function.

class CADAEditUndo {
public:

	enum EditCode {
		kEditVolumeAndPan
	};
	
	CADAEditUndo(
		EditCode			inCode,				// should only be kEditVolumeAndPan
		PropertyIndex		inPropertyIndex1,
		const Value&		inOldValue1,
		const Value&		inNewValue1,
		PropertyIndex		inPropertyIndex2,
		const Value&		inOldValue2,
		const Value&		inNewValue2) :
			mEditCode(inCode),
			mPropertyIndex1(inPropertyIndex1),
			mOldValue1(inOldValue1),
			mNewValue1(inNewValue1),
			mPropertyIndex2(inPropertyIndex2),
			mOldValue2(inOldValue2),
			mNewValue2(inNewValue2)
	{
	}

	EditCode		mEditCode;
	PropertyIndex	mPropertyIndex1;
	Value			mOldValue1;
	Value			mNewValue1;
	PropertyIndex	mPropertyIndex2;
	Value			mOldValue2;
	Value			mNewValue2;
};

// ---------------------------------------------------------------------------------
//		DACreateUndoRedoData
// ---------------------------------------------------------------------------------
// This callback function creates the private undo/redo data for changes made to
// the pan/volume control.

static void*
DACreateUndoRedoData(
	ActorInfo*					/* inActorInfo */,
	const ADAAudioControlsInfo*	inAudioCtlInfo,
	void*						/* inRefCon */)
{
	CADAEditUndo* data = new CADAEditUndo(CADAEditUndo::kEditVolumeAndPan, inAudioCtlInfo->mVolumeInputPropertyIndex1, inAudioCtlInfo->mInitialVolume, inAudioCtlInfo->mNewVolume, inAudioCtlInfo->mPanInputPropertyIndex1, inAudioCtlInfo->mInitialPan, inAudioCtlInfo->mNewPan);
	
	return data;
}

// ---------------------------------------------------------------------------------
//		DisposeAudioStream
// ---------------------------------------------------------------------------------
// Disposes any existing audio stream;

static void
DisposeAudioStream(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo)
{
	PluginInfo* info = GetPluginInfo_(inActorInfo);
	if (info->mAudioStreamRef) {
		IzzyAudio_DisposeAudioStream_(ip, info->mAudioStreamRef);
		info->mAudioStreamRef = NULL;
		inActorInfo->mDrawActorAreaFlags |= kDrawVolumePan | kDrawVolumeMatrix;
	}
}

// ---------------------------------------------------------------------------------
//		CreateAudioStream
// ---------------------------------------------------------------------------------
// Creates the audio stream using the sample rate, number of channels channels,
// volume and pan specified by the associated members in the PluginInfo struct

static void
CreateAudioStream(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo)
{
	DisposeAudioStream(ip, inActorInfo);

	PluginInfo* info = GetPluginInfo_(inActorInfo);

	info->mAudioStreamRef = IzzyAudio_CreateAudioStream_(ip, inActorInfo, GetAudioSamples, NULL, info->mActualChannels, info->mActualSampleRate, info->mVolume, info->mPan, TRUE);
	inActorInfo->mDrawActorAreaFlags |= kDrawVolumePan | kDrawVolumeMatrix;
}

// ---------------------------------------------------------------------------------
//		 CreateActor
// ---------------------------------------------------------------------------------
// Called once, prior to the first activation of an actor in its Scene. The
// corresponding DisposeActor actor function will not be called until the file
// owning this actor is closed, or the actor is destroyed as a result of being
// cut or deleted.

static void
CreateActor(
	IsadoraParameters*	ip,	
	ActorInfo*			ioActorInfo)		// pointer to this actor's ActorInfo struct - unique to each instance of an actor
{
	// creat the PluginInfo struct - initializing it to all zeroes
	PluginInfo* info = (PluginInfo*) IzzyMallocClear_(ip, sizeof(PluginInfo));
	PluginAssert_(ip, info != NULL);
	
	ioActorInfo->mActorDataPtr = info;
	info->mActorInfoPtr = ioActorInfo;

	// ### allocation and initialization of private member variables
	
}

// ---------------------------------------------------------------------------------
//		 DisposeActor
// ---------------------------------------------------------------------------------
// Called when the file owning this actor is closed, or when the actor is destroyed
// as a result of its being cut or deleted.
//
static void
DisposeActor(
	IsadoraParameters*	ip,
	ActorInfo*			ioActorInfo)		// pointer to this actor's ActorInfo struct - unique to each instance of an actor
{
	PluginInfo* info = GetPluginInfo_(ioActorInfo);
	PluginAssert_(ip, info != NULL);
	
	// ### destruction of private member variables
	
	// destroy the PluginInfo struct allocated with IzzyMallocClear_ the CreateActor function
	PluginAssert_(ip, ioActorInfo->mActorDataPtr != NULL);
	IzzyFree_(ip, ioActorInfo->mActorDataPtr);
}

// ---------------------------------------------------------------------------------
//		 ActivateActor
// ---------------------------------------------------------------------------------
//	Called when the scene that owns this actor is activated or deactivated. The
//	inActivate flag will be true when the scene is activated, false when deactivated.
//
static void
ActivateActor(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,		// pointer to this actor's ActorInfo struct - unique to each instance of an actor
	Boolean				inActivate)			// true when actor is becoming active, false otherwise.
{
	PluginInfo* info = GetPluginInfo_(inActorInfo);
	
	if (inActivate) {
		
		info->mMessageReceiver = CreateActorMessageReceiver_(
			ip,
			inActorInfo,
			ReceiveMessage,
			0,
			kWantIzzyTick,
			PtrToIzRefCon(inActorInfo));
		info->mNextSampleRateCheck_mS = 0;
		
		info->mAudioRouting = new std::string();

		memset(&info->mAudioCtlInfo, 0, sizeof(info->mAudioCtlInfo));
		
		info->mAudioCtlInfo.mVolumePanWidgetHeight = kVolPanIndicatorHeightPixels;
		info->mAudioCtlInfo.mMatrixTopOffset = kVolPanMatrixMargin;
		info->mAudioCtlInfo.mMatrixHeight = kMatrixIndicatorHeightPixels;
		info->mAudioCtlInfo.mVolumeInputPropertyIndex1 = kInputVolume;
		info->mAudioCtlInfo.mPanInputPropertyIndex1 = kInputPan;
		
		info->mAudioCtlInfo.mMinVolume = kMinVolume;
		info->mAudioCtlInfo.mMaxVolume = kMaxVolume;
		info->mAudioCtlInfo.mMinPan = kMinPan;
		info->mAudioCtlInfo.mMaxPan = kMaxPan;
		
		info->mAudioCtlInfo.mRoutingPropertyType = kInputProperty;
		info->mAudioCtlInfo.mRoutingPropertyIndex1 = kInputAudioRouting;

		if (info->mEnabled) {
			CreateAudioStream(ip, inActorInfo);
		} else {
			info->mAudioStreamRef = NULL;
		}
	} else {
	
		// dispose our any lingering audio stream
		DisposeAudioStream(ip, inActorInfo);
		
		if (info->mAudioRouting != NULL) {
			delete info->mAudioRouting;
			info->mAudioRouting = NULL;
		}

		if (info->mMessageReceiver != nil) {
			DisposeMessageReceiver_(ip, info->mMessageReceiver);
			info->mMessageReceiver = nil;
		}

		info->mReqSampleRate = 0;
		info->mReqChannels = 0;
		info->mActualSampleRate = 0;
		info->mActualChannels = 0;
		info->mFrequency = 0.0;
		info->mVolume = 0.0;
		info->mPan = 0.5;
		info->mPhase = 0;
		info->mEnabled = false;
		
		info->mAudioStreamRef = NULL;
	}
	
}

// ---------------------------------------------------------------------------------
//		 GetParameterString
// ---------------------------------------------------------------------------------
//	Returns the property definition string. Called when an instance of the actor
//	needs to be instantiated.

static const char*
GetParameterString(
	IsadoraParameters*	/* ip */,
	ActorInfo*			/* inActorInfo */)
{
	return sPropertyDefinitionString;
}

// ---------------------------------------------------------------------------------
//		 GetHelpString
// ---------------------------------------------------------------------------------
//	Returns the help string for a particular property. If you have a fixed number of
//	input and output properties, it is best to use the PropertyTypeAndIndexToHelpIndex_
//	function to determine the correct help string to return.

static void
GetHelpString(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	PropertyType		inPropertyType,			// kPropertyTypeInvalid when requesting help for the actor
												// or kInputProperty or kOutputProperty when requesting help for a specific property
	PropertyIndex		inPropertyIndex1,		// the one-based index of the property (when inPropertyType is not kPropertyTypeInvalid)
	char*				outParamaterString,		// receives the help string
	UInt32				inMaxCharacters)		// size of the outParamaterString buffer
{
	const char* helpstr = NULL;
	
	// The PropertyTypeAndIndexToHelpIndex_ converts the inPropertyType and
	// inPropertyIndex1 parameters to determine the zero-based index into
	// your list of help strings.
	UInt32 index1 = PropertyTypeAndIndexToHelpIndex_(ip, inActorInfo, inPropertyType, inPropertyIndex1);
	
	// get the help string
	helpstr = sHelpStrings[index1];
	
	// copy it to the output string
	strncpy(outParamaterString, helpstr, inMaxCharacters);
}
	
// ---------------------------------------------------------------------------------
//		 HandlePropertyChangeValue	[INTERRUPT SAFE]
// ---------------------------------------------------------------------------------
//	### This function is called whenever one of the input values of an actor changes.
//	The one-based property index of the input is given by inPropertyIndex1.
//	The new value is given by inNewValue, the previous value by inOldValue.
//
static void
HandlePropertyChangeValue(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	PropertyIndex		inPropertyIndex1,			// the one-based index of the property than changed values
	ValuePtr			/* inOldValue */,			// the property's old value
	ValuePtr			inNewValue,					// the property's new value
	Boolean				inInitializing)				// true if the value is being set when an actor is first initalized
{
	PluginInfo* info = GetPluginInfo_(inActorInfo);

	switch (inPropertyIndex1) {

	case kInputSampleRate:
	case kInputChannels:
		if (IsAtLeastIsadoraCallbackVersion(ip, kIsadoraCallbacksVersion_323b01)) {
			
			IzValue_Int oldSampleRate = info->mActualSampleRate;
			IzValue_Int oldChannels = info->mActualChannels;
			
			if (inPropertyIndex1 == kInputSampleRate) {
				info->mReqSampleRate = inNewValue->u.ivalue;
				if (info->mReqSampleRate == 0) {
					info->mActualSampleRate = IzzyAudio_GetSampleRate_(ip, inActorInfo);
				} else {
					info->mActualSampleRate = info->mReqSampleRate;
				}
			} else if (inPropertyIndex1 == kInputChannels) {
				info->mReqChannels = inNewValue->u.ivalue;
				if (info->mReqChannels == 0) {
					info->mActualChannels = IzzyAudio_GetSampleRate_(ip, inActorInfo);
				} else {
					info->mActualChannels = info->mReqChannels;
				}
			}
			info->mAudioCtlInfo.mMatrixSourceChannels = info->mActualChannels;
			inActorInfo->mDrawActorAreaFlags |= kDrawVolumePan | kDrawVolumeMatrix;
			// if we have an existing audio stream...
			if (info->mAudioStreamRef != NULL) {
				// and if we changed sample rate or channel count, we need to restart the audio
				if (oldSampleRate != info->mActualSampleRate || oldChannels != info->mActualChannels) {
					CreateAudioStream(ip, inActorInfo);
				}
			}
		}
		break;
		
	case kInputFrequency:
		info->mFrequency = inNewValue->u.fvalue;
		break;

	case kInputVolume:
		info->mVolume = inNewValue->u.fvalue/100.0;
		inActorInfo->mDrawActorAreaFlags |= kDrawVolumePan;
		if (info->mAudioStreamRef != NULL) {
			IzzyAudio_SetVolumeAndPan_(ip, info->mAudioStreamRef, info->mVolume, info->mPan);
		}
		break;

	case kInputPan:
		info->mPan = inNewValue->u.fvalue / 100.0;
		inActorInfo->mDrawActorAreaFlags |= kDrawVolumePan;
		if (info->mAudioStreamRef != NULL) {
			IzzyAudio_SetVolumeAndPan_(ip, info->mAudioStreamRef, info->mVolume, info->mPan);
		}
		break;

	case kInputAudioRouting:
		*info->mAudioRouting = inNewValue->u.str->strData;
		inActorInfo->mDrawActorAreaFlags |= kDrawVolumeMatrix;
		if (info->mAudioStreamRef != NULL) {
			IzzyAudio_SetRouting_(ip, info->mAudioStreamRef, info->mAudioRouting->c_str());
		}
		break;

	case kInputEnable:
		{
			if (IsAtLeastIsadoraCallbackVersion(ip, kIsadoraCallbacksVersion_323b01)) {
				// don't start up the sound during initialization... instead
				// let the ActivateActor() function handle it.
				if (!inInitializing) {
					if (inNewValue->u.ivalue != 0 && info->mAudioStreamRef == NULL) {
						CreateAudioStream(ip, inActorInfo);
					} else if (inNewValue->u.ivalue == 0 && info->mAudioStreamRef != NULL) {
						DisposeAudioStream(ip, inActorInfo);
					}
				}
			}
			info->mEnabled = (inNewValue->u.ivalue != 0);
		}
		break;

	default:
		PluginAssert_(ip, false);
		break;
	}
}

// ---------------------------------------------------------------------------------
//		 GetActorDefinedArea
// ---------------------------------------------------------------------------------
//	If the mGetActorDefinedAreaProc in the ActorInfo struct points to this function,
//	it indicates to Isadora that the object would like to draw either an icon or else
//	an graphic representation of its function.
//
//	### This function uses the 'PICT' 0 resource stored with the plugin to draw an icon.
//  You should replace this picture (located in the Plugin Resources.rsrc file) with
//  the icon for your actor.
// 
static ActorPictInfo	gPictInfo = { false, NULL, NULL, 0, 0 };

static Boolean
GetActorDefinedArea(
	IsadoraParameters*			ip,			
	ActorInfo*					inActorInfo,
	SInt16*						outTopAreaWidth,			// returns the width to reserve for the top Actor Defined Area
	SInt16*						outTopAreaMinHeight,		// returns the minimum height of the top area
	SInt16*						outBotAreaHeight,			// returns the width to reserve for the bottom Actor Defined Area
	SInt16*						outBotAreaMinWidth)			// returns the minimum width of the bottom area
{
	bool showedAudioMatrix = false;
	
	if (gPictInfo.mInitialized == false) {
		PrepareActorDefinedAreaPict_(ip, inActorInfo, 0, &gPictInfo);
	}
		
	if (IsAtLeastIsadoraCallbackVersion(ip, kIsadoraCallbacksVersion_323b01)) {
	
		*outTopAreaWidth = 0;
		*outTopAreaMinHeight = 0;
		*outBotAreaHeight = 0;
		*outBotAreaMinWidth = 0;

		*outTopAreaWidth = kVolPanMatrixIndicatorWidthPixels;
		*outTopAreaMinHeight = kVolPanIndicatorHeightPixels + 4 + kMatrixIndicatorHeightPixels;
		
		showedAudioMatrix = true;
	}
		
	if (!showedAudioMatrix) {
	
		// place picture in top area
		*outTopAreaWidth = gPictInfo.mWidth;
		*outTopAreaMinHeight = gPictInfo.mHeight;
		
		// don't draw anything in bottom area
		*outBotAreaHeight = 0;
		*outBotAreaMinWidth = 0;
	
	}
	
	return true;
}

// ---------------------------------------------------------------------------------
//		 DrawActorDefinedArea
// ---------------------------------------------------------------------------------
//	If GetActorDefinedArea is defined, then this function will be called whenever
//	your ActorDefinedArea needs to be drawn.
//
//	Beacuse we are using the PICT 0 resource stored with this plugin, we can use
//	the DrawActorDefinedAreaPict_ supplied by the Isadora callbacks.
//
//  DrawActorDefinedAreaPict_ is Alpha Channel aware, so you can have nice
//	shading if you like.

static void
DrawActorDefinedArea(
	IsadoraParameters*			ip,
	ActorInfo*					inActorInfo,
	void*						/* inDrawingContext */,		// unused at present
	ActorDefinedAreaPart		inActorDefinedAreaPart,		// the part of the actor that needs to be drawn
	ActorAreaDrawFlagsT			/* inAreaDrawFlags */,		// actor draw flags
	Rect*						inADAArea,					// rect enclosing the entire Actor Defined Area
	Rect*						/* inUpdateArea */,			// subset of inADAArea that needs updating
	Boolean						inSelected)					// TRUE if actor is currently selected
{
	if (inActorDefinedAreaPart == kActorDefinedAreaTop) {
	
		PluginInfo* info = GetPluginInfo_(inActorInfo);

		bool showedAudioMatrix = false;
		
		if (IsAtLeastIsadoraCallbackVersion(ip, kIsadoraCallbacksVersion_323b01)) {
		
			std::string routing;
			const ValuePtr v = GetInputPropertyValue_(ip, inActorInfo, kInputAudioRouting);
			if (v != NULL && v->type == kString && v->u.str != NULL) {
				routing = v->u.str->strData;
			}
		
			info->mAudioCtlInfo.mPlaying = (info->mAudioStreamRef != NULL);
			
			info->mAudioCtlInfo.mCurrentVolume = GetInputPropertyValue_(ip, inActorInfo, kInputVolume)->u.fvalue;
			info->mAudioCtlInfo.mCurrentPan = GetInputPropertyValue_(ip, inActorInfo, kInputPan)->u.fvalue;
			
			info->mAudioCtlInfo.mMatrixSourceChannels = info->mActualChannels;
			info->mAudioCtlInfo.mMatrixDestinationChannels = 0;
			info->mAudioCtlInfo.mRoutingString = routing.c_str();
			
			ADAAudioControls_Draw_(ip, inActorInfo, &info->mAudioCtlInfo, *inADAArea);
			showedAudioMatrix = true;
		}

		if (!showedAudioMatrix && gPictInfo.mInitialized) {
			DrawActorDefinedAreaPict_(ip, inActorInfo, inSelected, inADAArea, &gPictInfo);
		}
	}
}
	
// ---------------------------------------------------------------------------------
//		 MouseTrackInActorDefinedArea
// ---------------------------------------------------------------------------------
Boolean
MouseTrackInActorDefinedArea(
	IsadoraParameters*		ip,
	ActorInfo*				inActorInfo,
	IzzyMouseMessage		inMouseMessage,
	void*					inDrawingContext,
	ActorDefinedAreaPart	/* inActorDefinedAreaPart */,
	IzRect*					inArea,
	IzPoint					inMouse)
{
	Boolean handled = FALSE;
	if (IsAtLeastIsadoraCallbackVersion(ip, kIsadoraCallbacksVersion_323b01)) {
		PluginInfo* info = GetPluginInfo_(inActorInfo);
		handled = ADAAudioControls_AttemptToHandleClick_(ip, inActorInfo, &info->mAudioCtlInfo, *inArea, inMouseMessage, inMouse, DACreateUndoRedoData, NULL);
	}
	return handled;
}

// ---------------------------------------------------------------------------------
//		 HandleActorMessage
// ---------------------------------------------------------------------------------
void
HandleActorMessage(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	IzzyActorMessage	inActorMessage,
	void*				inUserData)
{
	CADAEditUndo* undoData = ((CADAEditUndo*) inUserData);
	
	switch (inActorMessage) {
	case kIzzyActorHandleUndo:
		if (undoData != NULL) {
			switch (undoData->mEditCode) {
			case CADAEditUndo::kEditVolumeAndPan:
				SetInputPropertyValue_(ip, inActorInfo, undoData->mPropertyIndex1, &undoData->mOldValue1);
				SetInputPropertyValue_(ip, inActorInfo, undoData->mPropertyIndex2, &undoData->mOldValue2);
				break;
			}
		}
		break;
	case kIzzyActorHandleRedo:
		if (undoData != NULL) {
			switch (undoData->mEditCode) {
			case CADAEditUndo::kEditVolumeAndPan:
				SetInputPropertyValue_(ip, inActorInfo, undoData->mPropertyIndex1, &undoData->mNewValue1);
				SetInputPropertyValue_(ip, inActorInfo, undoData->mPropertyIndex2, &undoData->mNewValue2);
				break;
			}
		}
		break;
	case kIzzyActorDisposeUndoRedo:
		if (undoData != NULL)
			delete undoData;
		break;
	}
}

// ---------------------------------------------------------------------------------
//		 GetActorInfo
// ---------------------------------------------------------------------------------
//	This is function is called by to get the actor's class and ID, and to get
//	pointers to the all of the plugin functions declared locally.
//
//	All members of the ActorInfo struct pointed to by outActorParams have been
//	set to 0 on entry. You only need set functions defined by your plugin
//	
EXPORT_ void
GetActorInfo(
	void*				/* inParam */,
	ActorInfo*			outActorParams)
{
	// REQUIRED information
	outActorParams->mActorName							= kActorName;
	outActorParams->mClass								= kActorClass;
	outActorParams->mID									= kActorID;
	outActorParams->mCompatibleWithVersion				= kActorCompatilibilityVersion_v323b01;
	
	// REQUIRED functions
	outActorParams->mGetActorParameterStringProc		= GetParameterString;
	outActorParams->mGetActorHelpStringProc				= GetHelpString;
	outActorParams->mCreateActorProc					= CreateActor;
	outActorParams->mDisposeActorProc					= DisposeActor;
	outActorParams->mActivateActorProc					= ActivateActor;
	outActorParams->mHandlePropertyChangeValueProc		= HandlePropertyChangeValue;
	
	// OPTIONAL FUNCTIONS
	outActorParams->mHandlePropertyChangeTypeProc		= NULL;
	outActorParams->mHandlePropertyConnectProc			= NULL;
	outActorParams->mPropertyValueToStringProc			= NULL;
	outActorParams->mPropertyStringToValueProc			= NULL;
	outActorParams->mGetActorDefinedAreaProc			= GetActorDefinedArea;
	outActorParams->mDrawActorDefinedAreaProc			= DrawActorDefinedArea;
	outActorParams->mMouseTrackInActorDefinedAreaProc	= MouseTrackInActorDefinedArea;
	
	outActorParams->mHandleActorMessageProc				= HandleActorMessage;
}

// ---------------------------------------------------------------------------------
//		 ReceiveMessage
// ---------------------------------------------------------------------------------
//	Handles incoming video frame changed messages

static void
ReceiveMessage(
	IsadoraParameters*	ip,
	MessageMask			inMessageMask,
	PortIndex			/* inIndex1 */,
	const MsgData*		inData,
	UInt32				/* inLen */,
	IzRefCon			inRefCon)
{
	// Grab the actor info and state of the leap
	ActorInfo* inActorInfo = (ActorInfo*) inRefCon;
	PluginInfo* info = GetPluginInfo_(inActorInfo);
	
	if (ip->mMillisecondCount > info->mNextSampleRateCheck_mS) {
	
		info->mNextSampleRateCheck_mS = ip->mMillisecondCount + 25;
	
		if (info->mReqSampleRate == 0 || info->mReqChannels == 0) {
		
			UInt32 curSampleRate = info->mReqSampleRate;
			if (curSampleRate == 0)
				curSampleRate = IzzyAudio_GetSampleRate_(ip, inActorInfo);
			
			UInt32 curChannels = info->mReqChannels;
			if (curChannels == 0)
				curChannels = IzzyAudio_GetOutputChannels_(ip, inActorInfo);
						
			if (info->mActualSampleRate != curSampleRate || info->mActualChannels != curChannels) {
			
				inActorInfo->mDrawActorAreaFlags |= kDrawVolumePan | kDrawVolumeMatrix;
				
				if (info->mAudioStreamRef != NULL) {
					info->mActualSampleRate = curSampleRate;
					info->mActualChannels = curChannels;
					CreateAudioStream(ip, inActorInfo);
				}
			}
		}
	}
}
