// ===========================================================================
//	IsadoraPlugin				   2001 Mark F. Coniglio. All rights reserved.
// ===========================================================================

// STANDARD INCLUDES
#include <string.h>
#include <stdio.h>
#include <stdint.h>
#include <float.h>
#include <ctype.h>
#include <vector>
#include <string>
#include <fstream>
#include <mutex>
#include <thread>


// INCLUDE JSON
#include "nlohmann/json.hpp"
using json = nlohmann::json;

// ISADORA INCLUDES
#include "IsadoraPluginPrefix.h"
#include "IsadoraCallbacks.h"

// ---------------------------------------------------------------------------------
//	OSC MESSAGE HELPER FUNCTIONS
// ---------------------------------------------------------------------------------

#define STRING_ALIGN_PAD 4

static const char* OSCDataAfterAlignedString(const char *string, const char *boundary, bool& outError)
{
	int i;

	if ((boundary - string) %4 == 0) {
		for (i = 0; string[i] != '\0'; i++) {
			if (string + i >= boundary) {
				// we have reached the boundary for this message without
				// completing the string - so we bail with an error
				outError = true;
				return NULL;
			}
		}

		// string -> first NULL character
		// there may be more due to padding
		i++;

		for (; (i % STRING_ALIGN_PAD) != 0; i++) {
		
			if (string + i >= boundary) {
				// we have reached the boundary for this message without
				// completing the string - so we bail with an error
				outError = true;
				return NULL;
			}
			
			if (string[i] != '\0') {
				// the string does not end with a null terminator
				// so we bail with an error
				outError = true;
				return NULL;
			}
		}
	}

	return (string+i);
}

// get an four-byte float from the OSC message
static inline const char*
GetOSCFloat(
	const char*	inPtr,
	const char*	inBoundary,
	Float32*	outValue,
	bool&		outError)
{
	outError = false;
	if (inPtr + sizeof(Float32) <= inBoundary) {
		*((UInt32*)outValue) = EndianU32_BtoN(*((UInt32*) inPtr));
		inPtr += sizeof(float);
		return inPtr;
	} else {
		outError = true;
		return NULL;
	}
}

// get an eight-byte double from the OSC message
static inline const char*
GetOSCDouble(
	const char*	inPtr,
	const char*	inBoundary,
	Float64*	outValue,
	bool&		outError)
{
	if (inPtr + sizeof(Float64) <= inBoundary) {
		*((UInt64*)outValue) = EndianU64_BtoN(*((UInt64*) inPtr));
		inPtr += sizeof(float);
		return inPtr;
	} else {
		outError = true;
		return NULL;
	}
}

// get an four-byte integer from the OSC message
static inline const char*
GetOSCInt(
	const char*	inPtr,
	const char*	inBoundary,
	SInt32*		outValue,
	bool&		outError)
{
	if (inPtr + sizeof(UInt32) <= inBoundary) {
		*((SInt32*)outValue) = EndianU32_BtoN(*((UInt32*) inPtr));
		inPtr += sizeof(float);
		return inPtr;
	} else {
		outError = true;
		return NULL;
	}
}

// get a multi-byte texty string from the OSC message
static const char*
GetOSCString(
	const char*		inPtr,
	const char*		inBoundary,
	std::string*	outString,
	bool&			outError)
{
	// remember start of the string
	const char* sptr = inPtr;
	// skip through the string, checking the padding and boundary
	inPtr = OSCDataAfterAlignedString(inPtr, inBoundary, outError);
	// if there was no error, then set the output string
	if (!outError && inPtr != NULL) {
		*outString = sptr;
	}
	return inPtr;
}

// ---------------------------------------------------------------------------------
//	OSCMessageBundle CLASS
// ---------------------------------------------------------------------------------
// This class "bundles" an OSC message and its values into a single object.
//
// For maximum efficiency, the class copies the raw data from an OSC message when it
// arrives, deferring the parsing of the data until the ParaeData() function is
// called.
//
// When ParseData is called, this object creatss a list of Isadora Value structures,
// each of which contains an integer, floating point number, or string found in the
// message parameters.
//
// The individual values can be accessed by their index using the GetBool, GetInteger,
// GetFloat and GetString methods.

class OSCMessageBundle : public std::vector<Value*>
{
public:
				OSCMessageBundle(
					IsadoraParameters*	inIP,
					std::string			inOSCMessage,
					size_t				inDataLen,
					const void*			inData) :
						std::vector<Value*>(),
						mIP(inIP),
						mOSCMessage(inOSCMessage),
						mData(malloc(inDataLen)),
						mDataLen(inDataLen),
						mParsed(false),
						mValid(false)
				{
					memcpy(mData, inData, inDataLen);
				}
	
				~OSCMessageBundle()
				{
					// release any string values we allocated
					OSCMessageBundle::iterator i;
					for (i=begin(); i<end(); i++) {
						if ((*i)->type == kString) {
							ReleaseValueString_(mIP, *i);
						}
						delete *i;
					}
					// release any lingering data (in case we never received the parse message)
					if (mData != NULL) {
						free(mData);
						mData = NULL;
						mDataLen = 0;
					}
				}
	
	bool		IsValid() const
					{ return mValid; }

	bool		ParseData()
				{
					if (!mParsed && mData != NULL && mDataLen > 0) {
					
						mParsed = true;
						
						const char* c = static_cast<const char*>(mData);
						const char* bounds = c + mDataLen;

						// if this message has type tags
						if (c != NULL && *c == ',') {
							
							const char* typeTag = c+1;
							bool error = false;
							c = ::OSCDataAfterAlignedString(c, bounds, error);
							
							while (!error && c != NULL && *typeTag != 0) {
							
								switch (*typeTag) {
								case 'i':
									{
										SInt32 ivalue;
										c = GetOSCInt(c, bounds, &ivalue, error);
										if (c != NULL && !error) {
											AddIntegerParam(ivalue);
										}
									}
									break;
								case 'f':
									{
										float fvalue;
										c = GetOSCFloat(c, bounds, &fvalue, error);
										if (c != NULL && !error) {
											AddFloatParam(fvalue);
										}
									}
									break;
								case 'd':
									{
										double fvalue;
										c = GetOSCDouble(c, bounds, &fvalue, error);
										if (c != NULL && !error) {
											AddFloatParam(fvalue);
										}
									}
									break;
								case 's':
									{
										std::string svalue;
										c = GetOSCString(c, bounds, &svalue, error);
										if (c != NULL && !error) {
											AddStringParam(svalue);
										}
									}
									break;
								}
								
								typeTag += 1;
							}

							// if there were no errors during parsing, then set the valid flag
							if (!error) {
								mValid = true;
							}
						}
					}
					
					if (mData != NULL) {
						free(mData);
						mData = NULL;
						mDataLen = 0;
					}
					
					return mValid;
				}
	
	void		AddIntegerParam(
					IzValue_Int		inValue)
				{
					Value* v = new Value;
					v->type = kInteger;
					v->u.ivalue = inValue;
					push_back(v);
				}

	void		AddFloatParam(
					IzValue_Flt		inValue)
				{
					Value* v = new Value;
					v->type = kFloat;
					v->u.fvalue = inValue;
					push_back(v);
				}

	void		AddStringParam(
					std::string		inValue)
				{
					Value* v = new Value;
					v->type = kString;
					v->u.str = NULL;
					AllocateValueString_(mIP, inValue.c_str(), v);
					push_back(v);
				}

	bool		GetBool(
					size_t			inIndex0,
					bool&			outValue)
				{
					if (inIndex0 < size() && at(inIndex0)->type == kInteger) {
						outValue = at(inIndex0)->u.ivalue != 0;
						return true;
					} else if (inIndex0 < size() && at(inIndex0)->type == kFloat) {
						outValue = at(inIndex0)->u.fvalue != 0.0;
						return true;
					} else {
						return false;
					}
				}
	
	bool		GetInteger(
					size_t			inIndex0,
					int&			outValue)
				{
					if (inIndex0 < size() && at(inIndex0)->type == kInteger) {
						outValue = (int) at(inIndex0)->u.ivalue;
						return true;
					} else {
						return false;
					}
				}
	
	bool		GetFloat(
					size_t			inIndex0,
					double&			outValue)
				{
					if (inIndex0 < size() && at(inIndex0)->type == kFloat) {
						outValue = at(inIndex0)->u.fvalue;
						return true;
					} else {
						return false;
					}
				}
	
	bool		GetString(
					size_t			inIndex0,
					std::string&	outValue)
				{
					if (inIndex0 < size() && at(inIndex0)->type == kString) {
						outValue = at(inIndex0)->u.str->strData;
						return true;
					} else {
						return false;
					}
				}
	
	std::string	ToString() const
				{
					char buffer[1024];
					std::string result = mOSCMessage + "\n";
					OSCMessageBundle::const_iterator i;
					for (i=begin(); i<end(); i++) {
						switch ((*i)->type) {
						case kInteger:
							sprintf(buffer, "-- #%d: int = %lld\n", (int)(i-begin()), (*i)->u.ivalue);
							break;
						case kFloat:
							sprintf(buffer, "-- #%d: float = %g\n", (int)(i-begin()), (*i)->u.fvalue);
							break;
						case kString:
							sprintf(buffer, "-- #%d: string = %s\n", (int)(i-begin()), (*i)->u.str->strData);
							break;
						}
						result += buffer;
					}
					return result;
				}
	
	IsadoraParameters*	mIP;
	std::string			mOSCMessage;
	void*				mData;
	size_t				mDataLen;
	bool				mParsed;
	bool				mValid;
};

// ---------------------------------------------------------------------------------
//	JSON Generation Flags
// ---------------------------------------------------------------------------------
// Or these flags together when calling the ToJSON() method to determine which fields
// will be added to the JSON output.

enum {
	kJSONGenFlag_CompID				= (1 << 0),
	kJSONGenFlag_UserName			= (1 << 1),
	kJSONGenFlag_TargetIndex		= (1 << 2),
	kJSONGenFlag_GalleryIndex		= (1 << 3),
	kJSONGenFlag_ZoomID				= (1 << 4),
	kJSONGenFlag_TargetCount		= (1 << 5),
	kJSONGenFlag_ListCount			= (1 << 6),
	kJSONGenFlag_Role				= (1 << 7),
	kJSONGenFlag_Online				= (1 << 8),
	kJSONGenFlag_CameraOn			= (1 << 9),
	kJSONGenFlag_MicrophoneOn		= (1 << 10),
	kJSONGenFlag_HandRaised			= (1 << 11)
};

typedef UInt32	JSONFlags;

// ---------------------------------------------------------------------------------
//	ZoomUser CLASS
// ---------------------------------------------------------------------------------
// The ZoomUser class encapsulates the all user data from the ZoomOSC list messasge
// for an individual user.
//
// To create this object the last, pass a parsed OSCMessageBundle that contains a
// ZoomOSC list message to the constructor.
//
// To convert the list to JSON, call the ToJSON method which will return a valid
// JSON text string reprepsenting the data associated with this user.

typedef std::vector<std::string> string_list;
#include <sstream>

class ZoomUser
{
public:
	ZoomUser(
		OSCMessageBundle&	inBundle)
	{
		if (inBundle.size() == 11) {
			
			string_list tokens;
			{
				std::istringstream iss(inBundle.mOSCMessage);
				std::string token;
				while (std::getline(iss, token, '/')) {
					tokens.push_back(token);
				}
			}
			
			bool success = true;
			
			if (success) {
				if (tokens.size() >= 3) {
					mComputerID = atoi(tokens.at(2).c_str());
				}
			}
			
    		if (success)
				success = inBundle.GetInteger(0, mTargetIndex1);
			if (success)
				success = inBundle.GetString(1, mUserName);
			if (success)
				success = inBundle.GetInteger(2, mGalleryIndex1);
			if (success)
				success = inBundle.GetInteger(3, mZoomID);

			if (success)
				success = inBundle.GetInteger(4, mTargetCount);
			if (success)
				success = inBundle.GetInteger(5, mListCount);
			if (success)
				success = inBundle.GetInteger(6, mRole);
			if (success)
				success = inBundle.GetBool(7, mOnline);
			if (success)
				success = inBundle.GetBool(8, mVideoOn);
			if (success)
				success = inBundle.GetBool(9, mAudioOn);
			if (success)
				success = inBundle.GetBool(10, mHandRaised);
			
			mValid = success;
		}
	}
	
	// determines if two OSC users are "equal" -- currenrly
	// we compare the computer ID and target index to make
	// this determination. If they are equal, then the users
	// are equal.
	bool	operator==(
				const ZoomUser&	inRhs) const
			{
				if (mComputerID == inRhs.mComputerID
				&& mTargetIndex1 == inRhs.mTargetIndex1) {
					return true;
				}
				
				return false;
			}
	
	// the fields that will be included are determined by the
	// indvivudal bits of the inJSONFlags parameter. See the
	// JSONFlags definition above for a list of the available
	// flags
	json	ToJSON(
				JSONFlags	inJSONFlags = 0xFFFFFFFF) const
			{
				json json;
				
				if ((inJSONFlags & kJSONGenFlag_CompID) != 0)
					json["computer-id"] = mComputerID;
				if ((inJSONFlags & kJSONGenFlag_UserName) != 0)
					json["user-name"] = mUserName;
				if ((inJSONFlags & kJSONGenFlag_TargetIndex) != 0)
					json["target-index"] = mTargetIndex1;
				if ((inJSONFlags & kJSONGenFlag_GalleryIndex) != 0)
					json["gallery-index"] = mGalleryIndex1;
				if ((inJSONFlags & kJSONGenFlag_ZoomID) != 0)
					json["zoom-id"] = mZoomID;
				if ((inJSONFlags & kJSONGenFlag_TargetCount) != 0)
					json["target-count"] = mTargetCount;
				if ((inJSONFlags & kJSONGenFlag_ListCount) != 0)
					json["list-count"] = mListCount;
				if ((inJSONFlags & kJSONGenFlag_Role) != 0)
					json["role"] = mRole;
				if ((inJSONFlags & kJSONGenFlag_Online) != 0)
					json["online"] = mOnline;
				if ((inJSONFlags & kJSONGenFlag_CameraOn) != 0)
					json["camera-on"] = mVideoOn;
				if ((inJSONFlags & kJSONGenFlag_MicrophoneOn) != 0)
					json["mic-on"] = mAudioOn;
				if ((inJSONFlags & kJSONGenFlag_HandRaised) != 0)
					json["hand-raised"] = mHandRaised;
				
				return json;
			}
	
	bool		mValid = false;
	int			mComputerID = 0;
	int 		mTargetIndex1 = 0;
	std::string	mUserName;
	int 		mGalleryIndex1 = 0;
	int 		mZoomID = 0;
	int			mTargetCount = 0;
	int			mListCount = 0;
	int			mRole = 0;
	bool		mOnline = false;
	bool		mVideoOn = false;
	bool		mAudioOn = false;
	bool		mHandRaised = false;
};

// ---------------------------------------------------------------------------------
//	ZoomUserList CLASS
// ---------------------------------------------------------------------------------
// Maintains a list of ZoomOSC users.
//
// New users can are added to the list using the SaveUser method.
//
// Existing users will be replaced if the SaveUser method determines that the new
// ZoomUser and one of the existing users are the same. This determination is made
// by virtue of the ZoomUser's equality operator (==).
//
// A JSON record representing all users on all computers will be generated by the
// ToJSON method.

class ZoomUserList : public std::vector<ZoomUser>
{
public:
			ZoomUserList(
				int		inComputerID) :
					std::vector<ZoomUser>(),
					mComputerID(inComputerID)
			{
			
			}
	
			~ZoomUserList()
			{
			}

	// returns the number of users in this list
	size_t	GetUserCount() const
				{ return size(); }
	
	// stores the user in this list, replacing the matching user
	// if one already exists
	void	SaveUser(
				const ZoomUser&		inUser)
			{
				ZoomUserList::iterator i;
				for (i=begin(); i<end(); i++) {
					if (inUser == *i) {
						*i = inUser;
						return;
					}
				}
				
				push_back(inUser);
			}


	// generate a JSON record that represents all users
	// within this list. The actual fields returned for
	// each individual record is determiend by the individual
	// bits of the inJSONFlags parameter. See JSONFlags
	// defintion above for a list of all flags
	
	json	ToJSON(
				JSONFlags	inJSONFlags) const
			{
				json user_list;
				user_list["user-count"] = (int) size();
				
				json user_array;
				
				for (size_t i=0; i<size(); i++) {
					try {
						AddUserToJSONArray(inJSONFlags, i, user_array);
					} catch (...) {
						// what to do?
					}
				}
				
				user_list["user-list"] = user_array;
				
				return user_list;
			}
	
	// used by superclasses when accessing a list of users
	void	AddUserToJSONArray(
				JSONFlags	inJSONFlags,
				size_t		inIndex0,
				json&		inUserArray) const
			{
				if (inIndex0 < size()) {
					inUserArray.push_back(at(inIndex0).ToJSON(inJSONFlags));
				}
			}
	
public:
	int		mComputerID;
};

// ---------------------------------------------------------------------------------
//	ZoomComputerList CLASS
// ---------------------------------------------------------------------------------

class ZoomComputerList : public std::vector<ZoomUserList>
{
public:
			ZoomComputerList() :
				std::vector<ZoomUserList>(),
				mUseCount(1),
				mTotalUserCount(0),
				mThreadRunning(false),
				mThread(NULL),
				mJSONFlags(0xFFFFFFFF),	// default to all flags on
				mJSONString()
			{
			
			}
	
private:
			~ZoomComputerList()
			{
				FinishToJSONStringThreaded();
			}

public:
	void	Retain()
			{
				mUseCount += 1;
			}
	
	void	Release()
			{
				mUseCount -= 1;
				if (mUseCount == 0) {
					delete this;
				}
			}
	
	size_t	GetUserCount() const
				{ return mTotalUserCount; }
	
	void	SaveUser(
				const ZoomUser&		inUser)
			{
				bool savedUser = false;
				ZoomComputerList::iterator i;
				for (i=begin(); i<end(); i++) {
					if (inUser.mComputerID == i->mComputerID) {
						int oldSize = (int) i->size();
						i->SaveUser(inUser);
						int newSize = (int) i->size();
						mTotalUserCount += (newSize - oldSize);
						savedUser = true;
						break;
					}
				}
				if (!savedUser) {
					push_back(ZoomUserList(inUser.mComputerID));
					size_t index0 = size()-1;
					at(index0).SaveUser(inUser);
					mTotalUserCount += 1;
				}
			}

	// generate a JSON record that represents all users
	// on all computers. The actual fields returned for
	// each individual record is determiend by the individual
	// bits of the inJSONFlags parameter. See JSONFlags
	// defintion above for a list of all flags
	
	json	ToJSON(
				JSONFlags	inJSONFlags) const
			{
				json users;
				users["user-count"] = mTotalUserCount;
				
				json user_array;
				
				try {
					for (size_t i=0; i<size(); i++) {
						const ZoomUserList& zul = at(i);
						size_t user_count = zul.GetUserCount();
						for (size_t j=0; j<user_count; j++) {
							zul.AddUserToJSONArray(inJSONFlags, j, user_array);
						}
					}
				} catch (...) {
					// what to do???
				}
				
				users["user-list"] = user_array;
				
				return users;
			}
	
	// THREADED CONVERT TO JSON STRING
	bool				IsThreadRunning() const
							{ return mThreadRunning; }
	
	// executes a thread to convert the user list to JSON
	// After starting the conversion process, the caller
	// should poll the IsThreadRunning method until it
	// returns false, at which time it should call
	// FinishToJSONStringThreaded() to clean up the thread
	// and GetJSONString() to get the resulting JSON
	
	void				ToJSONStringThreaded(
							UInt32		inJSONFlags)
						{
							if (!mThreadRunning) {
								mThreadRunning = true;
								Retain();
								mJSONFlags = inJSONFlags;
								mThread = new std::thread(StaticToJSONThreaded, this);
							}
						}
	
	// cleans up the thread after executing ToJSONStringThreaded
	const void			FinishToJSONStringThreaded()
						{
							if (mThread != NULL) {
								mThread->join();
								delete mThread;
								mThread = NULL;
							}
						}

	// retursn the JSON text string generated by ToJSONStringThreaded
	const std::string&	GetJSONString() const
						{
							return mJSONString;
						}

private:
	static void 		StaticToJSONThreaded(
							ZoomComputerList*	inList)
						{
							inList->ToJSONThreaded();
						}

	void 				ToJSONThreaded()
						{
							json userList = ToJSON(mJSONFlags);
							mJSONString = userList.dump();
							mThreadRunning = false;
							Release();
						}

public:
	int				mUseCount;
	int				mTotalUserCount;
	bool			mThreadRunning;
	std::thread*	mThread;
	JSONFlags		mJSONFlags;
	std::string		mJSONString;
};

// ---------------------------------------------------------------------------------
//	OSCMessageBundleList CLASS
// ---------------------------------------------------------------------------------

class OSCMessageBundleList : private std::vector<OSCMessageBundle*>, private std::mutex
{
public:
						OSCMessageBundleList() :
							std::vector<OSCMessageBundle*>(),
							std::mutex()
						{
						
						}
	
	void				push(
							OSCMessageBundle*	inBundle)
						{
							lock();
							push_back(inBundle);
							unlock();
						}
	
	OSCMessageBundle*	pull()
						{
							OSCMessageBundle* outBundle = NULL;
							lock();
							if (size() > 0) {
								outBundle = at(0);
								erase(begin());
							}
							unlock();
							return outBundle;
						}
};

// ---------------------------------------------------------------------------------
// MacOS Specific
// ---------------------------------------------------------------------------------
#if TARGET_OS_MAC
#define EXPORT_
#endif

#if DEBUG && TARGET_OS_MAC
	#define	log_info	printf
#else
	#ifdef log_info
	#undef log_info
	#endif
	#define	log_info(...) { }
#endif

// ---------------------------------------------------------------------------------
// Win32  Specific
// ---------------------------------------------------------------------------------
#if TARGET_OS_WIN
	#define EXPORT_ __declspec(dllexport)
#endif

// ---------------------------------------------------------------------------------
//	Exported Function Definitions
// ---------------------------------------------------------------------------------

#ifdef __cplusplus
extern "C" {
#endif

EXPORT_ void GetActorInfo(void* inParam, ActorInfo* outActorParams);

#ifdef __cplusplus
}
#endif

// static Boolean gInsideHandlePropertyChangeType = false;

// ---------------------------------------------------------------------------------
//	Function Prototypes
// ---------------------------------------------------------------------------------
static void
ReceiveMessage(
	IsadoraParameters*	ip,
	MessageMask			inMessageMask,
	PortIndex			inPortIndex1,
	const MsgData*		inData,
	UInt32				inLen,
	IzRefCon			inRefCon);
	
// ---------------------------------------------------------------------------------
// PluginInfo struct
// ---------------------------------------------------------------------------------
// This structure neeeds to contain all variables used by your plugin. Memory for
// this struct is allocated during the CreateActor function, and disposed during
// the DisposeActor function.

#define kMaxOutputNeedsUpdateFlags	100

typedef struct {
	IsadoraParameters*		mIP;
	ActorInfo*				mActorInfoPtr;
	MessageReceiverRef		mMessageReceiver;
	bool					mInitialized;
	std::string*			mOSCAddress;
	OSCMessageBundleList*	mPendindBundles;
	std::mutex*				mOSCAddressReceiverMutex;
	OSCAddressReceiver		mOSCAddressReceiver;
	IzValue_Int				mMessaagesReceived;
	ZoomComputerList*		mZoomComputerList;
	bool					mIsGeneratingJSON;
	JSONFlags				mJSONFlags;
} PluginInfo;

// A handy macro for casting the mActorDataPtr to PluginInfo*
#if __cplusplus
#define	GetPluginInfo_(actorDataPtr)		static_cast<PluginInfo*>((actorDataPtr)->mActorDataPtr);
#else
#define	GetPluginInfo_(actorDataPtr)		(PluginInfo*)((actorDataPtr)->mActorDataPtr);
#endif

// ---------------------------------------------------------------------------------
//	Constants
// ---------------------------------------------------------------------------------
//	Defines various constants used throughout the plugin.

// * GROUP ID
static const OSType	kActorClass 	= kGroupText;

// * PLUGIN IN
static const OSType	kActorID		= FOUR_CHAR_CODE('zula');

// * ACTOR NAME
static const char* kActorName		= "ZoomOSC Users ALPHA";

// * PROPERTY DEFINITION STRING
static const char* sPropertyDefinitionString =

	"INPROP		osc_addr			addr	string		text		*		*		\r"
	"INPROP		json_comp_id		jcid	bool		onoff		0		1		1\r"
	"INPROP		json_user_name		junm	bool		onoff		0		1		1\r"
	"INPROP		json_target_index	jtix	bool		onoff		0		1		1\r"
	"INPROP		json_gallery_index	jgix	bool		onoff		0		1		1\r"
	"INPROP		json_zoom_id		jzid	bool		onoff		0		1		1\r"
	"INPROP		json_target_count	jtgc	bool		onoff		0		1		1\r"
	"INPROP		json_list_count		jlcn	bool		onoff		0		1		1\r"
	"INPROP		json_role			jrol	bool		onoff		0		1		1\r"
	"INPROP		json_online			jonl	bool		onoff		0		1		1\r"
	"INPROP		json_camera_on		jcam	bool		onoff		0		1		1\r"
	"INPROP		json_mic_on			jmic	bool		onoff		0		1		1\r"
	"INPROP		json_hand_raised	jhra	bool		onoff		0		1		1\r"
	"INPROP		gen_json			updt	bool		trig		0		1		0\r"

	"OUTPROP	osc_addr			otad	string		text		*		*		\r"
	"OUTPROP	msg_count			mcnt	int			number		0		*		0\r"
	"OUTPROP	user_count			uscn	int			number		0		*		0\r"
	"OUTPROP	list_changed		ndat	bool		trig		0		1		0\r"
	"OUTPROP	json_list			user	string		text		*		*		\r";

// * Property Index Constants
enum
{
	kInputOSCAddress = 1,
	
	kInputJSONCompID,
	kInputJSONUserName,
	kInputJSONTargetIndex,
	kInputJSONGalleryIndex,
	kInputJSONZoomID,
	kInputJSONListCount,
	kInputJSONTargetCount,
	kInputJSONRole,
	kInputJSONOnline,
	kInputJSONCameraOn,
	kInputJSONMicrophoneOn,
	kInputJSONHandRaised,
	
	kInputGenerateJSON,

	kOutputOSCAddr = 1,
	kOutputMessageCount,
	kOutputUserCount,
	kOutputListChanged,
	kOutputUserList
};

// * Help Strings
const char* sHelpStrings[] =
{
	// ACTOR
	"Accumulates a list of users from the ZoomOSC 'list' message, keeping an internal copy of that is "
		"updated as new list messages arrive. When requested by triggering the 'gen json' input, "
		"the internal list that will be converted to a JSON text string that can be decoded by the JSON "
		"Parser (see https://troikatronix.com/add-ons).",

	// INPUTS
	"The OSC address used to receive ZoomOSC list messages, which can contain the '*' to indicate a 'wildcard' "
		"that matches any text. After ZoomOSC receives a '/zoom/list', it will respond with a messages "
		"that list the local ZoomOSC user (/zoomosc/me/list) as well as the remote users (/zoomosc/user/list). If "
		"the option to send the computer ID is turned on in ZoomOSC, then these messages will be sent in the form "
		"/zoomosc/COMP_ID/USER_TYPE/list where COMP_ID is the unique identifier of the computer and USER_TYPE is either "
		"'me' or 'user'. When COMP_ID is present, it will be used by this actor to group users who are on the same "
		"computer in the internally held list of users. By entering a wildcard for both of these elements of the "
		"address, this actor will receive all messages for all users on all computers.\r\r"
		"Example OSC addresses are:\r"
		"/zoomosc/me/list - receives information about the local ZoomOSC user.\r"
		"/zoomosc/user/list - receives information about the remote ZoomOSC users.\r"
		"/zoomosc/*/list - matches 'me' or 'user' for the second elemetn, giving a complete list of ZoomOSC users.\r"
		"/zoomosc/*/*/list - when the computer ID is present, receives information all users on all computers, both remote and local.\r",
	
	"When turned on, add 'computer-id' element to the generated JSON output for each user",
	"When turned on, add 'user-name' element to the generated JSON output for each user",
	"When turned on, add 'target-index' element to the generated JSON output for each user",
	"When turned on, add 'gallery-index' element to the generated JSON output for each user",
	"When turned on, add 'zoom-id' element to the generated JSON output for each user",
	"When turned on, add 'target-count' element to the generated JSON output for each user",
	"When turned on, add 'list-count' element to the generated JSON output for each user",
	"When turned on, add 'role' element to the generated JSON output for each user",
	"When turned on, add 'online' element to the generated JSON output for each user",
	"When turned on, add 'camera-on' element to the generated JSON output for each user",
	"When turned on, add 'mic-on' element to the generated JSON output for each user",
	"When turned on, add 'hand-raised' element to the generated JSON output for each user",

	"Trigger this input to generate the JSON outpupt that represents the current list of users. Generating "
		"the JSON output can take some time if you have thousands of users. (The generation process runs "
		"in a background 'thread' to ensure Isadora's rendering thread is not interrupted.)",

	// OUTPUTS

	"The most recent ZoomOSC 'list' message that matched the OSC address pattern specified at the 'osc addr' input.",

	"The total number of ZoomOSC 'list' messages received since entering the current scene.",
	
	"The current number of users in the list.",

	"Sends a trigger whenever the internal user list has changed.",

	"The list of users output as a JSON text string. Use the JSON Decoder plugin (see https://troikatronix.com/add-ons) "
		"to access the indvidual members of the JSON data. Please note, generating the JSON list can take a significant "
		"amount of time if you have thousands of users! Documentation on the JSON structure can be found in the "
		"documentation provided when you download this actor."
};

// ---------------------------------------------------------------------------------
//		 CreateActor
// ---------------------------------------------------------------------------------
static void
CreateActor(
	IsadoraParameters*	ip,
	ActorInfo*			ioActorInfo)
{
	PluginInfo* info = (PluginInfo*) IzzyMallocClear_(ip, sizeof(PluginInfo));
	PluginAssert_(ip, info != nil);
	
	ioActorInfo->mActorDataPtr = info;
	info->mIP = ip;
	info->mActorInfoPtr = ioActorInfo;
}

// ---------------------------------------------------------------------------------
//		 DisposeActor
// ---------------------------------------------------------------------------------
static void
DisposeActor(
	IsadoraParameters*	ip,
	ActorInfo*			ioActorInfo)
{
//	PluginInfo* info = (PluginInfo*) ioActorInfo->mActorDataPtr;
	PluginAssert_(ip, ioActorInfo->mActorDataPtr != nil);
	IzzyFree_(ip, ioActorInfo->mActorDataPtr);
}


// ---------------------------------------------------------------------------------
//		StaticOSCAddressListenerCallback
// ---------------------------------------------------------------------------------

static void
StaticOSCAddressListenerCallback(
	void*					inRefCon,				// user reference constant specified when the receiver was created
	const char*				inOSCAddress,			// the full OSC addressed that was received with all wildcards filled in
	uint64_t				inTimeStamp,			// the timestamp for the OSC message
	size_t					inDataLen,				// the length of the OSC data associated with this message
	const void*				inDataPtr)				// pointer to the OSC data associated with this message
{
	PluginInfo* info = static_cast<PluginInfo*>(inRefCon);
	if (info != NULL) {
		info->mOSCAddressReceiverMutex->lock();
		info->mPendindBundles->push(new OSCMessageBundle(info->mIP, inOSCAddress, inDataLen, inDataPtr));
		info->mOSCAddressReceiverMutex->unlock();
	}
}

// ---------------------------------------------------------------------------------
//		 SetOSCAddressReceiver
// ---------------------------------------------------------------------------------
static void
SetOSCAddressReceiver(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	std::string			inOSCAddress)
{
	PluginInfo* info = GetPluginInfo_(inActorInfo);

	*info->mOSCAddress = inOSCAddress;
	
	info->mOSCAddressReceiverMutex->lock();

	if (info->mOSCAddressReceiver != NULL) {
		DisposeOSCAddressReceiver_(ip, inActorInfo, info->mOSCAddressReceiver);
		info->mOSCAddressReceiver = NULL;
	}

	OSCAddressReceiver oscAddrReceiver = NULL;
	// this will crash if we're not in Isadora 3.0.9b01 or later!
	CreateOSCAddressReceiver_(ip, inActorInfo, inOSCAddress.c_str(), StaticOSCAddressListenerCallback, info, &oscAddrReceiver);
	info->mOSCAddressReceiver = oscAddrReceiver;
	
	info->mOSCAddressReceiverMutex->unlock();
}

// ---------------------------------------------------------------------------------
//		 ClearOSCAddressReceiver
// ---------------------------------------------------------------------------------
static void
ClearOSCAddressReceiver(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo)
{
	PluginInfo* info = GetPluginInfo_(inActorInfo);
	info->mOSCAddressReceiverMutex->lock();
	if (info->mOSCAddressReceiver != NULL) {
		DisposeOSCAddressReceiver_(ip, inActorInfo, info->mOSCAddressReceiver);
		info->mOSCAddressReceiver = NULL;
	}
	info->mOSCAddressReceiverMutex->unlock();
}


// ---------------------------------------------------------------------------------
//		 ActivateActor
// ---------------------------------------------------------------------------------

static void
ActivateActor(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	Boolean				inActivate)
{
	PluginInfo* info = (PluginInfo*) inActorInfo->mActorDataPtr;

	if (inActivate) {

		PluginAssert_(ip, info->mMessageReceiver == NULL);
		info->mMessageReceiver = CreateActorMessageReceiver_(
			ip,
			inActorInfo,
			ReceiveMessage,
			0,
			kWantIzzyTick,
			PtrToIzRefCon(info));

		info->mInitialized = false;
		info->mOSCAddress = new std::string();
		info->mPendindBundles = new OSCMessageBundleList();
		info->mOSCAddressReceiverMutex = new std::mutex();
		info->mOSCAddressReceiver = NULL;
		info->mMessaagesReceived = 0;
		info->mZoomComputerList = new ZoomComputerList();
		info->mIsGeneratingJSON = false;
		
	} else {
	
		ClearOSCAddressReceiver(ip, inActorInfo);
		
		delete info->mOSCAddressReceiverMutex;
		info->mOSCAddressReceiverMutex = NULL;
		
		delete info->mPendindBundles;
		info->mPendindBundles = NULL;
		
		delete info->mOSCAddress;
		info->mOSCAddress = NULL;
		
		info->mZoomComputerList->Release();
		info->mZoomComputerList = NULL;
		
		PluginAssert_(ip, info->mMessageReceiver != nil);
		DisposeMessageReceiver_(ip, info->mMessageReceiver);
		info->mMessageReceiver = NULL;
	}
}

// ---------------------------------------------------------------------------------
//		 GetParameterString
// ---------------------------------------------------------------------------------

static const char*
GetParameterString(
	IsadoraParameters*	/* ip */,
	ActorInfo*			/* inActorInfo */)
{
	return sPropertyDefinitionString;
}

// ---------------------------------------------------------------------------------
//		 GetHelpString
// ---------------------------------------------------------------------------------

static void
GetHelpString(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	PropertyType		inPropertyType,
	PropertyIndex		inPropertyIndex1,
	char*				outParamaterString,
	UInt32				inMaxCharacters)
{
	UInt32 index1 = PropertyTypeAndIndexToHelpIndex_(ip, inActorInfo, inPropertyType, inPropertyIndex1);

	const char* helpstr = sHelpStrings[index1];

	strcpy(outParamaterString, helpstr);
}
	

// ---------------------------------------------------------------------------------
//		 HandlePropertyChangeValue	[INTERRUPT SAFE]
// ---------------------------------------------------------------------------------

static void
HandlePropertyChangeValue(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	PropertyIndex		inPropertyIndex1,
	ValuePtr			/* inOldValue */,
	ValuePtr			inNewValue,
	Boolean				inInitializing)
{
	PluginInfo* info = (PluginInfo*) inActorInfo->mActorDataPtr;

	switch (inPropertyIndex1) {

	case kInputOSCAddress:
		if (inNewValue->u.str != NULL && strlen(inNewValue->u.str->strData) > 0) {
			SetOSCAddressReceiver(ip, inActorInfo, inNewValue->u.str->strData);
		}
		break;
		
	case kInputGenerateJSON:
		{
			info->mZoomComputerList->ToJSONStringThreaded(info->mJSONFlags);
			info->mIsGeneratingJSON = true;
		}
		break;
		
	case kInputJSONCompID:
	case kInputJSONUserName:
	case kInputJSONTargetIndex:
	case kInputJSONGalleryIndex:
	case kInputJSONZoomID:
	case kInputJSONListCount:
	case kInputJSONRole:
	case kInputJSONOnline:
	case kInputJSONCameraOn:
	case kInputJSONMicrophoneOn:
	case kInputJSONHandRaised:
		{
			UInt32 flag = (1 << (inPropertyIndex1 - kInputJSONCompID));
			if (inNewValue->u.ivalue != 0) {
				info->mJSONFlags |= flag;
			} else {
				info->mJSONFlags &= ~flag;
			}
			
		}
		break;


	}
}

// ---------------------------------------------------------------------------------
//		 PropertyValueToString
// ---------------------------------------------------------------------------------
//	Handles incoming video frame changed messages

//const char* sErrorFlag[] = {
//	"no",
//	"yes"
//};
//
//static Boolean
//PropertyValueToString(
//	IsadoraParameters*	ip,
//	ActorInfo*			/* inActorInfo */,
//	PropertyType		inPropertyType,
//	PropertyIndex		inPropertyIndex1,
//	ValuePtr			inValue,
//	char*				outString)
//{
//	Boolean result = false;
//
//	if (inPropertyType == kOutputProperty && inPropertyIndex1 == kOutputError) {
//		switch (inValue->u.ivalue) {
//		case 0:
//		case 1:
//			strcpy(outString, sErrorFlag[inValue->u.ivalue]);
//			break;
//		default:
//			PluginAssert_(ip, false);
//			break;
//		}
//		result = true;
//	}
//
//	return result;
//}

// ---------------------------------------------------------------------------------
//		 PropertyStringToValue
// ---------------------------------------------------------------------------------

//static Boolean
//PropertyStringToValue(
//	IsadoraParameters*	ip,
//	ActorInfo*			/* inActorInfo */,
//	PropertyType		inPropertyType,
//	PropertyIndex		inPropertyIndex1,
//	const char*			inString,
//	ValuePtr			outValue)
//{
//	Boolean encodeSuccessful = false;
//
//	if (inPropertyType == kOutputProperty && inPropertyIndex1 == kOutputError) {
//		if (strlen(inString) == 1 && inString[0] >= '0' && inString[0] <= '1') {
//			outValue->type = kInteger;
//			outValue->u.ivalue = (SInt32) inString[0] - '0';
//			encodeSuccessful = true;
//		} else {
//			SInt32 matchIndex = LookupPartialStringInList_(ip, 2, sErrorFlag, inString);
//			if (matchIndex >= 0) {
//				outValue->type = kInteger;
//				outValue->u.ivalue = matchIndex;
//				encodeSuccessful = true;
//			}
//		}
//	}
//
//	return encodeSuccessful;
//}

// ---------------------------------------------------------------------------------
//		 GetActorDefinedArea
// ---------------------------------------------------------------------------------

static ActorPictInfo	gPictInfo = { false, nil, nil, 0, 0 };

static Boolean
GetActorDefinedArea(
	IsadoraParameters*			ip,			
	ActorInfo*					inActorInfo,
	SInt16*						outTopAreaWidth,
	SInt16*						outTopAreaMinHeight,
	SInt16*						outBotAreaHeight,
	SInt16*						outBotAreaMinWidth)
{
	if (gPictInfo.mInitialized == false) {
		PrepareActorDefinedAreaPict_(ip, inActorInfo, 0, &gPictInfo);
	}
	
	*outTopAreaWidth = gPictInfo.mWidth;
	*outTopAreaMinHeight = gPictInfo.mHeight;
	*outBotAreaHeight = 0;
	*outBotAreaMinWidth = 0;

	
	return true;
}

// ---------------------------------------------------------------------------------
//		 DrawActorDefinedArea
// ---------------------------------------------------------------------------------

static void
DrawActorDefinedArea(
	IsadoraParameters*			ip,
	ActorInfo*					inActorInfo,
	void*						/* inDrawingContext */,
	ActorDefinedAreaPart		inActorDefinedAreaPart,
	ActorAreaDrawFlagsT			/* inAreaDrawFlags */,
	Rect*						inADAArea,
	Rect*						/* inUpdateArea */,
	Boolean						inSelected)
{
	if (inActorDefinedAreaPart == kActorDefinedAreaTop && gPictInfo.mInitialized) {
		DrawActorDefinedAreaPict_(ip, inActorInfo, inSelected, inADAArea, &gPictInfo);
	}
}
	
// ---------------------------------------------------------------------------------
//		 HandlePropertyChangeType
// ---------------------------------------------------------------------------------

static void
HandlePropertyChangeType(
	IsadoraParameters*					ip,
	ActorInfo*							inActorInfo,
	const PropertyDefinedValueSource*	/* inPropDefValueSrc */,
	PropertyType						inPropertyType,
	PropertyIndex						inPropertyIndex1,
	PropertyDispFormat					/* inFormats */,
	PropertyDispFormat					/* inCurFormat */,
	ValuePtr							/* inMinimum */,
	ValuePtr							/* inMaximum */,
	ValuePtr							/* inScaleMin */,
	ValuePtr							/* inScaleMax */,
	ValuePtr							/* inValue */)
{
//	if (gInsideHandlePropertyChangeType == false) {
//
//		gInsideHandlePropertyChangeType = true;
//
//		if (inPropertyType == kOutputProperty && inPropertyIndex1 >= kFirstVariableOutputProperty) {
//			// log_info("Changed Output %d - New Type = %d\n", (int) inPropertyIndex1, (int) inValue->type);
//			// UpdateOutputTypes(ip, inActorInfo);
//		}
//
//		gInsideHandlePropertyChangeType = false;
//	}
}

// ---------------------------------------------------------------------------------
//	Function Prototypes
// ---------------------------------------------------------------------------------
static void
ReceiveMessage(
	IsadoraParameters*	ip,
	MessageMask			inMessageMask,
	PortIndex			inPortIndex1,
	const MsgData*		inData,
	UInt32				inLen,
	IzRefCon			inRefCon)
{
	PluginInfo* info = static_cast<PluginInfo*>(inRefCon);
	ActorInfo* inActorInfo = info->mActorInfoPtr;

	Value v;
	
	if (!info->mInitialized) {
	
		v.type = kInteger;
		v.u.ivalue = info->mMessaagesReceived;
		SetOutputPropertyValue_(ip, inActorInfo, kOutputMessageCount, &v);
		info->mInitialized = true;
		
		v.type = kInteger;
		v.u.ivalue = (IzValue_Int) info->mZoomComputerList->GetUserCount();
		SetOutputPropertyValue_(ip, inActorInfo, kOutputUserCount, &v);
		info->mInitialized = true;
		
		v.type = kString;
		v.u.str = NULL;
		AllocateValueString_(ip, "", &v);
		SetOutputPropertyValue_(ip, inActorInfo, kOutputUserList, &v);
		ReleaseValueString_(ip, &v);
	}
	
	std::string lastOSCMessage;
	bool listChanged = false;
	
	bool getBundles = true;
	while (getBundles) {
	
		OSCMessageBundle* bundle = info->mPendindBundles->pull();

		if (bundle != NULL) {
		
			if (bundle->ParseData()) {
		
				lastOSCMessage = bundle->mOSCMessage;
				
				log_info("Received Bundle %s with %d values\n", bundle->mOSCMessage.c_str(), (int) bundle->size());
				log_info("%s", bundle->ToString().c_str());
				
				ZoomUser user(*bundle);
				info->mZoomComputerList->SaveUser(user);
				
				info->mMessaagesReceived += 1;
				listChanged = true;
				
			}

		} else {
			getBundles = false;
		}
	}
	
	if (listChanged) {
	
		v.type = kString;
		v.u.str = NULL;
		AllocateValueString_(ip, lastOSCMessage.c_str(), &v);
		SetOutputPropertyValue_(ip, inActorInfo, kOutputOSCAddr, &v);
		ReleaseValueString_(ip, &v);

		v.type = kInteger;
		v.u.ivalue = info->mMessaagesReceived;
		SetOutputPropertyValue_(ip, inActorInfo, kOutputMessageCount, &v);

		v.type = kInteger;
		v.u.ivalue = (IzValue_Int) info->mZoomComputerList->GetUserCount();
		SetOutputPropertyValue_(ip, inActorInfo, kOutputUserCount, &v);
		
		v.type = kInteger;
		v.u.ivalue = 1;
		SetOutputPropertyValue_(ip, inActorInfo, kOutputListChanged, &v);

	}
	
	if (info->mIsGeneratingJSON && !info->mZoomComputerList->IsThreadRunning()) {
		Value vstr;
		info->mZoomComputerList->FinishToJSONStringThreaded();
		AllocateValueString_(ip, info->mZoomComputerList->GetJSONString().c_str(), &vstr);
		SetOutputPropertyValue_(ip, inActorInfo, kOutputUserList, &vstr);
		ReleaseValueString_(ip, &vstr);
		info->mIsGeneratingJSON = false;
	}
}

// ---------------------------------------------------------------------------------
//		 GetActorInfo
// ---------------------------------------------------------------------------------
//	This is function is called by to get the actor's class and ID, and to get
//	pointers to the all of the plugin functions declared locally.
//
//	All members of the ActorInfo struct pointed to by outActorParams have been
//	set to 0 on entry. You only need set functions defined by your plugin
//	
EXPORT_ void
GetActorInfo(
	void*				/* inParam */,
	ActorInfo*			outActorParams)
{
	// REQUIRED information
	outActorParams->mActorName							= kActorName;
	outActorParams->mClass								= kActorClass;
	outActorParams->mID									= kActorID;
	outActorParams->mCompatibleWithVersion				= kActorCompatilibilityVersion_v1;
	outActorParams->mActorFlags							= kActorFlags_Plugin_CheckForUpdates;
	
	// REQUIRED functions
	outActorParams->mGetActorParameterStringProc		= GetParameterString;
	outActorParams->mGetActorHelpStringProc				= GetHelpString;
	outActorParams->mCreateActorProc					= CreateActor;
	outActorParams->mDisposeActorProc					= DisposeActor;
	outActorParams->mActivateActorProc					= ActivateActor;
	outActorParams->mHandlePropertyChangeValueProc		= HandlePropertyChangeValue;
	
	// OPTIONAL FUNCTIONS
	outActorParams->mHandlePropertyChangeTypeProc		= HandlePropertyChangeType;
	outActorParams->mHandlePropertyConnectProc			= NULL;
//	outActorParams->mPropertyValueToStringProc			= PropertyValueToString;
//	outActorParams->mPropertyStringToValueProc			= PropertyStringToValue;
	outActorParams->mGetActorDefinedAreaProc			= GetActorDefinedArea;
	outActorParams->mDrawActorDefinedAreaProc			= DrawActorDefinedArea;
	outActorParams->mMouseTrackInActorDefinedAreaProc	= NULL;

	outActorParams->mSetActorPrivateDataProc			= NULL;
	outActorParams->mGetActorPrivateDataLengthProc		= NULL;
	outActorParams->mGetActorPrivateDataProc			= NULL;
}
