// ===========================================================================
//	IsadoraCallbacks			(c) 1999 Mark F. Coniglio. All rights reserved
// ===========================================================================
//
//	Defines callback functions useful to Isadora plugins

#ifndef _H_IsadoraCallbacks
#define	_H_IsadoraCallbacks

#include "IZTypes.h"
#include "IZFile.h"

// ---------------------------------------------------------------
// Inline Versions of QuickTime Functions

#include "GraphicsHelpers.h"

// #include "_tchar.h"

// ----------------------------------------------------------------------------
// PLUGIN DEBUGGING
//
// After this, NDEBUG, DEBUG, _DEBUG, ISADORA_DEBUG, and ISADORA_NDEBUG, are 
// all consistent with eachother and defined as such:
//
//                       Debug     Release
//   NDEBUG                  0           1
//   DEBUG                   1           0
//   _DEBUG            defined   undefined
//   ISADORA_DEBUG     defined   undefined
//   ISADORA_NDEBUG  undefined     defined
//
// Be careful not to use #ifdef with DEBUG or NDEBUG, as they are still 
// "defined" even when they are defined as 0. If you just stick with #if 
// everywhere you'll be fine.
//
// IMPORTANT: This is the same logic as the internal DebugFlags.h (not a file 
// in the plugin SDK). If this logic is updated here, it should be changed in 
// that file as well!

#if defined(_DEBUG) || DEBUG || defined(ISADORA_DEBUG)
#  undef NDEBUG
#  define NDEBUG 0
#else
#  if !(NDEBUG || defined(ISADORA_NDEBUG))
#    pragma message ("Warning: NDEBUG, DEBUG, _DEBUG, ISADORA_DEBUG, and ISADORA_NDEBUG not set -- assuming NDEBUG.")
#  endif
#  undef NDEBUG
#  define NDEBUG 1
#endif

#undef DEBUG
#define DEBUG (!NDEBUG)

// These are either defined or undefined, not 0 or 1.
#undef _DEBUG
#undef ISADORA_DEBUG
#undef ISADORA_NDEBUG
#if NDEBUG
#  define ISADORA_NDEBUG
#else
#  define _DEBUG
#  define ISADORA_DEBUG
#endif

//-----------------------------------------------------------------------------

#include "IsadoraTypes.h"
#include "ImageBufferCommon.h"
#include "IzzyFrameBufferBase.h"
#include "TextBufferCommon.h"
#include "MessageReceiverCommon.h"
#include "CaptureConstants.h"
#include "ColorRefs.h"

#ifdef __cplusplus
extern "C" {
#endif

#include "ValueCommon.h"

#ifdef __cplusplus
class UGraphics;
#else
struct UGraphics;
#endif

/*

--------------------------
PROPERTY DEFINITION STRING
--------------------------

The GetActorParameterStringProc function is called to obtain the parameter 
string used to build the GUI representation of the actor and its properties. 

It is given as one long c-string, with carriage returns ('\r') to the mark
the end of lines. Each line has the following format, with capitalized
items being chosen from a list of possible constants.

PROPTYPE namestr id TYPE DISPLAYFMT minval maxval initval

(Note that the parser for the parameter does not care about case, so
'INT' is the same as 'int', etc.)

A description of each parameter follows.

Item		Description
----------- -------------------------------------------------------------------
PROPTYPE	is either INPROP or OUTPROP to indicate the type of
			property. You should list all of the input properties
			first, followed by all of the output properties

namestr		is the name of the property that will be displayed to the user.
			Note that since the name must be one continuous string, you must use
			underlines (_) to specify spaces. Thus "my_input" is shown as
			"my input" to the user

id			an identifier, unique to this actor, up to four characters long. This
			ID allows Isadora to identify a property in subsequent versions
			of your plugin, even if you change the order of inputs or outputs.
			If you delete a property, then you should retire its ID so that
			if a user attempts to load a file that was created with the older
			version of your plugin, it will know that it should remove the
			deleted property from the earlier version of the file.

TYPE		indicates the property type. it should be one of

			BOOL	= a boolean, 0 or 1 (off or on) value
			INT		= a 32 bit signed integer
			FLOAT	= a 32 bit floating point number
			RANGE	= two 16 bit signed integers indicating a range of values
			STRING	= a pointer to a String struct, which holds a pointer
					  to a c-string and a use count.
			DATA	= private data defined by the user. in this case, the
					  value must point to a block of data that begins with
					  DataValueInfo structure (see IsadoraTypes.h.) The first
					  member of that structure is an OSType that indicates the
					  type of data contained in the block of memory.

DISPLAYFMT	Indicates the one  or more possible formats that can be used to display
			the property value. It can be one or more of the following possibilities,
			with each separated by commas. The first item in the list of display
			formats will be the default when creating a new object.

			MUTEABLE	= Value will adapt to match type of input
			NUMBER		= Display as an signed integer
			ONOFF		= Displays as "on" or "off" for non-zero and zero values, respectively
			PITCH		= Displays value as a MIDI pitch, from C-2 to G8
			CHAR		= Displays value as a character surrounded by single quotes,
						  i.e., decimal 65 would be shown as 'A'
			SECONDS		= Displays value in seconds, with the actual value being an
						  an integer or float that represents milliseconds
			MILLISECS	= Displays value in milliseconds, with the actual value being
						  an integer or float that represents milliseconds
			HERTZ		= Displays value in hertz, with the actual value being
						  an integer or float that represents milliseconds.
			TIMECODE	= Displays a timecode value for position in a movie or 
						  other piece of media - it is up to the actor to figure
						  out how to determine this value.
			TEXT		= Valid only when TYPE is STRING, will display the value as a string
			FONT		= Valid only when TYPE is STRING, displays value as a string, and will
						  allow user to choose a font when they click on the input property
						  value edit box.
			TRIG		= Special display for types of BOOL. With this option, the value 
						  is normally displayed as a dash (-) to indicate that nothing is
						  happening. When a non-zero value is received, the display
						  changes momentarily to an 'X', returning to the dash after
						  100 milliseconds of inactivity.
			PROPDEF		= The actor will create the display string. If this option
						  is chosen, then the actor should define the 
						  PropertyValueToString function to convert from an numeric
						  value to a string that Isadora can display.

minval		the property's minimum allowable value. Use an asterisk (*) to
			indicate the default lowest possible minimum value for the
			property's type, i.e., for an INT the * here would mean -(2^32).
			The minimum and maximum value for a STRING are ignored.

maxval		the property's maximum allowable value. Use an asterisk (*) to
			indicate the default highest possible minimum value for the
			property's type, i.e., for an INT an * here would mean (2^32-1).
			The minimum and maximum value for a STRING are ignored.

initval		the default value for the property when creating a new actor. This
			value can be text for STRING property types. Note however, that if you
			initialize the value with a string, it cannot contain a space.

An example string might be

const char* MyDefintionString = 
	"INPROP		trigger		trig	bool		trig				0		1		0\r"
	"INPROP		frequency	f		int			hertz				0		100		0\r"
	"OUTPROP	output		oval	float		number				0		100		0\r"
	"OUTPROP	end_trigger	otrg	bool		trig				0		1		0\r";
*/

////////////////////////////////////////////////////////////////////////////
// DEFINES
////////////////////////////////////////////////////////////////////////////

// the following macros are deprecated.
// for FloatToRawLong use RealToIzRefCon
// for FloatFromRawLong use IzRefConToReal

// #define	FloatToRawLong(floatValue)			(*((SInt32*) &floatValue))			// returns a float as a long integer without conversion
// #define FloatFromRawLong(floatValueAsLong)	(*((float*) &floatValueAsLong))		// returns a float that was stored in a long integer

// when this flag is true, the render order of
// primary scenes can be changed using the render
// order input to the Jump functions. This is new
// functionality in Isadora v2.6

#define SCENE_RENDER_ORDER_ENABLE			TRUE
#define RENDER_ORDER_DEFAULT				0

////////////////////////////////////////////////////////////////////////////
// TYPES
////////////////////////////////////////////////////////////////////////////

// Actor Defined Area Types

typedef	UInt16				ActorAreaDrawFlagsT;
#define	kDrawAllActorAreas	0xFFFF

// PropertyIdentifier - a property type + a property ID
// that when combined uniquely identify an actor's property
// The property ID is guaranteed to identify the correct
// property even if the order of the properties is changed
// in a subsequent version of Isadora.

struct _PropertyIdentifier {
	PropertyType	mPropertyType;
	PropIDT			mPropertyID;
};

typedef struct _PropertyIdentifier PropertyIdentifier;

// PropTypeAndIndex - a property type + a property index
// that when combined uniquely identify an actor's property.
// This combination cannot be relied on to work if the
// order of the actor's inputs change, and so should not
// be stored between runs of the program

enum {
	kPropTimeHint_None,
	kPropTimeHint_PlayStart,
	kPropTimeHint_PlayLength,
	kPropTimeHint_PlayPosition
};

struct _TimePropInfo {
	PropertyType	mPropertyType;
	PropertyIndex	mPropertyIndex;
	UInt32			mPropTimeHint;		// see constants above
};

typedef struct _TimePropInfo TimePropInfo;


////////////////////////////////////////////////////////////////////////////
// ACTOR FUNCTIONS
////////////////////////////////////////////////////////////////////////////
// Below you will find typedefs for the functions that Isadora will call
// inside of your actor. Some of them are required, others are optional.
// Descriptions of what your actor should do when these actors are called
// are listed below.

	// GetActorInfoProc (Required)
	// 
	// When your actor's GetActorInfoProc is called, it must fill in the
	// ActorInfo structures with the calls that the actor supports.

typedef	void		(*GetActorInfoProc)(
						const void*					inGetActorInfoParam,
						struct ActorInfo*			outActorParams);

	// GetActorParameterStringProc (Required)
	//
	// The GetActorParameterStringProc function is called to obtain the parameter 
	// string used to build the GUI representation of the actor and its properties. 
	// See "Property Definition String" above for more information

typedef	const char*	(*GetActorParameterStringProc)(
						struct IsadoraParameters*	inIzzyParams,
						struct ActorInfo*			outActorParams);
	
	// GetDefaultHiddenPropertiesProc (Optional)
	//
	// This optional function allows an actor to specify which
	// inputs will be hidden by default. If implemented, this
	// function should set outPropertyIdentifierArray to a pointer
	// to an array of PropertyIdentifier items, and return the 
	// number of items in the list as the result
	
typedef	size_t		(*GetDefaultHiddenPropertiesProc)(
						const PropertyIdentifier**	outPropertyIdentifierArray);
	
	// CreateActorProc (Required)
	//
	// Create a new instance of the actor. Do not allocate or
	// load media resources required to execute the actor (that should be 
	// handled by the ActivateActorProc.) Instead, preallocate memory 
	// for your actor, initialize its internal values and setup
	// the ActorInfo struct with all of the required information.
typedef	void	(*CreateActorProc)(
					struct IsadoraParameters*	inIzzyParams,		
					struct ActorInfo*			outActorParams);
	
	// DisposeActorProc (Required)
	//
	// Dispose an existing instance of the actor created by
	// the CreateActorProc.
typedef	void	(*DisposeActorProc)(
					struct IsadoraParameters*	inIzzyParams,			
					struct ActorInfo*			inActorParams);

	// GetActorSpecificOptionsProc (Optional)
	//
	// Allows an actor to provide a list of options that can
	// be turned on or off. If an actor implements this function,
	// it will be called to list those options at the bottom of
	// the actor menu whenever one or more of matching actors of
	// the same type are selected
	//
	// TO GET THE TOTAL NUMBER OF OPTIONS
	//
	// Set 'inSet' to FALSE and ioOptionList to NULL. The number
	// of options will be returned in the value pointed to by
	// outActualOptionCount
	//
	//
	// TO GET THE CURRENT OPTION VALUES
	//
	// Set 'inSet' to FALSE and the value of ioOptionCount to the
	// number of items in the array pointed to by ioOptionList.
	// If the number of items is greater than the value specified
	// by inOptionCount, the result will be false and the value
	// pointed to outActualOptionCount will be updated to reflect
	// the number of options avilable.
	//
	// TO SET ONE OR MORE OPTIONS
	//
	// To set one or more options, set 'inSet' to TRUE and the
	// value pointed to by inOptionCount to the number of items
	// in the array pointed to by ioOptionList. Then fill in the
	// ioOptionList items with the option ID and new option value.
	// outActualOptionCount is not used during set operations.

	// Isadora calls the callback when the an the menu item representing
	// an actor specific option needs updating, or when the menu item
	// itself is chosen.

					struct ActorSpecificOption;								// forward declaration;

					// constants passed to the ActorOptionCallback
					enum ActorOptionAction
					{
						kActorOptionAction_UpdateMenuStatus,				// set outEnabled and outChecked as appopriate
						kActorOptionActoin_MenuItemChosenNotification,		// respond to the menu choice as appropriate
					};


	typedef	void	(*ActorOptionCallback)(
						struct IsadoraParameters*	inIzzyParams,			
						struct ActorInfo*			inActorParams,
						
						ActorOptionAction			inOptionAction,			// action code for the callback
						
						ActorSpecificOption*		ioASO,					// actor specific option data
						
						Boolean*					outEnabled,				// only used with kActorOptionAction_UpdateMenuStatus
																			// can be NULL otherwise

						Boolean*					outChecked);			// only used with kActorOptionAction_UpdateMenuStatus
																			// can be NULL otherwise
										   
	// ActorSpecificOption is used to pass data back and forth about
	// an actor specific option via the GetActorSpecificOptionsProc

	struct ActorSpecificOption {
		OSType				mOptionID;		// unique ID for the option
		int64_t				mOptionValue;	// value for this option
		bool				mSaveWithFile;	// if true, write out this option with the file and restore it when the file is read
		const char*			mOptionName;	// text to display in menu item; ignored on set
		UInt32				mMenuID;		// a unique menu ID value for this option, between 7501 and 7599
		ActorOptionCallback	mCallback;		// callback used to update the menu item and to notify the actor when the item is chosen
	};

	typedef struct ActorSpecificOption ActorSpecificOption;


typedef	Boolean		(*GetActorSpecificOptionsProc)(
						struct IsadoraParameters*	inIzzyParams,			
						struct ActorInfo*			inActorParams,
						UInt32						inOptionCount,
						ActorSpecificOption*		ioOptionList,
						UInt32*						outActualOptionCount);

	// GetActorTimePropertiesProc (Optional)
	//
	// Actors that need to switch between different time representations
	// for the media they play should implement this procedure. When called,
	// the target actor will return
	//
	// outMediaType:					the media type of the media played by this actor
	//
	// outMediaInputPropertyIndex1:		the input property index of the media index for
	//									the media played by this actor
	//
	// outTimePropCount:				the number of items in the PropertyAndTypeIndex
	//									array returned by outTimePropArray
	//
	// outTimePropArray:	pointer to an array of TimePropInfo,
	//									each representing and input or output property
	//									that reprsents time in relation to a media
	//									file being played
	
typedef	Boolean		(*GetActorTimePropertiesProc)(
						struct IsadoraParameters*	inIzzyParams,
						struct ActorInfo*			inActorParams,
						MediaCode*					outMediaType,
						PropertyIndex*				outMediaInputPropertyIndex1,
						UInt32*						outTimePropCount,
						const TimePropInfo**		outTimePropArray);

	// SpecifyInputsForExecuteProc_2_O (Optional)
	//
	// This function allows your actor to specify which
	// upstream actors will receive an Execute() message
	// when this actor is exectued.
	//
	// The default behavior is that, when an actor receives
	// an execute message, _all_ actors that are linked to this
	// actor's inputs receive an Execute() message. Your actor
	// should override this behavior if  executing particular
	// outputs should only execute one or more specific inputs.
	//
	// An example would be the Selector actor, which routes
	// one of many inputs to a single output. The Selector actor
	// sets the outActualActivatedInputCount to 3.
	//
	// Then, in the outArrayOfActivatedInputIndexes array, three
	// one-based indexes are returned. The first two are the indexes
	// for the first two inptus (the 'select' input and the 'inputs'
	// input, both of which should execute their upstream actors.) The
	// third element of outArrayOfActivatedInputIndexes will be set to
	// the input that is passing values through the Selector actor. By
	// doing this the Selector actor ensures that only the actor connected
	// the the active input is Executed(). Actors connected to the inputs
	// that are not passing through to the output will thus not receive
	// an Execute() message.
	//
	// Your function should set the outActualActivatedInputCount
	// value to the number of input indexes in the outArrayOfActivatedInputIndexes
	// to a pointer to an array that has the at least as many elements as
	// specified by outActualActivatedInputCount. (This array must remain valid
	// for as long as the actor is active.) The elements of this array must
	// be to the one-based index of the inputs that should receive activation
	// messages.
	
typedef	void	(*SpecifyInputsForExecuteProc_2_O)(
					struct IsadoraParameters*	inIzzyParams,			
					struct ActorInfo*			inActorInfo,
					const IzzyTime*				inIzzyTime,		
					UInt32						inOutputIndex1,
					UInt32*						outActualActivatedInputCount,
					PropertyIndex**				outArrayOfActivatedInputIndexes);
	
	// DEPRECATED DEPRECATED DEPRECATED DEPRECATED DEPRECATED DEPRECATED
	// This function was depecated for v2.0 of Isadora, because the manner
	// in which the function operates needed to be changed. You need to
	// change the function to use the new SpecifyInputsForExecuteProc_2_0
	// function listed above 

typedef	void	(*SpecifyInputsForExecuteProc_1_O)(
					struct IsadoraParameters*	inIzzyParams,			
					struct ActorInfo*			inActorInfo,
					const IzzyTime*				inIzzyTime,		
					UInt32						inOutputIndex1,
					UInt32						inMaxActivatedInputCount,
					UInt32*						outActualActivatedInputCount,		// actual number of inputs to execute
					PropertyIndex*				outArrayOfActivatedInputIndexes);	// a pointer to an array of one-based indexes, supplied by the actor
	
	// ExecuteActorProc (Optional)
	//
	// This optional function allows an actor to receive an execute
	// message just after all upstream actors (i.e. the actors connected
	// to this actor's inputs) have been executed.
	//
typedef	void	(*ExecuteActorProc)(
					struct IsadoraParameters*	inIzzyParams,			
					struct ActorInfo*			inActorInfo,
					const IzzyTime*				inIzzyTime);
	
	// HandleIzzyStatusMessageProc (Optional)
	//
	// This function allows an actor to be notified in certain
	// changes of status. Primarily this is used to be notified
	// of the activation of a scene, but you may also receive
	// notifications of changes to the Media List (i.e., the 
	// Media Window) and other messages. See IzzyStatusMessage
	// in IsadoraTypes.h for more details on the messsages you
	// can receive

typedef	void	(*HandleIzzyStatusMessageProc)(
					struct IsadoraParameters*	inIzzyParams,			
					struct ActorInfo*			inActorInfo,
					IzzyStatusMessage			inIzzyStatusMessage,
					const void*					inData);

	// set this define to 1 to have plugins use the IzzyStatusMessage mechanism
	#define USE_IZZY_STATUS_MESSAGE			1
	
	
	// PrepareActorForFirstUseProc (Optional)
	//
	// This function is called the _first_ time that the user
	// grabs the actor from the toolbox and drags it into the
	// Scene Editor. This is called immediately after the 
	// CreateActorProc, once Isadora's internal actor
	// information is initialized. You can feel free to call
	// any of Isadora's callback functions that reference
	// the actor or its properties.
	
typedef	void	(*PrepareActorForFirstUseProc)(
					struct IsadoraParameters*	inIzzyParams,			
					struct ActorInfo*			inActorParams);

	// ActivateActorProc (Optional)
	//
	// When the ActivateActorProc is called with the inActivate
	// flag set to true, your actor is becoming active. It should
	// allocate any memory and/or resources required to get the
	// actor to run. When inActivate is false, your actor is becoming
	// inactive - all memory and resources that were allocated
	// during the activate process should be released.
	// 
	// When the ActivateActorProc is called with the inActivate
	// flag set to false, your actor is being deactivated. It should
	// dispose of any memory or other resources that were allocated
	// when the actor was activated.
	
typedef	void	(*ActivateActorProc)(
					struct IsadoraParameters*	inIzzyParams,			
					struct ActorInfo*			inActorParams,
					Boolean						inActivate);
	
	// GetActorHelpStringProc (Optional, but strongly recommended)
	//		
	// the user has clicked requested help for this actor or one of
	// its properties.
	//
	// If inPropertyType is kPropertyTypeInvalid then you should
	// return help for the whole actor. otherwise, examine the
	// values of inPropertyType and inPropertyIndex to determine
	// the property for which the user is requesting help.
	// Return the resulting help string in buffer pointed to
	// by the outParamterString buffer. (The size of this output
	// buffer is specified by the inMaxCharacters parameter.)

typedef	void	(*GetActorHelpStringProc)(
					struct IsadoraParameters*	inIzzyParams,		
					struct ActorInfo*			inActorParams,
					PropertyType				inPropertyType,
					PropertyIndex				inPropertyIndex,
					char*						outParamaterString,
					UInt32						inMaxCharacters);

	// GetActorDefinedAreaProc (Optional)
	//
	// When called, this function should return the size of
	// area that will be drawn by the actor in _logical_ pixels
	// The values returned returned by this actor will be
	// scaled by the current scaling factor which is based on the
	// DPI setting for the window into which the graphics will be
	// drawn. If you wish to return unscaled, raw pixel values,
	// then use the GetActorDefinedAreaPixelsProc instead.
	//
	// There are two areas you may define: an area to be placed 
	// between the input and output values (the top area) and/or
	// an area that will be added on to the bottom of the actor.
	// You return non-zero values for the parameters if you 
	// would like to draw your own interface element inside
	// of the actor. If you define an actor defined area, the
	// DrawActorDefinedAreaProc will be called when it is time
	// to draw your area.
	//
	// For the top area, you may specified the width you desire
	// (which will always be given to you) and a minimum height.
	// The top area will always be at least as big as the width
	// and minimum height, though it may be taller than that if
	// the actor has many inputs or outputs.
	// 
	// For the bottom area, you specify a height and mimimum
	// width. The height will be added on to the actor, below
	// the input and output values. It is preferred to use
	// the top area when possible, as it keeps the screen 
	// dimensions of the actor smaller when it is collapsed.
	// However, certain interface elements want to be wide,
	// (see the Movie Player Actor) so it is reasonable to 
	// use the bottom area in this case.
	//
typedef	Boolean	(*GetActorDefinedAreaProc)(
					struct IsadoraParameters*	inIzzyParams,			
					struct ActorInfo*			inActorParams,
					SInt16*						outTopAreaWidth,
					SInt16*						outTopAreaMinHeight,
					SInt16*						outBotAreaHeight,
					SInt16*						outBotAreaMinWidth);

	// GetActorDefinedAreaUnscaledProc (Optional)
	//
	// Functions the same as GetActorDefinedAreaProc except
	// that the pixel values are 'raw', in the sense they will
	// not be scaled to reflect the urrent DPI setting of the
	// window into which the graphics will be drawn.
	
typedef	Boolean	(*GetActorDefinedAreaPixelsProc)(
					struct IsadoraParameters*	inIzzyParams,			
					struct ActorInfo*			inActorParams,
					SInt16*						outTopAreaWidth,
					SInt16*						outTopAreaMinHeight,
					SInt16*						outBotAreaHeight,
					SInt16*						outBotAreaMinWidth);

	// ActorResizedProc (Optional)
	//
	// This function is called if an actor is resized. This allows the
	// actor to adjust the size of its Actor Defined Area if needed.

typedef	void	(*ActorResizedProc)(
					struct IsadoraParameters*	inIzzyParams,			
					struct ActorInfo*			inActorParams,
					SInt16						inWidthDelta,
					SInt16						inHeightDelta);

	// DrawActorDefinedAreaProc (Optional)
	//
	// Draw the user interface element defined by your actor.
	// 
	// The inActorDefinedAreaPart will be set either to kActorDefinedAreaTop
	// or kActorDefinedAreaBottom, indicating which area needs to be drawn.
	//
	// The inAreaDrawFlags is a bitfield used to indicate which part of the
	// actor should be drawn. When Isadora is processing an update event to
	// draw all actors in a window, this value will be 0xFFFF, all parts 
	// need to be drawn. Otherwise, it will be a value that you passed back
	// in the mDrawActorAreaFlags member of the ActorInfo struct. (See
	// Getting Your Actor Area To Draw, below).
	//
	// The inADAArea is the rectangle that defines the part to be drawn.
	//
	// The inUpdateArea is the portion of inADAArea that needs to be drawn
	// in response to an update event.
	//
	// The inDrawingContext pointer is not used at this time, but may be
	// used in future cross-platform versions of Isadora.
	//
	// Getting Your Actor Area To Draw:
	// --------------------------------
	// 
	// To get a portion of your actor area to draw, you set one or more
	// bits in the mDrawActorAreaFlags member of the ActorInfo struct. This
	// can be done at any time. On a periodic basis, Isadora scans all actors
	// that are currently visible to see if their mDrawActorAreaFlags have been
	// set to a non-zero value. When it finds one, your DrawActorDefinedAreaProc
	// will be called, with the inAreaDrawFlags set to the value you stored
	// in mDrawActorAreaFlags. The idea is that mDrawActorAreaFlags (and
	// inAreaDrawFlags) are a bitfield, with each bit indicating one part of
	// your user interface. The hope here is that by only drawing the user
	// interface element that needs to be drawn, the amount of processing 
	// time required to update your user interface is minimized.
	//
	// As an example, the Sound Player actor has both the progress/playback
	// bar at the bottom, and the volume/pan indicator at the top. Each of
	// these has their own bit to indicate which item needs to be drawn. When
	// the volume or panning changes, the bit for that UI element is set. When
	// the sound position or playback selection changes, then the other bit 
	// is set.

typedef	void	(*DrawActorDefinedAreaProc)(
					struct IsadoraParameters*	inIzzyParams,			
					struct ActorInfo*			inActorParams,
					void*						inDrawingContext,
					ActorDefinedAreaPart		inActorDefinedAreaPart,
					ActorAreaDrawFlagsT			inAreaDrawFlags,
					IzRect*						inADAArea,
					IzRect*						inUpdateArea,
					Boolean						inSelected);

	// MouseTrackInActorDefinedAreaProc (Optional)
	//
	// The user has clicked in your user interface area. Respond
	// to the click as appropriate.
	//
	// See IzzyMouseMessage for a list of messages that can be
	// passed into this function
	
typedef	Boolean	(*MouseTrackInActorDefinedAreaProc)(
					struct IsadoraParameters*	inIzzyParams,			
					struct ActorInfo*			inActorParams,
					IzzyMouseMessage			inMouseMessage,
					void*						inDrawingContext,
					IzRect*						inArea,
					IzPoint						inMouse);

	// HandlePropertyChangeTypeProc (Optional)
	//
	// The user would like to display one of your properties
	// in an alternate format or some operation has changed
	// the basic type of this property (a prime example being
	// changing from kFloat to kTimecode)
	
typedef void	(*HandlePropertyChangeTypeProc)(
					struct IsadoraParameters*			inIzzyParams,
					struct ActorInfo*					inActorParams,
					const PropertyDefinedValueSource*	inPropDefSrc, 			// CAN BE NULL
					PropertyType						inPropertyType,
					PropertyIndex						inPropertyIndex1,
					PropertyDispFormat					inFormats,
					PropertyDispFormat					inCurFormat,
					ValuePtr							inMinimum,
					ValuePtr							inMaximum,
					ValuePtr							inScaleMin,
					ValuePtr							inScaleMax,
					ValuePtr							inValue);

	// HandlePropertyConnectProc (Optional)
	//
	// One of your input properties, identified by inPropertyIndex1,
	// has been connected or disconnected from an output to which it was previously
	// connected. Handling this callback allows you to take action when
	// a connection is lost
	
typedef void	(*HandlePropertyConnectProc)(
					struct IsadoraParameters*	inIzzyParams,
					struct ActorInfo*			outActorParams,
					PropertyType				inPropertyType,
					PropertyIndex				inPropertyIndex1,
					Boolean						inConnect);

	// HandlePropertyChangeProc (Required)
	//
	// One of your input properties, identified by inPropertyIndex1,
	// has changed value. The current value of the property is the
	// new value, but both the old and new values are passed into
	// this function for your examination.
	
typedef void	(*HandlePropertyChangeProc)(
					struct IsadoraParameters*	inIzzyParams,
					struct ActorInfo*			outActorParams,
					PropertyIndex				inPropertyIndex1,
					ValuePtr					oldValue,
					ValuePtr					newValue,
					Boolean						initializing);

	// PropertyValueToStringProc (Optional)
	//
	// Your parameter string indicated that you would handle converting
	// the internal value of a properties into a text reprsentation by
	// specifying the "PROPDEF" as one of its types in the parameter string
	//
	// When PropertyValueToStringProc is called, examine the inPropertyType and
	// inPropertyIndex1 parameters to determine which parameter to convert.
	// The current value of that paramter is also provided. Output the resulting
	// text into the buffer pointed to by outString, which has a maximum
	// length of 256 characters.
	
typedef Boolean	(*PropertyValueToStringProc)(
					struct IsadoraParameters*	ip,
					struct ActorInfo*			inActorInfo,
					PropertyType				inPropertyType,
					PropertyIndex				inPropertyIndex1,
					const ValuePtr				inValue,
					char*						outString);
	
	// PropertyStringToValueProc (Optional)
	//
	// Your parameter string indicated that you would handle converting
	// the internal value of a properties into a text reprsentation by
	// specifying the "PROPDEF" as one of its types in the parameter string
	//
	// When PropertyStringToValueProc is called, examine the inPropertyType and
	// inPropertyIndex1 parameters to determine which parameter to convert.
	// The text that the user typed is in the buffer pointed to by inString.
	// Your job is to convert that text representation into a value, which
	// is returned in the Value struct pointed to by outValue.
	
typedef Boolean	(*PropertyStringToValueProc)(
					struct IsadoraParameters*	ip,
					struct ActorInfo*			inActorInfo,
					PropertyType				inPropertyType,
					PropertyIndex				inPropertyIndex1,
					const char*					inString,
					ValuePtr					outValue);


	// PropertyGetPopupMenuInfo (Optional)
	//
	// If your parameter string includes the "POPUP" type it 
	// indicates that both the PropertyGetPopupMenuInfo and the
	// PropertyPopupMenuSelectionToValue functions will be
	// implemented.
	//
	// Before v2.6, the PropertyPopupMenuSelect function was
	// required to respond to the user's choice. This function
	// has been deprecated; now plugins should always used the
	// PropertyGetPopupMenuInfo instead.
	//
	// When PropertyGetPopupMenuInfo is called, the actor should provide
	// a list of text menu items separated by a carriage return character
	// i.e., ITEM1/rITEM2/rITEM3/r. The size of this bufer will be large,
	// the specific number of characters indicated by the inMenuItemsStringBufferSize
	// input parameter.
	//
	// Once this function returns, Isadora will show this menu to the 
	// user and allow them to make a selection. Subsequently, if the
	// user makes a choice from this menu, the PropertyPopupMenuSelect
	// function will be called, with the one-based index of the selection,
	// as well as the menu item text, allowing the actor to react to
	// the user selection
	
typedef void	(*PropertyGetPopupMenuInfoProc)(
					struct IsadoraParameters*	ip,
					struct ActorInfo*			inActorInfo,
					PropertyType				inPropertyType,
					PropertyIndex				inPropertyIndex1,
					char*						outMenuItemsString,
					UInt32						inMenuItemsStringBufferSize,
					UInt32*						outCurrentlySelectedItem);


	// PropertyPopupMenuSelect (Deprecated for v2.6 and later)
	// use PropertyPopupMenuSelectionToValue() instead.
	//
	// If your parameter string includes the "POPUP" type it 
	// indicates that both the PropertyGetPopupMenuInfo and the
	// PropertyPopupMenuSelect functions will be implemented.
	//
	// If the user successfully chooses an item from the popup menu,
	// this function receives both the one-based index of the selected
	// item as well as the text shown in the popup menu.
	
typedef void	(*PropertyPopupMenuSelectProc)(
					struct IsadoraParameters*	ip,
					struct ActorInfo*			inActorInfo,
					PropertyType				inPropertyType,
					PropertyIndex				inPropertyIndex1,
					SInt32						inMenuIndex1,
					const char*					inSelectedItem);

	// PropertyPopupMenuSelectionToValue (Optional)
	//
	// new in v2.6
	//
	// If your parameter string includes the "POPUP" type it 
	// indicates that both the PropertyGetPopupMenuInfo and the
	// PropertyPopupMenuSelectionToValue functions will be implemented.
	//
	// If the user successfully chooses an item from the popup menu,
	// this function receives both the one-based index of the selected
	// item as well as the text shown in the popup menu. If the choice
	// is valid, this function should convert that choice into a Value,
	// and store it in outValue. If this result was allocated using
	// AllocateStringValue_() it will be released by the caller.
	//
	// RETURN VALUE: If the called function returned a valid value in outValue,
	// then return return true. Otherwise, return false.
	
typedef Boolean	(*PropertyPopupMenuSelectionToValueProc)(
					struct IsadoraParameters*	ip,
					struct ActorInfo*			inActorInfo,
					PropertyType				inPropertyType,
					PropertyIndex				inPropertyIndex1,
					SInt32						inMenuIndex1,
					const char*					inSelectedItem,
					ValuePtr					outValue);

	// SetActorPrivateDataProc (Optional)
	//
	// This actor is about to be activated - this function is called
	// with data that was saved the last time the actor's
	// GetActorPrivateDataProc function was called.
	
typedef	void	(*SetActorPrivateDataProc)(
					struct IsadoraParameters*	inIzzyParams,			
					struct ActorInfo*			inActorParams,
					const void*					inActorData,
					UInt32						inDataDataLength);

// v2.6.2b4 ->
	// GetSaveActorPrivateDataWithFileProc (Optional)
	//
	// Return true if the actor's private data should
	// be written out with the Isadora document.
	
typedef	Boolean	(*GetSaveActorPrivateDataWithFileProc)(
					struct IsadoraParameters*	inIzzyParams,			
					struct ActorInfo*			inActorParams);

// v2.6.2b4 <-

	// GetActorPrivateDataLengthProc (Optional)
	//
	// Return the size of the private data that to be restored when
	// the actor is next loaded from disk.
	
typedef	UInt32	(*GetActorPrivateDataLengthProc)(
					struct IsadoraParameters*	inIzzyParams,			
					struct ActorInfo*			inActorParams);

	// GetActorPrivateDataProc (Optional)
	//
	// Return the private data that to be restored when this actor 
	// is next loaded from disk.
	// 
	// Store the actor's private data into the memory location pointed
	// to by inActorDataMaxLen. Do not exceed the length specified by
	// inActorDataMaxLen. This function should return the actual length of
	// the data copied. (This shouldn't be different than the value returned
	// by the actor's GetActorPrivateDataLengthProc function, but it is there
	// for absolute certainty.)
	
typedef	UInt32	(*GetActorPrivateDataProc)(
					struct IsadoraParameters*	inIzzyParams,			
					struct ActorInfo*			inActorParams,
					UInt32						inActorDataMaxLen,
					void*						outActorData);

	// ActorDoubleClickProc (Optional)
	//
	// Called when an actor has been double-clicked by the user
	// This function should return true if the double-click was
	// handled in some way.
	//
	// the function should set the outDirty flag if the actor "dirtied"
	// the document (i.e., changed data such a way that the document
	// needs to be saved) as a result of handling the double click.
	
typedef	Boolean	(*ActorDoubleClickProc)(
					struct IsadoraParameters*	inIzzyParams,			
					struct ActorInfo*			inActorParams,
					Boolean*					outDirty);


	// HandleActorMessage (Optional)
	//
	// Allows Isadora to send messages to the actor. Currently this
	// is only used to notify the actor that an Undo/Redo actions
	// after a actor has called SetActorUndoAction_(). If the
	// user specifies a user data parameter to that function, it
	// will be passed back to the caller via the inUserData parameter.
	// 
	// This list of IzzyActorMessage types is defined in IsadoraTypes.h
	// under the heading ACTOR MESSAGE DEFINES
	//
	// *NOTE*: when this function is called with the kIzzyActorDisposeUndoRedo
	// message, the plugin _must_ be able to dispose the data even when the
	// actor info (inActorParams) pointer is invalid
		
typedef	void		(*HandleActorMessageProc)(
						struct IsadoraParameters*	inIzzyParams,			
						struct ActorInfo*			inActorParams,
						IzzyActorMessage			inActorMessage,
						void*						inUserData);
						
	// GetMediaPlaybackInfoProc (Optional)
	//
	// If an actor plays media, it should respond to this call, providing
	// the media code, media index, and a filled in MediaInfo struct.
	//
	// The return value is TRUE if the call succeeded in providing
	// valid information
	
typedef	Boolean		(*GetMediaPlaybackInfoProc)(
						struct IsadoraParameters*	inIzzyParams,			
						struct ActorInfo*			inActorParams,
						MediaCode*					outMediaCode,
						MediaIndexT*				outMediaIndex1);

	// GetPreloadMediaInfoProc (Optional)
	//
	// If an actor plays media, it should respond to this call, providing
	// the media code, media index, and a filled in AcquireMediaInfo struct
	// that will allow the caller to initate preloading of media
	//
	// This call will be made just prior to activating a scene, to allow
	// the media to be preloaded (hopefully in the background) before the
	// scene is actually activated. The actor can depend on the fact that
	// it's initial values have been set. It will however not be active yet.
	// 
	// The called function should return true if the information was
	// filled in; otherwise return false to indicate it is invalid
	
typedef	Boolean		(*GetPreloadMediaInfoProc)(
						struct IsadoraParameters*	inIzzyParams,			
						struct ActorInfo*			inActorParams,
						MediaCode*					outMediaCode,
						UInt32*						outMediaIndex1,
						struct AcquireInfo*			outAcquireMediaInfo);


	// InitiateActorPreloadTask (Optional)
	//
	// Actors that play their own media should respond to this call
	//
	// This call will be made just prior to activating a scene, to allow
	// the media to be preloaded (hopefully in the background) before the
	// scene is actually activated. The actor can depend on the fact that
	// it's initial values have been set. It will however not be active yet.
	// 
	// If the preload process was successfully initiated, outPreloadTaskID
	// should be set to a pointer to a block of memory containing an 
	// ActorPreloadInfo struct so that Isadora can monitor the progress
	// of the preload
	
typedef	Boolean		(*InitiateActorPreloadTaskProc)(
						struct IsadoraParameters*	inIzzyParams,			
						struct ActorInfo*			inActorParams,
						UInt32						inPreloadUniqueID,
						ActorPreloadInfo**			outPreloadTaskInfo);


////////////////////////////////////////////////////////////////////////////
// Actor Flags
////////////////////////////////////////////////////////////////////////////
//
// Flags set in the mActorFlags member of the ActofInfo struct to indicate
// properties of this this actor. Actors who requires these flags should
// set them when they initializes the ActorInfo struct

enum {
	kActorFlags_IsVideoOutput					= 0x00000001,	// indicates this actor is a video output
	
	kActorFlags_IsDataOutput					= 0x00000002,	// indicates this actor outputs data of some kind,
																// e.g., MIDI, serial, OSC, etc.
	
	kActorFlags_DontForceInputValuesOnActivate	= 0x00000004,	// don't forcibly set the input values of this actor when
																// when it transitions from inactive to active unless
																// the propery's "initialize" flag is only
	
	kActorFlags_IsDeprecatedPlugin				= 0x00000008,	// indicates that the actor is a "classic" plugin that
																// is no longer recommended.
	
	kActorFlags_NeverUnloadPlugin				= 0x00000010,	// indicates that this plugin/dll should not be dynamically
																// unloaded once it has been loaded
	
	kActorFlags_UsesCActorClass					= 0x00000020,	// mActorDataPtr is a subclass of CActor*
	

	kActorFlags_IsMediaOutput					= 0x00000040,	// this actor generates/produces video, sound or other media

	
	kActorFlags_Plugin_CheckForUpdates			= 0x00000080	// only to be used by official TroikaTronix plugins!
																// this flag checks the web for updates using a special
																// system available only to the official plugins
};

#define	kActorFlags_IsOuptut	(kActorFlags_IsVideoOutput | kActorFlags_IsDataOutput | kActorFlags_IsMediaOutput)

typedef UInt32	ActorFlags;

////////////////////////////////////////////////////////////////////////////
// Actor Info struct
////////////////////////////////////////////////////////////////////////////
//
// Members marked REQ must be filled in by your actor when its
// GetActorInfo function is called.
//
// Members marked OPT can be set to NULL if your actor does not
// need/support the specified function.
//
// Members marked ISADORA are filled in by Isadora and should not
// be changed.
	
struct ActorInfo {

	ActorClassIDT						mClass;								// REQ: actor class ID
	ActorClassIDT						mSubClass;							// OPT: actor sub-class within class
	ActorIDT							mID;								// REQ: actor unique ID
	UInt32								mCompatibleWithVersion;				// REQ: version of Isadora Callbacks for which this
																			//      this plugin was built
																			//      ms byte = major revision, ls byte = minor revision
	ActorFlags							mActorFlags;						// OPT: flags giving information about this actor; importantly if it is an output or not
	UInt32								mRequiredFeatureFlags;				// OPT: if this version requires special features, set those flags here
																			//      (e.g., CoreVideo; see FeatureFlags.h)
	const char*							mActorName;							// REQ: pointer to c-string with actor's name
	
	CONST_ Boolean						mActive;							// ISADORA: true if this actor is active now
	CONST_ ThingPtr						mThing;								// ISADORA: opaque ptr to the thing object that owns this actor
	CONST_ long							mThingOptions;						// ISADORA: options for this thing that owns actor
	CONST_ IzSpec						mPluginFileSpec;					// ISADORA: file spec for this plugin
	void*								mActorDataPtr;						// ISADORA: private data pointer
	void*								mPluginUserDefinedPtr;				// ISADORA: private
	MessageReceiverRef					mVideoTickMessageReceiver;			// ISADORA: video tick mesage receiver for this actor
	MessageReceiverRef					mIzzyTickMessageReceiver;			// ISADORA: isadora tick mesage receiver for this actor

	// CREATION/INITIALIZATION/DESTRUCTION
	GetActorParameterStringProc			mGetActorParameterStringProc;		// REQ: returns the parameter string used to create actor
	GetActorHelpStringProc				mGetActorHelpStringProc;			// OPT (but highly desireable): returns the help string for a particular property
	CreateActorProc						mCreateActorProc;					// REQ: called to create the actor
	DisposeActorProc					mDisposeActorProc;					// REQ: called to dispose the actor
	ActivateActorProc					mActivateActorProc;					// OPT: called to activate/deactivate the actor (MFC: made this optional on 01/24/08)
	ActivateActorProc					DEPRECATED_mNotifyPostActivateProc;	// OPT: called after all actors in a scene are activated or deactivated

	// PROPERTIES
	HandlePropertyChangeProc			mHandlePropertyChangeValueProc;		// REQ: called when a property changes value
	HandlePropertyChangeTypeProc		mHandlePropertyChangeTypeProc;		// OPT: called when a property changes type
	HandlePropertyConnectProc			mHandlePropertyConnectProc;			// OPT: called when an input is connected or disconnected
	PropertyValueToStringProc			mPropertyValueToStringProc;			// OPT: called to allow actor to convert a property's value to a string
	PropertyStringToValueProc			mPropertyStringToValueProc;			// OPT: called to allow actor to convert string into a property's value
	
	// ACTOR DEFINED AREA
	ActorAreaDrawFlagsT					mDrawActorAreaFlags;				// OPT: actor should set non-zero when its draw area needs updating
	GetActorDefinedAreaProc				mGetActorDefinedAreaProc;			// OPT: allows actor to define a area for display/interaction. The result
																			//      of this function will be scaled to accomodate the current DPI
																			//      setting of the window into which the graphics will be drawn
	DrawActorDefinedAreaProc			mDrawActorDefinedAreaProc;			// OPT: called when actor defined area needs to be drawn
	MouseTrackInActorDefinedAreaProc	mMouseTrackInActorDefinedAreaProc;	// OPT: called when there is mouse activity inside actor defined area

	// ACTOR PRIVATE DATA
	SetActorPrivateDataProc				mSetActorPrivateDataProc;			// OPT: called prior to activating to restore private actor data
	GetActorPrivateDataLengthProc		mGetActorPrivateDataLengthProc;		// OPT: called prior to calling GetActorPrivateDataProc to get the length of the data
	GetActorPrivateDataProc				mGetActorPrivateDataProc;			// OPT: called after deactivating to store private actor data

	// MOUSE TRACKING
	ActorDoubleClickProc				mActorDoubleClickProc;				// OPT: called when an actor has been double-clicked by user
	PrepareActorForFirstUseProc			mPrepareActorForFirstUseProc;		// OPT: called the first time any actor is dragged from toolbox to the Scene Editor

	// DRAWING & SIZINGS
	ActorResizedProc					mActorResizedProc;					// OPT: called when actor has been resized
	
	// ACTOR EXECUTION
	SpecifyInputsForExecuteProc_1_O		DEPRECATED_mSpecifyInputsForExecuteProc;	// DO NOT USE! Change to mSpecifyInputsForExecuteProc_2_0

	// IZZY MESSAGE NOTIFICATION
	HandleIzzyStatusMessageProc			mHandleIzzyStatusMessageProc;		// OPT: called when an IzzyStatusMessage is available

	// ACTOR MESSAGE NOTIFICATION
	HandleActorMessageProc				mHandleActorMessageProc;			// OPT: called when an IzzyActorMessage is available
																			// *NOTE*: when this function is called with the kIzzyActorDisposeUndoRedo
																			// selector, the plugin must be able to dispose the data without the
																			// actor info pointer being valid

	// POPUP MENU HANDLERS
	PropertyGetPopupMenuInfoProc		mPropertyGetPopupMenuInfoProc;		// OPT: called if the user clicks in an input marked with the "POPUP" in the parameter string
	
	PropertyPopupMenuSelectProc			mPropertyPopupMenuSelectProc;		// DEPRECATED: use PropertyPopupMenuSelectionToValue from v2.6 forward
																			// called if the user chooses an item from a popup menu generated by mPropertyGetPopupMenuInfo

	GetPreloadMediaInfoProc				mGetPreloadMediaInfoProc;			// OPT: if the actor uses media, it should respond to this call providing information on the media it plays

	// ACTOR EXECUTION
	ExecuteActorProc					mExecuteActorProc;					// OPT: called when an actor is executed
	
	// PRELOADING
	InitiateActorPreloadTaskProc		mInitiateActorPreloadTaskProc;		// OPT: actors that play their own media should implement this function

	SpecifyInputsForExecuteProc_2_O		mSpecifyInputsForExecuteProc_2_0;	// OPT: allows actor to specify which inputs need to be executed prior to itself being executed 

	GetDefaultHiddenPropertiesProc		mGetDefaultHiddenPropertiesProc;	// OPT: allows actor to specify which inputs and outputs are hidden by default 

	GetActorDefinedAreaPixelsProc		mGetActorDefinedAreaPixelsProc;		// OPT: same as GetActorDefinedAreaProc except that the output should
																			//      be raw, unscaled pixel values.
	// POPUP MENU SELECTION	// v2.6.2b4
	
	PropertyPopupMenuSelectionToValueProc	mPropertyPopupMenuSelectionToValueProc;	// OPT: called if the user chooses an item from a popup menu generated by mPropertyGetPopupMenuInfo

	// ACTOR PRIVATE DATA EXTENSIONS
	GetSaveActorPrivateDataWithFileProc		mGetSaveActorPrivateDataWithFileProc;	// OPT: if present, called to determine if the actor's private data
																					//      should be written out with the document. If not present,
																					//      always save the private data with the file.

	GetMediaPlaybackInfoProc			mGetMediaPlaybackInfoProc;			// OPT: for "player" actors that play media from the
																			// media view, returns the media type for the currently
																			// chosen media file.
	
	GetActorSpecificOptionsProc			mGetActorSpecificOptionsProc;		// OPT: allows actors to provide a list of options that
																			// will appear as submenu items for Actors > Actor Options
																			// See the GetActorSpecificOptionsProc above for more detail
	
	GetActorTimePropertiesProc			mGetActorTimePropertiesProc;		// OPT: for actors that can switch between time
																			// representations, returns the media type, media
																			// input property index, and a list of properties
																			// that can switch between time representations
};
typedef struct ActorInfo ActorInfo;


////////////////////////////////////////////////////////////////////////////
// ISADORA CONTROLS FUNCTIONS
////////////////////////////////////////////////////////////////////////////
// Below you will find typedefs for the functions that Isadora will call
// inside of your control. Some of them are required, others are optional.
// Descriptions of what your control should do when these functions are 
// called are listed below.

	// GetControlInfoProc (Required)
	// 
	// When your control's GetControlInfoProc is called, it must fill in the
	// ControlInfo structure with the appropriate values. See the ControlInfo
	// definition below for details

typedef	void		(*GetControlInfoProc)(
						const void*				inGetActorInfoParam,
						struct ControlInfo*		outControlParams);

	// CreateControlProc (Required)
	//
	// Create a new instance of the control.
typedef	void		(*CreateControlProc)(
						struct IsadoraParameters*	inIzzyParams,		
						struct ControlInfo*			outControlParams);		// and array of availble control numbers
	
	// DisposeControlProc (Required)
	//
	// Dispose an existing instance of the control created by
	// the CreateControlProc.
typedef	void		(*DisposeControlProc)(
						struct IsadoraParameters*	inIzzyParams,			
						struct ControlInfo*			inControlParams);

	// UpdateControllerNumbers (Required)
	//
	// Requests that the control update it's controller numbers
	// picking from the list provided
typedef	UInt32		(*UpdateControllerNumbersProc)(								// return number of controller numbers used
						struct IsadoraParameters*	inIzzyParams,			
						struct ControlInfo*			inControlParams,
						UInt32						inAvailControlNumberCount,	// count of avaialable control numbers
						const UInt32*				inAvailControlNumbers);		// and array of availble control numbers

	// GetIndexedControllerNumber (Required)
	//
	// This function allows Isadora to ascertain which controller
	// numbers your controller is using. If inIndex1 is 0, then
	// the function should return the count of controller numbers
	// used by your contro. When inIndex1 > 0, you should return
	// the controller number for that index.
	
typedef	IzzyCtlNum	(*GetIndexedControllerNumberProc)(
						struct IsadoraParameters*	inIzzyParams,			
						struct ControlInfo*			inControlParams,
						UInt32						inIndex1);				// set to 0 to get number of controllers, > 0 to get controller number
					
	// SetIndexedControllerNumber (Required)
	//
	// This function allows Isadora to set the controller
	// numbers your controller is using. When inIndex1 > 0,
	// you should set controller number for that index.
	
typedef	void	(*SetIndexedControllerNumberProc)(
						struct IsadoraParameters*	inIzzyParams,			
						struct ControlInfo*			inControlParams,
						UInt32						inIndex1,
						IzzyCtlNum					inNewCtlNumber);
					

	// GetIndexedControllerParamInfo (Optional)
	//
	// This function allows the host to get information about the
	// parameters of the control, i.e., associated with a particular
	// controller number
	
typedef	Boolean	(*GetIndexedControllerParamInfo)(
						struct IsadoraParameters*	inIzzyParams,			
						struct ControlInfo*			inControlParams,
						UInt32						inIndex1,
						ValueType*					outValueType,
						DataType*					outDataType);		// only valid if outValueType is kData
					
	// SetControllerFloatValueProc (Required)
	//
	// When a controller is linked to an Actor property, and the property
	// changes value, then this function is called to inform the control
	// of the new value
	//
	// The function should return true if the controller value actually
	// changed.
	// 

typedef UInt8	CtlUpdateCode;

// any of the SetControllerXXXValueProc functions should return
// one of these values
enum {
	kControlUpdateCode_DontUpdate = 0,		// no need to update
	kControlUpdateCode_DrawControl = 1,		// just draw the control -- don't update underlying layers (faster - use if possible)
	kControlUpdateCode_UpdateControl = 2	// invalidate the area under the control, and redraw using a full update of underlying layers
};
			

typedef	CtlUpdateCode	(*SetControllerFloatValueProc)(
							struct IsadoraParameters*	inIzzyParams,			
							struct ControlInfo*			inCtlInfo,
							IzzyCtlNum					inCtlNum,
							const IzRect*				inCtlRect,
							IzReal						inMin,
							IzReal						inMax,
							IzReal						inValue,
							Boolean						inDirect);
					
typedef	CtlUpdateCode	(*SetControllerDataValueProc)(
							struct IsadoraParameters*	inIzzyParams,			
							struct ControlInfo*			inCtlInfo,
							IzzyCtlNum					inCtlNum,
							const IzRect*				inCtlRect,
							void*						inValue);
					
typedef	CtlUpdateCode	(*SetControllerStringValueProc)(
							struct IsadoraParameters*	inIzzyParams,			
							struct ControlInfo*			inCtlInfo,
							IzzyCtlNum					inCtlNum,
							const IzRect*				inCtlRect,
							const String*				inValue);

// 3.0.3b9 -- added to support new timecode control
typedef	CtlUpdateCode	(*SetControllerTimecodeValueProc)(
							struct IsadoraParameters*	inIzzyParams,			
							struct ControlInfo*			inCtlInfo,
							IzzyCtlNum					inCtlNum,
							const IzRect*				inCtlRect,
							IzValue_Timecode			inValue);
					

	// ControlDrawProc (Required)
	//
	// Draw the user interface for your Control
typedef	void		(*ControlDrawProc)(
						struct IsadoraParameters*	inIzzyParams,			
						struct ControlInfo*			inCtlInfo,
						const IzRect*				inControlRect,
						const IzRect*				inUpdateArea,
						const char*					inControlName,
						Boolean						inLocked,
						Boolean						inSelected);

	// ControlMouseMessageProc (Optional)
	//
	// The user has clicked in your user interface area. Respond
	// to the click as appropriate.
typedef	Boolean		(*ControlMouseMessageProc)(
						struct IsadoraParameters*	inIzzyParams,			
						struct ControlInfo*			inCtlInfo,
						const IzRect*				inControlRect,
						const char*					inControlName,
						Boolean						inSelected,
						IzzyMouseMessage			inMouseMessage,
						IzPoint						inMouse,
						UInt32*						outSleepTime);

// v2.6 ->

	// ControlMouseWhekProc (Optional)
	//
	// The mouse wheel has been moved in your user interface area.
	// Respond to the wheel movement as appropriate.
	
typedef	Boolean		(*ControlMouseWheelProc)(
						struct IsadoraParameters*	inIzzyParams,			
						struct ControlInfo*			inCtlInfo,
						const IzRect*				inControlRect,
						const char*					inControlName,
						Boolean						inSelected,
						IzzyMouseMessage			inMouseMessage,		// either kIzzyMouseWheelX or kIzzyMouseWheelY
						IzPoint						inMouse,
						SInt32						inDelta,			// delta units
						UInt32						inModifiers);		// modifier keys down when wheel was moved
// v2.6 <-

	// ControlEditOptionsProc (Optional)
	//
	// Supply this procedure if you have private options that the
	// user needs to edit via a dialog. It is this function's
	// responsibility to implement the user interface.
typedef	Boolean		(*ControlEditOptionsProc)(
						struct IsadoraParameters*	inIzzyParams,			
						struct ControlInfo*			inCtlInfo,
						const IzRect*				inControlRect,
						const char*					inControlName);
										
	// ControlEditMessageProc (Required)
	//
	// When values for this control are edited by the user
	// the new data is passed to the control via this
	// callback.

enum {
	kControlParamType_Int,
	kControlParamType_Float,
	kControlParamType_Text,
	kControlParamType_BinList,
	kControlParamType_TextBox
};
typedef UInt16	ControlParamType;

// structure definition for the data contained in a bin list
#define kBinListMaxUniqueIDs	256

struct IzzyBinList
{
	MediaCode	mMediaType;
	UInt16		mShowAllOfType;
	UInt32		mBinCount;
	UInt32		mUniqueIDs[kBinListMaxUniqueIDs];
};
typedef IzzyBinList		IzzyBinList;


typedef	Boolean		(*ControlSetParamValueProc)(
						struct IsadoraParameters*	inIzzyParams,			
						struct ControlInfo*			inCtlInfo,
						Boolean						inTestOnly,				// test only, don't store: return true if value is OK
						OSType						inParamID,				// four char code - param ID
						ControlParamType			inParamType,			// param type: int, float, text
						const void*					inParamData,			// pointer to param data - determined by param type
						UInt32						inParamDataLen);		// length of param type
					
typedef	void		(*ControlGetParamValueProc)(
						struct IsadoraParameters*	inIzzyParams,			
						struct ControlInfo*			inCtlInfo,
						OSType						inParamID,				// four char code - param ID
						ControlParamType*			outParamType,			// param type: int, float, text
						const void**				outParamData,			// pointer to param data - determined by param type
						UInt32*						outParamDataLen);		// length of param type

	// GetControlPrivateDataLenProc
	//
	// Called when the control data needs to be stored by Isadora
	//
	// When called, this function should return the number of bytes
	// required to store the Control's private data in a cross-platform
	// compatible way.
typedef	UInt32		(*GetControlPrivateDataLenProc)(
						struct IsadoraParameters*	inIzzyParams,			
						struct ControlInfo*			inControlParams);		

	// GetControlPrivateDataProc
	//
	// Called when the control data needs to be stored by Isadora
	//
	// On entry, outDataBuffer points to a block of data, and 
	// inDataBufferSize gives the size of that buffer. It is the
	// responsibility of the Control to copy the data into the 
	// buffer in a platform independent manner (i.e., the same 
	// for both Mac and PC) and then return the size of the actual
	// data that was copied. Return 0 if the data could not be copied
	// for some reason.
	// 
typedef	UInt32		(*GetControlPrivateDataProc)(
						struct IsadoraParameters*	inIzzyParams,			
						struct ControlInfo*			inControlParams,
						void*						outDataBuffer,			// pointer to a buffer to hold the data
						UInt32						inDataBufferSize);		// size of the data buffer

	// SetControlPrivateDataProc
	//
	// Called when the control data is being restored by Isadora
	//
	// On entry, inDataBuffer points to a block of data, and 
	// inDataBufferSize gives the size of that buffer. Since the
	// data was fetched by Isadora using GetControlPrivateDataProc,
	// the data should be in a platform independent format. It is
	// the responsibility to make a copy of the buffer, convert it
	// to the native format for this machine, and use it to restore
	// its state.
typedef	void		(*SetControlPrivateDataProc)(
						struct IsadoraParameters*	inIzzyParams,			
						struct ControlInfo*			inControlParams,
						const void*					inDataBuffer,
						UInt32						inDataBufferSize);


	// CompareControlPrivateDataProc
	//
	// Compares the private data of the target control with
	// the private data of another control with the same time
	//
	// On entry, inDataBuffer points to a block of data, and 
	// inDataBufferSize gives the size of that buffer. The 
	// function should compare the two buffers for equality,
	// returning TRUE if they are equal and FALSE otherwise.
	//
typedef	Boolean		(*CompareControlPrivateDataProc)(
						struct IsadoraParameters*	inIzzyParams,			
						struct ControlInfo*			inControlParams,
						const void*					inDataBuffer,
						UInt32						inDataBufferSize);


	// ActivateDeactivateControlProc (Optional)
	//
	// Activate or deactivate the control.
	//
	// The control is activated whenever it's owning scene becomes
	// activated, and deactivated when the owning scene is deactivated
typedef	void		(*ActivateDeactivateControlProc)(
						struct IsadoraParameters*	inIzzyParams,			
						struct ControlInfo*			inControlParams,
						Boolean						inActivate);

	// EnableDisableControlProc (Optional)
	//
	// Enables or disables the control.
	//
	// Controls are disabled whenever Edit Mode becomes active
	// in the Control Panel; otherwise they are enabled.
typedef	void		(*EnableDisableControlProc)(
						struct IsadoraParameters*	inIzzyParams,			
						struct ControlInfo*			inControlParams,
						Boolean						inEnable);

	// ControlSetFrameSizeProc (Optional)
	//
	// This function is called whenever the control is moved or resized
typedef	void		(*ControlSetFrameSizeProc)(
						struct IsadoraParameters*	inIzzyParams,			
						struct ControlInfo*			inCtlInfo,
						const IzRect*				inControlRect);


	// GetIndexedMediaIndex (Optional)
	//
	// This function allows Isadora to ascertain which media indexes
	// your controller is using. If inIndex1 is 0, then
	// the function should return the count of media indexes
	// used by your control in outMediaIndex1. When inIndex1 > 0, you should return
	// the media type and media index for that index, as well
	// as a pointer to any private data you would like to use
	// when the SetIndexedMediaIndexProc function is called
	// to actually remap the data
	
typedef	Boolean		(*GetIndexedMediaIndexProc)(
						struct IsadoraParameters*	inIzzyParams,			
						struct ControlInfo*			inControlParams,
						UInt32						inIndex1,				// set to 0 to get number of controllers, > 0 to get controller number
						MediaCode*					outMediaType,
						UInt32*						outMediaIndex1,
						void**						outUserData);

	// SetIndexedMediaIndex (Optional)
	//
	// Allows the media index used by this control to be remapped to
	// another value
	
typedef	void		(*SetIndexedMediaIndexProc)(
						struct IsadoraParameters*	inIzzyParams,			
						struct ControlInfo*			inControlParams,
						UInt32						inIndex1,
						MediaCode					inMediaType,
						UInt32						inOldMediaIndex1,
						UInt32						inNewMediaIndex1,
						void*						inUserData);
					

	// HandleControlMessage (Optional)
	//
	// Allows Isadora to send messages to the control. Currently this
	// is only used to notify the control that an Undo/Redo actions
	// after a control has called SetControlUndoAction_(). If the
	// user specifies a user data parameter to that function, it
	// will be passed back to the caller via the inUserData parameter.
	// 
	// This list of IzzyControlMessage types is defined in IsadoraTypes.h
	// under the heading CONTROL MESSAGE DEFINES
		
typedef	void		(*HandleControlMessageProc)(
						struct IsadoraParameters*	inIzzyParams,			
						struct ControlInfo*			inControlParams,
						IzzyControlMessage			inControlMessage,
						void*						inUserData);
					

////////////////////////////////////////////////////////////////////////////
// Control Flags
////////////////////////////////////////////////////////////////////////////
//
// Flags set in the mControlFlags member of the ControlInfo struct to indicate
// properties of this this Control. Controls who requires these flags should
// set them when they initializes the ControlInfo struct

enum {
	kControlFlags_IsVideoOutput				= 0x00000001,		// indicates this control is a video output
	
	kControlFlags_IsDataOutput				= 0x00000002,		// indicates this control outputs data of some kind

	kControlFlags_IsDeprecated				= 0x00000004,		// indicates this control outputs data of some kind

	kControlFlags_DoesntUseFont				= 0x00000008,		// indicates that this control does not draw any text
	
	kControlFlags_DontDeactivateOnEdit		= 0x00000010,		// control should not be deactivated on when edit box is visible

	kControlFlags_Unused1					= 0x00000020,
	kControlFlags_Unused2					= 0x00000040,
	
	kControlFlags_Plugin_CheckForUpdates	= 0x00000080		// only to be used by official TroikaTronix controls!
																// this flag checks the web for updates using a special
																// system available only to the official plugins
};

#define	kControlFlags_IsOuptut	(kControlFlags_IsVideoOutput | kControlFlags_IsDataOutput)

typedef UInt32	ControlFlags;

////////////////////////////////////////////////////////////////////////////
// Control Info struct
////////////////////////////////////////////////////////////////////////////
//
// Members marked REQ must be filled in by your actor when its
// GetControlInfo function is called.
//
// Members marked OPT can be set to zero or NULL if your actor does not
// need/support the specified function.
//
// Members marked ISADORA are filled in by Isadora and should not
// be changed.


struct ControlInfo {
	ControlIDT							mID;								// REQ:		actor unique ID
	UInt32								mCompatibleWithVersion;				// REQ:		version of Isadora Callbacks for which this this plugin was built
																			//     		ms byte = major revision, ls byte = minor revision
	ControlFlags						mControlFlags;						// OPT:		flags giving information about this control; importantly if it is an output or not

	CONST_ IzSpec						mPluginFileSpec;					// ISADORA:	file spec for this plugin
	CONST_ IzzyCtlPtr					mControl;							// ISADORA: thing that owns this actor
	
	SInt16								mDefaultWidth;						// OPT:		default control width when inserting new control - if set to 0, defaults to 32
	SInt16								mDefaultHeight;						// OPT: 	default control height when inserting new control - if set to 0, defaults to 32 
	
	const char*							mControlName;						// REQ:		pointer to c-string with control's name
	const char*							mEditDialogDefs;					// OPT:		pointer to string that defines the fields in your control's edit settings dialog
	void*								mControlDataPtr;					// OPT:		pointer to control's private data
	
	CreateControlProc					mCreateControlProc;					// REQ:		function called when control is to be created
	DisposeControlProc					mDisposeControlProc;				// REQ:		function called when control is to be disposed
	UpdateControllerNumbersProc			mUpdateControllerNumbersProc;		// REQ:		update controller numbers
	GetIndexedControllerParamInfo		mGetIndexedControllerParamInfo;		// OPT:		returns information about a specific parameter
	GetIndexedControllerNumberProc		mGetIndexedControllerNumberProc;	// REQ:		function called get number of/which controller numbers used by the control
	SetIndexedControllerNumberProc		mSetIndexedControllerNumberProc;	// REQ:		function called set controller numbers used by the control
	SetControllerFloatValueProc			mSetControllerFloatValueProc;		// REQ/OPT:	Required if this control can be linked to numeric properties and track their value
	SetControllerDataValueProc			mSetControllerDataValueProc;		// OPT/OPT:	Required if this control can be linked to data properties and track their value
	SetControllerStringValueProc		mSetControllerStringValueProc;		// OPT/OPT:	Required if this control can be linked to string properties and track their value
	ControlDrawProc						mControlDrawProc;					// REQ:		function called to draw your control
	ControlMouseMessageProc				mControlMouseMessageProc;			// OPT:		function called when mouse is clicked, moved, or released
	
	ControlGetParamValueProc			mControlGetParamValueProc;			// OPT:		function called when control params are edited
	ControlSetParamValueProc			mControlSetParamValueProc;			// OPT:		function called when control params are edited
	GetControlPrivateDataLenProc		mGetControlPrivateDataLenProc;		// OPT:		return the length of the control's private data
	GetControlPrivateDataProc			mGetControlPrivateDataProc;			// OPT:		get a pointer to the control's private data
	SetControlPrivateDataProc			mSetControlPrivateDataProc;			// OPT:		get a pointer to the control's private data
	
	ControlEditOptionsProc				mControlEditOptionsProc;			// OPT:		edit private options/settings
	
	ActivateDeactivateControlProc		mActivateDeactivateControlProc;		// OPT:		activate / deactivate the control

	ControlSetFrameSizeProc				mControlSetFrameSizeProc;			// OPT:		called when control moved or resized
	
	GetIndexedMediaIndexProc			mGetIndexedMediaIndexProc;			// OPT:		function called to get a number of/which media indexes used by this control
	SetIndexedMediaIndexProc			mSetIndexedMediaIndexProc;			// OPT:		function called to set a media indexes used by this control

	HandleControlMessageProc			mHandleControlMessageProc;			// OPT:		handles a message being sent to the control
	
	EnableDisableControlProc			mEnableDisableControlProc;			// OPT:		controls are enabled when Control Panel's Edit Mode is disabled and vice-versa
	
	CONST_ Boolean						mActive;							// ISADORA: true if this control is active now
	CONST_ Boolean						mEnabled;							// ISADORA: true if this control is enabled now

	CompareControlPrivateDataProc		mCompareControlPrivateDataProc;		// OPT:		compare private data for equality
																			//			only required if a memcmp of the two 
																			//			data blocks will not always return the
																			//			correct result

	ControlMouseWheelProc				mControlMouseWheelProc;				// OPT:		(v2.6) function called when wheel is moved over control
	
	SetControllerTimecodeValueProc		mSetControllerTimecodeValueProc;	// OPT:		(v3.0.3b9) function called to deliver timecode to control
};
typedef struct ControlInfo ControlInfo;


////////////////////////////////////////////////////////////////////////////
// ISADORA CALLBACK FUNCTIONS
////////////////////////////////////////////////////////////////////////////

// The ActorInfo struct has the field mCompatbileWithVersion. Since
// v3.0.9b05, this should be set to kActorCompatilibilityVersion_v1
// which is defined as 0x0100.
//
// In older versions of the SDK, kCurrentIsadoraCallbackVersion was
// defined here. The value had been 0x0100 since the inception of
// Isadora
//
// New plugins should use kActorCompatilibilityVersion_v1 unless they
// _always_ require callbacks that were added with a later version.
//
// Plugins can also use the IsAtLeastIsadoraCallbackVersion() macro
// to dynamically check for the existence of certain callbacks and
// make use of those callbacks only if they are actually present.

#define	kActorCompatilibilityVersion_v1				0x0100	
#define	kActorCompatilibilityVersion_v309b05		VERSION_NUMBER(3, 0, 9, kReleaseType_Beta, 5)	// 3.0.9b5
#define	kActorCompatilibilityVersion_v309b06		VERSION_NUMBER(3, 0, 9, kReleaseType_Beta, 6)	// 3.0.9b6
#define	kActorCompatilibilityVersion_v312b10		VERSION_NUMBER(3, 1, 2, kReleaseType_Beta, 10)	// 3.1.2b10
#define	kActorCompatilibilityVersion_v322b50		VERSION_NUMBER(3, 2, 2, kReleaseType_Beta, 50)	// 3.2.2b50
#define	kActorCompatilibilityVersion_v323b01		VERSION_NUMBER(3, 2, 3, kReleaseType_Beta, 1)	// 3.2.3b01

#define	kActorCompatilibilityVersion_Current		kActorCompatilibilityVersion_v323b01


// -----------------------------------------------------------------------------
// Callback Error Return Codes
// -----------------------------------------------------------------------------
enum {
	kIzzyNoError = 0,
	kIzzyErr_UnexpectedError,				// unexpected internal error
	kIzzyErr_InvalidParam,					// a parameter passed to a callback was invalid
	kIzzyErr_BadThingPtr,					// the thing ptr passed to a callback is invalid
	kIzzyErr_InvalidPropertyType,			// the property type passed to a callback is not kInputProperty or kOutputProperty
	kIzzyErr_PropertyIndexOutOfRange,		// the property index is not between 1 and the number of properties of the specified type
	kIzzyErr_ValueTypesDontMatch,			// when setting or changing min/max/value of a property, the types in the ValuePtrs don't all match
	kIzzyErr_PropertyTypeOrIndexInvalid,	// a bad property type code or index that is out of range
	kIzzyErr_InvalidJumpMode,				// caller passed an invalid jump mode
	kIzzyErr_SceneDoesntExist,				// scene in jump or get scene name call does not exist,
	kIzzyErr_ResourceNotFound,				// requested resource cannot be found
	kIzzyErr_SnapshotDoesntExist,			// requested snapshot cannot does not exist

	// v2.6->
	kIzzyErr_OSCAddrNotFound,
	kIzzyErr_OSCAddrPartialName,
	// v2.6 <-
	
	kIzzyErr_BadControlPtr,					// the control ptr passed to a callback is invalid
	
	kIzzyErr_LastError

};
typedef	SInt16		IzzyError;			// the isadora error code type


enum {
	kIzzyPluginFailed_BadValue = 1
};
typedef	SInt16		PluginFailureError;	// the plugin error code type

void
ShowIsadoraPluginFailedAlert(
	ActorInfo*			inActorInfo,
	PluginFailureError	inError);


// All of the callbacks below come with a macro that allows you to call
// the callback without explicity using the indirection operator (->) to
// access the function. The first parameter to the macro is always the
// IsadoraParameters pointer that is passed to your actor.

// -----------------------------------------------------------------------------
// ShowAlert_
// -----------------------------------------------------------------------------

#define	ShowAlert_(izzyParamsPtr, errStr, infoStr ) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mShowAlertProc)( (errStr), (infoStr) )

typedef	void			(*ShowAlertProc)(
								const char*	inErrorString,
								const char*	inExplanationString);
					
// -----------------------------------------------------------------------------
// CreateMessageReceiver_
// -----------------------------------------------------------------------------
// Watchers typically need to be notified about events coming in from the real
// world. These include, mouse, keyboard, MIDI, video frames, etc. To be notified
// of these events, your actor will need to register a message receiver function
// using the CreateMessageReceiver_ macro.
// 
// The second parameter (msgReceiverFunc) to the macro is the address of your
// message receiver function.
// 
// The third parameter (desiredPort) is only valid for MIDI messages - it
// specifies which port (i.e., cable) the message must arrive on for your message
// receiver function to be notified. Other messages ignore this parameter.
// 
// The fourth parameter (messageMask) is a mask that defines which messages will
// be delivered to your message receiver function. For example, if you wanted to
// be notified whenever the user pressed a key on the computer keyboard, you
// would use kWantKeyDown. Since this parameter is a mask, you can use any
// combination of the mask flags to receive messages of the various types. See
// MessageReceiverCommon.h for all of the possible masks. Note that you may call
// CreateMessageReceiver_ with a mask of kWantNone to register your message
// receiver function but not receive any messages yet. You can then later call
// UpdateMessageReceiver or UpdateMessageReceiverMask to change the mask and
// start receiving messages.
// 
// The fifth parameter (refCon) is a reference constant that will be passed to
// your message receiver proc whenever it is called.
// 
// The result of this function is a MessageReceiverRef. You need to keep track of
// this value as it is used to dispose the message receiver and to update the
// MessageMask, if so desired.
// 
// INTERRUPTS: This can be called at interrupt time

// this version can be called by non-actor items -- the actor info pointer is set to NULL
#define	CreateMessageReceiver_(izzyParamsPtr, msgReceiverFunc, desiredPort, messageMask, refCon) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mCreateMessageReceiverProc)( (msgReceiverFunc), (desiredPort), (messageMask), nil, nil, (refCon) )

// this version should be called by actors -- the actor info pointer is required
#define	CreateActorMessageReceiver_(izzyParamsPtr, inActorInfo, msgReceiverFunc, desiredPort, messageMask, refCon) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mCreateMessageReceiverProc)( (msgReceiverFunc), (desiredPort), (messageMask), (inActorInfo), (nil), (refCon) )

// this version should be called by controls -- the control info pointer is required
#define	CreateControlMessageReceiver_(izzyParamsPtr, inControlInfo, msgReceiverFunc, desiredPort, messageMask, refCon) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mCreateMessageReceiverProc)( (msgReceiverFunc), (desiredPort), (messageMask), (nil), (inControlInfo), (refCon) )

typedef	MessageReceiverRef	(*CreateMessageReceiverProc)(						// returns MessageReceiverRef opaque reference
								MessageReceiveFunction*	inReceiveFunction,		// address of receive function
								PortIndex				inDesiredPortIndex,		// for MIDI, desired input port index
								MessageMask				inMessageMask,			// mask specifies what messages, upon reception, will
																				// cause the receive function to be called
								ActorInfo*				inActorInfo,			// called by an actor, this should be the ActorInfo struct for this actor
																				// this parameter can be NULL if caller is not an actor
								ControlInfo*			inCtlInfo,				// called by aa control, this should be the ControlInfo struct for this control
																				// this parameter can be NULL if caller is not an actor
								IzRefCon				inRefCon);				// ref con passed to receive function
					

// -----------------------------------------------------------------------------
// DisposeMessageReceiver_
// -----------------------------------------------------------------------------
// Removes your message receiver fucntion from the Message Receiver list. This
// function should always be called to balance a call to the
// CreateMessageReceiver_ macro.
// 
// INTERRUPTS: This can be called at interrupt time

#define	DisposeMessageReceiver_(izzyParamsPtr, msgReceiver) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mDisposeMessageReceiverProc)( (msgReceiver) )

typedef void				(*DisposeMessageReceiverProc)(						// disconnect a message receiver proc	
								MessageReceiverRef			inMessageReceiver);	// address of receive function
					

// -----------------------------------------------------------------------------
// UpdateMessageReceiver_
// -----------------------------------------------------------------------------
// If you want to change what messages will cause your message receiver fucntion
// to be called, then call the UpdateMessageReceiver macro to update the settings
// for your proc in the message receiver queue.
// 
// INTERRUPTS: This can be called at interrupt time

#define	UpdateMessageReceiver_(izzyParamsPtr, msgReceiver, msgRcvFunc, portIndex, msgMask, refCon) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mUpdateMessageReceiverProc)((msgReceiver), (msgRcvFunc), (portIndex), (msgMask), 0, 0, (refCon))

#define	UpdateActorMessageReceiver_(izzyParamsPtr, msgReceiver, inActorInfo, msgRcvFunc, portIndex, msgMask, refCon) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mUpdateMessageReceiverProc)((msgReceiver), (msgRcvFunc), (portIndex), (msgMask), (inActorInfo), 0, (refCon))

#define	UpdateControlMessageReceiver_(izzyParamsPtr, msgReceiver, inControlInfo, msgRcvFunc, portIndex, msgMask, refCon) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mUpdateMessageReceiverProc)((msgReceiver), (msgRcvFunc), (portIndex), (msgMask), 0, (inControlInfo), (refCon))

typedef	void				(*UpdateMessageReceiverProc)(						// update parameters of message receiver proc
								MessageReceiverRef		inMessageReceiver,		// message receiver opaque reference
								MessageReceiveFunction*	inReceiveFunction,		// address of receive function
								PortIndex				inDesiredPortIndex,		// for MIDI, desired input port index
								MessageMask				inMessageMask,			// mask specifies what messages will be received
								ActorInfo*				inActorInfo,			// called by an actor, this should be the ActorInfo struct for this actor
																				// this parameter can be NULL if caller is not an actor
								ControlInfo*			inCtlInfo,				// called by aa control, this should be the ControlInfo struct for this control
																				// this parameter can be NULL if caller is not an actor
								IzRefCon				inRefCon);				// ref con passed when receiver is called
					

// -----------------------------------------------------------------------------
// UpdateMessageReceiverMask_
// -----------------------------------------------------------------------------
// Allows you to change what messages will cause your message receiver function
// be called by changing the MessageMask for your function.
// 
// INTERRUPTS: This can be called at interrupt time

#define	UpdateMessageReceiverMask_(izzyParamsPtr, msgReceiver, msgMask) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mUpdateMessageReceiverMaskProc)((msgReceiver), (msgMask))

typedef	void				(*UpdateMessageReceiverMaskProc)(					
								MessageReceiverRef		inMessageReceiver,		// message receiver opaque reference
								MessageMask				inMessageMask);			// mask specifies what messages will be received
					
// -----------------------------------------------------------------------------
// SetActorFlags_
// -----------------------------------------------------------------------------
// Allows the caller to set the actor flags, which determines if the actor
// is an output. Some modules may need to change these flags if they start
// outputing based on user input, or if they want to change from being
// a video output to a data output
// 
#define	SetActorFlags_(izzyParamsPtr, actorInfo, flags, mask) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetActorFlagsProc)((actorInfo)->mThing, (flags), (mask))

typedef void				(*SetActorFlagsProc)(								
								ThingPtr					inThing,				// thing that owns the property
								ActorFlags					inActorFlags,			// the new flags setting
								ActorFlags					inActorFlagsMask);		// the mask indicating which flags are affected

// -----------------------------------------------------------------------------
// SetPropertyNotApplicableFlag
// -----------------------------------------------------------------------------
// If an input or output of an actor is not meaningful in a certain context,
// you can set the "not applicable" flag to indicate visually to the user
// that the input or output has no function.

#define	SetPropertyNotApplicableFlag_(izzyParamsPtr, actorInfo, propType, propIndex1, notApplicable) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetPropertyNotApplicableFlagProc)((actorInfo)->mThing, (propType), (propIndex1), (notApplicable), NULL)

#define	SetPropertyNotApplicableFlagWithText_(izzyParamsPtr, actorInfo, propType, propIndex1, notApplicable, naText) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetPropertyNotApplicableFlagProc)((actorInfo)->mThing, (propType), (propIndex1), (notApplicable), (naText))

typedef IzzyError			(*SetPropertyNotApplicableFlagProc)(				// returns an non-zero error code if there is an error
								ThingPtr				inThing,				// thing that owns the property
								PropertyType			inPropertyType,			// the property type, either input or output
								SInt32					inPropertyIndex1,		// index of the input property
								Boolean					inNotApplicable,		// true if the property should be marked as not applicable
								const char*				inNotApplicableText);	// the text that will be shown when inNotApplicable is true


// -----------------------------------------------------------------------------
// SetPropertyDisallowWhitespaceFlag_
// -----------------------------------------------------------------------------
// For string properties, call this function with a parameter of true to prevent
// any whitepspace characters (space, tab, return, etc.) from being entered into
// the text editor that opens when the user clicks on this property.

#define	SetPropertyDisallowWhitespaceFlag_(izzyParamsPtr, actorInfo, propType, propIndex1, disallowWhitespaace) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetPropertyDisallowWhitespaceFlagProc)((actorInfo)->mThing, (propType), (propIndex1), (disallowWhitespaace))

typedef IzzyError			(*SetPropertyDisallowWhitespaceFlagProc)(				// returns an non-zero error code if there is an error
								ThingPtr				inThing,					// thing that owns the property
								PropertyType			inPropertyType,				// the property type, either input or output
								SInt32					inPropertyIndex1,			// index of the input property
								Boolean					inDisallowWhitespace);		// true if the property should not allow whitespace characters to be entered

// -----------------------------------------------------------------------------
// SetPropertyDisallowCarriageReturnsFlag_
// -----------------------------------------------------------------------------
// For string properties, call this function with a parameter of true to prevent
// any carriage return characters from being entered into the text editor that
// opens when the user clicks on this property.

#define	SetPropertyDisallowCarriageReturnsFlag_(izzyParamsPtr, actorInfo, propType, propIndex1, disallowWhitespaace) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetPropertyDisallowCarriageReturnsFlagProc)((actorInfo)->mThing, (propType), (propIndex1), (disallowWhitespaace))

typedef IzzyError			(*SetPropertyDisallowCarriageReturnsFlagProc)(			// returns an non-zero error code if there is an error
								ThingPtr				inThing,					// thing that owns the property
								PropertyType			inPropertyType,				// the property type, either input or output
								SInt32					inPropertyIndex1,			// index of the input property
								Boolean					inDisallowCarriageReturns);	// true if the property should not allow whitespace characters to be entered

// -----------------------------------------------------------------------------
// SetPropertyIsPasswordFlag
// -----------------------------------------------------------------------------
// To display a text property as a password, meaning the value is displayed as a
// series of 'bullet' characters, call SetPropertyIsPasswordFlag_() with the
// inIsPassword paraemter set to TRUE. Note that once you do this, the only
// way to see the password is to click on the text field and to open its editor.

#define	SetPropertyIsPasswordFlag_(izzyParamsPtr, actorInfo, propType, propIndex1, inIsPassword) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetPropertyIsPasswordFlagProc)((actorInfo)->mThing, (propType), (propIndex1), (inIsPassword))

typedef IzzyError			(*SetPropertyIsPasswordFlagProc)(					// returns an non-zero error code if there is an error
								ThingPtr				inThing,				// thing that owns the property
								PropertyType			inPropertyType,			// the property type, either input or output
								SInt32					inPropertyIndex1,		// index of the input property
								Boolean					inIsPassword);			// true if the property should be displayed as a password


// -----------------------------------------------------------------------------
// SetPropertyIsInvalidFlag
// -----------------------------------------------------------------------------
// To display a text property as invalid  call SetPropertyIsInvalidFlag_() with
// the inIsInvalid paraemter set to TRUE. By default this will color the text red.
// If you wish to set your own color, call SetPropertyIsInvalidFlagWithColor_()

#define	SetPropertyIsInvalidFlag_(izzyParamsPtr, actorInfo, propType, propIndex1, inIsInvalid) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetPropertyIsInvalidFlagProc)((actorInfo)->mThing, (propType), (propIndex1), (inIsInvalid), 0)

#define	SetPropertyIsInvalidFlagWithColor_(izzyParamsPtr, actorInfo, propType, propIndex1, inIsInvalid, inColor) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetPropertyIsInvalidFlagProc)((actorInfo)->mThing, (propType), (propIndex1), (inIsInvalid), (inColor))

typedef IzzyError			(*SetPropertyIsInvalidFlagProc)(					// returns an non-zero error code if there is an error
								ThingPtr				inThing,				// thing that owns the property
								PropertyType			inPropertyType,			// the property type, either input or output
								SInt32					inPropertyIndex1,		// index of the input property
								Boolean					inIsInvalid,			// true if the property should be displayed as a password
								UInt32					inInvalidColor);		// set to 0 to use default; otherwise set to a ARGB color, e.g., FF00FF00 for green


// -----------------------------------------------------------------------------
// SetPropertyTimecodeFPS
// -----------------------------------------------------------------------------
// Sets the timecode type of the specified property based on the frame rate
// given by inFPS. This method is deprecated. You should remove it from your code

#define	SetPropertyTimecodeFPS_DEPRECATED(izzyParamsPtr, actorInfo, propType, propIndex1, fps) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetPropertyTimecodeFPSProc)((actorInfo)->mThing, (propType), (propIndex1), (fps))

typedef IzzyError			(*SetPropertyTimecodeFPSProc_DEPRECATED)(			// returns an non-zero error code if there is an error
								ThingPtr				inThing,				// thing that owns the property
								PropertyType			inPropertyType,			// the property type, either input or output
								PropertyIndex			inPropertyIndex1,		// index of the input property
								double					inFPS);					// can be any of the rates in TimecodeType


// -----------------------------------------------------------------------------
// SetPropertyTimecodeType
// -----------------------------------------------------------------------------
// Sets the timecode type of the specified property. Note that this update is
// performed "silently" -- even though the property value changes because the
// timecode type has changes, the seconds value of the timecode will not change.
// Thus calling this proc will not cause HandlePropertyChangeValue() to be called.

#define	SetPropertyTimecodeType_(izzyParamsPtr, actorInfo, propType, propIndex1, inTimecodeType) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetPropertyTimecodeTypeProc)((actorInfo)->mThing, (propType), (propIndex1), (inTimecodeType))

typedef IzzyError			(*SetPropertyTimecodeTypeProc)(						// returns an non-zero error code if there is an error
								ThingPtr				inThing,				// thing that owns the property
								PropertyType			inPropertyType,			// the property type, either input or output
								PropertyIndex			inPropertyIndex1,		// index of the input property
								TimecodeType			inTimecodeType);		// see TimecodeType


// -----------------------------------------------------------------------------
// ChangePropertyType_
// -----------------------------------------------------------------------------
// ChangePropertyType allows you to reconfigure the type of one of your input or
// output properties. This macro is usually called in response to your
// HandlePropertyChangeType function being called for a muteable input or output
// property.
// 
// You must supply the Property Type (kInputProperty or kOutputProperty), and the
// one-based property index to identify which property will be changed. You then
// supply the possible display formats and default display formats (see
// IsadoraTypes.h for the possible values) and the value's new minimum, maximum,
// and current value.
// 
// INTERRUPTS: This should not be called at interrupt time

#define	ChangePropertyType_(izzyParamsPtr, actorInfo, propDefSrc, propType, propIndex1, dispFmts, curDispFmt, min, max, scaleMin, scaleMax, value) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mChangePropertyTypeProc)((actorInfo)->mThing, (propDefSrc), (propType), (propIndex1), (dispFmts), (curDispFmt), (min), (max), (scaleMin), (scaleMax), (value))

#define	ChangePropertyTypeAndName_(izzyParamsPtr, actorInfo, propDefSrc, propType, propIndex1, name, dispFmts, curDispFmt, min, max, scaleMin, scaleMax, value) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mChangePropertyTypeAndNameProc)((actorInfo)->mThing, (propDefSrc), (propType), (propIndex1), (name), (dispFmts), (curDispFmt), (min), (max), (scaleMin), (scaleMax), (value))

#define	MutateProperty_(izzyParamsPtr, actorInfo, propDefSrc, propType, propIndex1, name, customDataType, dispFmts, curDispFmt, mutableGroup, min, max, scaleMin, scaleMax, value) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mMutatePropertyProc)((actorInfo)->mThing, (propDefSrc), (propType), (propIndex1), (name), (customDataType), (dispFmts), (curDispFmt), (mutableGroup), (min), (max), (scaleMin), (scaleMax), (value))

typedef IzzyError			(*ChangePropertyTypeProc)(										// returns an non-zero error code if there is an error
								ThingPtr							inThing,				// thing that owns the property
								const PropertyDefinedValueSource*	inPropDefSrc,			// CAN BE NULL
								PropertyType						inPropertyType,			// the property type
								PropertyIndex						inPropertyIndex1,		// index of the property
								PropertyDispFormat					inDisplayFormats,
								PropertyDispFormat					inCurFormat,
								const ValuePtr						inMin,
								const ValuePtr						inMax,
								const ValuePtr						inScaleMin,
								const ValuePtr						inScaleMax,
								const ValuePtr						inValue);

typedef IzzyError			(*ChangePropertyTypeAndNameProc)(								// returns an non-zero error code if there is an error
								ThingPtr							inThing,				// thing that owns the property
								const PropertyDefinedValueSource*	inPropDefSrc,			// CAN BE NULL
								PropertyType						inPropertyType,			// the property type
								PropertyIndex						inPropertyIndex1,		// index of the property
								const char*							inName,					// NULL = no change
								PropertyDispFormat					inDisplayFormats,
								PropertyDispFormat					inCurFormat,
								const ValuePtr						inMin,
								const ValuePtr						inMax,
								const ValuePtr						inScaleMin,
								const ValuePtr						inScaleMax,
								const ValuePtr						inValue);

typedef IzzyError			(*MutatePropertyProc)(											// returns an non-zero error code if there is an error
								ThingPtr							inThing,				// thing that owns the property
								const PropertyDefinedValueSource*	inPropDefSrc,			// CAN BE NULL
								PropertyType						inPropertyType,			// the property type
								PropertyIndex						inPropertyIndex1,		// index of the property
								const char*							inName,					// NULL = no change
								PropIDT								inCustomDataType,
								PropertyDispFormat					inDisplayFormats,
								PropertyDispFormat					inCurFormat,
								MutableGroup						inMutableGroup,
								const ValuePtr						inMin,
								const ValuePtr						inMax,
								const ValuePtr						inScaleMin,
								const ValuePtr						inScaleMax,
								const ValuePtr						inValue);


// -----------------------------------------------------------------------------
// CopyPropDefValueSource_
// -----------------------------------------------------------------------------
#define	CopyPropDefValueSource_(izzyParamsPtr, actorInfo, srcPropType, srcPropIndex, dstPropType, dstPropIndex) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mCopyPropDefValueSourceProc)((actorInfo)->mThing, (srcPropType), (srcPropIndex), (dstPropType), (dstPropIndex))

typedef IzzyError			(*CopyPropDefValueSourceProc)(							// returns an non-zero error code if there is an error
								ThingPtr					inThing,				// thing that owns the property
								PropertyType				inSrcPropType,			// the source property type
								PropertyIndex				inSrcPropertyIndex1,	// index of the source property
								PropertyType				inDstPropType,			// the destination property type
								PropertyIndex				inDstPropertyIndex1);	// index of the destination property

// -----------------------------------------------------------------------------
// GetPropertyDisplayFormats_
// -----------------------------------------------------------------------------
// Gets the available and current property display formats for a property
// 
// INTERRUPT SAFE: This function can be called at interrupt time.

#define	GetPropertyDisplayFormats_(izzyParamsPtr, actorInfo, propType, propIndex1, outAvail, outCur) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetPropertyDisplayFormatsProc)((actorInfo)->mThing, (propType), (propIndex1), (outAvail), (outCur))

typedef IzzyError			(*GetPropertyDisplayFormatsProc)(					// returns an non-zero error code if there is an error
								ThingPtr				inThing,				// thing that owns the property
								PropertyType			inPropertyType,			// the property type, either input or output
								SInt32					inPropertyIndex1,		// index of the input property
								PropertyDispFormat*		outAvailFormats,		// the property's available display formats
								PropertyDispFormat*		outCurFormat);			// the property's current display format

// -----------------------------------------------------------------------------
// GetPropertyID_
// -----------------------------------------------------------------------------
// Gets the minimum and maximum value for the specified property
// 

#define	GetPropertyID_(izzyParamsPtr, actorInfo, propType, propIndex1, outPropID) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetPropertyIDProc)((actorInfo)->mThing, (propType), (propIndex1), (outPropID))

typedef IzzyError			(*GetPropertyIDProc)(								// returns an non-zero error code if there is an error
								ThingPtr				inThing,				// thing that owns the property
								PropertyType			inPropertyType,			// the property type, either input or output
								SInt32					inPropertyIndex1,		// index of the input property
								PropIDT*				outPropertyID);			// resulting property ID

// -----------------------------------------------------------------------------
// GetPropertyMinMax_
// -----------------------------------------------------------------------------
// Gets the minimum and maximum value for the specified property
// 
// INTERRUPT SAFE: This function can be called at interrupt time.

#define	GetPropertyMinMax_(izzyParamsPtr, actorInfo, propType, propIndex1, minValue, maxValue, initValue) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetPropertyMinMaxProc)((actorInfo)->mThing, (propType), (propIndex1), (minValue), (maxValue), (initValue))

typedef IzzyError			(*GetPropertyMinMaxProc)(							// returns an non-zero error code if there is an error
								ThingPtr				inThing,				// thing that owns the property
								PropertyType			inPropertyType,			// the property type, either input or output
								SInt32					inPropertyIndex1,		// index of the input property
								ValuePtr				outPropertyMinimum,		// the property's mimumum value
								ValuePtr				outPropertyMaximum,		// the property's maximum value
								ValuePtr				outPropertyInitValue);

// -----------------------------------------------------------------------------
// SetPropertyMinMax_
// -----------------------------------------------------------------------------
// Sets the minimum and maximum value for the specified property.
//
// *** SIDE EFFECTS ***
//
// If the value is not within the range specified by the min and max, then the
// value will be forced in range.
//
// The Scale To Minimum and Scale To Maximum are set to match the new Minimum
// and Maximum.
//
// INTERRUPT SAFE: This function can be called at interrupt time.

#define	SetPropertyMinMax_(izzyParamsPtr, actorInfo, propType, propIndex1, minValue, maxValue, initValue) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetPropertyMinMaxProc)((actorInfo)->mThing, (propType), (propIndex1), (minValue), (maxValue), (initValue))

typedef IzzyError			(*SetPropertyMinMaxProc)(							// returns an non-zero error code if there is an error
								ThingPtr				inThing,				// thing that owns the property
								PropertyType			inPropertyType,			// the property type, either input or output
								SInt32					inPropertyIndex1,		// index of the input property
								const ValuePtr			inPropertyMinimum,		// the new mimumum value
								const ValuePtr			inPropertyMaximum,		// the new maximum value
								const ValuePtr			inPropertyInit);		// the new init value -- can be nil

// -----------------------------------------------------------------------------
// SmartSetPropertyMinMax_
// -----------------------------------------------------------------------------
// Has the same function as SetPropertyMinMax_, except that this version looks
// at the current minimum and maximum values in relation to the current scale
// min and scale max values. The idea is to preserve inverse relationships if
// they have specified them as such, or to warn them that the scale min and
// scale max values will be reset.
//
// 1) If the old Min == old Scale Min, and old Max = old Scale Max, then set
// Min and Scale Min to the new minimum, and Max and Scale Max to the new
// maximum.
// 
// 2) If the old Min == old Scale Max, and old Max = old Scale Min, then set
// Min and Scale _Max_ to the new minimum, and Max and Scale _Min_ to the new
// maximum.
//
// 3) Otherwise, let the user know that the value of the specified property
// has been reset, indicating that this was because of a change in the 
// property identified by inChangedPropType and inChangedPropIndex1.
// 
// INTERRUPT SAFE: This function can be called at interrupt time.

#define	SmartSetPropertyMinMax_(izzyParamsPtr, actorInfo, propType, propIndex1, minValue, maxValue, notifyOfChangedValue, changedPropType, changedPropIndex) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSmartSetPropertyMinMaxProc)((actorInfo)->mThing, (propType), (propIndex1), (minValue), (maxValue), (notifyOfChangedValue), (changedPropType), (changedPropIndex))

typedef IzzyError			(*SmartSetPropertyMinMaxProc)(						// returns an non-zero error code if there is an error
								ThingPtr				inThing,				// thing that owns the property
								PropertyType			inPropertyType,			// the property type, either input or output
								SInt32					inPropertyIndex1,		// index of the input property
								const ValuePtr			inPropertyMinimum,		// the new mimumum value
								const ValuePtr			inPropertyMaximum,		// the new maximum value
								Boolean					inNotifyOfLimitedValue,	// if true and the actual value of the property changes, notify the input via HandlePropertyChangeValue
								PropertyType			inChangedPropType,		// type of property that caused this change - used when warning the user in case 3
								SInt32					inChangedPropIndex1);	// index of property that caused this change - used when warning the user in case 3

// -----------------------------------------------------------------------------
// GetInputPropertyValue_
// -----------------------------------------------------------------------------
// Retrieves the value of one of your actor's input properties. The result is
// returned as a ValuePtr struct, which gives the value's type and actual value.
// 
// INTERRUPT SAFE: This function can be called at interrupt time.

#define	GetInputPropertyValue_(izzyParamsPtr, actorInfo, propIndex1) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetInputPropertyValueProc)((actorInfo)->mThing, (propIndex1))

typedef const ValuePtr		(*GetInputPropertyValueProc)(						// returns the value of the input property
								ThingPtr				inThing,				// thing that owns the property
								SInt32					inPropertyIndex1);		// index of the input property

// -----------------------------------------------------------------------------
// SetInputPropertyValue_
// -----------------------------------------------------------------------------
// Sets the value of one of your input properties. Note that the value will be
// limited to the property's minimum and maximum values.
// 
// INTERRUPT SAFE: This function can be called at interrupt time.

#define	SetInputPropertyValue_(izzyParamsPtr, actorInfo, propIndex1, value) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetInputPropertyValueProc)((actorInfo)->mThing, (propIndex1), (value))

typedef IzzyError			(*SetInputPropertyValueProc)(						// returns an non-zero error code if there is an error
								ThingPtr				inThing,				// thing that owns the property
								PropertyIndex			inPropertyIndex1,		// index of the input property
								const ValuePtr			inValue);				// new property value

// -----------------------------------------------------------------------------
// SetOutputPropertyValue_
// -----------------------------------------------------------------------------
// Sets the value of one of your output properties. Note that the value will be
// limited to the property's minimum and maximum values.
// 
// INTERRUPT SAFE: This function can be called at interrupt time.

#define	SetOutputPropertyValue_(izzyParamsPtr, actorInfo, propIndex1, value) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetOutputPropertyValueProc)((actorInfo)->mThing, (propIndex1), (value))

typedef IzzyError			(*SetOutputPropertyValueProc)(						// returns an non-zero error code if there is an error
								ThingPtr				inThing,				// thing that owns the property
								PropertyIndex			inPropertyIndex1,		// index of the output property
								const ValuePtr			inValue);				// new property value

// -----------------------------------------------------------------------------
// MarkOutputPropertyAsNewForNextCycle_
// -----------------------------------------------------------------------------
// This marks the output property as being 'new' so that it's receivers will
// receive the value even it it's time slice matches the current time slice.
// You should not use this call unless you know what you're doing. It is
// intended only to be used by actors that allow inter scene communication
// like the Listner actor
//
// INTERRUPT SAFE: This function can be called at interrupt time.

#define	MarkOutputPropertyAsNewForNextCycle_(izzyParamsPtr, actorInfo, propIndex1) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mMarkOutputPropertyAsNewForNextCycleProc)((actorInfo)->mThing, (propIndex1))

typedef IzzyError			(*MarkOutputPropertyAsNewForNextCycleProc)(		// returns an non-zero error code if there is an error
								ThingPtr				inThing,				// thing that owns the property
								PropertyIndex			inPropertyIndex1);

// -----------------------------------------------------------------------------
// GetInputPropertyScaleToMinMax_
// -----------------------------------------------------------------------------
// Gets the minimum and maximum scale to value for the specified input property
// 
// INTERRUPT SAFE: This function can be called at interrupt time.

#define	GetInputPropertyScaleToMinMax_DEPRECATED(izzyParamsPtr, actorInfo, propIndex1, minValue, maxValue) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetInputPropertyScaleToMinMaxProc)((actorInfo)->mThing, (propIndex1), (minValue), (maxValue))

typedef IzzyError			(*GetInputPropertyScaleToMinMaxProc_DEPRECATED)(	// returns an non-zero error code if there is an error
								ThingPtr				inThing,				// thing that owns the property
								SInt32					inPropertyIndex1,		// index of the input property
								ValuePtr				inPropertyMinimum,		// the property's mimumum value
								ValuePtr				inPropertyMaximum);		// the property's maximum value

#define	GetPropertyScaleToMinMax_(izzyParamsPtr, actorInfo, propType, propIndex1, minValue, maxValue) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetInputPropertyScaleToMinMaxProc)((actorInfo)->mThing, (propType), (propIndex1), (minValue), (maxValue))

typedef IzzyError			(*GetPropertyScaleToMinMaxProc)(				// returns an non-zero error code if there is an error
								ThingPtr				inThing,				// thing that owns the property
								PropertyType			inPropertyType,
								SInt32					inPropertyIndex1,		// index of the input property
								ValuePtr				inPropertyMinimum,		// the property's mimumum value
								ValuePtr				inPropertyMaximum);		// the property's maximum value

// -----------------------------------------------------------------------------
// SetInputPropertyScaleToMinMax_
// -----------------------------------------------------------------------------
// Sets the scale to minimum and scale to maximum for the specified input
// property. Note that if the current input value of this property is not
// in range when this function is called, it will be forced into range
// 
// INTERRUPT SAFE: This function can be called at interrupt time.

#define	SetInputPropertyScaleToMinMax_DEPRECATED(izzyParamsPtr, actorInfo, propIndex1, minValue, maxValue) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetInputPropertyScaleToMinMaxProc)((actorInfo)->mThing, (propIndex1), (minValue), (maxValue))

typedef IzzyError			(*SetInputPropertyScaleToMinMaxProc_DEPRECATED)(	// returns an non-zero error code if there is an error
								ThingPtr				inThing,				// thing that owns the property
								SInt32					inPropertyIndex1,		// index of the input property
								const ValuePtr			inPropertyMinimum,		// the new mimumum value
								const ValuePtr			inPropertyMaximum);		// the new maximum value

#define	SetPropertyScaleToMinMax_(izzyParamsPtr, actorInfo, propType, propIndex1, minValue, maxValue) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetInputPropertyScaleToMinMaxProc)((actorInfo)->mThing, (propType), (propIndex1), (minValue), (maxValue))

typedef IzzyError			(*SetPropertyScaleToMinMaxProc)(				// returns an non-zero error code if there is an error
								ThingPtr				inThing,				// thing that owns the property
								PropertyType			inPropertyType,
								SInt32					inPropertyIndex1,		// index of the input property
								const ValuePtr			inPropertyMinimum,		// the new mimumum value
								const ValuePtr			inPropertyMaximum);		// the new maximum value

// -----------------------------------------------------------------------------
// GetOutputPropertyLimitMinMax_
// -----------------------------------------------------------------------------
// Gets the limit minimum and limit maximum for the specified output property.
//
// INTERRUPT SAFE: This function can be called at interrupt time.

#define	GetOutputPropertyLimitMinMax_(izzyParamsPtr, actorInfo, propIndex1, minValue, maxValue) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetOutputPropertyLimitMinMaxProc)((actorInfo)->mThing, (propIndex1), (minValue), (maxValue))

typedef IzzyError			(*GetOutputPropertyLimitMinMaxProc)(				// returns an non-zero error code if there is an error
								ThingPtr				inThing,				// thing that owns the property
								SInt32					inPropertyIndex1,		// index of the input property
								ValuePtr				outPropertyMinimum,		// the new mimumum value
								ValuePtr				outPropertyMaximum);	// the new maximum value

// -----------------------------------------------------------------------------
// SetOutputPropertyLimitMinMax_
// -----------------------------------------------------------------------------
// Sets the limit minimum and limit maximum for the specified output
// property. Note that if the current output value of this property is not
// in range when this function is called, it will be forced into range
/// 
// INTERRUPT SAFE: This function can be called at interrupt time.

#define	SetOutputPropertyLimitMinMax_(izzyParamsPtr, actorInfo, propIndex1, minValue, maxValue) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetOutputPropertyLimitMinMaxProc)((actorInfo)->mThing, (propIndex1), (minValue), (maxValue))

typedef IzzyError			(*SetOutputPropertyLimitMinMaxProc)(				// returns an non-zero error code if there is an error
								ThingPtr				inThing,				// thing that owns the property
								SInt32					inPropertyIndex1,		// index of the input property
								const ValuePtr			inPropertyMinimum,		// the new mimumum value
								const ValuePtr			inPropertyMaximum);		// the new maximum value

// -----------------------------------------------------------------------------
// GetOutputPropertyValue_
// -----------------------------------------------------------------------------
// Retrieves the current value of one of your actor's input properties. The
// result is returned as a ValuePtr struct, which gives the value's type and
// actual value.
// 
// INTERRUPT SAFE: This function can be called at interrupt time.

#define	GetOutputPropertyValue_(izzyParamsPtr, actorInfo, propIndex1) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetOutputPropertyValueProc)((actorInfo)->mThing, (propIndex1))

typedef const ValuePtr		(*GetOutputPropertyValueProc)(						// returns the value of the output property
								ThingPtr				inThing,				// thing that owns the property
								SInt32					inPropertyIndex1);		// index of the input property


// -----------------------------------------------------------------------------
// GetOutputPropertyMinimumValue_
// -----------------------------------------------------------------------------
// Returns the minimum value for an output property as a ValuePtr struct.
// 
// INTERRUPT SAFE: This function can be called at interrupt time.

#define	GetOutputPropertyMinimumValue_(izzyParamsPtr, actorInfo, propIndex1) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetOutputPropertyMinimumValueProc)((actorInfo)->mThing, (propIndex1))

typedef const ValuePtr		(*GetOutputPropertyMinimumValueProc)(				// returns the minimum value of the output property
								ThingPtr				inThing,				// thing that owns the property
								SInt32					inPropertyIndex1);		// index of the input property

// -----------------------------------------------------------------------------
// GetOutputPropertyMaximumValue_
// -----------------------------------------------------------------------------
// Returns the maximum value for an output property as a ValuePtr struct.
// 
// INTERRUPT SAFE: This function can be called at interrupt time.

#define	GetOutputPropertyMaximumValue_(izzyParamsPtr, actorInfo, propIndex1) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetOutputPropertyMaximumValueProc)((actorInfo)->mThing, (propIndex1))

typedef const ValuePtr		(*GetOutputPropertyMaximumValueProc)(				// returns the minimum value of the output property
								ThingPtr				inThing,				// thing that owns the property
								SInt32					inPropertyIndex1);		// index of the input property

// -----------------------------------------------------------------------------
// QueueMediaCommand_
// -----------------------------------------------------------------------------
// Adds a command to the Media Command Queue. At the next system task time (i.e.,
// when Isadora is running not at interrupt time) all pending commands in the
// queue will be processed. This allows you to queue non-interrupt time friendly
// commands for later processing. The result is an IzMCQID integer that uniquely
// identifies the command in the queue.
// 
// Parameters to the marco and their meanings are as follows.
// 
// izzyParamsPtr: pointer to the IsadoraParameter struct passed to your actor
// 
// actorInfo: a pointer to your actors ActorInfo struct.
// 
// completionProc: a procedure that will be called upon the completion of your
// request.
// 
// id: a unique ID that will identify the caller who made the request. If you
// wish, you may use the mThing member of the ActorInfo struct - this is
// guaranteed to be unique.
// 
// cmd: the media command. See IsadoraTypes.h for a list of all of the possible
// media commands.
// 
// type: the type of media that will be affected by this command. This should be
// one of kMediaTypeMovie, kMediaTypeAudio, or kMediaTypeMidi file. These three
// constants refer to the three sections of the Media Window into which the user
// imports these media.
// 
// num: the one-based index of the media that will be affected by this command,
// again, refering to the index listed in the Media Window for the specified
// media type.
// 
// p1 & p2: these are parameters that are specific to the media command being
// issued. The parameters and their meanings are documented in the list of Media
// Commands given in IsadoraTypes.h
// 
// INTERRUPT SAFE: This function can be called at interrupt time.

#if IZ_64BIT
typedef	UInt64	IzMCQID;
#elif IZ_32BIT
typedef	UInt64	IzMCQID;
#endif

#define	QueueMediaCommand_(izzyParamsPtr, actorInfo, completionProc, id, cmd, type, num, p1, p2) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mQueueMediaCommandProc)((actorInfo)->mThing, (completionProc), false, (id), (cmd), (type), (num), (p1), (p2) )

typedef	IzMCQID				(*QueueMediaCommandProc)(
								ThingPtr					inThing,				// opaque thing reference
								MediaCommandCompleteProc	inActionCompleteProc,	// completion proc called upon completion of the command
								Boolean						inReplace,				// if true, replace existing commands in queue with matching command & identifier
								IzMCQID						inIdentifier,			// unique identifier of caller who issued the command
								MediaCommand				inMediaCommand,			// media command to execute
								MediaCode					inMediaType,			// media type
								UInt32						inMediaObjectNumber,	// media object number
								IzRefCon					inParam1,				// first parameter
								IzRefCon					inParam2);				// second parameter

// -----------------------------------------------------------------------------
// QueueMediaCommandReplace_
// -----------------------------------------------------------------------------
// Same as DoQueueMediaCommand except that if there is already a Media Command in
// the queue with the same command and Identifier as the one passed to this
// macro, then the old media command and its parameters are replaced by the new
// command.
// 
// This function is useful for plugins that may receive data at interrupt time
// but queue media commands that must wait for the next system task time to
// execute. In this situations, there will be instances where you will want to
// replace the older, existing command with the newer command.
// 
// For example, at interrupt time, you queue a command to start playing Movie 2.
// A subsequent event causes the plugin to play Movie 3 instead. By calling
// QueueMediaCommandReplace_, if the movie 2 had already started playing, then
// the command would already be removed from the queue and movie 3 would start
// playing as expected. If movie 2 had not yet started playing, then the command
// would be replaced and movie 3 would start playing instead of movie 2. If you
// didn't do this, Movie 2 would play for an instant and then Movie 3 would start
// playing right afterward. This would cause a visible glitch.
// 
// The other place that this is useful is to prevent the overflow of System Task
// time commands in the queue. For instance, a MIDI continuous controller is
// connected to the Play Start parameter of Movie. When the controller is moved,
// a dense stream of MIDI messages comes in. If you didn't use
// QueueMediaCommandReplace_, your plugin would queue a Media Command for each
// control change message received, possibly overflowing the queue.
// 
// INTERRUPT SAFE: This function can be called at interrupt time.

#define	QueueMediaCommandReplace_(izzyParamsPtr, actorInfo, completionProc, id, cmd, type, num, p1, p2) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mQueueMediaCommandProc)((actorInfo)->mThing, (completionProc), true, (id), (cmd), (type), (num), (p1), (p2) )

// -----------------------------------------------------------------------------
// ExecuteMediaCommand_
// -----------------------------------------------------------------------------
// Immediately executes the specified media command if possible. If Isadora is at
// System Task time, then the command will be executed immediately. If Isadora is
// at Interrupt Time, then the command will be executed if the specified media
// command can legally be processed at interrupt time. In both of these cases,
// the result of the function call will be non-zero to indicate success. If,
// however, you attempt to exceute a non-interrupt-friendly command at interrupt
// time, the result will be zero and the command will not be executed.
// 
// INTERRUPTS: see description immediately above.
// 
// See DoQueueMediaCommand above for a list of the parameters and their meanings.

#define	ExecuteMediaCommand_(izzyParamsPtr, actorInfo, id, cmd, type, num, p1, p2) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mExecuteMediaCommandProc)((actorInfo)->mThing, (id), (cmd), (type), (num), (p1), (p2) )

typedef	Boolean				(*ExecuteMediaCommandProc)(
								ThingPtr					inThing,				// opaque thing reference
								IzMCQID						inIdentifier,			// unique identifier of caller who issued the command
								MediaCommand				inMediaCommand,			// media command to execute
								MediaCode					inMediaType,			// media type
								UInt32						inMediaObjectNumber,	// media object number
								IzRefCon					inParam1,				// first parameter
								IzRefCon					inParam2);				// second parameter

// -----------------------------------------------------------------------------
// SendMidiMessage_
// -----------------------------------------------------------------------------
// Sends a midi message of arbitrary length to the specified MIDI output port
// 
// INTERRUPTS: This callback can be executed at interrupt time.

#define	SendMidiMessage_(izzyParamsPtr, inPortIndex1, len, data) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSendMidiMessageProc)((inPortIndex1), (len), (data))

typedef void				(*SendMidiMessageProc)(
								PortIndex					inPortIndex1,			// one-based port index - as shown in the MIDI setup window
								UInt8						inLen,					// length of the message
								const void*					inData);				// pointer to midi data to send

#define	SendLongMidiMessage_(izzyParamsPtr, inPortIndex1, len, data) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSendLongMidiMessageProc)((inPortIndex1), (len), (data))

typedef void				(*SendLongMidiMessageProc)(
								PortIndex					inPortIndex1,			// one-based port index - as shown in the MIDI setup window
								UInt32						inLen,					// length of the message
								const void*					inData);				// pointer to midi data to send

// -----------------------------------------------------------------------------
// GetCurrentMTC
// -----------------------------------------------------------------------------
// Returns the current MIDI timecode from the specified port

#define	GetCurrentMTC_(izzyParamsPtr, inPortIndex1, outFormat, outHMSF, outLocked) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetCurrentMTCProc)((inPortIndex1), (outFormat), (outHMSF), (outLocked))

typedef Boolean				(*GetCurrentMTCProc)(
								PortIndex					inPortIndex1,			// one-based port index - as shown in the MIDI setup window
								MTCType*					outFormat,				// format of the timecode: 0 = 24, 1 = 25, 2 = 30, 3 = 30 drop frame
								TimecodeHMSF*				outHMSF,
								Boolean*					outLocked);

// -----------------------------------------------------------------------------
// TestMTCTimecodeTime
// -----------------------------------------------------------------------------
// Compares the specified timecode time (in HHMMSSFF format) to the current
// MTC time received by Isadora. Result is 0 if the two times are equal,
// < 0 if the specified time is less than the current time, or > 0 if the
// specified time is greater than the current MTC time.

#define	TestMTCTimecodeTime_(izzyParamsPtr, inPortIndex1, inHMSF) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mTestMTCTimecodeTimeProc)((inPortIndex1), (inHMSF))

typedef SInt32				(*TestMTCTimecodeTimeProc)(
								PortIndex					inPortIndex1,			// one-based port index - as shown in the MIDI setup window
								TimecodeHMSF				inTestTimeHMSF);

// -----------------------------------------------------------------------------
// AllNotesOff_
// -----------------------------------------------------------------------------
// Turns off all notes that are currently sounding. All MIDI messages sent
// using the SendMidiMessage_ function are logged. Note ons and note offs are
// kept track of, so that Isadora knows how many sound notes there are at any
// given time. When this command is executed, all of the sounding notes are
// sent a corresponding note off message.
// 
// INTERRUPTS: This callback can be executed at interrupt time.

#define	AllNotesOff_(izzyParamsPtr) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mAllNotesOffProc)()

typedef void				(*AllNotesOffProc)();				// pointer to midi data to send

// -----------------------------------------------------------------------------
// GetAudioDeviceList_
// -----------------------------------------------------------------------------
// Returns a list of all Audio Devices currently recognized by the system.
// The format of the list is a series of names separated by a carriage return,
// which would be appropriate to pass to the PropertyGetPopupMenuInfo calback.
//
// Currently, under windows or on systems that do not have Core Audio, the result
// will always be a single item: 'default'. On other systems, the result will be
// the list of available devices.
//
// outBuffer should point to a chracter buffer large enough to accomodate all
// of the devices in the list.
//
// outBufferMaxSize indicates the size of the buffer. if the list is larger
// than this byte count, the resulting list will be incomplete.
//
// The return value indicates the number of bytes used, not including the
// terminating null (0) character.
//
// Make this call with outBuffer set to NULL and outBufferMaxSize set to 0 to
// determine the size of the buffer required to hold the list

#define	GetAudioDeviceList_(izzyParamsPtr, inOrOut, buffer, bufferSize) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetAudioDeviceListProc)((inOrOut), (buffer), (bufferSize))

enum AudioDeviceType {
	kGetAudioDeviceList_Input,
	kGetAudioDeviceList_Output
};

typedef UInt32				(*GetAudioDeviceListProc)(
								AudioDeviceType	inType,
								char*			outBuffer,
								UInt32			outBufferMaxSize);

// -----------------------------------------------------------------------------
// AudioDeviceNameToAudioDeviceID_
// -----------------------------------------------------------------------------
// Given the name of an audio device insalled on the system, returns the
// native audio device ID for that device.

#if TARGET_OS_MAC
	typedef	UInt32			NativeAudioDeviceID;
#elif TARGET_OS_WIN
	typedef	DWORD			NativeAudioDeviceID;
#endif

#define	AudioDeviceNameToAudioDeviceID_(izzyParamsPtr, inOrOut, buffer) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mAudioDeviceNameToAudioDeviceIDProc)((inOrOut), (buffer))

typedef NativeAudioDeviceID	(*AudioDeviceNameToAudioDeviceIDProc)(
								AudioDeviceType	inType,
								const char*		inName);

// -----------------------------------------------------------------------------
// CopyImageBufferToStage_ (DEPRECATED for v3.0)
// -----------------------------------------------------------------------------
// Copies an Image Buffer to the stage. The destination rect specifies the
// location and dimensions of where the to which the contents of the image buffer
// will be copied. The dimensions of this rectangle can be different than the
// dimensions of the image buffer source image. Optionally, you may also specify
// a region that will be erased to the background color of the stage after the
// copy is complete.
// 
// ** NOT ** INTERRUPT SAFE: This function must not be called at interrupt time.

/*
enum {
	kCopyToStageCopyBits,					// use CopyBits to copy to stage
	kCopyToStageDirectFullScreenOnly		// use direct blitting if full stage
};
typedef SInt16	CopyToStageMode;

typedef struct {
	IzRect		mDstRect;		// destination rect before zoom
	float		mZoom;			// zoom factor -- from 0 to 100%
	NativeRgn	mUpdateRgn;		// if not nil, this should be a valid NativeRgn.
								// In the latter case, erase the specified area
								// to black after copying the image buf to the
								// stage
} CopyToStageInfo;

#define	CopyImageBufferToStage_(izzyParamsPtr, actorInfo, inBlitMode, destStageIndex, imgBuffer, inCopyToStageInfo) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mCopyImageBufferToStageProc)((actorInfo)->mThing, (inBlitMode), (destStageIndex), (imgBuffer), (inCopyToStageInfo))

typedef	void				(*CopyImageBufferToStageProc)(
								ThingPtr					inThing,				// opaque thing reference
								CopyToStageMode				inBlitMode,				// if true, blit directly to stage
								StageIndexT					inDestStageIndex0,		// destination stage index
								ImageBufferPtr				inImageBufferPtr,		// pointer to an image buffer struct
								CopyToStageInfo*			inCopyToStageInfo);		// information used to copy to the stage
*/

// -----------------------------------------------------------------------------
// CopyImageBufferToRecordedOutput_ (DEPRECATED for v3.0)
// -----------------------------------------------------------------------------
// 
// ** NOT ** INTERRUPT SAFE: This function must not be called at interrupt time.

/*

#define	CopyImageBufferToRecordedOutput_(izzyParamsPtr, actorInfo, destStageIndex, imgBuffer, inTop, inLeft, inWidth, inHeight, inZoom) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mCopyImageBufferToRecordedOutputProc)((actorInfo)->mThing, (destStageIndex), (imgBuffer), (inTop), (inLeft), (inWidth), (inHeight), (inZoom))

typedef	void				(*CopyImageBufferToRecordedOutputProc)(
								ThingPtr					inThing,				// opaque thing reference
								StageIndexT					inDestStageIndex0,		// destination stage index
								ImageBufferPtr				inImageBufferPtr,		// pointer to an image buffer struct
								IzReal						inTop,
								IzReal						inLeft,
								IzReal						inWidth,
								IzReal						inHeight,
								IzReal						inZoom);
*/

// -----------------------------------------------------------------------------
// GetStageBounds_
// -----------------------------------------------------------------------------
// Returns a rectangle that gives the bounds of the stage.
// 
// ** NOT ** INTERRUPT SAFE: This function must not be called at interrupt time.

#define	GetStageBounds_(izzyParamsPtr, actorInfo, destStageIndex, inDrawableAreaOnly, outRect) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetStageBoundsProc)((actorInfo)->mThing, (destStageIndex), (inDrawableAreaOnly), (outRect))

typedef	Boolean	(*GetStageBoundsProc)(
					ThingPtr					inThing,
					StageIndexT					inDestStageIndex0,
					Boolean						inDrawableAreaOnly,
					IzRect*						inInvalRect);

// -----------------------------------------------------------------------------
// CanRenderToStage_ (REMOVED for v3.0)
// -----------------------------------------------------------------------------
// Returns true if a plugin can render to a specified stage
// 
// ** NOT ** INTERRUPT SAFE: This function must not be called at interrupt time.

/*

#define	CanRenderToStage_(izzyParamsPtr, actorInfo, destStageIndex) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mCanRenderToStageProc)((actorInfo)->mThing, (destStageIndex))

typedef	Boolean	(*CanRenderToStageProc)(
					ThingPtr					inThing,
					StageIndexT					inDestStageIndex0);
*/

// -----------------------------------------------------------------------------
// GetStageNameList
// -----------------------------------------------------------------------------
// Returns a list of all stage names separated by return characters

#define	GetStageNameList_(izzyParamsPtr, actorInfo) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetStageNameListProc)((actorInfo)->mThing, 0)

#define	GetStageOrRendererNameList_(izzyParamsPtr, actorInfo, rendererInputPort) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetStageNameListProc)((actorInfo)->mThing, (rendererInputPort))

typedef	const char*	(*GetStageNameListProc)(
						ThingPtr					inThing,
						PortIndex					inRendererInputPortIndex);

// -----------------------------------------------------------------------------
// GetIndexedStageName
// -----------------------------------------------------------------------------
// Returns the name of a specific stage based on the provided one-based stage
// index. If the stage doesn't exist, the result is an empty string

#define	GetIndexedStageName_(izzyParamsPtr, actorInfo, stageIndex1, includeIndexInName) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetIndexedStageNameProc)((actorInfo)->mThing, (stageIndex1), (includeIndexInName))

typedef	const char*	(*GetIndexedStageNameProc)(
						ThingPtr					inThing,
						StageIndexT					inStageIndex1,
						Boolean						inIncludeIndexInName);

// -----------------------------------------------------------------------------
// HandleStageValueToString_
// -----------------------------------------------------------------------------
// Given a stage proprery and, optionally, a render destination property, convert
// the current value to a stage name

#define	HandleStageValueToString_(izzyParamsPtr, actorInfo, propType, stageValuePtr, renderPropIndex, outString, outStringMaxLen) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mHandleStageValueToStringProc)((actorInfo)->mThing, (propType), (stageValuePtr), (renderPropIndex), (outString), (outStringMaxLen))

typedef	Boolean		(*HandleStageValueToStringProc)(
						ThingPtr					inThing,
						PropertyType				inPropertyType,
						const ValuePtr				inStageIndex,
						PropertyIndex				inRenderDestPropertyIndex,		// 0 if not applicable
						char*						outString,
						size_t						outStringMaxLen);

// -----------------------------------------------------------------------------
// HandleStageValueToString_
// -----------------------------------------------------------------------------
// Given a  proprery and, optionally, a render destination property, convert
// the current value to a stage name

#define	HandleStageStringToValue_(izzyParamsPtr, actorInfo, propType, stagePropIndex, renderPropIndex, inString, outStageIndex) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mHandleStageStringToValueProc)((actorInfo)->mThing, (propType), (stagePropIndex), (renderPropIndex), (inString), (outStageIndex))

typedef	Boolean		(*HandleStageStringToValueProc)(
						ThingPtr					inThing,
						PropertyType				inPropertyType,
						PropertyIndex				inStagePropertyIndex,
						PropertyIndex				inRenderDestPropertyIndex,		// 0 if not applicable
						const char*					inString,
						ValuePtr					outValue);

// -----------------------------------------------------------------------------
// InvalAreaOnStage_ (DEPRECATED for v3.0)
// -----------------------------------------------------------------------------
// Invalidates the specified area on the stage so that it will be erased to the
// background color the next time an update event can be processed.
// 
// ** NOT ** INTERRUPT SAFE: This function must not be called at interrupt time.

/*

#define	InvalAreaOnStage_(izzyParamsPtr, actorInfo, destStageIndex, inInvalRect, inInvalRgn) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mInvalAreaOnStageProc)((actorInfo)->mThing, (destStageIndex), (inInvalRect), (inInvalRgn))

typedef	void	(*InvalAreaOnStageProc)(
					ThingPtr					inThing,
					StageIndexT					inDestStageIndex0, 
					const IzRect*				inInvalRect,
					NativeRgn					inInvalRgn);

*/

// -----------------------------------------------------------------------------
// ValidAreaOnStage_ (DEPRECATED for v3.0)
// -----------------------------------------------------------------------------
// Validates the specified area on the stage so that it will be erased to the
// background color the next time an update event can be processed.
// 
// ** NOT ** INTERRUPT SAFE: This function must not be called at interrupt time.

/*

#define	ValidAreaOnStage_(izzyParamsPtr, actorInfo, destStageIndex, inValidRect, inValidRgn) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mValidAreaOnStageProc)((actorInfo)->mThing, (destStageIndex), (inValidRect), (inValidRgn))

typedef	void	(*ValidAreaOnStageProc)(
					ThingPtr					inThing,
					StageIndexT					inDestStageIndex0, 
					const IzRect*				inValidRect,
					NativeRgn					inValidRgn);

*/

// -----------------------------------------------------------------------------
// SetForceCursorVisibleOnStage
// -----------------------------------------------------------------------------
// Allows the caller to set or clear the flag that forces the cursor to be
// visible when it is over a full screen stage. Note that every call to
// SetForceCursorVisibleOnStage() where inForceCursorVisible is TRUE must
// be balanced by a call where inForceCursorVisible is FALSE. Otherwise, the
// cursor will remain visible even after the actor is disabled, etc. 
// 
// ** NOT ** INTERRUPT SAFE: This function must not be called at interrupt time.

#define	SetForceCursorVisibleOnStage_(izzyParamsPtr, actorInfo, destStageIndex, forceCursorVisible) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetForceCursorVisibleOnStageProc)((actorInfo)->mThing, (destStageIndex), (forceCursorVisible))

typedef	void	(*SetForceCursorVisibleOnStageProc)(
					ThingPtr					inThing,
					StageIndexT					inDestStageIndex0,
					Boolean						inForceCursorVisible);

// -----------------------------------------------------------------------------
// SetForceCursorVisibleOnStage
// -----------------------------------------------------------------------------
// Allows the caller to set or clear the flag that forces the cursor to be
// visible when it is over a full screen stage. Note that every call to
// SetForceCursorVisibleOnStage() where inForceCursorVisible is TRUE must
// be balanced by a call where inForceCursorVisible is FALSE. Otherwise, the
// cursor will remain visible even after the actor is disabled, etc. 
// 
// ** NOT ** INTERRUPT SAFE: This function must not be called at interrupt time.

#define	SetForceCursorVisibleOnStage_(izzyParamsPtr, actorInfo, destStageIndex, forceCursorVisible) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetForceCursorVisibleOnStageProc)((actorInfo)->mThing, (destStageIndex), (forceCursorVisible))

typedef	void	(*SetForceCursorVisibleOnStageProc)(
					ThingPtr					inThing,
					StageIndexT					inDestStageIndex0,
					Boolean						inForceCursorVisible);

// -----------------------------------------------------------------------------
// SetCursor_
// -----------------------------------------------------------------------------
// Sets the cursor to the using the specified cursor code
// 
// ** NOT ** INTERRUPT SAFE: This function must not be called at interrupt time.

#define	SetCursor_(izzyParamsPtr, inCursorID) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetCursorProc)((inCursorID))

typedef	void			(*SetCursorProc)(
							SInt16			inCursorID);

// -----------------------------------------------------------------------------
// EnableStagePreviewFromControlProc_
// -----------------------------------------------------------------------------
// Enables or disables the stage preview for the specified stage
// 
// ** NOT ** INTERRUPT SAFE: This function must not be called at interrupt time.

#define	ActorEnableStagePreviewProc_(izzyParamsPtr, actorInfo, inEnable, stageIndex0, keepAspect, borderSize, maxFPS) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mActorControlEnableStagePreview)((actorInfo), NULL, (inEnable), (stageIndex0), (keepAspect), (borderSize), (maxFPS))

#define	ControlEnableStagePreviewProc_(izzyParamsPtr, controlInfo, inEnable, stageIndex0, keepAspect, borderSize, maxFPS) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mActorControlEnableStagePreview)(NULL, (controlInfo), (inEnable), (stageIndex0), (keepAspect), (borderSize), (maxFPS))

typedef	void		(*ActorControlEnableStagePreview)(
						ActorInfo*			inActorInfo,
						ControlInfo*		inControlInfo,
						Boolean				inEnable,
						StageIndexT			inStageIndex0,
						Boolean				inKeepAspect,
						SInt32				inBorderSize,
						float				inMaxFPS);

// -----------------------------------------------------------------------------
// ActorCreatePreviewProc_ & ControlCreatePreviewProc_
// -----------------------------------------------------------------------------
// Creates or disables a preview for the specified actor or control. Returns
// a VideoPreviewRef which can be used with the DrawVideoIntoPreview_() call
// defined below

typedef struct OpaqueVideoPreview*	VideoPreviewRef;

typedef struct
{
	Boolean		mKeepAspect;
	SInt32		mBorderSize;
	
} VideoPreviewInfo;

#define	ActorCreateVideoPreviewProc_(izzyParamsPtr, actorInfo, videoPreviewInfo, outPreviewRef, inVideoValue) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mActorControlCreateVideoPreviewProc)((actorInfo), NULL, (videoPreviewInfo), (outPreviewRef), (inVideoValue))

#define	ControlCreateVideoPreviewProc_(izzyParamsPtr, controlInfo, videoPreviewInfo, outPreviewRef, inVideoValue) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mActorControlCreateVideoPreviewProc)(NULL, (controlInfo), (videoPreviewInfo), (outPreviewRef), (inVideoValue))

typedef	void		(*ActorControlCreateVideoPreviewProc)(
						ActorInfo*				inActorInfo,
						ControlInfo*			inControlInfo,
						const VideoPreviewInfo*	inVideoPreviewInfo,
						VideoPreviewRef*		outPreviewRef,
						const Value*			inVideoValue);			// CAN BE NULL

// -----------------------------------------------------------------------------
// DisposeVideoPreviewProc_
// -----------------------------------------------------------------------------
// Disposes a video preview created byActorEnablePreviewProc_ or
// ControlEnablePreviewProc_

#define	DisposeVideoPreviewProc_(izzyParamsPtr, inPreviewRef) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mDisposeVideoPreviewProc)((inPreviewRef))

typedef	void		(*DisposeVideoPreviewProc)(
						VideoPreviewRef			inPreviewRef);

// -----------------------------------------------------------------------------
// UpdateVideoPreviewProc_
// -----------------------------------------------------------------------------
// A video preview client should call this function if it changes the geometry
// or other characteristics of its video preview. This ensures the video preview
// stays in sync with the client's settings.

#define	UpdateVideoPreviewProc_(izzyParamsPtr, previewRef, videoPreviewInfo) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mUpdateVideoPreviewProc)((previewRef), (videoPreviewInfo))

typedef	void		(*UpdateVideoPreviewProc)(
						VideoPreviewRef			inPreviewRef,
						const VideoPreviewInfo*	inVideoPreviewInfo);

// -----------------------------------------------------------------------------
// DrawVideoIntoPreview_
// -----------------------------------------------------------------------------
// Draws the video specified by inValue into the valid video preview reference
// specified by inPreviewRef. The ActorInfo or ControlInfo parameter provided
// should match that provided when the reference was created.

#define	DrawVideoIntoPreview_(izzyParamsPtr, previewRef, videoValue) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mDrawVideoIntoPreviewProc)((previewRef), (videoValue))

typedef	void		(*DrawVideoIntoPreviewProc)(
						VideoPreviewRef			inPreviewRef,
						const Value*			inVideoValue);

// -----------------------------------------------------------------------------
// GetAggregateBufferFormat_
// -----------------------------------------------------------------------------
// Given an array of image buffer ptrs, uses the user preferences to determine
// the common  size and format at which those images should be processed.
//
// INTERRUPTS: This function may be called at interrupt time.

#define	GetAggregateBufferFormat_(izzyParamsPtr, inProcessingModeOverride, inDefaultPixelFormat, inImageBufferCount, inImageBufferPtrs, outIBFormatInfo) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetAggregateBufferFormatProc)((inProcessingModeOverride), (inDefaultPixelFormat), (inImageBufferCount), (inImageBufferPtrs), (outIBFormatInfo))

typedef	Boolean	(*GetAggregateBufferFormatProc)(
					ImageProcessingMode			inProcessingModeOverride,		// see notes above regarding this parameter
					OSType						inDefaultPixelFormat,			// default pixel format
					SInt32						inImageBufferCount,				// number of image buffer ptrs in the array inImageBufferPtrs[]
					ImageBufferPtr*				inImageBufferPtrs,				// the array of image buffer ptrs to test
					IBFormatInfoPtr				outIBFormatInfo);				// the resulting common format and size

// -----------------------------------------------------------------------------
// IsIntermediaryImageBufferRequired_
// -----------------------------------------------------------------------------
// Given a list of image buffer ptrs, uses the user preferences to determine the
// size at which those images should be processed. if the image buffer indicated
// by inImageBufferToTestIndex does not match the size or bit depth of calculated
// based on the user prefs and image buffers to be combined, then the result will
// be true.
// 
// Normally you should always pass kImageProcessUseDefault as the image processing
// mode to respect the settings made by the User in the preferences. But, since
// it may be useful to force a particular setting in certain situations, you can
// optionally set this parameter to one of kImageProcessLargerSlower,
// kImageProcessSmallerFaster or kImageProcessFixedSize
// 
// INTERRUPTS: This function may be called at interrupt time.

#define	IsIntermediaryImageBufferRequired_(izzyParamsPtr, inProcessingModeOverride, inDefaultPixelFormat, inImageBufferCount, inImageBufferPtrs, inImageBufferToTestIndex) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIsIntermediaryImageBufferRequiredProc)((inProcessingModeOverride), (inDefaultPixelFormat),(inImageBufferCount), (inImageBufferPtrs), (inImageBufferToTestIndex))

typedef	Boolean	(*IsIntermediaryImageBufferRequired)(
					ImageProcessingMode			inProcessingModeOverride,		// see notes above regarding this parameter
					OSType						inDefaultPixelFormat,			// default pixel format
					SInt32						inImageBufferCount,				// number of image buffer ptrs pointed to by inImageBuffers
					ImageBufferPtr*				inImageBufferPtrs,				// an array of image buffer ptrs
					SInt32						inImageBufferToTestIndex);		// the image buffer to test, to see if an intermediary
																				// buffer is required

// -----------------------------------------------------------------------------
// CreateComplimentaryImageBuffer_
// -----------------------------------------------------------------------------
// Creates a new image buffer that matches the size and depth of the specified
// image buffers. It is up to the caller to dispose the resulting image buffer
// with the DisposeImageBuffer_ function/macro.
// 
// Note that this function respects the mImgProcPrefs preference settings given
// in the IsadoraParams struct. Thus the resulting ImageBuffer may not be of the
// same dimensions as the source. Your function should take this possibility into
// account. You can test to see whether or not you need to make an intediary
// buffer with the DoIntermediaryImageBufferRequired macro.
// 
// Normally you should always pass kImageProcessUseDefault as the image processing
// mode to respect the settings made by the User in the preferences. But, since
// it may be useful to force a particular setting in certain situations, you can
// optionally set this parameter to one of kImageProcessLargerSlower,
// kImageProcessSmallerFaster or kImageProcessFixedSize, or kImageProcessMatchFormatOnly
//
// If there is not enough memory available to create the image buffer, the
// resulting value is nil.
// 
// ** NOT ** INTERRUPT SAFE: This function must not be called at interrupt time.

#define	CreateComplimentaryImageBuffer_(izzyParamsPtr, inProcessingModeOverride, inDefaultPixelFormat, inImageBufferCount, inImageBufferPtrs) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mCreateComplimentaryImageBufferProc)((inProcessingModeOverride), (inDefaultPixelFormat), (inImageBufferCount), (inImageBufferPtrs))

typedef	ImageBufferPtr		(*CreateComplimentaryImageBufferProc)(
								ImageProcessingMode			inProcessingModeOverride,	// see notes above regarding this parameter
								OSType						inDefaultPixelFormat,		// default pixel format
								SInt32						inImageBufferCount,			// number of image buffer ptrs pointed to by inImageBuffers
								ImageBufferPtr*				inImageBufferPtrs);			// an array of image buffer ptrs

// -----------------------------------------------------------------------------
// DisposeImageBuffer_
// -----------------------------------------------------------------------------
// Dispose an ImageBuffer created with the CreateComplimentaryImageBufferProc
// function or DoCreateComplimentaryImageBuffer macro.
// 
// ** NOT ** INTERRUPT SAFE: This function must not be called at interrupt time.

#define	DisposeImageBuffer_(izzyParamsPtr, inImageBuffer) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mDisposeImageBufferProc)((inImageBuffer))

typedef	void			(*DisposeImageBufferProc)(
								ImageBufferPtr	inImageBuffer);			// the image buffer to dispose

// -----------------------------------------------------------------------------
// CopyImageBuffer_
// -----------------------------------------------------------------------------
// Copy one all or part of one image buffer to another. You may optionally
// provide a pointer to source and/or destination rectangle. If either of these
// parameters are set to nil, then the respective bounds rectangle for the source
// and/or destination buffer is used. If the source and destination rectangles
// differ in size, then the source image will be scaled to fit the destination
// bounds.
// ** NOT ** INTERRUPT SAFE: This function must not be called at interrupt time.

#define	CopyImageBuffer_(izzyParamsPtr, inImageBufferSrc, inImageBufferDst, inSrcRect, inDstRect) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mCopyImageBufferProc)((inImageBufferSrc), (inImageBufferDst), (inSrcRect), (inDstRect), kIzGraphicsMode_SrcCopy, nil)

#define	CopyImageBufferMode_(izzyParamsPtr, inImageBufferSrc, inImageBufferDst, inSrcRect, inDstRect, inMode, inRGBBlendColor) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mCopyImageBufferProc)((inImageBufferSrc), (inImageBufferDst), (inSrcRect), (inDstRect), (inMode), (inRGBBlendColor))

#define	CopyImageBufferInvertColors_(izzyParamsPtr, inImageBufferSrc, inImageBufferDst, inSrcRect, inDstRect) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mCopyImageBufferInvertColorsProc)((inImageBufferSrc), (inImageBufferDst), (inSrcRect), (inDstRect))

typedef	void			(*CopyImageBufferProc)(
								ImageBufferPtr	inImageBufferSrc,
								ImageBufferPtr	inImageBufferDst,
								const IzRect*	inSrcRect,			// can be nil for entire buffer
								const IzRect*	inDstRect,			// can be nil for entire buffer
								SInt16			inMode,				// transfer mode (kIzGraphicsMode_SrcCopy, kIzGraphicsMode_SrcOr, etc.)
								RGBColor*		inBlendColor);		// blend color when transfer mode is blend

typedef	void			(*CopyImageBufferInvertColorsProc)(
								ImageBufferPtr	inImageBufferSrc,
								ImageBufferPtr	inImageBufferDst,
								const IzRect*	inSrcRect,			// can be nil for entire buffer
								const IzRect*	inDstRect);			// can be nil for entire buffer

// -----------------------------------------------------------------------------
// FillImageBuffer_
// -----------------------------------------------------------------------------
// Fills the image buffer with the specified color
// 
#define	FillImageBuffer_(izzyParamsPtr, inImageBufferDst, inColorRef) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mFillImageBufferProc)((inImageBufferDst), (inColorRef))

typedef	void			(*FillImageBufferProc)(
								ImageBufferPtr	inImageBufferDst,
								UInt32			inColorRef);
								
// -----------------------------------------------------------------------------
// CreateAndRetainColorImageBuffer_
// -----------------------------------------------------------------------------
// Creates a black image buffer with the specified format and size
// 
// ** NOT ** INTERRUPT SAFE: This function must not be called at interrupt time.

#define	CreateAndRetainColorImageBuffer_(izzyParamsPtr, inFormat, inWidth, inHeight, inColorRef) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mCreateAndRetainColorImageBufferProc)((inFormat), (inWidth), (inHeight), (inColorRef))

typedef	ImageBufferPtr		(*CreateAndRetainColorImageBufferProc)(OSType inFormat, SInt32 inWidth, SInt32 inHeight, UInt32 inColorRef);

// -----------------------------------------------------------------------------
// ReleaseColorImageBuffer__
// -----------------------------------------------------------------------------
// Releases a black image buffer created with CreateAndRetainColorImageBuffer_
// 
// ** NOT ** INTERRUPT SAFE: This function must not be called at interrupt time.

#define	ReleaseColorImageBuffer_(izzyParamsPtr, inImageBufferPtr) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mReleaseColorImageBufferProc)((inImageBufferPtr))

typedef	void		(*ReleaseColorImageBufferProc)(ImageBufferPtr inImageBufferPtr);

// -----------------------------------------------------------------------------
// GetSceneName_
// -----------------------------------------------------------------------------
// Returns the name of the specified scene
// 
// The scene to be whose name will be returend is determined by the jump mode
// and scene index parameters -- see the JumpToScene_ command for more info.
// 
// ** NOT ** INTERRUPT SAFE: This function must not be called at interrupt time.

#define	GetSceneName_(izzyParamsPtr, actorInfo, mode, sceneIndex, maxLen, sceneNamePtr) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetSceneNameProc)((actorInfo)->mThing, (mode), (sceneIndex), (maxLen), (sceneNamePtr))

typedef	IzzyError			(*GetSceneNameProc)(									// returns an non-zero error code if there is an error
								ThingPtr					inThing,
								JumpMode					inMode,
								SInt32						inSceneIndex,
								UInt32						inSceneNameMaxLen,
								char*						outSceneName);

// -----------------------------------------------------------------------------
// GetSceneInfo_
// -----------------------------------------------------------------------------
// Returns the name of the scene specified by a one-based index.
// 
// ** NOT ** INTERRUPT SAFE: This function must not be called at interrupt time.

// v2.6 ->

typedef UInt32 GetSceneInfoOptions;

enum {

	// GET CURRENT PRIMARY ACTIVE SCENE
	//
	// return the currently active primary scene index
	// note that if this option is used in the ActivateScene
	// or ActivateControl proc of an actor or control, the
	// current primary scene may not be accurate because
	// it has not yet been actiavted
	
	kGetSceneInfoOption_GetPrimarySceneIndex		= 0x0001,
	
	// GET RELATIVE SCENE INDEX
	//
	// for Actors, return the absolute scene index relative
	// to the scene owning the Actor that made the GetSceneInfo
	// call.
	//
	// for Controls, return the absolute scene index relative
	// to the current primary activec scene
	//
	// mInputGetSceneIndex should contain the desired the offset
	// from the absolute scene index given by the mInputGetSceneIndex
	// member of the mInputGetSceneIndex struct
	//
	// should not be used in combination with any other option
	// except kGetSceneInfoOption_GetFullList
	
	kGetSceneInfoOption_GetRelativeSceneIndex		= 0x0002,
	
	// GET SCENE NAME(S) / CUE NUMBER(S)
	
	// when using these options, set the mInputGetSceneIndex member
	// of the IzzyGetSceneInfo struct to the absolute index of the
	// desire scene
	
	kGetSceneInfoOption_SceneName					= 0x0004,		// get scene name
	kGetSceneInfoOption_CueNumber					= 0x0008,		// get cue number
	kGetSceneInfoOption_GetCueNumberAndSceneName	= kGetSceneInfoOption_SceneName | kGetSceneInfoOption_CueNumber,
	
	// GET FULL SCENE LIST
	//
	// when using this options, the kGetSceneInfoOption_SceneName
	// or kGetSceneInfoOption_CueNumber flags used only to determine
	// if the scene name, cue number, or both are returned in the list

	kGetSceneInfoOption_GetFullList					= 0x0010,		// when set, return list of cue/scene names separated by a CR ('\r')
																	// respects the above options regarding cue number and scene name
	// FIND SCENE BY NAME

	kGetSceneInfoOption_FindScene					= 0x0020,		// search for the scene name specified by mInputSearchTerm
	kGetSceneInfoOption_FindCue						= 0x0040,		// search for the cue number specified by mInputSearchTerm
	kGetSceneInfoOption_FindCue_PartialOK			= 0x0080		// or with kGetSceneInfoOption_FindCue to allow a partial match
};

typedef struct {

	UInt32				mSceneInfoSize;				// set to sizeof(IzzySceneInfo)
	
	
	// OPTIONS
	
	// options field indicating what information to get
	// or which action to take
	
	GetSceneInfoOptions	mOptions;							// options - see above
	
	// GET SCENE INDEX

	SceneIndexT			mOutCurrentPrimarySceneIndex1;		// output when options includes kGetSceneInfoOption_GetPrimarySceneIndex
	SceneIndexT			mOutRelativeSceneIndex1;			// output when options includes kGetSceneInfoOption_GetRelativeSceneIndex

	// GET RELATIVE SCENE NUMBER, SCENE NAME or CUE NUMBER
	
	// mInputGetSceneIndex member:
	//
	// mInputGetSceneIndex is used for getting a specific scene name or cue number
	// or for calculating a scene relative to the scene owning the specified actor
	// or control
	//
	// if the kGetSceneInfoOption_GetFullList option is enabled, this
	// input is ignored
	//
	// if the kGetSceneInfoOption_GetRelativeSceneIndex option is enabled, then
	// this input is the index of the scene from which to calculate the relative
	// scene number; kGetSceneInfoOption_SceneName or kGetSceneInfoOption_CueNumber
	// are ignored in this case
	//
	// otherwise, when kGetSceneInfoOption_SceneName or kGetSceneInfoOption_CueNumber
	// are specified, then it is an absolute scene index
	
	SInt32				mInputGetSceneIndex;
	
	UInt32				mOutputBufferSize;					// specify size of buffer pointed to by mOutputBuffer;
	char*				mOutputBuffer;						// the resulting data allocated with IzzyMalloc() - if
															// this member returns a non-null result, it must be
															// disposed by caller using IzzyFree_()
	// FIND
	// the following two fields are used if the mOptions
	// member is kGetSceneInfoOption_FindCue
	
	const char*			mInputSearchTerm;					// utf8 search string when using a find option; allocated by caller
	
	SceneIndexT			mOutFoundSceneIndex;				// set when options include is kGetSceneInfoOption_Find... and the
															// the search term is found
	
} IzzyGetSceneInfo, *IzzyGetSceneInfoPtr;

#define	GetSceneInfoForActor_(izzyParamsPtr, actorInfo, sceneInfoStruct) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetSceneInfoProc)(TRUE, (actorInfo)->mThing, (sceneInfoStruct))

#define	GetSceneInfoForControl_(izzyParamsPtr, controlInfo, sceneInfoStruct) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetSceneInfoProc)(FALSE, (controlInfo)->mControl, (sceneInfoStruct))

typedef	IzzyError			(*GetSceneInfoProc)(									// returns an non-zero error code if there is an error
								Boolean						inCalledByActor,		// true if called by an actor, false if by a control
								void*						inObjectPtr,			// ActorInfo* if Actor, ControlInfo* if Control
								IzzyGetSceneInfo*			ioSceneInfo);
// v2.6 <-

// -----------------------------------------------------------------------------
// JumpToScene_
// -----------------------------------------------------------------------------
// Jumps from the current scene to the scene specified by the combination of the
// JumpMode specified by inMode and the value of inJumpValue. The default behavior
// is to do an additive crossfade to the new scene (kTransition_Additive.)
// You can change this behavior by specifing kTransition_TransparentAbove, or
// kTransition_TransparentBelow to do a transparent fade above or below the
// current scene.
//
// If the JumpMode is kJumpRelative, then inJumpValue will will be added to the
// current Scene Index to determine the new scene to be activated. Thus a value
// of +1 would jump to the next scene, a value of -1 would jump to the previous
// scene.
// 
// If the JumpMode is kJumpAbsolute, then Isadora will attempt to set the Scene
// Index to the value given by inJumpValue. Note that the Scene Index is one-
// based.
// 
// In either case, if the calculated Scene Index is out of bounds (i.e., less
// than 1 or greater than the number of scenes in the current Isadora document)
// then Isadora does nothing.
//
// v2.6: added render order parameter
// 
// ** NOT ** INTERRUPT SAFE: This function must not be called at interrupt time.

// v2.6 ->

enum TransitionMode {
	kTransition_Invalid = 0,
	kTransition_Additive,
	kTransition_TransparentAbove,
	kTransition_TransparentBelow
};

#define	TransitionModeRendersAbove(xm)	((xm) == kTransition_Additive || (xm) == kTransition_TransparentAbove)
#define	TransitionModeRendersBelow(xm)	((xm) == kTransition_TransparentBelow)

#define	JumpToScene_(izzyParamsPtr, actorInfo, mode, sceneIndex) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mJumpToSceneProc)((actorInfo)->mThing, (mode), (sceneIndex), 0, 0, kTransition_Additive)

#define	JumpToSceneWithFade_(izzyParamsPtr, actorInfo, mode, sceneIndex, duration_mS) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mJumpToSceneProc)((actorInfo)->mThing, (mode), (sceneIndex), (duration_mS), (duration_mS), kTransition_Additive)

#define	JumpToSceneWithFadeInFadeOut_(izzyParamsPtr, actorInfo, mode, sceneIndex, fadeInDuration_mS, fadeOutDuration_mS) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mJumpToSceneProc)((actorInfo)->mThing, (mode), (sceneIndex), (fadeInDuration_mS), (fadeOutDuration_mS), kTransition_Additive)

#define	JumpToSceneWithFadeInFadeOutWithTransition_(izzyParamsPtr, actorInfo, mode, sceneIndex, fadeInDuration_mS, fadeOutDuration_mS, transitionMode) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mJumpToSceneProc)((actorInfo)->mThing, (mode), (sceneIndex), (fadeInDuration_mS), (fadeOutDuration_mS), (transitionMode))
	
typedef	IzzyError			(*JumpToSceneProc)(										// returns an non-zero error code if there is an error
								ThingPtr					inThing,
								JumpMode					inMode,
								SInt32						inSceneIndex,
								UInt32						inFadeOutDuration_mS,
								UInt32						inFadeInDuration_mS,
								TransitionMode				inTransitionMode);


#define	JumpToSceneWithBuiltInFadeForActor_(izzyParamsPtr, actorInfo, sceneIndex1, inRespectNextSceneFadeIn) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mJumpToSceneWithBuiltInFadeProc)(TRUE, (actorInfo)->mThing, (inRespectNextSceneFadeIn), FALSE)
	
#define	JumpToSceneWithBuiltInFadeForControl_(izzyParamsPtr, controlInfo, sceneIndex1, inRespectNextSceneFadeIn) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mJumpToSceneWithBuiltInFadeProc)(FALSE, (controlInfo)->mControl, (sceneIndex1), (inRespectNextSceneFadeIn), FALSE)
	
#define	JumpRelativeWithBuiltInFadeForActor_(izzyParamsPtr, actorInfo, sceneIndex1, inRespectNextSceneFadeIn) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mJumpToSceneWithBuiltInFadeProc)(TRUE, (actorInfo)->mThing, (inRespectNextSceneFadeIn), TRUE)
	
#define	JumpRelativeWithBuiltInFadeForControl_(izzyParamsPtr, controlInfo, sceneIndex1, inRespectNextSceneFadeIn) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mJumpToSceneWithBuiltInFadeProc)(FALSE, (controlInfo)->mControl, (sceneIndex1), (inRespectNextSceneFadeIn), TRUE)
	
typedef	IzzyError			(*JumpToSceneWithBuiltInFadeProc)(							// returns an non-zero error code if there is an error
								Boolean						inCalledByActor,			// true if called by an actor, false if by a control
								void*						inObjectPtr,				// ActorInfo* if Actor, ControlInfo* if Control
								SInt32						inSceneIndex1,				// one-based index of scene to jump to or relative jump if inRelative is true
								Boolean						inRespectNextSceneFadeIn,	// if true, respect to fade in time of the target scene
								Boolean						inRelative);				// if true, inSceneIndex1 specifies a relative jump
// v2.6 <-
// -----------------------------------------------------------------------------
// JumpToCue_
// -----------------------------------------------------------------------------
// v2.6 - added
// Jumps from the current scene to the cue specified by the specified cue
// number string.
// 
// v2.6: added render order parameter
// 
// ** NOT ** INTERRUPT SAFE: This function must not be called at interrupt time.

// v2.6 ->

#define	JumpToCueWithBuiltInFade_(izzyParamsPtr, actorInfo, cueNumberStr) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mJumpToCueProc)((actorInfo)->mThing, (cueNumberStr), FALSE, 0, 0, kTransition_Additive)

#define	JumpToCue_(izzyParamsPtr, actorInfo, cueNumberStr) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mJumpToCueProc)((actorInfo)->mThing, (cueNumberStr), TRUE, 0, 0, kTransition_Additive)

#define	JumpToCueWithFade_(izzyParamsPtr, actorInfo, cueNumberStr, duration_mS) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mJumpToCueProc)((actorInfo)->mThing, (cueNumberStr), TRUE, (duration_mS), (duration_mS), kTransition_Additive)

#define	JumpToCueWithFadeInFadeOut_(izzyParamsPtr, actorInfo, cueNumberStr, fadeInDuration_mS, fadeOutDuration_mS) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mJumpToCueProc)((actorInfo)->mThing, (cueNumberStr), TRUE, (fadeInDuration_mS), (fadeOutDuration_mS), kTransition_Additive)

#define	JumpToCueWithFadeInFadeOutWithTransition_(izzyParamsPtr, actorInfo, cueNumberStr, fadeInDuration_mS, fadeOutDuration_mS, transitionMode) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mJumpToCueProc)((actorInfo)->mThing, (cueNumberStr), TRUE, (fadeInDuration_mS), (fadeOutDuration_mS), (transitionMode))
	
typedef	IzzyError			(*JumpToCueProc)(										// returns an non-zero error code if there is an error
								ThingPtr					inThing,
								const char*					inCueNumber,
								Boolean						inFadeOverride,
								UInt32						inFadeOutDuration_mS,
								UInt32						inFadeInDuration_mS,
								TransitionMode				inTransitionMode);
// v2.6 <-

// -----------------------------------------------------------------------------
// ActivateDeactivateScene_
// -----------------------------------------------------------------------------
// Activates or deactivates a scene other than the current scene, referred to 
// here as a 'secondary' scene. Once a secondary scene is activated, it will
// remain active until 1) it becomes the primary scene (via a jump) and is
// subsequently deactivated by jumping to another scene or 2) it is deactivated
// by this calling this function with the activate parameter set to false.
// 
// The scene to be activated or deactivated is determined by the jump mode
// and scene index parameters -- see the JumpToScene_ command for more info.
// 
// v2.6: added render order parameter
// 
// ** NOT ** INTERRUPT SAFE: This function must not be called at interrupt time.

// v2.6 ->

#define	ActivateDeactivateScene_(izzyParamsPtr, actorInfo, mode, sceneIndex, activate) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mActivateDeactivateSceneProc)((actorInfo)->mThing, (mode), (sceneIndex), (activate), 0, kTransition_Additive)

#define	ActivateDeactivateSceneWithFade_(izzyParamsPtr, actorInfo, mode, sceneIndex, activate, duration_mS) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mActivateDeactivateSceneProc)((actorInfo)->mThing, (mode), (sceneIndex), (activate), (duration_mS), kTransition_Additive)

#define	ActivateDeactivateSceneWithFadeWithTransition_(izzyParamsPtr, actorInfo, mode, sceneIndex, activate, duration_mS, transitionMode) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mActivateDeactivateSceneProc)((actorInfo)->mThing, (mode), (sceneIndex), (activate), (duration_mS), (transitionMode))

typedef	IzzyError			(*ActivateDeactivateSceneProc)(							// returns an non-zero error code if there is an error
								ThingPtr					inThing,
								JumpMode					inMode,
								SInt32						inSceneIndex,
								Boolean						inActivate,
								UInt32						inFadeDuration_mS,
								TransitionMode				inCrossafadeMode);

// v2.6 <-

// -----------------------------------------------------------------------------
// PreloadScene_
// -----------------------------------------------------------------------------
// Preloads all of the media within a scene
// 
// The scene to be preloaded is determined by the jump mode
// and scene index parameters -- see the JumpToScene_ command for more info.
// 
// ** NOT ** INTERRUPT SAFE: This function must not be called at interrupt time.

#define	PreloadScene_(izzyParamsPtr, actorInfo, mode, sceneIndex, inDelayBetweenPreloads_mS) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mPreloadSceneProc)((actorInfo)->mThing, (mode), (sceneIndex), (inDelayBetweenPreloads_mS))

typedef	IzzyError			(*PreloadSceneProc)(											// returns an non-zero error code if there is an error
								ThingPtr					inThing,
								JumpMode					inMode,
								SInt32						inSceneIndex,
								UInt32						inDelayBetweenPreloads_mS);

// -----------------------------------------------------------------------------
// AddToLoadedMediaList_
// RemoveFromLoadedMediaList_
// -----------------------------------------------------------------------------
// Plugins that load media should call these functions to ensure information
// about the number of loaded media items in the user interface is accurate
//
// When the plugin loads media, it should call AddToLoadedMediaList_. If the
// media is not from the Media Window, then use pass kInvalidMediaIndex for
// the inMediaIndex1 parameter. inUniqueID should be an identifier that is
// unique to the loaded media and that will not change.
// 
// When the plugin unloads the media, it should call RemoveFromLoadedMediaList_
// with the same media type, media index, and unique ID specified when the
// plugin called AddToLoadedMediaList_
//
// ** NOT ** INTERRUPT SAFE: This function must not be called at interrupt time.

#define	AddToLoadedMediaList_(izzyParamsPtr, actorInfo, uniqueID, mediaType, mediaIndex1) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mAddRemoveMediaFromLoadedMediaListProc)((actorInfo)->mThing, TRUE, (uniqueID), (mediaType), (mediaIndex1))

#define	RemoveFromLoadedMediaList_(izzyParamsPtr, actorInfo, uniqueID, mediaType, mediaIndex1) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mAddRemoveMediaFromLoadedMediaListProc)((actorInfo)->mThing, FALSE, (uniqueID), (mediaType), (mediaIndex1))

typedef	IzzyError			(*AddRemoveMediaFromLoadedMediaListProc)(				// returns an non-zero error code if there is an error
								ThingPtr					inThing,	
								Boolean						inAdd,					// TRUE when adding media, false when removing it
								void*						inUniqueID,				// REQUIRE: unique identifier
								MediaCode					inMediaType,			// REQUIRED: the media type
								MediaIndexT					inMediaIndex1);			// OPT: specify if media is related to media window otherwise use kInvalidMediaIndex

// -----------------------------------------------------------------------------
// SetSceneIntensity_
// -----------------------------------------------------------------------------
// Allows you to set the intensity of a secondary scene. The intensity controls
// the visual brightness and aural volume of a secondary scene. Note that it is
// essential that you call tne ClearSceneIntensity_() to deactivate a scene that
// was activated via SetSceneIntensity_() or SetSceneIntensityWithTransition_().
// D not use the ActivateDeactivateScene_() function!

// v3.0.8b17 ->
enum IzAutoDeactivate
{
	kIzAutoDeactivate_Off = 0,
	kIzAutoDeactivate_On,
	kIzAutoDeactivate_DontCare
};
// <-

// v2.6 ->
#define	SetSceneIntensity_(izzyParamsPtr, actorInfo, mode, sceneIndex, intensity) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetSceneIntensityProc)((actorInfo)->mThing, (mode), (sceneIndex), (intensity), kTransition_Additive, kIzAutoDeactivate_DontCare)

#define	SetSceneIntensityWithTransition_(izzyParamsPtr, actorInfo, mode, sceneIndex, intensity, transitionMode) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetSceneIntensityProc)((actorInfo)->mThing, (mode), (sceneIndex), (intensity), (transitionMode), kIzAutoDeactivate_DontCare)

#define	ClearSceneIntensity_(izzyParamsPtr, actorInfo, mode, sceneIndex, autoDeacMode) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetSceneIntensityProc)((actorInfo)->mThing, (mode), (sceneIndex), 0.0, kTransition_Invalid, (autoDeacMode))

typedef	IzzyError			(*SetSceneIntensityProc)(								// returns an non-zero error code if there is an error
								ThingPtr					inThing,
								JumpMode					inMode,
								SInt32						inSceneIndex,
								IzReal						inIntensity,
								TransitionMode				inTransitionMode,
								IzAutoDeactivate			inAutoDeactivateMode);
// v2.6 <-

// -----------------------------------------------------------------------------
// GetSceneIntensity_
// -----------------------------------------------------------------------------
// Returns the master intensity for the scene that owns this actor

#define	GetSceneIntensity_(izzyParamsPtr, actorInfo) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetSceneIntensityProc)((actorInfo)->mThing)

typedef	float			(*GetSceneIntensityProc)(
							ThingPtr				inThing);

#define	GetContextSceneIntensity_(ip, actorInfo, ctx) ((ctx)->mIgnoreFade ? 1.0 : GetSceneIntensity_((ip), (actorInfo)))

// -----------------------------------------------------------------------------
// GetSceneIndex_
// -----------------------------------------------------------------------------
// Returns the one-based index of the scene that owns the specified actor

#define	GetSceneIndex_(izzyParamsPtr, actorInfo) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetSceneIndexProc)((actorInfo)->mThing)

typedef	SceneIndexT		(*GetSceneIndexProc)(
							ThingPtr				inThing);

// -----------------------------------------------------------------------------
// GetSceneCountForActor_
// GetSceneCountForControl_
// -----------------------------------------------------------------------------
// Returns the number of scenes in the current Isadora document
// added 3.0.8b18

#define	GetSceneCountForActor_(izzyParamsPtr, actorInfo, outSceneCount) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetSceneCountProc)(TRUE, (actorInfo)->mThing, (outSceneCount))

#define	GetSceneCountForControl_(izzyParamsPtr, actorInfo, outSceneCount) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetSceneCountProc)(FALSE, (actorInfo)->mThing, (outSceneCount))

typedef	IzzyError		(*GetSceneCountProc)(
							Boolean					inCalledByActor,
							void*					inObjectPtr,
							SceneIndexT*			outSceneCount);

// -----------------------------------------------------------------------------
// IsActiveAsPrimaryScene_
// -----------------------------------------------------------------------------

#define	IsActiveAsPrimaryScene_(izzyParamsPtr, actorInfo, sceneIndex1) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIsActiveAsPrimarySceneProc)((actorInfo)->mThing, (sceneIndex1))

typedef	Boolean			(*IsActiveAsPrimarySceneProc)(
							ThingPtr					inThing,
							SceneIndexT					inSceneIndex1);

// -----------------------------------------------------------------------------
// IsActiveAsSecondaryScene_
// -----------------------------------------------------------------------------

#define	IsActiveAsSecondaryScene_(izzyParamsPtr, actorInfo, sceneIndex1) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIsActiveAsSecondarySceneProc)((actorInfo)->mThing, (sceneIndex1))

typedef	Boolean			(*IsActiveAsSecondarySceneProc)(
							ThingPtr					inThing,
							SceneIndexT					inSceneIndex1);

// -----------------------------------------------------------------------------
// PropertyTypeAndIndexToHelpIndex_
// -----------------------------------------------------------------------------
// Converts a Property Type and Property Index into a zero-based index that can
// be used to access a help string.
// 
// This utility function is usually called from within the GetHelpString function
// inside of your plugin to determine which help string to return to the caller.
// If you organize your help strings as follows
// 
// Index 0 = General Help String Index 1 = Input Property 1 Help String Index 2 =
// Input Property 2 Help String ... Index n = Input Property n Help String Index
// n+1 = Output Property 1 Help String Index n+2 = Output Property 2 Help String
// ...
// 
// Then you can use this function to convert the Property Type and Property Index
// passed to GetHelpString to generate the proper index into that array.

#define	PropertyTypeAndIndexToHelpIndex_(izzyParamsPtr, actorInfo, propMode, propIndex) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mPropertyTypeAndIndexToHelpIndexProc)((actorInfo)->mThing, (propMode), (propIndex))

typedef UInt32				(*PropertyTypeAndIndexToHelpIndexProc)(
								ThingPtr					inThing,
								PropertyType				inPropertyType,
								PropertyIndex				inPropertyIndex1);

// -----------------------------------------------------------------------------
// GetPropertyCount_
// -----------------------------------------------------------------------------
// Returns the number of input or output properties for a given actor.

#define GetPropertyCount_(izzyParamsPtr, actorInfo, propType, intPtr) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetPropertyCount)((actorInfo)->mThing, (propType), (intPtr))

typedef IzzyError	(*GetPropertyCountProc)(										// returns an non-zero error code if there is an error
						ThingPtr					inThingPtr,
						PropertyType				inPropertyType,
						UInt32*						outCount);


// -----------------------------------------------------------------------------
// AddProperty_
// -----------------------------------------------------------------------------
// Adds a new property to the actor.
//
#define	AddProperty_(izzyParamsPtr, actorInfo, propType, propID, conformPropID, name, dispFmts, initFmt, inMutableGroup, min, max, value) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mAddPropertyProc)((actorInfo)->mThing, (propType), (propID), (conformPropID), (name), 0, (dispFmts), (initFmt), (inMutableGroup), (min), (max), (value), 0, 0, 0)

#define	AddPropertyOptions_(izzyParamsPtr, actorInfo, propType, propID, conformPropID, name, dispFmts, initFmt, inMutableGroup, min, max, value, options) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mAddPropertyProc)((actorInfo)->mThing, (propType), (propID), (conformPropID), (name), 0, (dispFmts), (initFmt), (inMutableGroup), (min), (max), (value), 0, 0, (options))

#define	AddPropertyCustomDataType_(izzyParamsPtr, actorInfo, propType, propID, conformPropID, name, customDataType, dispFmts, initFmt, inMutableGroup, min, max, value) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mAddPropertyProc)((actorInfo)->mThing, (propType), (propID), (conformPropID), (name), (customDataType), (dispFmts), (initFmt), (inMutableGroup), (min), (max), (value), 0, 0, 0)

// AddPropertyCustomDataTypeAfterIndex_:
//	take note of the special values for inAfterIndex1
//	>= 1 ::: insert after specified by the one-based index
//	   0 ::: ignore this parameter
//	  -1 ::: insert before current first property

#define	AddPropertyCustomDataTypeAfterIndex_(izzyParamsPtr, actorInfo, propType, propID, conformPropID, name, customDataType, dispFmts, initFmt, inMutableGroup, min, max, value, inAfterIndex1) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mAddPropertyProc)((actorInfo)->mThing, (propType), (propID), (conformPropID), (name), (customDataType), (dispFmts), (initFmt), (inMutableGroup), (min), (max), (value), (inAfterIndex1), 0, 0)

#define	AddPropertyCustomDataTypeAfterPropWithID_(izzyParamsPtr, actorInfo, propType, propID, conformPropID, name, customDataType, dispFmts, initFmt, inMutableGroup, min, max, value, inAfterPropID) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mAddPropertyProc)((actorInfo)->mThing, (propType), (propID), (conformPropID), (name), (customDataType), (dispFmts), (initFmt), (inMutableGroup), (min), (max), (value), 0, (inAfterPropID), 0)

typedef IzzyError	(*AddPropertyProc)(													// returns an non-zero error code if there is an error
						ThingPtr					inThingPtr,	
						PropertyType				inPropertyType,
						PropIDT						inPropID,
						PropIDT						inConformToPropID,
						const char*					inName,
						PropIDT						inCustomDataType,
						PropertyDispFormat			inDisplayFormats,
						PropertyDispFormat			inInitialDisplayFormat,
						MutableGroup				inMutableGroup,
						const ValuePtr				inMinimum,
						const ValuePtr				inMaximum,
						const ValuePtr				inInitialValue,
						SInt32						inAfterIndex1,				// 0 means ignore, >= 1 means insert after index one based index; -1 meas insert as first item
						PropIDT						inAfterPropID,				// zero -> ignore; non-zero, insert after specified property if it exists
						PropertyOptions				inOptionFlags);


// -----------------------------------------------------------------------------
// RemoveProperty_
// -----------------------------------------------------------------------------
// Removes an existing property from an actor. If there are any connections to
// that property ad the time that this call is made, they are disconnected
// prior to the property being removed.
// 
// Note: in previous versions of the SDK this function was named RemovePropertyProc_
// which was incorrect; but both names are now offered.

#define	RemoveProperty_(izzyParamsPtr, actorInfo, propType, propIndex1) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mRemovePropertyProc)((actorInfo)->mThing, (propType), (propIndex1))

#define	RemovePropertyProc_(izzyParamsPtr, actorInfo, propType, propIndex1) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mRemovePropertyProc)((actorInfo)->mThing, (propType), (propIndex1))

typedef IzzyError	(*RemovePropertyProc)(										// returns an non-zero error code if there is an error
						ThingPtr					inThingPtr,
						PropertyType				inPropertyType,
						PropertyIndex				inPropertyIndex1);

// -----------------------------------------------------------------------------
// MoveProperty_
// -----------------------------------------------------------------------------
// Moves a property from it's current index in the list to a new location

#define	MoveProperty_(izzyParamsPtr, actorInfo, propType, inCurrentPropIndex1, inDesiredPropIndex1) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mMovePropertyProc)((actorInfo)->mThing, (propType), (inCurrentPropIndex1), (inDesiredPropIndex1))

typedef IzzyError	(*MovePropertyProc)(										// returns an non-zero error code if there is an error
						ThingPtr					inThingPtr,
						PropertyType				inPropertyType,
						PropertyIndex				inCurrentPropIndex1,
						PropertyIndex				inDesiredPropIndex1);


// -----------------------------------------------------------------------------
// SetPropertyValueToStringCallback_
// -----------------------------------------------------------------------------
// Sets the title of the specified property

typedef bool (*CustomValueToStringProc)(
				struct IsadoraParameters*	inIzzyParams,
				struct ActorInfo*			inActorParams,
				PropertyType				inPropertyType,
				PropertyIndex				inPropertyIndex,
				const ValuePtr				inValue,
				size_t						outBufferSize,
				char*						outBuffer,
				bool*						outDisplayAsInvalid);
	
#define	SetPropertyValueToStringCallback_(izzyParamsPtr, actorInfo, propType, propIndex1, callback) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetPropertyValueToStringCallbackProc)((actorInfo)->mThing, (propType), (propIndex1), (callback))

typedef IzzyError	(*SetPropertyValueToStringCallbackProc)(										// returns an non-zero error code if there is an error
						ThingPtr					inThingPtr,
						PropertyType				inPropertyType,
						PropertyIndex				inPropertyIndex1,
						CustomValueToStringProc		inCustomConverterProc);

// -----------------------------------------------------------------------------
// SetPropertyName_
// -----------------------------------------------------------------------------
// Sets the title of the specified property

#define	SetPropertyName_(izzyParamsPtr, actorInfo, propType, propIndex1, propName) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetPropertyNameProc)((actorInfo)->mThing, (propType), (propIndex1), (propName))

typedef IzzyError	(*SetPropertyNameProc)(										// returns an non-zero error code if there is an error
						ThingPtr					inThingPtr,
						PropertyType				inPropertyType,
						PropertyIndex				inPropertyIndex1,
						const char*					inPropertyName);


// -----------------------------------------------------------------------------
// SetPropertyAllowedMutableFormats_
// -----------------------------------------------------------------------------
// Adds a new property to the actor.
//
#define	SetPropertyAllowedMutableFormats_(izzyParamsPtr, actorInfo, propType, propIndex1, allowedFmts) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetPropertyAllowedMutableFormatsProc)((actorInfo)->mThing, (propType), (propIndex1), (allowedFmts))

typedef IzzyError	(*SetPropertyAllowedMutableFormatsProc)(										// returns an non-zero error code if there is an error
						ThingPtr					inThingPtr,
						PropertyType				inPropertyType,
						PropertyIndex				inPropertyIndex1,
						PropertyDispFormat			inAllowedMutableFormats);
	
// -----------------------------------------------------------------------------
// GetPropertyCustomDataType_
// -----------------------------------------------------------------------------
// Returns the custom data type of the specified property. If the property is not
// of type kData, and if it does not have the kDisplayFormatPropertyDefined set,
// the result will be zero. Otherwise, the custom data type for the property
// will be returned.

#define	GetPropertyCustomDataType_(izzyParamsPtr, actorInfo, propType, propIndex1) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetPropertyCustomDataTypeProc)((actorInfo)->mThing, (propType), (propIndex1))

typedef PropIDT	(*GetPropertyCustomDataTypeProc)(
						ThingPtr					inThingPtr,	
						PropertyType				inPropertyType,
						PropertyIndex				inPropertyIndex);
	
// -----------------------------------------------------------------------------
// SetPropertyCustomDataType_
// -----------------------------------------------------------------------------
// Sets the custom data type for a specified property, which must already
// be of type kData and must have the kDisplayFormatCustomData bit set. If
// these conditions are not met, an error will be returned

#define	SetPropertyCustomDataType_(izzyParamsPtr, actorInfo, propType, propIndex1, customDataType) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetPropertyCustomDataTypeProc)((actorInfo)->mThing, (propType), (propIndex1), (customDataType))

typedef IzzyError	(*SetPropertyCustomDataTypeProc)(
						ThingPtr					inThingPtr,	
						PropertyType				inPropertyType,
						PropertyIndex				inPropertyIndex,
						PropIDT						inCustomDataType);
	
// -----------------------------------------------------------------------------
// GetPropertyAllowedMutableFormats_
// -----------------------------------------------------------------------------
// Adds a new property to the actor.
//
#define	GetPropertyAllowedMutableFormats_(izzyParamsPtr, actorInfo, propType, propIndex1, allowedFmts) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetPropertyAllowedMutableFormatsProc)((actorInfo)->mThing, (propType), (propIndex1), (allowedFmts))

typedef IzzyError	(*GetPropertyAllowedMutableFormatsProc)(										// returns an non-zero error code if there is an error
						ThingPtr					inThingPtr,
						PropertyType				inPropertyType,
						PropertyIndex				inPropertyIndex1,
						PropertyDispFormat*			outAllowedMutableFormats);
	
// -----------------------------------------------------------------------------
// GetPropertyName_
// -----------------------------------------------------------------------------
// Gets the title of the specified property

#define	GetPropertyName_(izzyParamsPtr, actorInfo, propType, propIndex1, propName, bufferSize) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetPropertyNameProc)((actorInfo)->mThing, (propType), (propIndex1), (propName), (bufferSize))

typedef IzzyError	(*GetPropertyNameProc)(										// returns an non-zero error code if there is an error
						ThingPtr					inThingPtr,
						PropertyType				inPropertyType,
						PropertyIndex				inPropertyIndex1,
						char*						outPropertyName,
						UInt32						inBufferSize);


// -----------------------------------------------------------------------------
// GetPropertyConnectionInfo_
// -----------------------------------------------------------------------------
// Retrieves information about a specific connection to/from a particular property.
// The caller supplies the Property Type, Property Index and Connection Index; if
// there is a valid connection to/from this property with the specified connection
// index, the result will be true and outConenctedThing and outConnectedPropertyIndex1
// will identify the connection. The caller may use this information to create
// or restore a connection using the LinkProperties_ macro below

#define	GetPropertyConnectionInfo_(izzyParamsPtr, actorInfo, propType, propIndex1, connectionIndex1, outConnectedThing, outConnectedPropertyIndex1) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetPropertyConnectionInfoProc)((actorInfo)->mThing, (propType), (propIndex1), (connectionIndex1), (outConnectedThing), (outConnectedPropertyIndex1))

typedef Boolean	(*GetPropertyConnectionInfoProc)(
						ThingPtr					inThingPtr,
						PropertyType				inPropertyType,
						PropertyIndex				inPropertyIndex1,
						UInt32						inConnectionIndex1,
						ThingPtr*					outConenctedThing,
						PropertyIndex*				outConnectedPropertyIndex1);


// -----------------------------------------------------------------------------
// LinkProperties_
// -----------------------------------------------------------------------------
// Creates a link between two propereties. One property is specified using the
// actorInfo, property type, and property index parameters. The other is specified
// by the "other thing" and "other property index" parameters. If the first property
// is an input, then the other property is assumed to be an output and vice-versa

#define	LinkProperties_(izzyParamsPtr, actorInfo, propType, propIndex1, inOtherThing, inOtherPropertyIndex1) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mLinkPropertiesProc)((actorInfo)->mThing, (propType), (propIndex1), (inOtherThing), (inOtherPropertyIndex1))

typedef IzzyError	(*LinkPropertiesProc)(										// returns an non-zero error code if there is an error
						ThingPtr					inThingPtr,
						PropertyType				inPropertyType,
						PropertyIndex				inPropertyIndex1,
						ThingPtr					inOtherThing,
						PropertyIndex				inOtherPropertyIndex1);


// v2.6 ->
// -----------------------------------------------------------------------------
// GetActorResourcePICT_ & GetControlResourcePICT_
// -----------------------------------------------------------------------------
// Gets a Macintosh PICT resource from an actor and converts it to an AlphaBitmap.
// This function will open the resource fork of the plugin and return a rendered
// version of the PICT in an AlphaBitmap struct if it is found. If the resource
// is not found, the AlphaBitmap struct passed into the function will remain
// unchanged.
//
// WIN32 NOTE:
//
// Note that if Isadora is running under windows, and the plugin file itself
// does not have a resource fork, Izzy will attempt to open a companion resource
// file with the .qtr extension. So, if a plugin called "myplugin.dll",
// calls GetActorMacResource_ and it has not been RezWack'd to include the
// Mac style resources as part of the plugin, then Isadora will look for
// a file in the same folder as myplugin.dll called "myplugin.qtr" and attempt
// to open it. If it can find the file it will attempt to read the resources
// from there.
//
// It is up to the plugin to insure that the returned ImageBuffer is disposed!

#define	GetActorResourcePICT_(izzyParamsPtr, actorInfo, resID, outRenderedImage) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetPluginResourcePICTProc)((actorInfo), NULL, (resID), (outRenderedImage))

#define	GetControlResourcePICT_(izzyParamsPtr, controlInfo, resID, outRenderedImage) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetPluginResourcePICTProc)(NULL, (controlInfo), (resID), (outRenderedImage))

typedef IzzyError	(*GetPluginResourcePICTProc)(								// returns an non-zero error code if there is an error
						const ActorInfo*			inActorInfo,				// non NULL for actors, NULL for controls
						const ControlInfo*			inControlInfo,				// non NULL for controls, NULL for actors
						SInt16						inResID,
						struct AlphaBitmap*			outRenderedImage);

// v2.6 <-

// -----------------------------------------------------------------------------
// LookupPartialStringInList_
// -----------------------------------------------------------------------------
// Searches a list of c-strings for the one whose starting characters
// most closely match the specified match string. Returns -1 if no matches
// could be made, otherwise, returns the zero based index of the matched 
// string in the list.
//
// This is useful if you have a property defined value string with a several
// options (e.g., the wave generator, which has the sine, square, triangle, etc.
// wave options). You can use this function to determine which item to
// chose based as the user types by calling it from your PropertyStringToValue
// function.

#define	LookupPartialStringInList_(izzyParamsPtr, inMaxStrings, inStringList, inStringToMatch) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mLookupPartialStringInListProc)((inMaxStrings), (inStringList), (inStringToMatch))

typedef SInt32									// returns -1 if no matches, otherwize zero-based index to match
(*LookupPartialStringInListProc)(
	UInt32				inMaxStrings,			// max strings to match
	const char* const *	inStringList,			// an array of c-strings (e.g., char* str[3] = { "one", "two", "three" }
	const char*			inStringToMatch);		// the string you are attempting to match
	

// -----------------------------------------------------------------------------
// ModKeyStringToMaskProc_
// -----------------------------------------------------------------------------
// Converts a c-string of one or more modifer keys string codes into a mod 
// key mask. Illegal or unknown codes are ignored and stop the parsing of
// the string.
//
// The valid codes are:
//
// "cmd" or "d" for command		mask = 0x0001
// "ctl" or "c" for control		mask = 0x0002
// "opt" or "o" for option		mask = 0x0004
// "shf" or "s" for shift		mask = 0x0010
//
// Example: string = "cmd opt" - result mask = 0x0005

#define	ModKeyStringToMask_(izzyParamsPtr, inString, outMask) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mModKeyStringToMaskProc)((inString), (outMask))

typedef void	(*ModKeyStringToMaskProc)(
						const char* 				inString,
						UInt32*						outMask);

// -----------------------------------------------------------------------------
// ModKeyMaskToString_
// -----------------------------------------------------------------------------
// Converts a modifer keys mask into a c-string representation. See above for
// mask codes and their string equivalents.

#define	ModKeyMaskToString_(izzyParamsPtr, inModKeyMask, outString) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mModKeyMaskToStringProc)((inModKeyMask), (outString))

typedef void	(*ModKeyMaskToStringProc)(
						UInt32						inModKeyMask,
						char*						outString);
	
// -----------------------------------------------------------------------------
// GetScreenBounds_
// -----------------------------------------------------------------------------
// Returns the bounding rectangle of any screen currently available on the
// the system. Screen 1 is always the primary start up screen. The numbering
// of other external monitors varies from system to system.
//
// The return value is true if the requested screen exists. Otherwise it is
// false.

#define	GetScreenBounds_(izzyParamsPtr, inScreenIndex1, outRect) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetScreenBoundsProc)((inScreenIndex1), (outRect))

typedef Boolean	(*GetScreenBoundsProc)(
						UInt32						inScreenIndex1,
						IzRect*						outScreenBounds);
	
// -----------------------------------------------------------------------------
// SendMidiNoteOnOff_
// -----------------------------------------------------------------------------
// Sends a midi note on message. The note is sent out of the specified port,
// with the channel, pitch, and velocity values given. If the inNoteOffMs
// parameter is greater than 0, then a corresponding Note Off message will
// be sent after the specified number of milliseconds has elapsed.

#define	SendMidiNoteOnOff_(izzyParamsPtr, inScreenIndex1, inPort, inChan, inPitch, inVel, inNoteOffMs, inRefCon) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSendMidiNoteOnOffProc)((inPort), (inChan), (inPitch), (inVel), (inNoteOffMs), (inRefCon))

typedef void	(*SendMidiNoteOnOffProc)(
						UInt16					inPortIndex1,							// one-based midi port - as shown in the MIDI setup window
						UInt8					inChannel1,								// one-based midi channel
						UInt8					inPitch,								// midi pitch, 0-127
						UInt8					inVelocity,								// midi velocity, 0-127
						UInt32					inMatchingNoteOffMilliseconds,			// send note off message after this duration in milliseconds
						IzRefCon				inRefCon);								// ref con, used with CancelMidiNoteOff_
	
// -----------------------------------------------------------------------------
// CancelMidiNoteOff_
// -----------------------------------------------------------------------------
// Cancels a note off command scheduled with the SendMidiNoteOnOff_ function
// The port, channel, pitch, and refcon specified in the SendMidiNoteOnOff_
// must all match for the note off to be properly cancelled.

#define	CancelMidiNoteOff_(izzyParamsPtr, inScreenIndex1, inPort, inChan, inPitch, inRefCon) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mCancelMidiNoteOffProc)((inPort), (inChan), (inPitch), (inRefCon))

typedef void	(*CancelMidiNoteOffProc)(
						UInt16					inPortIndex1,
						UInt8					inChannel1,
						UInt8					inPitch,
						IzRefCon				inRefCon);
	
// -----------------------------------------------------------------------------
// GetMidiDancerEnergy_
// -----------------------------------------------------------------------------
// Returns the overall energy of any MidiDancer activity seen on the specified
// MIDI input port.

#define	GetMidiDancerEnergy_(izzyParamsPtr, inPort, inLimbCount) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetMidiDancerEnergyProc)((inPort), (inLimbCount))

typedef float	(*GetMidiDancerEnergyProc)(
					UInt16						inPortIndex1,
					SInt32						inLimbCount);
	
// -----------------------------------------------------------------------------
// GetMidiDancerEnergyPlus_
// -----------------------------------------------------------------------------
// Returns the overall energy of any MidiDancer activity seen on the specified
// MIDI input port.

#define	GetMidiDancerEnergyPlus_(izzyParamsPtr, inPort, inMaxMilliseconds, inMinLimbIndex1, inMaxLimbIndex1) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetMidiDancerEnergyPlusProc)((inPort), (inMaxMilliseconds), (inMinLimbIndex1), (inMaxLimbIndex1))

typedef float	(*GetMidiDancerEnergyPlusProc)(
					UInt16						inPortIndex1,
					SInt32						inMaxMilliseconds,
					SInt32						inMinLimbIndex1,
					SInt32						inMaxLimbIndex1);
	
// -----------------------------------------------------------------------------
// GetTextFromDialog_
// -----------------------------------------------------------------------------
// Shows a dialog with the intial text provided by ioTextHandle. After the user
// is done editing, returns the new text in the same handle. The result is true
// if the user clicked OK in the dialog, false otherwise.

#define	GetTextFromDialog_(izzyParamsPtr, actorInfo, ioTextHandle, inDialogTitle, inValidateTextProc, inValidateTextRefCon, inMakeDocDirtyOnSuccess, helpURL) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetTextFromDialogProc)((actorInfo)->mThing, (ioTextHandle), (inDialogTitle), (inValidateTextProc), (inValidateTextRefCon), (inMakeDocDirtyOnSuccess), (helpURL))

typedef Boolean (*ValidateTextProc)(										// return true if text is valid, false if there is an error
					IzzyHandle	inTextHandle,								// text handle to validate
					SInt32*		outSelStart,								// when returning FALSE, selection start in dialog
					SInt32*		outErrSelEnd,								// when returning FALSE, selection end in dialog
					char* 		outErrorMessageBuffer,						// when returning FALSE, copy an error message here. set to null string for no error message
					SInt32		inErrorMessageBufferMaxLen,
					void*		inValidateTextRefCon);						// maximum length for the error message buffer

typedef Boolean	(*GetTextFromDialogProc)(
					ThingPtr					inThingPtr,					// owning ThingPtr
					IzzyHandle					ioTextHandle,				// on entry, the handle to the text; on exit, the edited text
					const char*					inDialogTitle,				// can be NULL (defaults to "Text Editor")
					ValidateTextProc			inValidateTextProc,			// can be NULL
					void*						inValidateTextRefCon,
					Boolean						inMakeDocDirtyOnSuccess,	// if TRUE, will make the document "dirty" if the text is changed
					const char*					inHelpURL);					// if non NULL, the URL to a help file for this dialog

// -----------------------------------------------------------------------------
// GetTextFromDialogAdvanced_
// -----------------------------------------------------------------------------
// Shows a dialog with the intial text provided by ioTextHandle. After the user
// is done editing, returns the new text in the same handle. The result is true
// if the user clicked OK in the dialog, false otherwise.

enum {
	kGetTextFromDialogFlags_None					= 0,
	kGetTextFromDialogFlags_ShowFontMenu			= (1 << 0),
	kGetTextFromDialogFlags_ShowFontSizeMenu		= (1 << 1),
	kGetTextFromDialogFlags_DisableWordWrap			= (1 << 2),
	
	// for editing code
	kGetTextFromDialogFlags_JavascriptEditor		= (1 << 3),
	kGetTextFromDialogFlags_CompileButton			= (1 << 4),
	kGetTextFromDialogFlags_ErrorsDisplay			= (1 << 5),
	kGetTextFromDialogFlags_LineNumberDisplay		= (1 << 6),
	kGetTextFromDialogFlags_DoubleString			= (1 << 7),

	// IS-1250 - Code Editor Syntax Coloring Is Now Language Aware
	// callers can only use this flag if the callback version number
	// in IsadoraParameters->mCallbacks->mCallbackVersionNumber is
	// is kIsadoraCallbacksVersion_322b50 or greater!
	kGetTextFromDialogFlags_PythonEditor			= (1 << 8),		// recognized in kIsadoraCallbacksVersion_322b50 or later!

	// warning messages
	kGetTextFromDialogFlags_GLSLEditor				= (1 << 31),
	
	kGetTextFromDialogFlags_Default					= (kGetTextFromDialogFlags_ShowFontMenu | kGetTextFromDialogFlags_ShowFontSizeMenu)
};

enum {

	kGetTextFromDialog_Options_GLSL_AutoDetect					= (1 << 0),
	kGetTextFromDialog_Options_GLSL_GLSLStandard				= (1 << 1),
	kGetTextFromDialog_Options_GLSL_GLSLSandbox					= (1 << 2),
	kGetTextFromDialog_Options_GLSL_ShaderToy					= (1 << 3),
	kGetTextFromDialog_Options_GLSL_Unused1						= (1 << 4),
	kGetTextFromDialog_Options_GLSL_Unused2						= (1 << 5),
	kGetTextFromDialog_Options_GLSL_Unused3						= (1 << 6),
	kGetTextFromDialog_Options_GLSL_Unused4						= (1 << 7),
	
	kGetTextFromDialog_Options_GLSL_VertexShader_Auto			= (1 << 8),
	kGetTextFromDialog_Options_GLSL_VertexShader_UserSupplied	= (1 << 9),
	
	kGetTextFromDialog_Options_GLSL_ShowFragmentSource			= (1 << 10),
	kGetTextFromDialog_Options_GLSL_ShowVertexSource			= (1 << 11)
};

#define	GetTextFromDialogAdvanced_(izzyParamsPtr, actorInfo, ioTextHandle, inDialogTitle, inValidateTextProc, inValidateTextRefCon, inMakeDocDirtyOnSuccess, helpURL, inFlags, ioOptionFlags) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetTextFromDialogAdvancedProc)((actorInfo)->mThing, (ioTextHandle), (inDialogTitle), (inValidateTextProc), (inValidateTextRefCon), (inMakeDocDirtyOnSuccess), (helpURL), (inFlags), (ioOptionFlags))

typedef Boolean (*ValidateTextAdvancedProc)(								// return true if text is valid, false if there is an error
					IzzyHandle					inTextHandle,				// text handle to validate
					SInt32*						outSelStart,				// when returning FALSE, selection start in dialog
					SInt32*						outErrSelEnd,				// when returning FALSE, selection end in dialog
					char*						outErrorMessageBuffer,		// when returning FALSE, copy an error message here. set to null string for no error message
					SInt32						inErrorMessageBufferMaxLen,	// maximum length for the error message buffer
					void*						inRefCon,					// reference constant passed to GetTextFromDialogAdvanced_
					UInt32*						inGetTextOptionsFlags);		// current kGetTextFromDialog_Options flags
					
typedef Boolean	(*GetTextFromDialogAdvancedProc)(
					ThingPtr					inThingPtr,					// owning ThingPtr
					IzzyHandle					ioTextHandle,				// on entry, the handle to the text; on exit, the edited text
					const char*					inDialogTitle,				// can be NULL (defaults to "Text Editor")
					ValidateTextAdvancedProc	inValidateTextProc,			// can be NULL
					void*						inValidateTextRefCon,		// reference constant passed back to the validate text proc
					Boolean						inMakeDocDirtyOnSuccess,	// if TRUE, will make the document "dirty" if the text is changed
					const char*					inHelpURL,					// if non NULL, the URL to a help file for this dialog
					UInt32						inGetTextFromDialogFlags,	// see kGetTextFromDialogFlags flags above
					UInt32*						ioGetTextOptionsFlags);		// see kGetTextFromDialog_Options above
	
// -----------------------------------------------------------------------------
// GetChannelMapDialog_
// -----------------------------------------------------------------------------
// Shows a dialog that allows the user to edit a list of single input channels
// that are linked with corresponding multiple output channels

#define	GetChannelMapDialog_(izzyParamsPtr, actorInfo, ioTextHandle, inDialogTitle, inMakeDocDirtyOnSuccess, helpURL) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetChannelMapDialogProc)((actorInfo)->mThing, (ioTextHandle), (inDialogTitle), (inMakeDocDirtyOnSuccess), (helpURL))

typedef Boolean	(*GetChannelMapDialogProc)(
					ThingPtr					inThingPtr,
					IzzyHandle					ioTextHandle,
					const char*					inDialogTitle,				// can be NULL
					Boolean						inMakeDocDirtyOnSuccess,
					const char*					inHelpURL);					// if non NULL, the URL to a help file for this dialog
	
// -----------------------------------------------------------------------------
// EditLanBoxRectListDialog_
// -----------------------------------------------------------------------------
// Shows a dialog where the user may edit a list of rectangles that determine
// where the LanBox RGB actor will acquire it's RGB values.

typedef struct {
	IzRect	mRect;				// coordinates of rect in raw pixels
	UInt8	mAverage;			// 1 = true, 0 = false
	UInt8	mSelected;			// 1 = selected, 0 = not
	UInt16	mBaseDMXChannel;	// base dmx channel - based on previous rects in list
} LanBoxRect, *LanBoxRectPtr;

typedef struct {
	SInt32		mVersion;		// version number
	SInt32		mHRes;			// resolution used when rect list
	SInt32		mVRes;			// was last edited
	SInt32		mMagnification;	// code for magnification
	LanBoxRect	mRectList[1];
} LanBoxRectList, *LanBoxRectListPtr, **LanBoxRectListHandle;

#define	kLanBoxRectListHeaderSizeBytes	(sizeof(SInt32) * 4)

#define	EditLanBoxRectListDialog_(izzyParamsPtr, actorInfo, ioRectHandle, inMakeDocDirtyOnSuccess, inDialogTitle, inSourceImageBuffer, inChannelOffset, inRectChangedCallback) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mEditLanBoxRectListDialogProc)((actorInfo)->mThing, (ioRectHandle), (inMakeDocDirtyOnSuccess), (inDialogTitle), (inSourceImageBuffer), (inChannelOffset), (inRectChangedCallback))

typedef void	(*LBRectanglesChangedCallback)(								// return true if text is valid, false if there is an error
					IzzyHandle					inRectListHandle);			// text handle to validate

typedef Boolean	(*EditLanBoxRectListDialog)(
					ThingPtr					inThingPtr,
					LanBoxRectListHandle		ioRectHandle,
					Boolean						inMakeDocDirtyOnSuccess,
					const char*					inDialogTitle,					// can be NULL
					ImageBufferPtr				inSourceImageBuffer,			// can be NULL
					SInt32						inChannelOffset,				// can be 0
					LBRectanglesChangedCallback	inRectChangedCallbackProc);		// can be NULL
	
// -----------------------------------------------------------------------------
// PrepareActorDefinedAreaPict_
// -----------------------------------------------------------------------------
// Takes the specified picture and makes two NativeBitmaps from it: one drawn an
// unselected background and one on a selected background. This makes it 
// possible to draw the actor's icon pict as fast as possible, without needed
// to convert the picture into a NativeBitmap each time.

typedef struct OpaqueUnivBitmap*	ActorIconBitmap;

typedef struct {
	Boolean			mInitialized;
	ActorIconBitmap	mUnselBitmap;
	ActorIconBitmap	mSelBitmap;
	SInt16			mWidth;
	SInt16			mHeight;
} ActorPictInfo;

#define	PrepareActorDefinedAreaPict_(izzyParamsPtr, actorInfo, pictID, outActorPictInfoPtr) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mPrepareActorDefinedAreaPictProc)((actorInfo), (pictID), (outActorPictInfoPtr))

typedef void	(*PrepareActorDefinedAreaPictProc)(
					ActorInfo*					inActorInfo,
					SInt16						inPictID,
					ActorPictInfo*				inActorPictInfo);
	
// -----------------------------------------------------------------------------
// DrawActorDefinedAreaPict_
// -----------------------------------------------------------------------------
// Draws a picture referenced by an ActorPictInfo struct, which is created
// by the PrepareActorDefinedAreaPict_ function.

#define	DrawActorDefinedAreaPict_(izzyParamsPtr, actorInfo, inSelected, inAreaRect, inActorPictInfoPtr) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mDrawActorDefinedAreaPictProc)((actorInfo), (inSelected), (inAreaRect), (inActorPictInfoPtr))

typedef void	(*DrawActorDefinedAreaPictProc)(
					ActorInfo*					inActorInfo,
					Boolean						inSelected,
					const IzRect*				inAreaRect,
					ActorPictInfo*				inActorPictInfo);

// -----------------------------------------------------------------------------
// ActorDefinedAreaSizeChanged_
// -----------------------------------------------------------------------------
// If the dimensions of your Actor Defined Area is about to change, call this
// function to ensure the dimensions of the actor are updated in the GUI

#define	ActorDefinedAreaSizeChanged_(izzyParamsPtr, actorInfo) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mActorDefinedAreaSizeChangedProc)((actorInfo)->mThing)

typedef void	(*ActorDefinedAreaSizeChangedProc)(
					ThingPtr					inThingPtr);

// -----------------------------------------------------------------------------
// GetWidthForActorDefinedAreaPropertyIndicator_
// -----------------------------------------------------------------------------
// If you're actor uses the DrawActorDefinedAreaPropertyIndicator_ function,
// your GetActorDefinedAreaPixelsProc proc, use use this function to determine the
// width in screen pixels.

#define	GetWidthForActorDefinedAreaPropertyIndicator_(izzyParamsPtr, actorInfo, type, inPropIndex1, outPropIndex1) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetWidthForActorDefinedAreaPropertyIndicatorProc)((actorInfo)->mThing, (type), (inPropIndex1), (outPropIndex1))

typedef SInt16	(*GetWidthForActorDefinedAreaPropertyIndicatorProc)(
					ThingPtr					inThingPtr,
					ActorIndicatorType			inIndicatorType,
					PropertyIndex				inInputPropertyIndex1,		// optional for kActorIndicatorType_Bullet, required for kActorIndicatorType_Link
					PropertyIndex				inOutputPropertyIndex1);	// optional for kActorIndicatorType_Bullet, required for kActorIndicatorType_Link

// -----------------------------------------------------------------------------
// DrawActorDefinedAreaPropertyIndicator_
// -----------------------------------------------------------------------------
// If the dimensions of your Actor Defined Area is about to change, call this
// function to ensure the dimensions of the actor are updated in the GUI.
//
// When the ActorIndicatorType is kActorIndicatorType_Bullet, either
// inInputPropertyIndex1 or inOutputPropertyIndex1 can be valid. If
// neither if valid, then nothing is drawn.
//
// For the type kActorIndicatorType_Link, both inInputPropertyIndex1 and
// inOutputPropertyIndex1 must be valid. If both are not valid property
// indexes for the specified actor, then nothing is drawn.

#define	DrawActorDefinedAreaPropertyIndicator_(izzyParamsPtr, actorInfo, type, inEraseRect, inPropIndex1, outPropIndex1) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mDrawActorDefinedAreaPropertyIndicatorProc)((actorInfo)->mThing, (type), (inEraseRect), (inPropIndex1), (outPropIndex1))

typedef void	(*DrawActorDefinedAreaPropertyIndicatorProc)(
					ThingPtr					inThingPtr,
					ActorIndicatorType			inIndicatorType,
					const IzRect*				inEraseRect,				// if NULL, don't erase
					PropertyIndex				inInputPropertyIndex1,		// optional for kActorIndicatorType_Bullet, required for kActorIndicatorType_Link
					PropertyIndex				inOutputPropertyIndex1);	// optional for kActorIndicatorType_Bullet, required for kActorIndicatorType_Link

// -----------------------------------------------------------------------------
// DrawTextInRect_
// -----------------------------------------------------------------------------
// Draws the specified text inside a rectangle without erasing the background.

#define	DrawTextInRect_(izzyParamsPtr, inGraphics, inText, inLength, inRect, inHorzJustification, inVertJustification, inConstrainHorizontally) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mDrawTextInRectProc)((inGraphics), (inText), (inLength), (inRect), (inHorzJustification), (inVertJustification), (inConstrainHorizontally), NULL)

typedef struct {
	IzReal		mOffsetH;			// horizontal offset applied before rotation
	IzReal		mOffsetV;			// vertical offset applied before rotation
	IzReal		mRotationCenterH;	// x coordinate around which the rotation will occur
	IzReal		mRotationCenterV;	// y coordinate around which the rotation will occur
	IzReal		mAngle;				// angle of rotation.
	Boolean		mDontClip;			// disable clipping beyond bounds
} TextOffsetRotation;

#define	DrawTextInRectWithOffsetAndRotation_(izzyParamsPtr, inGraphics, inText, inLength, inRect, inHorzJustification, inVertJustification, inConstrainHorizontally, inTextOffsetRotation) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mDrawTextInRectProc)((inGraphics), (inText), (inLength), (inRect), (inHorzJustification), (inVertJustification), (inConstrainHorizontally), inTextOffsetRotation)

// horz just: Default = 0, Force Center = 1, Force Right = -1, Force Left = -2
// vert just: Top = 1, Middle = 2, Bottom = 3

typedef void	(*DrawTextInRectProc)(
					UGraphics*			inGraphics,
					const void*			inText,
					SInt32				inLength,
					const IzRect*		inRect,
					SInt16				inHorzJustification,		// should match QDHorzAlign: default = 0, center = 1, right = 2, force left = -2
					SInt16				inVertJustification,		// should match QDVertAlign: top = 1, middle = 2, bottom = 3
					Boolean				inConstrainHorizontally,
					TextOffsetRotation*	inTextRotationInfo);		// CAN BE NULL

//-----------------------------------------------------------------------------
// More Plugin Debugging

#define	kForcePluginAssert				1

#if DEBUG || kForcePluginAssert
	#define	PluginAssert_(ip, cond)			if (!(cond)) (*((ip)->kConditionalCallbacks->mPluginAssertProc))( #cond, kActorName, __LINE__)
	#define PluginSignalString_(ip, str)	(*((ip)->kConditionalCallbacks->mPluginAssertProc))( str, kActorName, __LINE__)
#else
	#define	PluginAssert_(ip, cond)			{ }
	#define PluginSignalString_(ip, str)	{ }
#endif

typedef void	(*PluginAssertProc)(
					const char*			inCond,
					const char*			inFile,
					int					inLine);


#if DEBUG
	#define	PluginLogInfo_(ip, ...)		(*((ip)->kConditionalCallbacks->mPluginLogInfoProc)) (__VA_ARGS__)
#else
	#define	PluginLogInfo_(ip, ...)		{ }
#endif

typedef void	(*PluginLogInfoProc)(
					const char*			inString,
					...);

// -----------------------------------------------------------------------------
// WarnPluginLowMemory_
// -----------------------------------------------------------------------------
// Shows a dialog that indicates a plugin ran out of memory

#define	WarnPluginLowMemory_(izzyParamsPtr, inPluginName) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mWarnLowMemoryProc)((inPluginName))

typedef void	(*WarnLowMemoryProc)(
					const char*		inPluginName);
	
// -----------------------------------------------------------------------------
// RecallSnapshot_
// -----------------------------------------------------------------------------
// Recals the specified snapshot, restoring the input values of all actors
// in the specified scene

#define	RecallSnapshot_(izzyParamsPtr, actorInfo, inSnapshotID) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mRecallSnapshotProc)((actorInfo)->mThing, (inSnapshotID))

typedef IzzyError	(*RecallSnapshotProc)(								// returns an non-zero error code if there is an error
						ThingPtr					inActorInfo,
						SInt32						inSnapshotID);
	
// -----------------------------------------------------------------------------
// TakeSnapshot_
// -----------------------------------------------------------------------------
// Takes a new snapshot of the current input values of all actors in the
// associated scene

#define	TakeSnapshot_(izzyParamsPtr, actorInfo) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mTakeSnapshotProc)((actorInfo)->mThing)

typedef IzzyError	(*TakeSnapshotProc)(								// returns an non-zero error code if there is an error
						ThingPtr					inActorInfo);
	
// -----------------------------------------------------------------------------
// UpdateSnapshot_
// -----------------------------------------------------------------------------
// Updates the specified snapshot so that it reflects the current input values
// for all actors in the associated scene.

#define	UpdateSnapshot_(izzyParamsPtr, actorInfo, inSnapshotID) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mUpdateSnapshotProc)((actorInfo)->mThing, (inSnapshotID))

typedef IzzyError	(*UpdateSnapshotProc)(								// returns an non-zero error code if there is an error
						ThingPtr					inActorInfo,
						SInt32						inSnapshotID);
	
// -----------------------------------------------------------------------------
// CreateImageBuffer_
// -----------------------------------------------------------------------------
// Creates an image buffer with the specified width, height, and depth

#define	CreateImageBuffer_(izzyParamsPtr, inWidth, inHeight, inFormat) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mCreateImageBufferProc)((inWidth), (inHeight), (inFormat))

typedef	ImageBufferPtr		(*CreateImageBufferProc)(
								SInt32			inWidth,			// width of image buffer
								SInt32			inHeight,			// height of image buffer
								OSType			inFormat);			// buffer format

// -----------------------------------------------------------------------------
// EnterVideoProcessing_
// -----------------------------------------------------------------------------
// Called just prior to entering code that does video processing - returns an
// integer that is passed back to ExitVideoProcessing so that we can determine
// how much time is spent on video processing.

#define	EnterVideoProcessing_(izzyParamsPtr) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mEnterVideoProcessingProc)()

typedef UInt64				(*EnterVideoProcessingProc)();
	
// -----------------------------------------------------------------------------
// ExitVideoProcessing_
// -----------------------------------------------------------------------------
// Called just after to exiting code that does video processing - you should
// pass in the 64 bit integer returned by EnterVideoProcessing_

#define	ExitVideoProcessing_(izzyParamsPtr, time) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mExitVideoProcessingProc)((time))

typedef void				(*ExitVideoProcessingProc)(
								UInt64 inStartTime);
	
// -----------------------------------------------------------------------------
// BroadcastControlValue_
// -----------------------------------------------------------------------------
// Broadcasts the the control value for the controller number specfied. The
// min, max, and current value are given, so that listeners can scale the
// values appropriately.

#define	BroadcastControlValue_(izzyParamsPtr, inControlInfo, inControlNum, inMin, inMax, inValue, inRedrawControl) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mBroadcastControlValueProc)((inControlInfo), (inControlNum), (inMin), (inMax), (inValue), (inRedrawControl))

typedef void	(*BroadcastControlValueProc)(
					ControlInfo*	inControlInfo,
					UInt32			inControlNum,
					IzReal			inMin,
					IzReal			inMax,
					IzReal			inValue,
					Boolean			inRedrawControl);
	
// -----------------------------------------------------------------------------
// BroadcastControlValueDirect_
// -----------------------------------------------------------------------------
// Broadcasts the the control value for the controller number specfied. The
// "direct" version of this call prevents any scaling -- controls that are
// listenting to these messages will either limit or ignore the values to 
// the available range, depending on the setting of the inIgnoreOutOfRange
// parameter

#define	BroadcastControlValueDirect_(izzyParamsPtr, inControlInfo, inControlNum, inValue, inIgnoreOutOfRange, inRedrawControl)  \
	(*(izzyParamsPtr)->kConditionalCallbacks->mBroadcastControlValueDirectProc)((inControlInfo), (inControlNum), (inValue), (inIgnoreOutOfRange), (inRedrawControl))

typedef void	(*BroadcastControlValueDirectProc)(
					ControlInfo*	inControlInfo,
					UInt32			inControlNum,
					IzReal			inValue,
					Boolean			inIgnoreOutOfRange,
					Boolean			inRedrawControl);
	
// -----------------------------------------------------------------------------
// BroadcastControlTimecode_
// -----------------------------------------------------------------------------
// Broadcasts the the control timecode value for the controller number specfied.

#define	BroadcastControlTimecode_(izzyParamsPtr, inControlInfo, inControlNum, inValue, inRedrawControl)  \
	(*(izzyParamsPtr)->kConditionalCallbacks->mBroadcastControlTimecodeProc)((inControlInfo), (inControlNum), (inValue), (inRedrawControl))

typedef void	(*BroadcastControlTimecodeProc)(
					ControlInfo*		inControlInfo,
					UInt32				inControlNum,
					IzValue_Timecode	inValue,
					Boolean				inRedrawControl);
	
// -----------------------------------------------------------------------------
// GetMediaListFromDialog_
// -----------------------------------------------------------------------------
// Shows/edits a list of indexes into a media array -- allows the caller to 
// keep a list of items chosen from within the media types supported by the
// media window.
//
// The function returns true if the data was changed

#define	GetMediaIndexListFromDialog_(izzyParamsPtr, inControlInfo, mediaType, ioIndexHandle) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetMediaIndexListFromDialogProc)((inControlInfo), (mediaType), (ioIndexHandle))

typedef Boolean	(*GetMediaIndexListFromDialogProc)(
					ControlInfo* 				inControlInfo,
					MediaCode					inMediaType,
					IzzyHandle					ioIndexHandle);
	
// -----------------------------------------------------------------------------
// GetMediaObjectInfo_
// -----------------------------------------------------------------------------
// Gets information about a particular media object, including its name and 
// its thumbnail, if one exists.
//
//	The function returns true if the media object actually exists

#define	GetMediaObjectInfo_(izzyParamsPtr, inControlInfo, mediaType, index1, outName, inFullSizeImage, outImage, outDuration_mS) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetMediaObjectInfoProc)((inControlInfo), (mediaType), (index1), (outName), (inFullSizeImage), (outImage), (outDuration_mS))

typedef Boolean	(*GetMediaObjectInfoProc)(
					ControlInfo* 	inControlInfo,
					MediaCode		inMediaType,
					SInt32			inMediaObjectIndex1,	// the one-based media object index
					char*			outName,				// buffer should be at least 256 characters
					Boolean			outFullSizeImage,		// set to true to get image in full resolution
					ImageBufferPtr*	outImage,				// the object's thumbnail or full size image if it exists - otherwise outThumbnail is set to nil
					UInt32*			outDuration_mS);		// Can be NULL -- Duration of the object in mS
	
// -----------------------------------------------------------------------------
// TrackPopupMenu_
// -----------------------------------------------------------------------------
// Given a c-string with a list of carriage return terminated items,
// shows a popup menu with those items and let's the user choose.
// The return value is the one-based index of the item that the user
// chose. If the user doesn't choose an item, then the return value
// is zero.

#define	TrackPopupMenu_(izzyParamsPtr, inControlInfo, inItemString, inCurrentItemIndex1) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mTrackPopupMenuProc)((inControlInfo), (inItemString), (inCurrentItemIndex1))

typedef SInt32	(*TrackPopupMenuProc)(
					ControlInfo* 				inControlInfo,
					const char*					inItemString,
					SInt32						inCurrentItemIndex1);
	
// -----------------------------------------------------------------------------
// GetRandomSeed_
// -----------------------------------------------------------------------------
// Returns a number to be used when seeding a random number generator
//

#define	GetRandomSeed_(izzyParamsPtr) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetRandomSeed)()

typedef UInt32	(*GetRandomSeedProc)();
	
// -----------------------------------------------------------------------------
// SetCaptureMode_
// -----------------------------------------------------------------------------
// Starts video and/or audio capture
//
// To enable, use the following flags when calling SetCaptureMode
//
//	Audio + Video	kIzzyCaptureVideo | kIzzyCaptureAudio
//	Vido			kIzzyCaptureVideo
//	Audio			kIzzyCaptureAudio
//
// To disable capture use kIzzyCaptureNone
//
// Returns true if successful, false otherwise
//

#define	SetCaptureMode_(izzyParamsPtr, inCaptureChannelIndex1, inMode, inVideoInputSelect, inAudioInputSelect, inSize) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetCaptureModeProc)((inCaptureChannelIndex1), (inMode), (inVideoInputSelect), (inAudioInputSelect), (inSize), false, 0,0,0,0,0,0)

#define	SetCaptureModeCustom_(izzyParamsPtr, inCaptureChannelIndex1, inMode, inVideoInputSelect, inAudioInputSelect, inSize, inUseCustom, inTop, inLeft, inWidth, inHeight, inDstWidth, inDstHeight) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetCaptureModeProc)((inCaptureChannelIndex1), (inMode), (inVideoInputSelect), (inAudioInputSelect), (inSize), (inUseCustom), (inTop), (inLeft), (inWidth), (inHeight), (inDstWidth), (inDstHeight))

typedef void	(*SetCaptureModeProc)(
					CaptureChannelIndex		inCaptureChannelIndex1,				// one based capture channel
					CaptureModeFlags		inMode,
					VideoCaptureInputType	inVideoInputSelect,
					AudioCaptureInputType	inAudioInputSelect,
					CaptureSize				inCaptureSize,
					Boolean					inUseCustomSize,
					IzReal					inTop,				// 0 - 100 %
					IzReal					inLeft,				// 0 - 100 %
					IzReal					inWidth,			// 0 - 100 %
					IzReal					inHeight,			// 0 - 100 %
					SInt32					inDstWidthPixels,
					SInt32					inDstHeightPixels);
	
// -----------------------------------------------------------------------------
// StartRecordCaptureInputToDisk_
// -----------------------------------------------------------------------------
// Starts recording video and/or audio to disk via the capture mechanism
//
// Before calling this function, you must have already enabled capturing for
// the specified channel by calling SetCaptureMode_
//
// inDestinationMovieIndex1 is the one based index where the new movie
// will be stored in the media window. An item must already exist in that
// location, or it must be the next available slot for this function to
// succeed. (i.e., if you specifiy index 5, there must already be an item
// in that index or you must have 4 items already in the list.)
//

#define	StartRecordCaptureInputToDisk_(izzyParamsPtr, actorInfo, inCaptureChannelIndex1, inDestinationMovieIndex1, inFileNameString) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mStartRecordCaptureInputToDiskProc)((actorInfo)->mThing, (inCaptureChannelIndex1), (inDestinationMovieIndex1), (inFileNameString), TRUE)

#define	StartRecordCaptureInputToDiskReplacing_(izzyParamsPtr, actorInfo, inCaptureChannelIndex1, inDestinationMovieIndex1, inFileNameString, inReplaceExisting) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mStartRecordCaptureInputToDiskProc)((actorInfo)->mThing, (inCaptureChannelIndex1), (inDestinationMovieIndex1), (inFileNameString), (inReplaceExisting))

typedef void	(*StartRecordCaptureInputToDiskProc)(
					ThingPtr				inThingPtr,
					CaptureChannelIndex		inCaptureChannelIndex,
					MediaIndexT				inDestinationMovieIndex,
					const char*				inFileNameString,			// can be NULL or empty string
					Boolean					inReplaceExistingFile);		// only valid if inFileNameString is not NULL or empty string
	
// -----------------------------------------------------------------------------
// StopRecordCaptureInputToDisk_
// -----------------------------------------------------------------------------
// Starts recording video and/or audio to disk via the capture mechanism
//
// Before calling this function, you must have already enabled capturing for
// the specified channel by calling SetCaptureMode_
//
// inDestinationMovieIndex1 is the one based index where the new movie
// will be stored in the media window. An item must already exist in that
// location, or it must be the next available slot for this function to
// succeed. (i.e., if you specifiy index 5, there must already be an item
// in that index or you must have 4 items already in the list.)
//

#define	StopRecordCaptureInputToDisk_(izzyParamsPtr, actorInfo, inCaptureChannelIndex1) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mStopRecordCaptureInputToDiskProc)((actorInfo)->mThing, (inCaptureChannelIndex1))

typedef void	(*StopRecordCaptureInputToDiskProc)(
					ThingPtr				inThing,
					CaptureChannelIndex		inCaptureChannelIndex1);
	
// -----------------------------------------------------------------------------
// StartRecordStageToDisk_
// -----------------------------------------------------------------------------
// Starts recording video from the specified stage to disk. If the stage is not
// visible when this function is called, no action will be taken.
//
// inDestinationMovieIndex1 is the one based index where the new movie
// will be stored in the media window. An item must already exist in that
// location, or it must be the next available slot for this function to
// succeed. (i.e., if you specifiy index 5, there must already be an item
// in that index or you must have 4 items already in the list.)
//

#define	StartRecordStageToDisk_(izzyParamsPtr, actorInfo, inStageIndex1, inDestinationMovieIndex1, inFileNameString, TRUE) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mStartRecordStageToDiskProc)((actorInfo)->mThing, (inStageIndex1), (inDestinationMovieIndex1), (inFileNameString))

#define	StartRecordStageToDiskReplacing_(izzyParamsPtr, actorInfo, inStageIndex1, inDestinationMovieIndex1, inFileNameString, inReplaceExistingFile) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mStartRecordStageToDiskProc)((actorInfo)->mThing, (inStageIndex1), (inDestinationMovieIndex1), (inFileNameString), (inReplaceExistingFile))

typedef void	(*StartRecordStageToDiskProc)(
					ThingPtr				inThingPtr,
					StageIndexT				inCaptureChannelIndex,
					MediaIndexT				inDestinationMovieIndex,
					const char*				inFileNameString,			// can be NULL or empty string
					Boolean					inReplaceExistingFile);		// only valid if inFileNameString is not NULL or empty string
	
// -----------------------------------------------------------------------------
// StopRecordStageToDisk_
// -----------------------------------------------------------------------------
// Starts recording video and/or audio to disk via the capture mechanism
//
// Before calling this function, you must have already enabled capturing for
// the specified channel by calling SetCaptureMode_
//
// inDestinationMovieIndex1 is the one based index where the new movie
// will be stored in the media window. An item must already exist in that
// location, or it must be the next available slot for this function to
// succeed. (i.e., if you specifiy index 5, there must already be an item
// in that index or you must have 4 items already in the list.)
//

#define	StopRecordStageToDisk_(izzyParamsPtr, actorInfo, inCaptureChannelIndex1) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mStopRecordStageToDiskProc)((actorInfo)->mThing, (inCaptureChannelIndex1))

typedef void	(*StopRecordStageToDiskProc)(
					ThingPtr				inThing,
					StageIndexT				inCaptureChannelIndex1);
	
// -----------------------------------------------------------------------------
// CaptureStageToPictureFile_
// -----------------------------------------------------------------------------
// Captures the current image on the specified stage and writes it to a still
// picture file. The valid file format types are given in the CapturePictureFormat
// enumeration below
//

typedef UInt8	CapturePictureFormat;

enum {
	kCapturePictureFormat_Invalid,
	kCapturePictureFormat_JPEG,
	kCapturePictureFormat_PNG,
	kCapturePictureFormat_BMP,
	kCapturePictureFormat_TIFF,
	kCapturePictureFormat_GIF
};

#define	CaptureStageToPictureFile_(izzyParamsPtr, actorInfo, inFormat, inStageIndex1, inDestinationPictureIndex1, inFileNameString, inReplaceExistingFile, inPreserveAlpha) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mCaptureStageToPictureFileProc)((actorInfo)->mThing, (inFormat), (inStageIndex1), (inDestinationPictureIndex1), (inFileNameString), (inReplaceExistingFile), (inPreserveAlpha))

typedef void	(*CaptureStageToPictureFileProc)(
					ThingPtr				inThingPtr,
					CapturePictureFormat	inCapturePictureFormat,
					StageIndexT				inCaptureChannelIndex,
					MediaIndexT				inDestinationPictureIndex1,
					const char*				inFileNameString,			// can be NULL or empty string
					Boolean					inReplaceExistingFile,		// only valid if inFileNameString is not NULL or empty string
					Boolean					inPreserveAlpha);			// if true preserve the alpha channel; otherwise render on black
	
// -----------------------------------------------------------------------------
// EncodeString
// -----------------------------------------------------------------------------
// Takes a 32 bit number and encodes it in a particular manner - so that 
// plugins can check to see if they are running in the isadora environment
//

#define	EncodeString_(izzyParamsPtr, inStr) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mEncodeStringProc)(inStr)

typedef UInt32	(*EncodeStringProc)(
					const char*		inStr);
	
// -----------------------------------------------------------------------------
// GetUserName
// -----------------------------------------------------------------------------
// Returns the name of the user of this copy of Isadora.
//

#define	GetUserName_(izzyParamsPtr, outStr, inStrMaxLen) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetUserNameProc)((outStr), (inStrMaxLen))

typedef void	(*GetUserNameProc)(
					char*			outStr,
					UInt32			inStrMaxLen);
	
// -----------------------------------------------------------------------------
// BroadcastValue_
// -----------------------------------------------------------------------------
// Broadcasts the a pointer to a Value struct on the specified channel.

#define	BroadcastValue_(izzyParamsPtr, inActorInfo, inOverLan, inMachineID, inChannel, inValuePtr) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mBroadcastValueProc)((inActorInfo)->mThing, false, 0, (inChannel), (inValuePtr))

typedef void	(*BroadcastValueProc)(
					ThingPtr		inThingPtr,
					Boolean			inOverLAN,
					UInt32			inMachineID,
					UInt32			inChannel,
					const ValuePtr	inValuePtr);
	
// -----------------------------------------------------------------------------
// BroadcastValueFromControl_
// -----------------------------------------------------------------------------
// Broadcasts the a pointer to a Value struct on the specified channel.

#define	BroadcastValueFromControl_(izzyParamsPtr, inControlInfo, inOverLan, inMachineID, inChannel, inValuePtr) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mBroadcastValueFromControlProc)((inControlInfo), (inOverLan), (inMachineID), (inChannel), (inValuePtr))

typedef void	(*BroadcastValueFromControlProc)(
					ControlInfo*	inControlInfo,
					Boolean			inOverLAN,
					UInt32			inMachineID,
					UInt32			inChannel,
					const ValuePtr	inValuePtr);
	
// -----------------------------------------------------------------------------
// BroadcastValueOverLAN_
// -----------------------------------------------------------------------------
// Broadcasts the a pointer to a Value struct on the specified channel.

#define	BroadcastValueOverLAN_(izzyParamsPtr, inActorInfo, inMachineID, inChannel, inValuePtr) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mBroadcastValueProc)((inActorInfo)->mThing, true, (inMachineID), (inChannel), (inValuePtr))

// -----------------------------------------------------------------------------
// IPAddressStringToInetAddress
// -----------------------------------------------------------------------------
// Converts the specified IP Address string to a series of one-byte values
// contained within a four-byte integer.

#define	IPAddressStringToInetAddress_(izzyParamsPtr, inIPAddr, outInetAddr) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIPAddressStringToInetAddressProc)((inIPAddr), (outInetAddr))

typedef Boolean	(*IPAddressStringToInetAddressProc)(
					const char*		inIPAddress,
					UInt32*			outInetAddress);
	
// -----------------------------------------------------------------------------
// SendOSCValues_
// -----------------------------------------------------------------------------
// Sends a multiple integers or floats, as contained in an array of Value structs
// via Open Sound Control

#define	SendOSCValues_(izzyParamsPtr, inIPAddr, inPort, inOSCAddr, inSendType, inValueCount, inValueArray) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSendOSCValuesProc)((inIPAddr), (inPort), (inOSCAddr), (inSendType), (inValueCount), (inValueArray))

typedef void	(*SendOSCValuesProc)(
					UInt32			inIPAddr,
					UInt16			inPort,
					const char*		inOSCAddr,
					Boolean			inSendType,
					SInt32			inValueCount,
					const Value*	inValueArray);
	
// -----------------------------------------------------------------------------
// SendArbitraryOSCData_
// -----------------------------------------------------------------------------
// EXPERIMENTAL: this function allows you to send non-recognized OSC data types
// other applications need to know the type tag passed in the inTypeTag parameter
// before they will recognize OSC messages sent with this function
// 
// Note that this function always sends a type tag, as specified by the inTypeTag

typedef struct {
	char	mTypeTag;
	UInt32	mDataLen;
	void*	mData;
} ArbOSCData;
	
#define	SendArbitraryOSCData_(izzyParamsPtr, inIPAddr, inPort, inOSCAddr, inArbOSCCount, inArbOSCItems) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSendArbitraryOSCDataProc)((inIPAddr), (inPort), (inOSCAddr), (inArbOSCCount), (inArbOSCItems))

typedef void	(*SendArbitraryOSCDataProc)(
					UInt32				inIPAddr,
					UInt16				inPort,
					const char*			inOSCAddr,
					UInt32				inArbOSCCount,
					const ArbOSCData*	inArbOSCItems);
	
// -----------------------------------------------------------------------------
// GetOSCChanelAddress_
// -----------------------------------------------------------------------------
// Returns the OSC address for the specified port. If the the specified port
// number has no OSC adddress, the result is kIzzyErr_OSCAddrNotFound. If the OSC
// address is found but it is longer than inOSCAddrBufferLen, the result is
// kIzzyErr_OSCAddrPartialName. Otherwise, if the full address was returned, then
// the result is kIzzyNoError.

#define	GetOSCChanelAddress_(izzyParamsPtr, inActorInfo, inPort, inOSCAddrBufferLen, inOSCAddrBuffer, inAddEllipsesIfTooLong) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetOSCChanelAddressProc)((inActorInfo)->mThing, (inPort), (inOSCAddrBufferLen), (inOSCAddrBuffer), (inAddEllipsesIfTooLong))

typedef IzzyError	(*GetOSCChanelAddressProc)(
						ThingPtr		inThingPtr,
						UInt16			inPort,
						UInt32			inOSCAddrBufferLen,
						char*			inOSCAddrBuffer,
						bool			inAddEllipsesIfTooLong);
	
// -----------------------------------------------------------------------------
// OSCAddressReceiver
// -----------------------------------------------------------------------------
// 3.0.9b01
// Opaque reference to the an OSC receiver created with CreateOSCAddressReceiver_

typedef struct OpaqueOSCAddressReceiver*	OSCAddressReceiver;

// -----------------------------------------------------------------------------
// IzzyOSCCallback
// -----------------------------------------------------------------------------
// 3.0.9b01
// This callback is used by an OpaqueOSCAddressReceiver created with the
// CreateOSCAddressReceiver_() function. When an OSC message matching the
// pattern specified by the OSC address given to that function, this callback
// will be called from a background thread with the fully formed OSC address
// that triggered the callback and the data associated with that message.

typedef void (*IzzyOSCCallback)(
				void*					inRefCon,				// user reference constant specified when the receiver was created
				const char*				inOSCAddress,			// the full OSC addressed that was received with all wildcards filled in
			    uint64_t				inTimeStamp,			// the timestamp for the OSC message
				size_t					inDataLength,			// the number of bytes of data for this message, including type tags if present
				const void*				inArgumentData);		// the data for this message, including type tags if present

// -----------------------------------------------------------------------------
// CreateOSCAddressReceiver
// -----------------------------------------------------------------------------
// 3.0.9b01
// Creates a receiver for the specified OSC address. When on OSC message is
// received on the specified address, the IzzyOSCCallback supplied by the caller
// wil be called, giving information about the received message as well as the
// user reference constant supplied when the receiver was created.

#define	CreateOSCAddressReceiver_(izzyParamsPtr, inActorInfo, inOSCAddress, inIzzyOSCCallback, inRefCon, outOSCAddrRecv) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mCreateOSCAddressReceiverProc)((inActorInfo)->mThing, (inOSCAddress), (inIzzyOSCCallback), (inRefCon), (outOSCAddrRecv))

typedef IzzyError	(*CreateOSCAddressReceiverProc)(
						ThingPtr			inThingPtr,
						const char*			inOSCAddress,
						IzzyOSCCallback		inIzzyOSCCallback,
						void*				inRefCon,
						OSCAddressReceiver*	outOSCAddressReceiver);
	
// -----------------------------------------------------------------------------
// DisposeOSCAddressReceiver
// -----------------------------------------------------------------------------
// 3.0.9b01
// Disposes an OSC receiver created using the CreateOSCAddressReceiverProc
// callback

#define	DisposeOSCAddressReceiver_(izzyParamsPtr, inActorInfo, inOSCAddressReceiver) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mDisposeOSCAddressReceiverProc)((inActorInfo)->mThing, (inOSCAddressReceiver))

typedef IzzyError	(*DisposeOSCAddressReceiverProc)(
						ThingPtr			inThingPtr,
						OSCAddressReceiver	inOSCAddressReceiver);
	
// -----------------------------------------------------------------------------
// SendLanBoxValues_
// -----------------------------------------------------------------------------
// Sends multiple channels of data to a LanBox LC at the specified UDP address
// and port

#define	SendLanBoxValues_(izzyParamsPtr, inIPAddr, inPort, inMsgType, inTargetBufferID, inChannelOffset, inChannelCount, inChannelArray) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSendLanBoxProc)((inIPAddr), (inPort), (inMsgType), (inTargetBufferID), (inChannelOffset), (inChannelCount), (inChannelArray))

typedef void	(*SendLanBoxProc)(
					UInt32			inIPAddr,
					UInt16			inPort,
					UInt16			inMsgType,
					UInt16			inTargetBufferID,
					UInt16			inChannelOffset,
					SInt32			inChannelCount,
					const UInt8*	inChannelArray);
	
// -----------------------------------------------------------------------------
// BroadcastStreamingData
// -----------------------------------------------------------------------------
// Broadcasts streaming data of the specified type to the specified port.

#define	BroadcastStreamingData_(izzyParamsPtr, inStreamingDataTypeAndPort, inData, inDataLength) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mBroadcastStreamingDataProc)((inStreamingDataTypeAndPort), (inData), (inDataLength))

typedef void	(*BroadcastStreamingDataProc)(
					UInt32			inStreamingDataTypeAndPort,
					const void*		inData,
					UInt32			inDataLen);
	
// -----------------------------------------------------------------------------
// StartRecordingLanBoxDMXInput
// StopRecordingLanBoxDMXInput
// -----------------------------------------------------------------------------
// Starts or stops recording the LanBox DMX Input channels

#define	StartRecordingLanBoxDMXInput(izzyParamsPtr, inAddress, inPort, inBufferID, inPassword, inChannelOffset, inChannelCount, inFileName) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mStartStopRecordingLanBoxDMXInputProc)(TRUE, (inAddress), (inPort), (inBufferID), (inPassword), (inChannelOffset), (inChannelCount), (inFileName))

#define	StopRecordingLanBoxDMXInput(izzyParamsPtr) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mStartStopRecordingLanBoxDMXInputProc)(FALSE, NULL, 0, 0, NULL, 0, 0, NULL)

typedef void	(*StartStopRecordingLanBoxDMXInputProc)(
					Boolean			inRecord,
					const char*		inServerAddress,
					UInt16			inServerPort,
					UInt16			inBufferID,
					const char*		inPassword,
					UInt32			inChannelOffset,
					SInt32			inChannelCount,
					const char*		inFileName);
	
// -----------------------------------------------------------------------------
// EnableDisableMIDI_
// -----------------------------------------------------------------------------
// Enables or disables MIDI input and output for the specified port.

#define	EnableDisableMIDI_(izzyParamsPtr, inActorInfo, inEnable, inPort) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mEnableDisableMIDIProc)((inActorInfo)->mThing, (inEnable), (inPort))

typedef void	(*EnableDisableMIDIProc)(
					ThingPtr		inThingPtr,
					Boolean			inEnable,
					PortIndex		inPortIndex1);		// 0 = all ports, > 0 = enable/disable the specified port
	
// -----------------------------------------------------------------------------
// ShowNumericValueEditBox_
// -----------------------------------------------------------------------------
// Shows a value edit box that allows the user to edit a number

typedef void (*NumericValueChangedProc)(IsadoraParameters* ip, ValuePtr inNewValue, IzRefCon inRefCon);

#define	ShowNumericValueEditBox_(izzyParamsPtr, inCtlInfo, inDispRect, inCallback, inRefCon, inMin, inMax, inValue, inAlignment, inMaxDecimalPoints) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mShowNumericValueEditBoxProc)((inCtlInfo), (inDispRect), (inCallback), (inRefCon), (inMin), (inMax), (inValue), (inAlignment), (inMaxDecimalPoints))

typedef void	(*ShowNumericValueEditBoxProc)(
					ControlInfo* 				inControlInfo,
					const IzRect*				inDisplayRect,
					NumericValueChangedProc		inProc,
					IzRefCon					inRefCon,
					ValuePtr					inMin,
					ValuePtr					inMax,
					ValuePtr					inValue,
					SInt32						inAlignment, 			// -1 = left, 0 = center, +1 = right
					SInt32						inMaxDecimalPoints);	// number of decimals for floats; non-zero to show timecode rate for timecode
	
// -----------------------------------------------------------------------------
// DrawPascalString_
// -----------------------------------------------------------------------------
//	Draws a pascal string into the current graphics port.
//
//	Because of issues with QuickTime under windows, you should use this function
//	to draw text in your actor's Actor Defined Area or in your Custom Control.

/*
DEPRECATED! Use DrawTextAt_
#define	DrawPascalString_(izzyParamsPtr, inPString) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mDrawPascalStringProc)((inPString))

typedef void	(*DrawPascalStringProc)(
					StringPtr			inString);
*/

// -----------------------------------------------------------------------------
// DrawCString_
// -----------------------------------------------------------------------------
//	Draws a c-string into the current graphics port
//
//	Because of issues with QuickTime under windows, you should use this function
//	to draw text in your actor's Actor Defined Area or in your Custom Control.

/*
DEPRECATED! Use DrawTextAt_
#define	DrawCString_(izzyParamsPtr, inCString) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mDrawCStringProc)((inCString))

typedef void	(*DrawCStringProc)(
					const char*			inString);
*/

// -----------------------------------------------------------------------------
// DrawText_
// -----------------------------------------------------------------------------
//	Draws the specified text into the current graphics port
//
//	Because of issues with QuickTime under windows, you should use this function
//	to draw text in your actor's Actor Defined Area or in your Custom Control.

/*
DEPRECATED! Use DrawTextAt_
#define	DrawText_(izzyParamsPtr, inCString, inLength) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mDrawTextProc)((inCString), (inLength))

typedef void	(*DrawTextProc)(
					const char*			inString,
					SInt32				inLength);
*/

// -----------------------------------------------------------------------------
// TextWidthPascalString_
// -----------------------------------------------------------------------------
//	Returns the width in pixels of a pascal string for the current graphics port.
//
//	Because of issues with QuickTime under Windows, you should use this function
//	to measure text in your actor's Actor Defined Area or in your Custom Control.

/*

DEPRECATED! Use GetTextWidth_

#define	TextWidthPascalString_(izzyParamsPtr, inPString) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mTextWidthPascalStringProc)((inPString))

typedef SInt16	(*TextWidthPascalStringProc)(
					StringPtr			inString);
*/

// -----------------------------------------------------------------------------
// TextWidthCString_
// -----------------------------------------------------------------------------
//	Returns the width in pixels of a c string for the current graphics port.
//
//	Because of issues with QuickTime under windows, you should use this function
//	to draw text in your actor's Actor Defined Area or in your Custom Control.

/*

DEPRECATED! Use GetTextWidth_

#define	TextWidthCString_(izzyParamsPtr, inCString) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mTextWidthCStringProc)((inCString))

typedef SInt16	(*TextWidthCStringProc)(
					const char*			inString);
#endif
*/

// -----------------------------------------------------------------------------
// TextWidth_ (REMOVED for v3.0)
// -----------------------------------------------------------------------------
//	Returns the width in pixels of a a block of text for the current graphics port.
//
//	Because of issues with QuickTime under windows, you should use this function
//	to draw text in your actor's Actor Defined Area or in your Custom Control.

/*

DEPRECATED! Use GetTextWidth_

#define	TextWidth_(izzyParamsPtr, inCString, inLength) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mTextWidthProc)((inCString), (inLength))

typedef SInt16	(*TextWidthProc)(
					const char*			inString,
					SInt32				inLength);
*/

// -----------------------------------------------------------------------------
// GetMediaObjectProperties
// -----------------------------------------------------------------------------
// Returns the full path and file name for the specified media file
// in UTF-8 format.

#define	GetMediaObjectProperties_(izzyParamsPtr, inActorInfo, inMediaType, inMediaIndex1, outMediaPlayHints, inPathBufLen, outPathBuf, inNameBufLen, outNameBuf) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetMediaObjectPropertiesProc)((inActorInfo), (inMediaType), (inMediaIndex1), (outMediaPlayHints), (inPathBufLen), (outPathBuf), (inNameBufLen), (outNameBuf))

typedef Boolean	(*GetMediaObjectPropertiesProc)(
					ActorInfo*			inActorInfo,
					MediaCode			inMediaType,
					MediaIndexT			inMediaIndex1,
					IzzyMediaPlayHints*	outIzzyMediaPlayHints,	// can be NULL
					size_t				inPathNameBufferLen,	// can be 0 if path buffer is NULL
					char*				outPathNameBuffer,		// can be NULL
					size_t				inFileNameBufferLen,	// can be 0 if name buffer is NULL
					char*				outFileNameBuffer);		// can be NULL
	
// -----------------------------------------------------------------------------
// MediaObjectIndexToUniqueID
// -----------------------------------------------------------------------------
// Given a media type and a media index, return the unique ID for the
// specified media object

#define	MediaObjectIndexToUniqueID(izzyParamsPtr, inActorInfo, inMediaType, inMediaIndex1, outUniqueID) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mMediaObjectIndexToUniqueIDProc)((inActorInfo), (inMediaType), (inMediaIndex1), (outUniqueID))

typedef UInt32	(*MediaObjectIndexToUniqueIDProc)(
					ActorInfo*		inActorInfo,
					MediaCode		inMediaType,
					UInt32			inMediaIndex1);
	
// -----------------------------------------------------------------------------
// UniqueIDToMediaObjectIndex
// -----------------------------------------------------------------------------
// Given a media type and a media index, return the unique ID for the
// specified media object

#define	UniqueIDToMediaObjectIndex(izzyParamsPtr, inActorInfo, inMediaType, inUniqueID, outMediaIndex1) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mUniqueIDToMediaObjectIndexProc)((inActorInfo), (inMediaType), (inUniqueID), (outMediaIndex1))

typedef UInt32	(*UniqueIDToMediaObjectIndexProc)(
					ActorInfo*		inActorInfo,
					MediaCode		inMediaType,
					UInt32			inUniqueID);
	
// -----------------------------------------------------------------------------
// GetBinNameAndUniqueIDFromIndex_
// -----------------------------------------------------------------------------
// Returns the name of the bin at the specified one based index. If no bin
// exists at that location, the result is false

#define	GetBinNameAndUniqueIDFromIndex_(izzyParamsPtr, inControlInfo, inMediaType, inBinIndex1, outBinName, outBinUniqueID) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetBinNameAndUniqueIDFromIndexProc)((inControlInfo), (inMediaType), (inBinIndex1), (outBinName), (outBinUniqueID))

typedef Boolean	(*GetBinNameAndUniqueIDFromIndexProc)(
					ControlInfo* 	inControlInfo,
					MediaCode		inMediaType,
					UInt32			inBinIndex1,		// one-based index
					char*			outBinName,			// buffer should be at least 256 characters
					UInt32*			outBinUniqueID);
	
// -----------------------------------------------------------------------------
// GetBinNameAndTypeFromUniqueID_
// -----------------------------------------------------------------------------
// Returns the name of the bin with the specified unique ID

#define	GetBinNameAndTypeFromUniqueID_(izzyParamsPtr, inControlInfo, inBinUniqueID, outBinName, outMediaType) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetBinNameAndTypeFromUniqueIDProc)((inControlInfo), (inBinUniqueID), (outBinName), (outMediaType))

typedef Boolean	(*GetBinNameAndTypeFromUniqueIDProc)(
					ControlInfo* 	inControlInfo,
					UInt32			inBinUniqueID,		// one-based index
					char*			outBinName,			// buffer should be at least 256 characters
					MediaCode*		outMediaType);
	
// -----------------------------------------------------------------------------
// CountMediaObjectsInBin_
// -----------------------------------------------------------------------------
// Returns the number of objects within a named bin in the Media Window.
//
// The function searches the Media Window for a bin that matches the specified
// name and media type. If one is found, the result is the number of objects
// in that bin. If the bin can't be found, the result is 0.

#define	CountMediaObjectsInBin_(izzyParamsPtr, inControlInfo, inBinUniqueID) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mCountMediaObjectsInBinProc)((inControlInfo), (inBinUniqueID))

typedef SInt32	(*CountMediaObjectsInBinProc)(
					ControlInfo* 	inControlInfo,
					UInt32			inBinUniqueID);
	
// -----------------------------------------------------------------------------
// GetInfoForIndexedBin_
// -----------------------------------------------------------------------------
// Returns the name, unique ID, and media object count for a specified bin

#define	GetInfoForIndexedBin_(izzyParamsPtr, inActorInfo, inMediaType, inBinIndex1, outBinName, outBinUniqueID, outMediaObjectCount) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetInfoForIndexedBinProc)((inActorInfo), (inMediaType), (inBinIndex1), (outBinName), (outBinUniqueID), (outMediaObjectCount))

typedef Boolean	(*GetInfoForIndexedBinProc)(
					ActorInfo* 		inActorInfo,
					MediaCode		inMediaType,
					UInt32			inBinIndex1,		// one-based index
					char*			outBinName,			// buffer should be at least 256 characters
					UInt32*			outBinUniqueID,
					UInt32*			outMediaObjectCount);
	
// -----------------------------------------------------------------------------
// GetMediaObjectInfoFromBin_
// -----------------------------------------------------------------------------
// Gets information about a media object within a specified bin.
//
// If a bin with the specified unique ID exists, and if a media object
// at the specified index exists within that bin, then the result is true and
// the media object index, name and thumbnail for the requested object are
// returned.
//
// If the bin or media object do not exist, the result is false.

#define	GetMediaObjectInfoFromBin_(izzyParamsPtr, inControlInfo, inBinUniqueID, inObjectIndexWithinBin1, outMediaObjectIndex1, outMediaName, outThumbnail) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetMediaObjectInfoFromBinProc)((inControlInfo), (inBinUniqueID), (inObjectIndexWithinBin1), (outMediaObjectIndex1), (outMediaName), (outThumbnail))

typedef Boolean	(*GetMediaObjectInfoFromBinProc)(
					ControlInfo* 	inControlInfo,
					UInt32			inBinUniqueID,				// the unique ID of the bin
					UInt32			inObjectIndexWithinBin1,	// one based index of item within its bin
					UInt32*			outMediaObjectIndex1,		// the media object index for this item
					char*			outName,					// buffer to hold name should be 256 characters
					ImageBufferPtr*	outThumbnail);				// the object's thumbnail if it exists - otherwise outThumbnail is set to nil
	
// -----------------------------------------------------------------------------
// GetMediaIndexForItemInBin
// -----------------------------------------------------------------------------
// Returns the absolute media index of the specified media object within the
// specified bin

#define	GetMediaIndexForItemInBin_(izzyParamsPtr, inActorInfo, inMediaType, inBinIndex1, inMediaIndex1, outMediaIndex1) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetMediaIndexForItemInBinProc)((inActorInfo), (inMediaType), (inBinIndex1), (inMediaIndex1), (outMediaIndex1))

typedef Boolean	(*GetMediaIndexForItemInBinProc)(
					ActorInfo* 		inActorInfo,
					MediaCode		inMediaType,
					UInt32			inBinIndex1,		// one-based index
					UInt32			inMediaIndex1,
					UInt32*			outMediaIndex1);
	
// -----------------------------------------------------------------------------
// ConvertTimeToPercentForMediaObject
// -----------------------------------------------------------------------------
// Receives an array of millisecond values, and converts them to a percentage
// of the objects duration.

#define	ConvertTimeToPercentForMediaObject_(izzyParamsPtr, inActorInfo, inMediaType, inBinIndex1, inMediaIndex1, inTimeArraySize, inMillisecondTimeArray, outPercentages) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mConvertTimeToPercentForMediaObject)((inActorInfo), (inMediaType), (inBinIndex1), (inMediaIndex1), (inTimeArraySize), (inMillisecondTimeArray), (outPercentages))

typedef Boolean	(*ConvertTimeToPercentForMediaObjectProc)(
					ActorInfo* 		inActorInfo,
					MediaCode		inMediaType,
					UInt32			inBinIndex1,			// one-based index: 0 = use inMediaIndex only - ignore bin
					UInt32			inMediaIndex1,			// one-based index of the media within the bin, or absolute index if inBinIndex1 is 0
					UInt32			inTimeArraySize,		// size of array of mS times
					const UInt32*	inMillisecondTimeArray,	// array of mS times
					IzReal*			outPercentages);		// array of percentages from 0 to 100%
	
// -----------------------------------------------------------------------------
// ConvertPercentToTimeForMediaObject
// -----------------------------------------------------------------------------
// Receives an array of millisecond values, and converts them to a percentage
// of the objects duration.

#define	ConvertPercentToTimeForMediaObject_(izzyParamsPtr, inActorInfo, inMediaType, inBinIndex1, inMediaIndex1, inTimeArraySize, inPercentages, outMillisecondTimeArray) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mConvertPercentToTimeForMediaObject)((inActorInfo), (inMediaType), (inBinIndex1), (inMediaIndex1), (inTimeArraySize), (inPercentages), (outMillisecondTimeArray))

typedef Boolean	(*ConvertPercentToTimeForMediaObjectProc)(
					ActorInfo* 		inActorInfo,
					MediaCode		inMediaType,
					UInt32			inBinIndex1,			// one-based index: 0 = use inMediaIndex only - ignore bin
					UInt32			inMediaIndex1,			// one-based index of the media within the bin, or absolute index if inBinIndex1 is 0
					UInt32			inTimeArraySize,		// size of array of mS times
					const IzReal*	outPercentages,			// array of percentages from 0 to 100%
					UInt32*			inMillisecondTimeArray);// array of mS times
	
// -----------------------------------------------------------------------------
// EnableDisableActors_
// -----------------------------------------------------------------------------
// Gets information about a media object within a specified bin.
//
// If a bin with the specified unique ID exists, and if a media object
// at the specified index exists within that bin, then the result is true and
// the media object index, name and thumbnail for the requested object are
// returned.
//
// If the bin or media object do not exist, the result is false.

#define	EnableDisableActors_(izzyParamsPtr, inActorInfo, inEnable) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mEnableDisableActorsProc)((inActorInfo), (inEnable))

typedef void	(*EnableDisableActorsProc)(
					ActorInfo* 		inActorInfo,
					Boolean			inEnable);
	
// -----------------------------------------------------------------------------
// OutputTextFile_
// -----------------------------------------------------------------------------
//	Writes the data pointed to by dataPtr to a text file. If you want the file
//	to appear on the desktop, then pass a file name in 'fileName'. Otherwise,
//	pass nil for fileName, and a standard dialog will appear allowing the user
//	to choose the save location
//
#define	OutputTextFile_(izzyParamsPtr, fileName, dataPtr, dataLen, deleteExisting, appendDate, appendTime) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mOutputTextFileProc)((fileName), (dataPtr), (dataLen), (deleteExisting), (appendDate), (appendTime))

typedef Boolean	(*OutputTextFileProc)(
					const char*		inFileName,
					const void*		inDataPtr,
					SInt32			inLen,
					Boolean			inDeleteExistingFile,
					Boolean			inAppendDate,
					Boolean			inAppendTime);
	
// -----------------------------------------------------------------------------
// WriteToLogFile_
// -----------------------------------------------------------------------------
//	Writes the c-string to the log file associated with the document that 
//	owns the actor making this request. The log file will be writtent to
//	the same folder as the document.
//
#define	WriteToLogFile_(izzyParamsPtr, inActorInfo, inEventID, inCString) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mWriteToLogFileProc)(((inActorInfo)->mThing), (inEventID), (inCString))

typedef void	(*WriteToLogFileProc)(
					ThingPtr						inThingPtr,
					SInt32							inEventID,
					const char*						inCString);
	
// -----------------------------------------------------------------------------
// WriteToSerialPort_
// -----------------------------------------------------------------------------
//	Write data to the serial port

#define	WriteToSerialPort_(izzyParamsPtr, inActorInfo, inPortIndex1, inDataLen, inData) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mWriteToSerialPortProc)((inActorInfo), (inPortIndex1), (inDataLen), (inData))

typedef UInt32	(*WriteToSerialPortProc)(				// returns bytes written
					ActorInfo*			inActorInfo,
					SerialStreamIndexT	inPortIndex1,	// one based serial port index
					UInt32				inDataLen,		// length of data in bytes
					const void*			inData);		// pointer to the data
	
// -----------------------------------------------------------------------------
// ReadFromSerialPort_
// -----------------------------------------------------------------------------
//	Read data from the serial port

#define	ReadFromSerialPort_(izzyParamsPtr, inActorInfo, inPortIndex1, inDataLen, inData) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mReadFromSerialPortProc)((inActorInfo), (inPortIndex1), (inDataLen), (inData))

typedef UInt32	(*ReadFromSerialPortProc)(				// returns bytes read
					ActorInfo*			inActorInfo,
					SerialStreamIndexT	inPortIndex1,	// one based serial port index
					UInt32				inDataLen,		// size of data buffer to receive data
					void*				inData);		// pointer to the data buffer
	
// -----------------------------------------------------------------------------
// SetTCPStreamInfo_
// -----------------------------------------------------------------------------
//	Sets the TCP address, port, and buffer size for the specified TCP stream

#define	SetTCPStreamInfo_(izzyParamsPtr, inActorInfo, inStreamIndex1, inTCPAddressStr, inTCPPort, inBufferSize) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetTCPStreamInfoProc)((inActorInfo), (inStreamIndex1), (inTCPAddressStr), (inTCPPort), (inBufferSize))

typedef Boolean	(*SetTCPStreamInfoProc)(					// returns TRUE if successful
					ActorInfo*		inActorInfo,
					TCPStreamIndexT	inStreamIndex1,			// one based TCP stream Index
					const char*		inTCPAddressStr,		// tcp address (string)
					UInt16			inTCPPort,				// tcp port number
					UInt32			inBufferSize);			// tcp buffer size
	
// -----------------------------------------------------------------------------
// EnableDisableTCPStream
// -----------------------------------------------------------------------------
//	Enables or disables a TCP port; enabling opens the connection; disabling
//	closes it

#define	EnableDisableTCPStream_(izzyParamsPtr, inActorInfo, inStreamIndex1, inEnable) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mEnableDisableTCPStreamProc)((inActorInfo), (inStreamIndex1), (inEnable))

typedef Boolean	(*EnableDisableTCPStreamProc)(				// returns TRUE if successful
					ActorInfo*		inActorInfo,
					TCPStreamIndexT	inStreamIndex1,			// one based TCP stream Index
					Boolean			inEnable);				// pointer to the data
	
// -----------------------------------------------------------------------------
// GetTCPStreamStatus
// -----------------------------------------------------------------------------
//	Returns the current connection status of the TCP Stream

#define	GetTCPStreamStatus_(izzyParamsPtr, inActorInfo, inStreamIndex1, outParam1, outParam2, outParam3) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetTCPStreamStatusProc)((inActorInfo), (inStreamIndex1), (outParam1), (outParam2), (outParam3))

typedef Boolean	(*GetTCPStreamStatusProc)(						// returns TRUE if successful
					ActorInfo*			inActorInfo,
					TCPStreamIndexT		inStreamIndex1,			// one based TCP stream index
					TCPStreamStatus*	outTCPStreamStatus,		// pointer to the current TCP stream status
					void*				outFutureUse1,			// currently unused; pass NULL
					void*				outFutureUse2);			// currently unused; pass NULL

// -----------------------------------------------------------------------------
// WriteToTCPStream_
// -----------------------------------------------------------------------------
//	Write data to the TCP port

#define	WriteToTCPStream_(izzyParamsPtr, inActorInfo, inStreamIndex1, inDataLen, inData) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mWriteToTCPStreamProc)((inActorInfo), (inStreamIndex1), (inDataLen), (inData))

typedef UInt32	(*WriteToTCPStreamProc)(					// returns bytes written
					ActorInfo*		inActorInfo,
					TCPStreamIndexT	inStreamIndex1,			// one based TCP stream Index
					UInt32			inDataLen,				// length of data in bytes
					const void*		inData);				// pointer to the data
	
// -----------------------------------------------------------------------------
// ReadFromTCPStream_
// -----------------------------------------------------------------------------
//	Read data from the TCP port

#define	ReadFromTCPStream_(izzyParamsPtr, inActorInfo, inStreamIndex1, inDataLen, inData, inTimeSeconds) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mReadFromTCPStreamProc)((inActorInfo), (inStreamIndex1), (inDataLen), (inData), (inTimeSeconds))

typedef UInt32	(*ReadFromTCPStreamProc)(					// returns bytes actually read
					ActorInfo*		inActorInfo,
					TCPStreamIndexT	inStreamIndex1,			// one based TCP stream Index
					UInt32			inDataLen,				// size of data buffer to receive data
					void*			inData,					// pointer to the data buffer
					IzReal			inTimeout);				// number of seconds to wait before returning
	
// -----------------------------------------------------------------------------
// SetLocalTranslationRotationAndFOV_
// -----------------------------------------------------------------------------
//	Sets the translation and rotation for OpenGL rendering for the currently
//	active scene. This setting does not affect other scenes.

#define	SetLocalTranslationAndRotation_(izzyParamsPtr, inActorInfo, inChannel1, transX, transY, transZ, rotX, rotY, rotZ) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetLocalTranslationRotationAndFOVProc)(((inActorInfo)->mThing), (inChannel1), (transX), (transY), (transZ), (rotX), (rotY), (rotZ), 0.0)

#define	SetLocalTranslationRotationAndFOV_(izzyParamsPtr, inActorInfo, inChannel1, transX, transY, transZ, rotX, rotY, rotZ, fov) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetLocalTranslationRotationAndFOVProc)(((inActorInfo)->mThing), (inChannel1), (transX), (transY), (transZ), (rotX), (rotY), (rotZ), fov)

typedef void	(*SetLocalTranslationRotationAndFOVProc)(
					ThingPtr	inActorInfo,
					SInt32		inChannelIndex1,
					GLfloat		inTranslateX,
					GLfloat		inTranslateY,
					GLfloat		inTranslateZ,
					GLfloat		inRotateX,
					GLfloat		inRotateY,
					GLfloat		inRotateZ,
					GLfloat		inFOV);
	
// -----------------------------------------------------------------------------
// GetLocalTranslationRotationAndFOV_
// -----------------------------------------------------------------------------
//	Returns the translation and rotation for OpenGL rendering for the currently
//	active scene. See also SetLocalTranslationRotationAndFOV_()

#define	GetLocalTranslationAndRotation_(izzyParamsPtr, inActorInfo, inChannel1, transX, transY, transZ, rotX, rotY, rotZ) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetLocalTranslationRotationAndFOVProc)(((inActorInfo)->mThing), (inChannel1), (transX), (transY), (transZ), (rotX), (rotY), (rotZ), NULL)

#define	GetLocalTranslationRotationAndFOV_(izzyParamsPtr, inActorInfo, inChannel1, transX, transY, transZ, rotX, rotY, rotZ, fov) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetLocalTranslationRotationAndFOVProc)(((inActorInfo)->mThing), (inChannel1), (transX), (transY), (transZ), (rotX), (rotY), (rotZ), (fov))

typedef void	(*GetLocalTranslationRotationAndFOVProc)(
					ThingPtr	inActorInfo,
					SInt32		inChannelIndex1,
					GLfloat*	inTranslateX,
					GLfloat*	inTranslateY,
					GLfloat*	inTranslateZ,
					GLfloat*	inRotateX,
					GLfloat*	inRotateY,
					GLfloat*	inRotateZ,
					GLfloat*	inFOV);				// can be NULL
	
// -----------------------------------------------------------------------------
// SetLocalLighting_
// -----------------------------------------------------------------------------
//	Sets the 3D Lighting orientation for the scene that owns the calling actor.
//  If the inInfo parameter is NULL, then clear the local lighting for this scene

#define	SetLocalLighting_(izzyParamsPtr, inActorInfo, inChannel1, lightingInfo) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetLocalLightingProc)(((inActorInfo)->mThing), (inChannel1), (lightingInfo))

typedef void	(*SetLocalLightingProc)(
					ThingPtr							inActorInfo,
					SInt32								inChannelIndex1,
					const LightingInfo*					inInfo);				// CAN BE NULL (means 'clear local lighting')
	
// -----------------------------------------------------------------------------
// SetLocalBackgroundLighting_
// -----------------------------------------------------------------------------
//	Sets background lighting information for the scene that owns the calling actor
//  If the inInfo parameter is NULL, then clear the local lighting for this scene

#define	SetLocalBackgroundLighting_(izzyParamsPtr, inActorInfo, inChannel1, backgroundLightingInfo) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetLocalBackgroundLightingProc)(((inActorInfo)->mThing), (inChannel1), (backgroundLightingInfo))

typedef void	(*SetLocalBackgroundLightingProc)(
					ThingPtr							inActorInfo,
					SInt32								inChannelIndex1,
					const BackgroundLightingInfo*		inBackgroundLightingInfo);	// CAN BE NULL (means 'clear local background lighting')
	
// -----------------------------------------------------------------------------
// SetLocalScalingOptions
// -----------------------------------------------------------------------------
//	Sets the Stage Scaling Options for the scene that owns the calling actor.

#define	SetLocalScalingOptions_(izzyParamsPtr, inActorInfo, inChannel1, scalingOptions) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetLocalScalingOptionsProc)(((inActorInfo)->mThing), (inChannel1), (scalingOptions))

typedef void	(*SetLocalScalingOptionsProc)(
					ThingPtr						inThingPtr,
					SInt32							inChannelIndex1,
					const StageScalingOptions*		inInfo);
	
// -----------------------------------------------------------------------------
// SetLocalStageBackgroundOptions
// -----------------------------------------------------------------------------
//	Sets the Stage Background Options for the scene that owns the calling actor.

/*
#define	SetLocalStageBackgroundOptions_(izzyParamsPtr, inActorInfo, inChannel1, stageBackgroundOptions) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetLocalStageBackgroundOptionsProc)(((inActorInfo)->mThing), (inChannel1), (stageBackgroundOptions))

typedef void	(*SetLocalStageBackgroundOptionsProc)(
					ThingPtr						inThingPtr,
					SInt32							inChannelIndex1,
					const StageBackgroundOptions*	inInfo);
*/

// -----------------------------------------------------------------------------
// AllocateValueString_
// -----------------------------------------------------------------------------
//	Allocates a Value object that points to a string. It is the caller's
//	responsibiltiy to release the string using the ReleaseValueString_
//	function.

#define	AllocateValueString_(izzyParamsPtr, inString, outValue) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mAllocateValueStringProc)((inString), (outValue))

typedef void	(*AllocateValueStringProc)(	
					const char*		inString,	// pointer to a null terminated string
					ValuePtr		outValue);
	
// -----------------------------------------------------------------------------
// RetainValueString_
// -----------------------------------------------------------------------------
//	Retains a string allocated with the AllocateValueString_ function. Note that
//	you can pass any value ValuePtr to this function. If it is not a string, then
//	this function will do nothing.

#define	RetainValueString_(izzyParamsPtr, inValue) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mRetainValueStringProc)(inValue)

typedef void	(*RetainValueStringProc)(
					ValuePtr		outValue);
	
// -----------------------------------------------------------------------------
// ReleaseValueString_
// -----------------------------------------------------------------------------
//	Releases a string allocated with the AllocateValueString_ function. Note that
//	you can pass any value ValuePtr to this function. If it is not a string, then
//	this function will do nothing.

#define	ReleaseValueString_(izzyParamsPtr, inValue) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mReleaseValueStringProc)(inValue)

typedef void	(*ReleaseValueStringProc)(	
					ValuePtr		outValue);
	
// -----------------------------------------------------------------------------
// ControlAcquireImage_
// -----------------------------------------------------------------------------
//	Allows a control to acquire a media object in the media window, and returns
//	an image buffer to the bitmap associated with that piece of media. Currently
//	this should only be used for pictures.

#define	ControlAcquireImage_(izzyParamsPtr, inCtlInfo, inMediaType, inMediaIndex1, inID, outImageBuffer) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mControlAcquireImageProc)((inCtlInfo), (inMediaType), (inMediaIndex1), (inID), (outImageBuffer))

typedef void	(*ControlAcquireImageProc)(
					ControlInfo*	inCtlInfo,
					MediaCode		inMediaType,		// should always be kMediaTypePicture
					UInt32			inMediaIndex1,		// media index of the piece of media
					AcquireIDT		inID,				// unique acquire ID -- often the pointer to the control object
					ImageBufferPtr*	outImageBuffer);	// resulting image buffer with the image data
	
// -----------------------------------------------------------------------------
// ControlReleaseImage_
// -----------------------------------------------------------------------------
//	Releases the media acquired using the ControlAcquireImage_ function above

#define	ControlReleaseImage_(izzyParamsPtr, inCtlInfo, inMediaType, inMediaIndex1, inID) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mControlReleaseImageProc)((inCtlInfo), (inMediaType), (inMediaIndex1), (inID))

typedef void	(*ControlReleaseImageProc)(
					ControlInfo*	inCtlInfo,
					MediaCode		inMediaType,
					UInt32			inMediaIndex1,
					AcquireIDT		inID);
	
// -----------------------------------------------------------------------------
// GetActorDefaultPixelFormat_
// -----------------------------------------------------------------------------
//	Get the default pixel format for the specified Actor.

#define	GetActorDefaultPixelFormat_(izzyParamsPtr, inActorInfo) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetActorDefaultPixelFormatProc)((inActorInfo))

typedef OSType	(*GetActorDefaultPixelFormatProc)(
					ActorInfo*		inActorInfo);
	
// -----------------------------------------------------------------------------
// GetActorGraphics_
// -----------------------------------------------------------------------------
//	Get the Graphics Context for the specified Actor. This object is required
//	when using the graphics routine that follow

#define	GetActorGraphics_(izzyParamsPtr, inActorInfo) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetActorGraphicsProc)((inActorInfo))

typedef UGraphics*	(*GetActorGraphicsProc)(
						ActorInfo*		inActorInfo);
	
// -----------------------------------------------------------------------------
// GetControlGraphics_
// -----------------------------------------------------------------------------
//	Get the Graphics Context for the specified Control so that it can draw
//	into the current window. Callers should use GetControlGraphics_ when
//	responding to an update event via the ControlDraw callback, and use
//	GetControlGraphicsForceRefocus_ when drawing from within a Message Receiver
//	callback. (This is because the graphics context may not be focused properly
//	during such a callback.)

#define	GetControlGraphics_(izzyParamsPtr, inControlInfo) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetControlGraphicsProc)((inControlInfo), false)

#define	GetControlGraphicsForceRefocus_(izzyParamsPtr, inControlInfo) \
(*(izzyParamsPtr)->kConditionalCallbacks->mGetControlGraphicsProc)((inControlInfo), true)
	
typedef UGraphics*	(*GetControlGraphicsProc)(
						ControlInfo*		inControlInfo,
						bool				inForceFocus);

// -----------------------------------------------------------------------------
// GetGraphicsScaling
// -----------------------------------------------------------------------------
// Returns the scaling factor for the current UGraphics drawing context, which
// includes any scaling required for the current display DPI but does not take
// into account the amount the zoom factor of the Scene Editor or Control Panel
// editor.

#define	GetGraphicsScaling_(izzyParamsPtr, inUGraphics) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetGraphicsScalingProc)((inUGraphics))

typedef float	(*GetGraphicsScalingProc)(
					UGraphics*		inGraphics);

// -----------------------------------------------------------------------------
// GetActorGraphicScaling_
// -----------------------------------------------------------------------------
//	Returns the scale factor required when drawing graphics into the actor defined
//	area for a given actor. This scale factor takes into account the DPI of the
//	display on which the actor is being drawn, as well as the current zoom factor
//	of the Scene Editor

#define	GetActorGraphicScaling_(izzyParamsPtr, inActorInfo) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetActorGraphicScalingProc)((inActorInfo))

typedef float		(*GetActorGraphicScalingProc)(
						ActorInfo*		inActorInfo);

// Handy macro to scale pixels to screen if you call
// your scale factor variable gsScaleFactor

#define ActorPixelToScreen(a)	((a) * gsScaleFactor)

// Because fonts sizes are automatically scaled by the system to take display
// DPI into account, use this scale factor for the font size when drawing text
// into an Actor Defined area. It ignores the DPI of the display but respects
// the zoom factor of the Scene Editor.

#define GetActorFontScaling_(ip, inActorInfo, inUGraphics) \
	GetActorGraphicScaling_((ip), (inActorInfo)) / GetGraphicsScaling_((ip), (inUGraphics))

// -----------------------------------------------------------------------------
// GetControlGraphicScaling_
// -----------------------------------------------------------------------------
//	Returns the scale factor required when drawing graphics for a Control. This
//	scale factor takes into account the DPI of the display on which the actor
//	is being drawn, as well as the current zoom factor of the Control Panel.

#define	GetControlGraphicScaling_(izzyParamsPtr, inControlInfo) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetControlGraphicScalingProc)((inControlInfo))

// v2.6 <-

typedef float		(*GetControlGraphicScalingProc)(
						ControlInfo*	inControlinfo);

// Handy macro to scale pixels to screen if you call
// your scale factor variable gsScaleFactor

#define ControlPixelToScreen(a)	((a) * gsScaleFactor)

// Because fonts sizes are automatically scaled by the system to take display
// DPI into account, use this scale factor for the font size when drawing text
// into an Actor Defined area. It ignores the DPI of the display but respects
// the zoom factor of the Scene Editor.

#define GetControlFontScaling_(ip, inControlinfo, inUGraphics) \
	GetControlGraphicScaling_((ip), (inControlinfo)) / GetGraphicsScaling_((ip), (inUGraphics))

// -----------------------------------------------------------------------------
// SetForegroundColorRef_
// -----------------------------------------------------------------------------
//	Sets the current foreground color
#define	SetForegroundColorRef_(izzyParamsPtr, inGraphics, inColorRef) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetForegroundColorRefProc)((inGraphics), (inColorRef))

typedef void	(*SetForegroundColorRefProc)(
					UGraphics*	inGraphics,
					ColorRef	inColorRef);
	
// -----------------------------------------------------------------------------
// GetColorRefsForCurrentTheme_
// -----------------------------------------------------------------------------
//	Returns the default colors for selected and unselected text, as well as
//	the background color for the current theme

typedef struct {
	ColorRef		mEditorBackgroundColor;
	ColorRef		mGeneralBackgroundColor;
	ColorRef		mGeneralUnselectedTextColor;
	ColorRef		mGeneralSelectedTextColor;
	ColorRef		mMediaItemUnselectedBackgroundColor;
	ColorRef		mMediaItemSelectedBackgroundColor;
	ColorRef		mMediaItemTextColor;
	// v2.6->
	ColorRef		mPrimarySceneColor;
	ColorRef		mSecondarySceneColor;
	ColorRef		mInactiveSceneColor;
	ColorRef		mScrollBarBackgroundColor;
	ColorRef		mScrollBarThumbColor;
	// v2.6 <-
} ThemeColorRefs;

#define	GetColorRefsForCurrentTheme_(izzyParamsPtr, outThemeColorRefsPtr) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetColorRefsForCurrentThemeProc)((outThemeColorRefsPtr))

typedef void	(*GetColorRefsForCurrentThemeProc)(
					ThemeColorRefs*			outThemeColorRefs);
	
// -----------------------------------------------------------------------------
// PenMode_
// -----------------------------------------------------------------------------
//	Sets the specified pen mode

#define	PenMode_(izzyParamsPtr, inGraphics, inPenMode) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mPenModeProc)((inGraphics), (inPenMode), 1.0)

#define	BeginAlphaBlend_(izzyParamsPtr, inGraphics, inAlphaAmount) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mPenModeProc)((inGraphics), kIzGraphicsMode_Blend, (inAlphaAmount))

#define	EndAlphaBlend_(izzyParamsPtr, inGraphics) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mPenModeProc)((inGraphics), kIzGraphicsMode_SrcCopy, 1.0)

typedef void	(*PenModeProc)(
					UGraphics*		inGraphics,
					IzPenMode		inPenMode,
					float			inAlpha);
					
// -----------------------------------------------------------------------------
// DrawLine_
// -----------------------------------------------------------------------------
//	Draws a line from x1/y1 to x2/y2 in the current foreground color

#define	DrawLine_(izzyParamsPtr, inGraphics, x1, y1, x2, y2, inWidth) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mDrawLineProc)((inGraphics), (x1), (y1), (x2), (y2), (inWidth))

typedef void	(*DrawLineProc)(
					UGraphics*	inGraphics,
					IzCoord		x1,
					IzCoord		y1,
					IzCoord		x2,
					IzCoord		y2,
					IzCoord		inWidth);
	
// -----------------------------------------------------------------------------
// PaintRect_/FrameRect_
// -----------------------------------------------------------------------------
//	Each of these variations allows an Actor or Control to draw the specified
//	shape

// fill a rectangle with the current foreground color
#define	PaintRect_(izzyParamsPtr, inGraphics, inRect) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mPaintRectProc)((inGraphics), (inRect), 0, 0, 0)

// draw the outline of a rectangle with the current foreground color
#define	FrameRect_(izzyParamsPtr, inGraphics, inRect, inSize) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mPaintRectProc)((inGraphics), (inRect), (inSize), 0, 0)

// fill a rounded rectangle with the current foreground color
#define	PaintRoundRect_(izzyParamsPtr, inGraphics, inRect, inRoundRectCornerH, inRoundRectCornerV) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mPaintRectProc)((inGraphics), (inRect), 0, (inRoundRectCornerH), (inRoundRectCornerV))

// draw the outline of a rounded rectangle with the current foreground color
#define	FrameRoundRect_(izzyParamsPtr, inGraphics, inRect, inSize, inRoundRectCornerH, inRoundRectCornerV) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mPaintRectProc)((inGraphics), (inRect), (inSize), (inRoundRectCornerH), (inRoundRectCornerV))

// fill an oval with the current foreground color
#define	PaintOval_(izzyParamsPtr, inGraphics, inRect) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mPaintRectProc)((inGraphics), (inRect), 0, -1, -1)

// draw the outline of an oval with the current foreground color
#define	FrameOval_(izzyParamsPtr, inGraphics, inRect, inSize) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mPaintRectProc)((inGraphics), (inRect), (inSize), -1, -1)

typedef void	(*PaintRectProc)(
					UGraphics*		inGraphics,
					const IzRect*	inRect,
					SInt32			inFrameSize,
					SInt32			inRoundRectCornerH,		// set these last two to 0 for a normal rect, -1 for an oval, > 0 for round rect
					SInt32			inRoundRectCornerV);
					
// -----------------------------------------------------------------------------
// PaintRgn_
// -----------------------------------------------------------------------------
//	Fills the specified region using the specified graphics port

// fill a rectangle with the current foreground color
#define	PaintRgn_(izzyParamsPtr, inGraphics, inRegion) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mPaintRgnProc)((inGraphics), (inRegion))

typedef void	(*PaintRgnProc)(
					UGraphics*		inGraphics,
					NativeRgn		inRegion);
					

// -----------------------------------------------------------------------------
// PrepareUnprepareToDrawActorInUserInterface
// PrepareUnprepareToDrawControlInUserInterface
// -----------------------------------------------------------------------------
//	Actors or controls that need to draw to the user interface during a
//	ReceiveMessage callback should bracket their ReceiveMessage function
//	with a call to this function. It ensures that drawing will appear in
//	the correct position and with the correct clipping.

#define PrepareUnprepareToDrawActorInUserInterface(izzyParamsPtr, inActorInfo, inPrepare) \
	(*(izzyParamsPtr)->kConditionalCallbacks->PrepareUnprepareToDrawActorInUserInterfaceProc)((inActorInfo)->mThing, (inPrepare))

typedef void	(*PrepareUnprepareToDrawActorInUserInterfaceProc)(
					ThingPtr		inActorInfo,
					Boolean			inPrepare);
					
#define PrepareUnprepareToDrawControlInUserInterface(izzyParamsPtr, inControlInfo, inPrepare) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mPrepareUnprepareToDrawControlInUserInterfaceProc)((inControlInfo), (inPrepare))

typedef void	(*PrepareUnprepareToDrawControlInUserInterfaceProc)(
					ControlInfo*		inControlInfo,
					Boolean				inPrepare);
					
// -----------------------------------------------------------------------------
// GetNativeContextInfo
// -----------------------------------------------------------------------------

typedef struct {
	NativeView			mScreenDrawable;	// MacOS: NSView* - Windows - HWND
	NativeBitmap		mBitmapDrawable;	// MacOS: Always NULL; Windows: HBITMAP (only if mScreenContext is NULL);
	NativeContextRef	mContext;			// MacOS: CGContextRef; Windows:HDC
} NativeContextInfo;

#define	GetNativeContextInfo_(izzyParamsPtr, inGraphics, outNativeContextInfo) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetNativeContextInfoProc)((inGraphics), (outNativeContextInfo))

typedef void		(*GetNativeContextInfoProc)(
						UGraphics*			inGraphics,
						NativeContextInfo*	outNativeContextInfo);

// -----------------------------------------------------------------------------
// DrawNativeBitmap
// -----------------------------------------------------------------------------
//	Draws a NativeBitmap into the specified graphics context

#define GetNativeBitmapBounds_(izzyParamsPtr, inGraphics, inNativeBitmap, outRect) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetNativeBitmapBoundsProc)((inGraphics), (inNativeBitmap), (outRect))

typedef void	(*GetNativeBitmapBoundsProc)(
					UGraphics*		inGraphics,
					NativeBitmap	inNativeBitmap,
					IzRect*			outRec);
					
#define	DrawNativeBitmap_(izzyParamsPtr, inGraphics, inNativeBitmap, dstRect) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mDrawNativeBitmapProc)((inGraphics), (inNativeBitmap), nil, (dstRect), false)

#define	DrawAlphaNativeBitmap_(izzyParamsPtr, inGraphics, inNativeBitmap, dstRect) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mDrawNativeBitmapProc)((inGraphics), (inNativeBitmap), nil, (dstRect), true)

#define	DrawNativeBitmapSection_(izzyParamsPtr, inGraphics, inNativeBitmap, srcRect, dstRect) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mDrawNativeBitmapProc)((inGraphics), (inNativeBitmap), (srcRect), (dstRect), false)

#define	DrawAlphaNativeBitmapSection_(izzyParamsPtr, inGraphics, inNativeBitmap, srcRect, dstRect) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mDrawNativeBitmapProc)((inGraphics), (inNativeBitmap), (srcRect), (dstRect), true)

typedef void	(*DrawNativeBitmapProc)(
					UGraphics*		inGraphics,
					NativeBitmap	inNativeBitmap,
					const IzRect*	inSrcRect,
					const IzRect*	inDstRect,
					Boolean			inHasAlpha);
					
#define	DrawImageBuffer_(izzyParamsPtr, inGraphics, inImageBufferPtr, dstRect) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mDrawImageBufferProc)((inGraphics), (inImageBufferPtr), nil, (dstRect), false)

#define	DrawAlphaImageBuffer_(izzyParamsPtr, inGraphics, inImageBufferPtr, dstRect) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mDrawImageBufferProc)((inGraphics), (inImageBufferPtr), nil, (dstRect), true)

#define	DrawImageBufferSection_(izzyParamsPtr, inGraphics, inImageBufferPtr, srcRect, dstRect) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mDrawImageBufferProc)((inGraphics), (inImageBufferPtr), (srcRect), (dstRect), false)

#define	DrawAlphaImageBufferSection_(izzyParamsPtr, inGraphics, inImageBufferPtr, srcRect, dstRect) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mDrawImageBufferProc)((inGraphics), (inImageBufferPtr), (srcRect), (dstRect), true)

typedef void	(*DrawImageBufferProc)(
					UGraphics*		inGraphics,
					ImageBufferPtr	inImageBufferPtr,
					const IzRect*	inSrcRect,
					const IzRect*	inDstRect,
					Boolean			inHasAlpha);
					
#define	DrawPicture_(izzyParamsPtr, inGraphics, inPicture, inRect) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mDrawPictureProc)((inGraphics), (inPicture), (inRect), false)

typedef void	(*DrawPictureProc)(
					UGraphics*		inGraphics,
					PicHandle		inPicture,
					const IzRect*	inRect,
					Boolean			inHasAlpha);
					

// -----------------------------------------------------------------------------
// PushAndSetRectClip_ / PopClip_
// -----------------------------------------------------------------------------

// push the current clipping region onto the stack and then set the current
// clipping region using the specified rectangle within the specified UGraphics
// context

#define	PushAndSetRectClip_(izzyParamsPtr, inGraphics, inRect) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mPushPopSetClipProc)((inGraphics), true, (inRect))

// pop the clipping region from the top of the stack, using it to set the
// current clipping region for the specified UGraphics context
#define	PopClip_(izzyParamsPtr, inGraphics) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mPushPopSetClipProc)((inGraphics), false, nil)

typedef void	(*PushPopSetClipProc)(
					UGraphics*		inGraphics,
					Boolean			inPush,
					const IzRect*	inRect);		// can be nil if poping clip
					

// -----------------------------------------------------------------------------
// PaintPoly/FramePoly
// -----------------------------------------------------------------------------
//	Given an array of points, fills, frames or fills and frames the specified
//	region using the specified UGraphics context

// fill a poly with the current foreground color
#define	PaintPoly_(izzyParamsPtr, inGraphics, inPointCount, inPointList, inFillColor) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mPaintFramePolyProc)((inGraphics), (inPointCount), (inPointList), (inFillColor), nil, 0)

// fill a poly with the current foreground color
#define	FramePoly_(izzyParamsPtr, inGraphics, inPointCount, inPointList, inFrameColor, inFrameSize) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mPaintFramePolyProc)((inGraphics), (inPointCount), (inPointList), nil, (inFrameColor), (inFrameSize))

// fill a poly with the current foreground color
#define	PaintAndFramePoly_(izzyParamsPtr, inGraphics, inPointCount, inPointList, inFillColor, inFrameColor, inFrameSize) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mPaintFramePolyProc)((inGraphics), (inPointCount), (inPointList), (inFillColor), (inFrameColor), (inFrameSize))

typedef void	(*PaintFramePolyProc)(
					UGraphics*		inGraphics,
					SInt32			inPointCount,
					const IzPoint*	inPointList,
					const RGBColor*	inFillColor,
					const RGBColor*	inFrameColor,
					SInt32			inFrameSize);
					
// -----------------------------------------------------------------------------
// EraseRect_
// -----------------------------------------------------------------------------
// Fills a rectangle with the background color using the specified UGraphics context.

#define	EraseRect_(izzyParamsPtr, inGraphics, inRect) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mEraseRectProc)((inGraphics), (inRect))

typedef void	(*EraseRectProc)(
					UGraphics*		inGraphics,
					const IzRect*	inRect);
					
// -----------------------------------------------------------------------------
// DrawShadedBoxPercent_
// -----------------------------------------------------------------------------
// Draws a rectangle with shaded edges into the specified UGraphics context.

#define	DrawShadedBoxPercent_(izzyParamsPtr, inGraphics, inRect, inColorRef, inTopPercent, inBottomPercent, inOutline) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mDrawShadedBoxPercentProc)((inGraphics), (inRect), (inColorRef), (inTopPercent), (inBottomPercent), (inOutline))

typedef void	(*DrawShadedBoxPercentProc)(
					UGraphics*		inGraphics,
					const IzRect*	inRect,				// rectangle to be drawn
					ColorRef		inColorRef,			// color ref used to draw the rectangle
					float			inTopPecent,		// how much lighter the top edges will be, 0-100%
					float			inBottomPercent,	// how much darker the bottom edges will be, 0-100%
					Boolean			inOutline);			// set to true to draw the outline only

// -----------------------------------------------------------------------------
// Draw3DRect_
// -----------------------------------------------------------------------------
// Draws bezeled rectangle into the specified UGraphics context.

#define	Draw3DRect_(izzyParamsPtr, inGraphics, inRect, inFillColor, inLightenPercent, inDarkenPercent, inBorderWidth, inBorderFlags, inInset) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mDraw3DRectProc)((inGraphics), (inRect), (inFillColor), (inLightenPercent), (inDarkenPercent), (inBorderWidth), (inBorderFlags), (inInset))

typedef void	(*Draw3DRectProc)(
					UGraphics*		inGraphics,
					const IzRect*	inRect,
					ColorRef		inFillColor,
					float			inLightenPercent,
					float			inDarkenPercent,
					SInt16			inBorderWidth,
					UInt8			inBorderFlags,
					Boolean			inInset);

// -----------------------------------------------------------------------------
// DrawColorRefFeedbackInRect_
// -----------------------------------------------------------------------------
//
#define	DrawColorRefFeedbackInRect_(izzyParamsPtr, inGraphics, inRect, inColorRef, inDrawWithAlphaOverCheckerboard, inDrawRGBvsARGBSplit) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mDrawColorRefFeedbackInRectProc)((inGraphics), (inRect), (inColorRef), (inDrawWithAlphaOverCheckerboard), (inDrawRGBvsARGBSplit))

typedef void	(*DrawColorRefFeedbackInRectProc)(
					UGraphics*		inGraphics,
					const IzRect*	inRect,
					ColorRef		inColorRef,
					Boolean			inDrawWithAlphaOverCheckerboard,
					Boolean			inDrawRGBvsARGBSplit);

// -----------------------------------------------------------------------------
// DrawTextAt_
// -----------------------------------------------------------------------------
// Using the specified UGraphics context, draws the specified block of text
// at the specified horizontal and vertical location. If inClipRect is not NULL,
// then that rectangle is used to clip the text when drawn.

#define	DrawTextAt_(izzyParamsPtr, inGraphics, h, v, inString, inStringLen, inClipRect) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mDrawTextAtProc)((inGraphics), (h), (v), (inString), (inStringLen), (inClipRect))

typedef void	(*DrawTextAtProc)(
					UGraphics*			inGraphics,
					IzCoord				h,
					IzCoord				v,
					const void*			inString,
					SInt32				inStringLen,
					const IzRect*		inClipRect);		// can be NULL

// -----------------------------------------------------------------------------
// GetTextWidth_
// -----------------------------------------------------------------------------
// Using the specified UGraphics context, returns the width, in pixels, of the
// specified block of text

#define	GetTextWidth_(izzyParamsPtr, inGraphics, inString, inStringLen) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetTextWidthProc)((inGraphics), (inString), (inStringLen))

typedef SInt32	(*GetTextWidthProc)(
					UGraphics*			inGraphics,
					const void*			inString,		// pointer to the text
					SInt32				inStringLen);	// length of the text

// -----------------------------------------------------------------------------
// GetTextHeight_
// -----------------------------------------------------------------------------
// Using the specified UGraphics context, returns the height, in pixels, of the
// specified block of text

#define	GetTextHeight_(izzyParamsPtr, inGraphics, inString, inStringLen) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetTextHeightProc)((inGraphics), (inString), (inStringLen))

typedef SInt32	(*GetTextHeightProc)(
					UGraphics*			inGraphics,
					const void*			inString,		// pointer to the text
					SInt32				inStringLen);	// length of the text

// -----------------------------------------------------------------------------
// GetFontInfo_
// -----------------------------------------------------------------------------
// Using the specified UGraphics context, returns information about the currently
// selected font using a FontInfo structure

#define	GetFontInfo_(izzyParamsPtr, inGraphics, outFontInfo) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetFontInfoProc)((inGraphics), (outFontInfo))

typedef void	(*GetFontInfoProc)(
					UGraphics*			inGraphics,
					FontInfo*			outFontInfo);

// -----------------------------------------------------------------------------
// GetFamilyAndStyleFromFullFontName_ (DEPRECATED for v3.0)
// -----------------------------------------------------------------------------
// Returns the font family and font style values for a specified font name.

/*

#define	GetFamilyAndStyleFromFullFontName_(izzyParamsPtr, inFontName, outFamily, outStyle) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetFamilyAndStyleFromFullFontNameProc)((inFontName), (outFamily), (outStyle))

typedef bool	(*GetFamilyAndStyleFromFullFontName)(
					const char*			inFontName,
					IZFontFamily*		outFontInfo,
					IzFontStyle*		outFontStyle);
*/

// -----------------------------------------------------------------------------
// SetupFont_
// -----------------------------------------------------------------------------
// Prepares to draw text into the specified UGraphics context using the font
// parameters specified by inFontName, inFontSize, inFontStyle, and inFontMode

#define	SetupFont_(izzyParamsPtr, inGraphics, inFontName, inFontSize, inFontStyle, inFontMode) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetupFontProc)((inGraphics), (inFontName), (inFontSize), (inFontStyle), (inFontMode) )

typedef void	(*SetupFontProc)(
					UGraphics*			inGraphics,
					const char*			inFontName,
					SInt16				inFontSize,
					SInt16				inFontStyle,
					SInt16				inFontMode);

// -----------------------------------------------------------------------------
// InvalidateActorDefinedArea_
// -----------------------------------------------------------------------------
// Invalidates the actor defined drawing area of the calling actor. If you call
// InvalidateActorDefinedAreaRect_ you can specify a rectangle within that area
// to avoid having the entire area redrawn.

#define	InvalidateActorDefinedArea_(izzyParamsPtr, inActorInfo) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mInvalidateActorDefinedAreaProc)((inActorInfo), NULL)

#define	InvalidateActorDefinedAreaRect_(izzyParamsPtr, inActorInfo, inInvalRect) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mInvalidateActorDefinedAreaProc)((inActorInfo), (inInvalRect))

typedef void	(*InvalidateActorDefinedAreaProc)(
					ActorInfo*			inActorInfo,
					const IzRect*		inInvalRect);		// CAN BE NULL

// -----------------------------------------------------------------------------
// InvalidateActor_
// -----------------------------------------------------------------------------
// Forces the specified actor to be redrawn. Use the InvalidateActorProperty_
// variant to invalidate a specific property within the actor

#define	InvalidateActor_(izzyParamsPtr, inActorInfo) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mInvalidateActorProc)((inActorInfo), kPropertyTypeInvalid, 0)

#define	InvalidateActorProperty_(izzyParamsPtr, inActorInfo, inPropertyType, inPropertyIndex) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mInvalidateActorProc)((inActorInfo), (inPropertyType), (inPropertyIndex))

typedef void	(*InvalidateActorProc)(
					ActorInfo*			inActorInfo,
					PropertyType		inPropertyType,
					PropertyIndex		inPropertyIndex);

// -----------------------------------------------------------------------------
// InvalidateControl_
// -----------------------------------------------------------------------------
//
// If you wish to invalidate the entire control, then use
// InvalidateControl_
//
// If you wish to invalidate a specific area of the control
// then use InvalidateControlRect_
//
// In Isadora 3, the inInvalidateFlags parameter is ignored. Passing
// kControlUpdateCode_DrawControl or kControlUpdateCode_UpdateControl
// makes no difference.

#define	InvalidateControl_(izzyParamsPtr, inControlInfo, inInvalidateFlags) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mInvalidateControlProc)((inControlInfo), (inInvalidateFlags), NULL)

#define	InvalidateControlRect_(izzyParamsPtr, inControlInfo, inInvalidateFlags, inInvalidateRect) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mInvalidateControlProc)((inControlInfo), (inInvalidateFlags), (inInvalidateRect))

typedef void	(*InvalidateControlProc)(
					ControlInfo*		inControlInfo,
					UInt32				inInvalidateFlags,
					const IzRect*		inInvalidateRect);		// CAN BE NULL

// -----------------------------------------------------------------------------
// IsControlVisible_
// -----------------------------------------------------------------------------
//	Returns true if the specified control is visible

#define	IsControlVisible_(izzyParamsPtr, inControlInfo) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIsControlVisibleProc)(inControlInfo)

typedef bool	(*IsControlVisibleProc)(
					ControlInfo*		inControlInfo);

// -----------------------------------------------------------------------------
// GetFPS_
// -----------------------------------------------------------------------------
// Returns the current Frames Per Second at which Isadora is rendering

#define	GetFPS_(izzyParamsPtr) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetFPSProc)()

typedef IzReal	(*GetFPSProc)();
	
// -----------------------------------------------------------------------------
// GetPerformanceData_
// -----------------------------------------------------------------------------
// Returns the current Video Processing Overhead value

#define	GetPerformanceData_(izzyParamsPtr, outTargetFPS, outFPS, outCycles, outVPO) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetPerformanceDataProc)((outTargetFPS), (outFPS), (outCycles), (outVPO))

typedef void	(*GetPerformanceDataProc)(
					IzReal*				outTargetFPS,
					IzReal*				outFPS,
					IzReal*				outCycles,
					IzReal*				outVPO);
	
// -----------------------------------------------------------------------------
// MarkActorDocumentModified_
// -----------------------------------------------------------------------------
// Makes the curret document dirty so the user will be asked to save

#define	MarkActorDocumentModified_(izzyParamsPtr, inActorInfo) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mMarkActorDocumentModifiedProc)((inActorInfo)->mThing)

typedef void	(*MarkActorDocumentModifiedProc)(
					ThingPtr			inThing);
	
// -----------------------------------------------------------------------------
// MarkControlDocumentModified_
// -----------------------------------------------------------------------------
// Makes the curret document dirty so the user will be asked to save

#define	MarkControlDocumentModified_(izzyParamsPtr, inControlInfo) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mMarkControlDocumentModifiedProc)(inControlInfo)

typedef void	(*MarkControlDocumentModifiedProc)(
					ControlInfo*		inControlInfo);
	
// -----------------------------------------------------------------------------
// SetActorUndoAction_
// -----------------------------------------------------------------------------
// Allows a Actor to handle undo/redo operations.
//
// The calling function passes in a pointer to a block of data. This data
// will be passed back to the caller via the HandleActorMessage. The
// data may be allocated using any method desired (IzzyMalloc_() or new)
// as the caller is repsonsible for disposing the data when the 
// kIzzyDisposeUndoRedo message is passed to the callback. Typically this
// data will include the data required to both undo _and_ redo the
// user action taken.
//
// An optional c-string may also be provided to specify the
// text that will appear next to the undo/redo menu item.
//
// if the user chooses undo or redo, a message will be sent to the Actor's
// HandleActorMessage function, with a mesasge of kIzzyUndoChosen, kIzzyRedoChosen, 
// and will also pass back the user data given to this function. The 
// HandleActorMessage should take the respective action.
//
// When the undo/redo data is no longer needed, the message will be
// kIzzyDisposeUndoRedo. At this point the callback should dispose and allocated
// data.

#define	SetActorUndoAction_(izzyParamsPtr, inActorInfo, inUserData, inText) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetActorUndoActionProc)((inActorInfo), (inUserData), (inText))

typedef void	(*SetActorUndoActionProc)(
					ActorInfo*			inActorInfo,
					void*				inUserData,
					const char*			inUndoText);	// can be NULL -- shows next to Undo/Redo items in Edit Menu
	
// -----------------------------------------------------------------------------
// SetControlUndoAction_
// -----------------------------------------------------------------------------
// Allows a control to handle undo/redo operations.
//
// The calling function passes in a pointer to a block of data. This data
// will be passed back to the caller via the HandleControlMessage. The
// data may be allocated using any method desired (IzzyMalloc_() or new)
// as the caller is repsonsible for disposing the data when the 
// kIzzyDisposeUndoRedo message is passed to the callback. Typically this
// data will include the data required to both undo _and_ redo the
// user action taken.
//
// An optional c-string may also be provided to specify the
// text that will appear next to the undo/redo menu item.
//
// if the user chooses undo or redo, a message will be sent to the Control's
// HandleControlMessage function, with a mesasge of kIzzyUndoChosen, kIzzyRedoChosen, 
// and will also pass back the user data given to this function. The 
// HandleControlMessage should take the respective action.
//
// When the undo/redo data is no longer needed, the message will be
// kIzzyDisposeUndoRedo. At this point the callback should dispose and allocated
// data.

#define	SetControlUndoAction_(izzyParamsPtr, inControlInfo, inUserData, inText) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetControlUndoActionProc)((inControlInfo), (inUserData), (inText))

typedef void	(*SetControlUndoActionProc)(
					ControlInfo*		inControlInfo,
					void*				inUserData,
					const char*			inUndoText);	// can be NULL -- shows next to Undo/Redo items in Edit Menu
	
// -----------------------------------------------------------------------------
// PostEditActorPrivateDataAction
// -----------------------------------------------------------------------------
// Allows a control to handle undo/redo operations.
//
// The calling function passes in a pointers and lengths to the old and
// new data. These data blocks _must_ be allocated with IzzyMalloc_(),
// as Isadora disposes the data using IzzyFree_() when the data is no
// longer needed.
//
// An optional c-string may also be provided to specify the
// text that will appear next to the undo/redo menu item.
//
// if the user chooses undo or redo, the actor's SetPrivateData function
// wil be called with the old or new data, as appropriate.
//
// The data blocks are disposed using IzzyFree_() when the data is no
// longer needed.


#define	PostEditActorPrivateDataAction_(izzyParamsPtr, inActorInfo, inOldData, inOldLength, inNewData, inNewLength, inUndoText) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mPostEditActorPrivateDataActionProc)((inActorInfo), (inOldData), (inOldLength), (inNewData), (inNewLength), (inUndoText))

typedef void	(*PostEditActorPrivateDataActionProc)(
					ActorInfo*			inActorInfo,
					void*				inOldData,
					UInt32				inOldLength,
					void*				inNewData,
					UInt32				inNewLength,
					const char*			inUndoText);	// can be NULL -- shows next to Undo/Redo items in Edit Menu
	
// -----------------------------------------------------------------------------
// ShowHideStage_
// -----------------------------------------------------------------------------
// Allows a control or an actor to show or hide the stage.

#define	ShowHideStage_(izzyParamsPtr, inThing, inControlInfo, inShow, inDontWarn) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mShowHideStageProc)((inThing), (inControlInfo), (inShow), (inDontWarn))

typedef void	(*ShowHideStageProc)(
					ThingPtr			inThing,		// only one of either inThing or inControl info should be non-nulll
					ControlInfo*		inControlInfo,
					Boolean				inShow,
					Boolean				inDontWarn);
	
// -----------------------------------------------------------------------------
// FileSavedOnBigEndianMachine_
// -----------------------------------------------------------------------------
// Returns 1 if current document saved on Big Endian machine; 0 if saved on
// a little endian machine, and 0xFF the endianness of the file is unknown

#define	FileSavedOnBigEndianMachine_(izzyParamsPtr, inThingPtr) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mFileSavedOnBigEndianMachineProc)(inThingPtr)

typedef UInt8	(*FileSavedOnBigEndianMachineProc)(
					ThingPtr			inThing);
	
// -----------------------------------------------------------------------------
// GetUTF8PathToCurrentDocument_
// -----------------------------------------------------------------------------
// Returns a path in UTF8 format to the current document. If the document has
// not yet saved, then the path to root level drive is given. This function
// returns true if a valid path was returned; false otherwise.
//
// The returned path ends with the a file separator character - forward
// slash for MacOS (/), backward slash for Windows. (\)

#define	GetUTF8PathToCurrentDocument_(izzyParamsPtr, inThingPtr, inMaxLen, outString) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetUTF8PathToCurrentDocumentProc)((inThingPtr), (inMaxLen), (outString))

typedef Boolean	(*GetUTF8PathToCurrentDocumentProc)(
					ThingPtr			inThing,
					UInt32				inMaxLen,
					char*				outBuffer);
	
// -----------------------------------------------------------------------------
// GetUTF8PathToTroikaTronixFolder_
// -----------------------------------------------------------------------------
// Returns a path in UTF8 format to the the TroikaTronix folder, which contains
// the Isadora Plugins, Controls, GLSL Shaders, Themes and (as of 3.2) the
// global Javascript folder.
//
// The returned path ends with the a file separator character - forward
// slash for MacOS (/), backward slash for Windows. (\)
//
// Added for 3.1.2b10

#define	GetUTF8PathToTroikaTronixFolder_(izzyParamsPtr, inMaxLen, outString) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetUTF8PathToTroikaTronixFolderProc)((inMaxLen), (outString))

typedef Boolean	(*GetUTF8PathToTroikaTronixFolderProc)(
					UInt32				inMaxLen,
					char*				outBuffer);
	
// -----------------------------------------------------------------------------
// SetPropertyFilePathModeProc_
// -----------------------------------------------------------------------------
// Sets the PropFilePathMode for an input or output property, allowing the
// callaer to ensure the cleanliness of file names or file paths passed on
// to an actor.
//
// Note that this callback will fail if the data type of the specified property
// is not a string.

#define	SetPropertyFilePathModeProc_(izzyParamsPtr, actorInfo, propType, propIndex1, inFilePathMode) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetPropertyFilePathModeProc)((actorInfo)->mThing, (propType), (propIndex1), (inFilePathMode))

typedef IzzyError	(*SetPropertyFilePathModeProc)(					// returns an non-zero error code if there is an error
						ThingPtr			inThing,				// thing that owns the property
						PropertyType		inPropertyType,			// the property type, either input or output
						SInt32				inPropertyIndex1,		// index of the input property
						PropFilePathMode	inFilePathMode);		// the file path mode for this property

// -----------------------------------------------------------------------------
// IzzyGetUTF8PathToActorResources_ and IzzyGetUTF8PathToControlResources_
// -----------------------------------------------------------------------------
// 

#define	IzzyGetUTF8PathToActorResources_(izzyParamsPtr, actorInfo, outPath, outPathMaxLen) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetUTF8PathToPluginResourcesProc)((actorInfo), NULL, (outPath), (outPathMaxLen))

#define	IzzyGetUTF8PathToControlResources_(izzyParamsPtr, controlInfo, resID, outRenderedImage) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetUTF8PathToPluginResourcesProc)(NULL, (controlInfo), (outPath), (outPathMaxLen))

typedef IzzyError	(*GetUTF8PathToPluginResourcesProc)(						// returns kIzzyNoError on success
						const ActorInfo*			inActorInfo,				// non NULL for actors, NULL for controls
						const ControlInfo*			inControlInfo,				// non NULL for controls, NULL for actors
						char*						outUTF8Path,
						uint32_t					inUT8PathMaxLen);

// -----------------------------------------------------------------------------
// UpdateTextureBufferWithOwnedTexture
// -----------------------------------------------------------------------------
// If the mTexture is NULL, then a new IzzyTexture object with be created with
// the specified values; otherwise, the existing texture's parameters will
// be set to the specified values. Before the texture is deleted, the owner
// shall always call DisposeTextureBufferWithOwnedTexture_ to dispose the
// reference to the texture held by the TextureBuffer
enum {
	kUpdateTextureOptions_OriginTopLeft		= (1 << 0),
	kUpdateTextureOptions_OriginBottomLeft	= (1 << 1)
};

typedef UInt32 UpdateTextureOptions;

#define	UpdateTextureBufferWithOwnedTexture_(izzyParamsPtr, inThingPtr, stageIndex0, txBufferPtr, txTarget, txName, txGPUWidth, txGPUHeight, txImageWidth, txImageHeight, txImageOffsetH, txImageOffsetV, mipmapEnable) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mUpdateTextureBufferWithOwnedTextureProc)((inThingPtr), (stageIndex0), (txBufferPtr), \
		(txTarget), (txName), (txGPUWidth), (txGPUHeight), (txImageWidth), (txImageHeight), (txImageOffsetH), (txImageOffsetV), (mipmapEnable))

typedef Boolean	(*UpdateTextureBufferWithOwnedTextureProc)(
					ThingPtr				inThing,
					StageIndexT				inTargetStageIndex0,
					TextureBuffer*			inTextureBuffer,
					UInt32					inTexTarget,
					UInt32					inTexName,
					UInt32					inTextureWidth,				// hardware width on GPU
					UInt32					inTextureHeight,			// hardware height on GPU
					UInt32					inImageWidth,				// width of the image within texture
					UInt32					inImageHeight,				// height of the image within texture
					UInt32					inImageOffsetH,				// offset to the left edge of the texture
					UInt32					inImageOffsetV,				// offset to the top edge of the texture
					Boolean					inMipmapEnable);			// mipmap enable flag
	
// -----------------------------------------------------------------------------
// DisposeTextureBufferWithOwnedTexture_
// -----------------------------------------------------------------------------
// Disposes a texture buffer created with UpdateTextureBufferWithOwnedTexture_

#define	DisposeTextureBufferWithOwnedTexture_(izzyParamsPtr, txBufferPtr) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mDisposeTextureBufferWithOwnedTextureProc)((txBufferPtr))

typedef void	(*DisposeTextureBufferWithOwnedTextureProc)(
					TextureBuffer*		inTextureBuffer);
	
// -----------------------------------------------------------------------------
// GetNativeGLContextForStage_
// -----------------------------------------------------------------------------
// Returns the native CGLContext for the specified stage. On Mac OS this will
// be a CGLContextObj; under Windows it will be an HGLRC along with its 
// associated HDC. Actors that reference this information must receive the 
// kWantStageAboutToDispose message and invalidate this information if the
// stage is disposed. If the stage is not visible when the call is made,
// the result will be FALSE.

#define	GetNativeGLContextForStage_(izzyParamsPtr, thingPtr, stageIndex0, outCtx, outCtxParam) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetNativeGLContextForStageProc)((thingPtr), (stageIndex0), (outCtx), (outCtxParam))

typedef Boolean	(*GetNativeGLContextForStageProc)(
					ThingPtr			inThing,
					StageIndexT			inTargetStageIndex0,
					void**				outContext,					// MAC OS = CGLContextObj**; Windows = HGLRC**
					void**				outContextParam);			// MAC OS = n/a; Windows = HDC**

// -----------------------------------------------------------------------------
// CreateIzzyUnicodeStringWithUTF8Text
// -----------------------------------------------------------------------------
// Converts a UTF8 encoded text into a native unicode string. On Mac OS this
// is a CFStringRef, on Windows it is a pointer to an array of WCHARs. The caller
// dispose of the returned string using DisposeIzzyUnicodeString. If the
// function fails it will return NULL

#define	CreateIzzyUnicodeStringWithUTF8Text_(izzyParamsPtr, inUTF8String, inUTF8StringLength) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mCreateIzzyUnicodeStringWithUTF8TextProc)((inUTF8String), (inUTF8StringLength))
	
typedef IzzyUnicodeStr	(*CreateIzzyUnicodeStringWithUTF8TextProc)(
							const void*		inUTF8Text,
							size_t			inUTF8TextLen);
	
// -----------------------------------------------------------------------------
// DisposeIzzyUnicodeString
// -----------------------------------------------------------------------------
// Dipsoses an IzzyUnicode string.

#define	DisposeIzzyUnicodeString_(izzyParamsPtr, inIzzyUnicodeString) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mDisposeIzzyUnicodeStringProc)(inIzzyUnicodeString)
	
typedef void			(*DisposeIzzyUnicodeStringProc)(
							IzzyUnicodeStr	inString);
					

// -----------------------------------------------------------------------------
// IzzyUnicodeStringLength
// -----------------------------------------------------------------------------
// Returns the length of an IzzyUnicodeStr in characters

#define	IzzyUnicodeStringLength_(izzyParamsPtr, inIzzyUnicodeString) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIzzyUnicodeStringLengthProc)(inIzzyUnicodeString)
	
typedef size_t			(*IzzyUnicodeStringLengthProc)(
							const IzzyUnicodeStr	inString);
					

// -----------------------------------------------------------------------------
// IzzyUnicodeStringLineCount
// -----------------------------------------------------------------------------
// Returns the number of lines of an IzzyUnicodeStr in characters. The lines may
// be separated by carriage returns (CR), linefeeds (LF), or CR/LF pairs.

#define	IzzyUnicodeStringLineCount_(izzyParamsPtr, inIzzyUnicodeString) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIzzyUnicodeStringLineCountProc)(inIzzyUnicodeString)
	
typedef size_t			(*IzzyUnicodeStringLineCountProc)(
							const IzzyUnicodeStr	inString);
					

// -----------------------------------------------------------------------------
// CreateUTF8StringWithIzzyUnicodeString
// -----------------------------------------------------------------------------
// Converts a null terminated UTF8 encoded string into a native unicode string.
// The caller is responsible for the memory created by this function, and must
// call the IzzyFree_ to free the memory when it is no longer needed

#define	CreateUTF8StringWithIzzyUnicodeString_(izzyParamsPtr, inIzzyUnicodeStr) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mCreateUTF8StringWithIzzyUnicodeStringProc)(inIzzyUnicodeStr)
	
typedef char*		(*CreateUTF8StringWithIzzyUnicodeStringProc)(
							const IzzyUnicodeStr	inIzzyUnicodeString);
					
// -----------------------------------------------------------------------------
// IzzyUnicodeStringCompare
// -----------------------------------------------------------------------------
// Compares two Izzy Unicode Strings, returning -1 for less than, 0 for equal,
// and +1 for greater than.

#define	IzzyUnicodeStringCompare_(izzyParamsPtr, inString1, inString2, inCaseSensitive) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIzzyUnicodeStringCompareProc)((inString1), (inString2), (inCaseSensitive))
	
typedef int				(*IzzyUnicodeStringCompareProc)(
							IzzyUnicodeStr	inStr1,
							IzzyUnicodeStr	inStr2,
							bool			inCaseSensitive);

// -----------------------------------------------------------------------------
// IzzyUnicodeStringGetSubstring
// -----------------------------------------------------------------------------
// Creates a new IzzyUnicodeStr from using the specified ranges. The caller is
// responsible for disposing the returned IzzyUnicodeStr using

#define	IzzyUnicodeStringGetSubstring_(izzyParamsPtr, inString1, inMode, inOffset, inLength) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIzzyUnicodeStringGetSubstringProc)((inString1), (inMode), (inOffset), (inLength))
	
typedef IzzyUnicodeStr	(*IzzyUnicodeStringGetSubstringProc)(
							IzzyUnicodeStr	inStr1,
							IzzySubstrMode	inMode,
							int				inOffset,
							int				inLength);

// -----------------------------------------------------------------------------
// SetGlobalMoviePlaybackOptions
// -----------------------------------------------------------------------------
// Only for use when debugging by Mark. Has no effect in the release version
// of the program.

typedef struct {

	bool	mForceGWorldBitmapMovies;
	
	bool	mPostPreloadPlay;
	
	bool	mMovieDrawingCompleteCallback;

	bool	mCVBufferQueue;
	bool	mPreloadInitialFrames;
	bool	mForceUpdateMoviePostPreload;
	bool	mTaskMoviePostPreload;
	
	bool	mBackgroundMoviePreload;
	bool	mBackgroundMovieDispose;
	bool	mPlayDisposeMutex;
	bool	mAttachDetachMutex;
	
	bool	mVisualContextPool;
	bool	mAudioContextPool;
	
	bool	mSpecialDirectFeaturesDeviceOptions;
	
	bool	mGLSwapTopOfCycle;
	
	bool	mFreqTaskMovie;
	bool	mGlobalMoviesTask;
	bool	mGlobalMoviesTask_TopOfCycle;
	bool	mGlobalMoviesTask_BottomOfCycle;

} MoviePlaybackOptions;

#define	SetGlobalMoviePlaybackOptions_(izzyParamsPtr, moviePlaybackOptions) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetGlobalMoviePlaybackOptionsProc)((moviePlaybackOptions))

typedef void	(*SetGlobalMoviePlaybackOptionsProc)(
					MoviePlaybackOptions*		inOptions);
	
// -----------------------------------------------------------------------------
// AcquireIzzyMoviePlayerTask
// -----------------------------------------------------------------------------
// Acquires ownership of a movie player task for use by a client. This 
// functionality is undocumented, and is only for use within Isadora.

typedef void (*PlayerDiedCallback)(
	IsadoraParameters*	ip,
	IzzyMoviePlayerTask	inPlayerTask,
	void*				inCallerRefCon);

#define	AcquireIzzyMoviePlayerTask_(izzyParamsPtr, inType, inTaskHint, inPlayerDiedCallback, inPlayerDiedCallbackRefCon, outRestartOnFailtureFlag) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mAcquireIzzyMoviePlayerTaskProc)((inType), (inTaskHint), (inPlayerDiedCallback), (inPlayerDiedCallbackRefCon), (outRestartOnFailtureFlag))

typedef IzzyMoviePlayerTask	(*AcquireIzzyMoviePlayerTaskProc)(
								MPOptimizeMode			inYype,
								UInt32					inTaskHint,
								PlayerDiedCallback		inPlayerDiedCallback,			// CAN BE NULL
								void*					inPlayerDiedCallbackRefCon,		// CAN BE NULL
								Boolean*				outRestartOnFailtureFlag);		// CAN BE NULL
	
// -----------------------------------------------------------------------------
// RelinquishIzzyMoviePlayerTask
// -----------------------------------------------------------------------------
// Relinquishes ownership of a movie player task for use by a client. This 
// functionality is undocumented, and is only for use within Isadora.

#define	RelinquishIzzyMoviePlayerTask_(izzyParamsPtr, inTask) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mRelinquishIzzyMoviePlayerTaskProc)((inTask))

typedef void				(*RelinquishIzzyMoviePlayerTaskProc)(
								IzzyMoviePlayerTask		inTask);
	
// -----------------------------------------------------------------------------
// PrintMessageToMonitorWindow
// -----------------------------------------------------------------------------
// If the Monitor Window is visible, and if the specified message type is
// is enabled, this function will print a string to the window using the
// printf format

#define	PrintMessageToMonitorWindow_(izzyParamsPtr, messageDirection, messageType, printfStyleFmtString, ...) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mPrintMessageToMonitorWindowProc)((messageDirection), (messageType), (printfStyleFmtString), __VA_ARGS__)

typedef void	(*PrintMessageToMonitorWindowProc)(
					MWDirection		inMsgDirecction,
					MWMsgType		inMsgType,
					const char*		inPrintfStyleFormatString,
					...);

// -----------------------------------------------------------------------------
// LogErrorToSystemConsole_
// -----------------------------------------------------------------------------
// v2.6
// Allows the caller to log an error message to the system console. In the case
// of MacOS, such errors could be found in the Console utility application. For
// Windows, the system log can be examined using the Event Viewer application, 
// and will be stored in a section entitled "isadora-log"
// (See https://technet.microsoft.com/en-us/library/aa996634(v=exchg.65).aspx
// for details)

enum ErrorLogLevel
{
	kErrorLogLevele_Information,
	kErrorLogLevel_Warning,
	kErrorLogLevel_Error
};

#define	LogErrorToSystemConsole_(izzyParamsPtr, level, printfStyleFmtString, ...) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mLogErrorToSystemConsoleProc)((level), (printfStyleFmtString), __VA_ARGS__)

// the Debug version below only outputs in the Debug version of Isadora
#if DEBUG
	
	#define	DebugLogErrorToSystemConsole_(izzyParamsPtr, level, printfStyleFmtString, ...) \
		(*(izzyParamsPtr)->kConditionalCallbacks->mLogErrorToSystemConsoleProc)((level), (printfStyleFmtString), __VA_ARGS__)

#else

	#define	DebugLogErrorToSystemConsole_(...) { }

#endif

typedef void	(*LogErrorToSystemConsoleProc)(
					ErrorLogLevel	inLevel,
					const char*		inPrintfStyleFormatString,
					...);


// -----------------------------------------------------------------------------
// IsModernGUIActiveProc
// -----------------------------------------------------------------------------
// Returns TRUE if the Modern GUI (2.0 look) is active

#define	IsModernGUIActive_(izzyParamsPtr) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIsModernGUIActiveProc)()

typedef Boolean	(*IsModernGUIActiveProc)();

// -----------------------------------------------------------------------------
// CreateIzzyFrameBuffer
// -----------------------------------------------------------------------------
// Creates a framebuffer object for use by your plugin.
//
// IMPORTANT: If you intend to output this framebuffer as a Texture video
// stream, then it is essential that you set the inForcePowerOfTwo flag to TRUE.
// To stay compatible with FreeFrameGL, all texture video streams must have

enum {

	kCreateFBOOptions_None			= 0,
	kCreateFBOOptions_PowerOfTwo	= (1 << 0),
	kCreateFBOOptions_MipmapEnable	= (1 << 1),
	kCreateFBOOptions_CIContext		= (1 << 2),

	kCreateFBOOptions_InitToBlack	= (1 << 7)
};

typedef UInt32	CreateFBOOptions;

#define	CreateIzzyFrameBuffer_(izzyParamsPtr, inActorInfo, inTargetStageIndex0, inWidth, inHeight, inForcePowerOfTwo, inClearColorArray) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mCreateIzzyFrameBufferProc)((inActorInfo)->mThing, (inTargetStageIndex0), (inWidth), (inHeight), ((inForcePowerOfTwo) ? kCreateFBOOptions_PowerOfTwo : kCreateFBOOptions_None), (inClearColorArray))

#define	CreateIzzyFrameBufferWithOptions_(izzyParamsPtr, inActorInfo, inTargetStageIndex0, inWidth, inHeight, inOptions, inClearColorArray) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mCreateIzzyFrameBufferProc)((inActorInfo)->mThing, (inTargetStageIndex0), (inWidth), (inHeight), (inOptions), (inClearColorArray))

typedef IzzyFrameBuffer	(*CreateIzzyFrameBufferProc)(
							ThingPtr			inThing,
							StageIndexT			inTargetStageIndex0,
							UInt32				inWidth,
							UInt32				inHeight,
							CreateFBOOptions	inCreateFBOOptions,		// See CreateFBOOptions above
							GLfloat*			inClearColorArray);		// Four element array with RGBA; if NULL, then use BLACK

// -----------------------------------------------------------------------------
// UpdateIzzyFrameBuffer
// -----------------------------------------------------------------------------
// Checks to ensure that it is possible to call the BindIzzyFrameBuffer function.
// If this function returns FALSE, it generally means that the target stage
// associated with the framebuffer has been destroyed (i.e., the window has been
// cloased.)
//
// If this function returns false, and if you have output the texture associated
// with this framebuffer as a video output, then you must send a NULL texture
// buffer to the output to ensure other plugins to not attempt to use the
// no longer existent texture

#define	UpdateIzzyFrameBuffer_(izzyParamsPtr, inActorInfo, inIzzyFrameBuffer, inTargetStageIndex0, inWidth, inHeight, inForcePowerOfTwo, inClearColorArray) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mUpdateIzzyFrameBufferProc)((inActorInfo)->mThing, (inIzzyFrameBuffer), (inTargetStageIndex0), (inWidth), (inHeight), ((inForcePowerOfTwo) ? kCreateFBOOptions_PowerOfTwo : kCreateFBOOptions_None), (inClearColorArray))

#define	UpdateIzzyFrameBufferWithOptions_(izzyParamsPtr, inActorInfo, inIzzyFrameBuffer, inTargetStageIndex0, inWidth, inHeight, inOptions, inClearColorArray) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mUpdateIzzyFrameBufferProc)((inActorInfo)->mThing, (inIzzyFrameBuffer), (inTargetStageIndex0), (inWidth), (inHeight), (inOptions), (inClearColorArray))

typedef Boolean		(*UpdateIzzyFrameBufferProc)(
						ThingPtr			inThing,
						IzzyFrameBuffer		inIzzyFrameBuffer,
						StageIndexT			inTargetStageIndex0,
						UInt32				inWidth,
						UInt32				inHeight,
						CreateFBOOptions	inCreateFBOOptions,			// See CreateFBOOptions above
						GLfloat*			inClearColorArray);			// Four element array with RGBA; if NULL, then use BLACK
	
// -----------------------------------------------------------------------------
// DisposeIzzyFrameBuffer
// -----------------------------------------------------------------------------
// Disposes a framebuffer object for use by your plugin.

#define	DisposeIzzyFrameBuffer_(izzyParamsPtr, inActorInfo, inIzzyFrameBuffer) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mDisposeIzzyFrameBufferProc)((inActorInfo)->mThing, (inIzzyFrameBuffer))

typedef void	(*DisposeIzzyFrameBufferProc)(
					ThingPtr			inThing,
					IzzyFrameBuffer		inIzzyFrameBuffer);
	
// -----------------------------------------------------------------------------
// BindIzzyFrameBuffer
// -----------------------------------------------------------------------------
// Binds the framebuffer so that you may draw into it.
//
// Once bound, the framebuffer state is as follows:
// -- The OpenGL context associated with the framebuffer is 
//    the current context and it has been locked (on Mac OS)
// -- A glViewport call has been made to properly configure
//    the viewportr
// -- The identity matrix has been set for the GL_PROJECTION, 
//    GL_MODELVIEW, and GL_TEXTURE matrices
// -- The remaining OpenGL parameters are all at their initial
//    values.

#define	BindIzzyFrameBuffer_(izzyParamsPtr, inActorInfo, inIzzyFrameBuffer, inEraseToBlack) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mBindIzzyFrameBufferProc)((inActorInfo)->mThing, (inIzzyFrameBuffer), (inEraseToBlack))

typedef Boolean	(*BindIzzyFrameBufferProc)(
					ThingPtr			inThing,
					IzzyFrameBuffer		inIzzyFrameBuffer,
					Boolean				inEraseToBlack);
	
// -----------------------------------------------------------------------------
// UnbindIzzyFrameBuffer
// -----------------------------------------------------------------------------
// Unbinds the framebuffer. Call this function after you have called 
// BindIzzyFrameBuffer and have completed the OpenGL calls required
// to draw into the framebuffer.

#define	UnbindIzzyFrameBuffer_(izzyParamsPtr, inActorInfo, inIzzyFrameBuffer) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mUnbindIzzyFrameBufferProc)((inActorInfo)->mThing, (inIzzyFrameBuffer))

typedef TextureBuffer*	(*UnbindIzzyFrameBufferProc)(
							ThingPtr			inThing,
							IzzyFrameBuffer		inIzzyFrameBuffer);

// -----------------------------------------------------------------------------
// ConvertVideoStreamToBitmap
// -----------------------------------------------------------------------------
// Converts any video stream type (texture, core-image, bitmap) to a bitmap video
// image.
//
// The first time you call this function, set inIzzyBitmapConverter to NULL to
// to indicate the function should create the video converter. Subsequent calls
// should provide the value returned by the first call to ConvertVideoStreamToBitmap()
// as the parameter to inIzzyBitmapConverter. You must store the return value each
// time you call ConvertVideoStreamToBitmap() because the converter might be
// reallocated because the format or parameters have changed.
//
// If you can tolerate a one frame delay, set inAllowDelayedReadback to TRUE. This
// allows asynchronous readback of GPU based images, improving performance
// substantially.
//
// Set inVideoStreamValuePtr to a ValuePtr that references a valid video stream,
// where v.u.data points to an  a Texture Buffer, a Core Image Buffer, or an Image
// Buffer.
//
// To convert a subset of the image to a bitmap, set the inROI parameter to define
// the "region of interest", i.e., the area you wish to convert. Set this parameter
// to an array of four IzReals, giving the percentange of left, top, width and
// height from 0.0 to 1.0 for each paraemters.
//
// To force specific width and height, set both the inFixedOutputWidth and
// inFixedOutputHeight parameters to non-zero values.
//
// On return, the converted bitmap image will be contained in the Value passed back
// via the outputBitmapBuffer Value pointer.

typedef void*	IzzyBitmapConverter;

typedef struct
{
	IzFloat32	left;		// 0.0 to 1.0
	IzFloat32	top;		// 0.0 to 1.0
	IzFloat32	right;		// 0.0 to 1.0
	IzFloat32	bottom;		// 0.0 to 1.0

} ROIRect;

#define	ConvertVideoStreamToBitmap_(izzyParamsPtr, inActorInfo, inIzzyBitmapConverter, inAllowDelayedReadback, inVideoStreamValuePtr, inROI, outVideoStreamValuePtr) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mConvertStreamVideoToBitmapProc)((inActorInfo)->mThing, (inIzzyBitmapConverter), (inAllowDelayedReadback), (inVideoStreamValuePtr), (inROI), 0, 0, (outVideoStreamValuePtr))

#define	ConvertVideoStreamToBitmapFixedSize_(izzyParamsPtr, inActorInfo, inIzzyBitmapConverter, inAllowDelayedReadback, inVideoStreamValuePtr, inROI, fixedOutputWidth, fixedOutputHeight, outVideoStreamValuePtr) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mConvertStreamVideoToBitmapProc)((inActorInfo)->mThing, (inIzzyBitmapConverter), (inAllowDelayedReadback), (inVideoStreamValuePtr), (inROI), (fixedOutputWidth), (fixedOutputHeight), (outVideoStreamValuePtr))

typedef IzzyBitmapConverter (*ConvertVideoStreamToBitmapProc)(
								ThingPtr				inThing,
								IzzyBitmapConverter		inIzzyBitmapConverter,
								Boolean					inAllowDelayedReadback,
								const ValuePtr			inVideoStreamValuePtr,
								const ROIRect*			inROI,						// CAN BE NULL
								UInt32					inFixedOutputWidth,			// CAN BE ZERO
								UInt32					inFixedOutputHeight,		// CAN BE ZERO
								ValuePtr				outVideoStreamValuePtr);
	
// -----------------------------------------------------------------------------
// ConvertBitmapToVideoStream
// -----------------------------------------------------------------------------
// Converts a bitmap video stream input to any video stream output type
// (texture, core-image, bitmap)
//
// The first time you call this function, set inIzzyBitmapConverter to NULL to
// to indicate the function should create the video converter. Subsequent calls
// should provide the value returned by the first call to ConvertVideoStreamToBitmap()
// as the parameter to inIzzyBitmapConverter. You must store the return value each
// time you call ConvertVideoStreamToBitmap() because the converter might be
// reallocated because the format or parameters have changed.
//
// Set inOutputVideoBufferType to indicate what output type the converter should
// produce
//
// Set the inTargetStageIndex to the desired target stage for Texture or Core Image
// outputs; set it to kInvalidStageIndex for bitmap output
//
// Set inImageBuffer to the bitmap source buffer to be converted.
//
// To convert a subset of the image to a bitmap, set the inROI parameter to define
// the "region of interest", i.e., the area you wish to convert. Set this parameter
// to an array of four IzReals, giving the percentange of left, top, width and
// height from 0.0 to 1.0 for each paraemters.
//
// To force specific width and height, set both the inFixedOutputWidth and
// inFixedOutputHeight parameters to non-zero values.
//
// On return, the converted bitmap image will be contained in the Value passed back
// via the outputBitmapBuffer Value pointer.

#define	ConvertBitmapToVideoStream_(izzyParamsPtr, inActorInfo, inIzzyBitmapConverter, inOutputVideoBufferType, inTargetStageIndex, inImageBuffer, inROI, outVideoStreamValuePtr) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mConvertBitmapToVideoStreamProc)((inActorInfo)->mThing, (inIzzyBitmapConverter), (inOutputVideoBufferType), (inTargetStageIndex), (inImageBuffer), (inROI), 0, 0, (outVideoStreamValuePtr))

#define	ConvertBitmapToVideoStreamFixedSize_(izzyParamsPtr, inActorInfo, inIzzyBitmapConverter, inOutputVideoBufferType, inTargetStageIndex, inImageBuffer, inROI, fixedWidth, fixedHeight, outVideoStreamValuePtr) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mConvertBitmapToVideoStreamProc)((inActorInfo)->mThing, (inIzzyBitmapConverter), (inOutputVideoBufferType), (inTargetStageIndex), (inImageBuffer), (inROI), (fixedWidth), (fixedHeight), (outVideoStreamValuePtr))

typedef IzzyBitmapConverter (*ConvertBitmapToVideoStreamProc)(
								ThingPtr				inThing,
								IzzyBitmapConverter		inIzzyBitmapConverter,
								VideoBufferType			inOutputVideoBufferType,
								StageIndexT				inTargetStageIndex,
								ImageBufferPtr			inImageBuffer,
								const ROIRect*			inROI,						// CAN BE NULL
								UInt32					inFixedOutputWidth,		// CAN BE ZERO
								UInt32					inFixedOutputHeight,		// CAN BE ZERO
								ValuePtr				outVideoStreamValuePtr);
	
// -----------------------------------------------------------------------------
// DisposeIzzyBitmapConverter
// -----------------------------------------------------------------------------
// Disposes a BitmapConverter object for use by your plugin.

#define	DisposeIzzyBitmapConverter_(izzyParamsPtr, inActorInfo, inIzzyBitmapConverter) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mDisposeIzzyBitmapConverterProc)((inActorInfo)->mThing, (inIzzyBitmapConverter))

typedef void				(*DisposeIzzyBitmapConverterProc)(
								ThingPtr				inThing,
								IzzyBitmapConverter		inIzzyBitmapConverter);
	
// -----------------------------------------------------------------------------
// EnableDisableBitmapConverterAsyncReadback
// -----------------------------------------------------------------------------
// Added for 3.0.9b05
// Enables or disasble asyncrhonous readback from an IzzyBitmapConverter

#define	EnableDisableBitmapConverterAsyncReadback_(izzyParamsPtr, inActorInfo, inIzzyBitmapConverter, inEnable) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mEnableDisableBitmapConverterAsyncReadbackProc)((inActorInfo)->mThing, (inIzzyBitmapConverter), (inEnable))

typedef void				(*EnableDisableBitmapConverterAsyncReadback)(
								ThingPtr				inThing,
								IzzyBitmapConverter		inIzzyBitmapConverter,
								Boolean					inEnable);
	
// -----------------------------------------------------------------------------
// SimulateActorDoubleClick_
// -----------------------------------------------------------------------------

#define	SimulateActorDoubleClick_(izzyParamsPtr, inActorInfo) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSimulateActorDoubleClickProc)((inActorInfo)->mThing)

typedef void		(*SimulateActorDoubleClickProc)(
							ThingPtr			inThing);
// -----------------------------------------------------------------------------
// BitmapDrawFramedPolygon_ (DEPRECATED for v3.0)
// -----------------------------------------------------------------------------
// Draws a framed poiygoin into a bitmap using a list of inner and outer points

/*

DEPRECATED!

#define	BitmapDrawFramedPolygon_(izzyParamsPtr, ib, inWidth, inHeight, inStrokeColor, inFillColor, inPosX, inPosY, inRotation, inFloatPointCount, inFloatPointPairs) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mBitmapDrawFramedPolygonProc)((ib), (inWidth), (inHeight), (inStrokeColor), (inFillColor), (inPosX), (inPosY), (inRotation), (inFloatPointCount), (inFloatPointPairs))

typedef void	(*BitmapDrawFramedPolygonProc)(
					ImageBufferPtr		ib,
					const IzReal		inWidth,
					const IzReal		inHeight,
					ColorRef			inStrokeColor,
					ColorRef			inFillColor,
					const IzReal		inPosX,
					const IzReal		inPosY,
					const IzReal		inRotation,
					const UInt32		inFloatPointCount,
					const IzReal*		inFloatPointPairs);		// x/y outer + x/y inner
*/

// -----------------------------------------------------------------------------
// SetGlobalVariableValue_
// -----------------------------------------------------------------------------
//	Sets the named global variable to the specified value
//
#define	SetGlobalVariableValue_(izzyParamsPtr, inActorInfo, inGlobalVariableName, inValueCount, inValuePtr) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mSetGlobalVariableValueProc)( (inActorInfo)->mThing, (inGlobalVariableName), (inValueCount), (inValuePtr) )

typedef void			(*SetGlobalVariableValueProc)(
							ThingPtr						inThing,
							const char*						inGlobalVariableName,
							size_t							inValueCount,
							const Value*					inValueArray);
	
// -----------------------------------------------------------------------------
// GetGlobalVariableValue_
// -----------------------------------------------------------------------------
//	Sets the named global variable to the specified value
//
#define	GetGlobalVariableValue_(izzyParamsPtr, inActorInfo, inGlobalVariableName, outValueArray) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetGlobalVariableValueProc)((inActorInfo)->mThing, (inGlobalVariableName), (outValueArray))

typedef const size_t	(*GetGlobalVariableValueProc)(
							ThingPtr						inThing,
							const char*						inGlobalVariableName,
							const Value**					outValueArray);

// ============================================================================
//
//	MEMORY ALLOCATION AND MANAGEMENT FUNCTIONS
//
// ============================================================================

	
// -----------------------------------------------------------------------------
// IzzyMalloc_ & IzzyMallocClear_
// -----------------------------------------------------------------------------
// Asks Isadora to allocate memory, a la malloc from the standard library. The
// IzzyMallocClear_ clears the memory to 0 after the allocation is complete.

#define	IzzyMalloc_(izzyParamsPtr, inSize) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIzzyMallocProc)((inSize), FALSE)

#define	IzzyMallocClear_(izzyParamsPtr, inSize) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIzzyMallocProc)((inSize), TRUE)

typedef void*		(*IzzyMallocProc)(
						UInt32						inSize,
						Boolean						inClear);

// -----------------------------------------------------------------------------
// IzzyFree_
// -----------------------------------------------------------------------------
// Frees memory allocated with IzzyMalloc_

#define	IzzyFree_(izzyParamsPtr, inPtr) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIzzyFreeProc)((inPtr))

typedef void		(*IzzyFreeProc)(
						void*						inPtr);

// -----------------------------------------------------------------------------
// IzzyNewHandle_
// -----------------------------------------------------------------------------
//	Creates a new IzzyHandle of the specified size. To access the underlying
//	data, see IzzyLockHandle_

#define	IzzyNewHandle_(izzyParamsPtr, inHandleSize) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIzzyNewHandleProc)(inHandleSize, FALSE)

#define	IzzyNewHandleClear_(izzyParamsPtr, inHandleSize) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIzzyNewHandleProc)(inHandleSize, TRUE)

typedef IzzyHandle	(*IzzyNewHandleProc)(
						size_t							inHandleSize,
						Boolean							inClear);
	
// -----------------------------------------------------------------------------
// IzzyDisposeHandle_
// -----------------------------------------------------------------------------
//	Disposes the memory occupied by an IzzyHandle allocated with IzzyNewHandle_

#define	IzzyDisposeHandle_(izzyParamsPtr, inHandle) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIzzyDisposeHandleProc)(inHandle)

typedef void		(*IzzyDisposeHandleProc)(
						IzzyHandle						inHandle);
	
// -----------------------------------------------------------------------------
// IzzyGetHandleSize_
// -----------------------------------------------------------------------------
//	Returns the size of an IzzyHandle
//
#define	IzzyGetHandleSize_(izzyParamsPtr, inHandle) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIzzyGetHandleSizeProc)(inHandle)

typedef size_t	(*IzzyGetHandleSizeProc)(
					IzzyHandle							inHandle);
	
// -----------------------------------------------------------------------------
// IzzySetHandleSize_
// -----------------------------------------------------------------------------
//	Returns the size of an IzzyHandle
//
#define	IzzySetHandleSize_(izzyParamsPtr, inHandle, inSize) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIzzySetHandleSizeProc)((inHandle), (inSize))

typedef void	(*IzzySetHandleSizeProc)(
					IzzyHandle							inHandle,
					size_t								inSize);
	
// -----------------------------------------------------------------------------
// IzzyLockHandle_
// -----------------------------------------------------------------------------
//	Locks an IzzyHandle and returns a pointer to the locked data in memory

#define	IzzyLockHandle_(izzyParamsPtr, inHandle) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIzzyLockHandleProc)(inHandle)

typedef IzzyPtr		(*IzzyLockHandleProc)(
						IzzyHandle						inHandle);
	
// -----------------------------------------------------------------------------
// IzzyUnlockHandle_
// -----------------------------------------------------------------------------
//	Locks an IzzyHandle and returns a pointer to the locked data in memory.
//	After unlocking an IzzyHandle, the pointer returned by IzzyLockHandle_
//	should be considered invalid.

#define	IzzyUnlockHandle_(izzyParamsPtr, inHandle) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIzzyUnlockHandleProc)(inHandle)

typedef void		(*IzzyUnlockHandleProc)(
						IzzyHandle						inHandle);


// -----------------------------------------------------------------------------
// MTCGeneratorCreate_
// -----------------------------------------------------------------------------
//	Creates a MIDI timecode Generator that will broadcast MIDI timecode to
//	the specified MIDI output port. Returns a reference to the MTC Generator
//	which the caller must destroy using MTCGeneratorDispose

#define	MTCGeneratorCreate_(izzyParamsPtr, inMTCGenMode, inMTCType, inMidiOutputPort, inOffsetHMSV, inStopGenerating_mS) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mMTCGeneratorCreateProc)((inMTCGenMode), (inMTCType), (inMidiOutputPort), (inOffsetHMSV), (inStopGenerating_mS))

typedef MTCGeneratorRef	(*MTCGeneratorCreateProc)(
							MTCGenMode			inMTCGenMod,
							MTCType				inMTCType,
							PortIndex			inMidiOutputPort,
							TimecodeHMSF		inOffsetHMSF,			// bits 24-31 = hours, 16-23 = minutes, 8-15 = seconds, 0-7 = frames
							UInt32				inStopGenerating_mS);	// if MTCGeneratorSetTime_() is not called after this duration elapses,
																		// stop auto generating timecode. Set to 0 for 'never'
// -----------------------------------------------------------------------------
// MTCGeneratorDispose_
// -----------------------------------------------------------------------------
//	Creates a MIDI timecode Generator that will broadcast MIDI timecode to
//	the specified MIDI output port

#define	MTCGeneratorDispose_(izzyParamsPtr, inMTCGenerator) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mMTCGeneratorDisposeProc)((inMTCGenerator))

typedef void			(*MTCGeneratorDisposeProc)(
							MTCGeneratorRef		inMTCGenerator);

// -----------------------------------------------------------------------------
// MTCGeneratorControl_
// -----------------------------------------------------------------------------
//	Allows the caller to control generation of timecode for the speciified MTC
//	generator

enum MTCGeneratorCommand {
	kMTCGenerator_Run,
	kMTCGenerator_Pause,
	kMTCGenerator_Reset
};

#define	MTCGeneratorRun_(izzyParamsPtr, inMTCGenerator) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mMTCGeneratorControlProc)((inMTCGenerator), kMTCGenerator_Run, TimecodeHMSF());
	
#define	MTCGeneratorPause_(izzyParamsPtr, inMTCGenerator) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mMTCGeneratorControlProc)((inMTCGenerator), kMTCGenerator_Pause, TimecodeHMSF());
	
#define	MTCGeneratorReset_(izzyParamsPtr, inMTCGenerator, inHMSFValue) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mMTCGeneratorControlProc)((inMTCGenerator), kMTCGenerator_Reset, (inHMSFValue));
	

typedef void			(*MTCGeneratorControlProc)(
							MTCGeneratorRef		inMTCGenerator,
							MTCGeneratorCommand	inCommand,
							TimecodeHMSF		inHMSFResetOffset);

// -----------------------------------------------------------------------------
// MTCGeneratorSetTime_
// -----------------------------------------------------------------------------
//	Creates a MIDI timecode Generator that will broadcast MIDI timecode to
//	the specified MIDI output port

#define	MTCGeneratorSetTime_(izzyParamsPtr, inMTCGenerator, inTimeSeconds) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mMTCGeneratorSetTimeProc)((inMTCGenerator), (inTimeSeconds))

typedef void			(*MTCGeneratorSetTimeProc)(
							MTCGeneratorRef		inMTCGenerator,
							double				inTimeSeconds);

// -----------------------------------------------------------------------------
// MTCGeneratorGetCurrentTime_
// -----------------------------------------------------------------------------
//	Creates a MIDI timecode Generator that will broadcast MIDI timecode to
//	the specified MIDI output port

#define	MTCGeneratorGetCurrentTime_(izzyParamsPtr, inMTCGenerator, outHMSFValue, outSecondsValue) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mMTCGeneratorGetCurrentTimeProc)((inMTCGenerator), (outHMSFValue), (outSecondsValue))

typedef void			(*MTCGeneratorGetCurrentTimeProc)(
							MTCGeneratorRef		inMTCGenerator,
							TimecodeHMSF*		outHMSFValue,
							IzReal*				outSecondsValue);

// -----------------------------------------------------------------------------
// MTCGeneratorSetOffset_
// -----------------------------------------------------------------------------
//	Sets the offset using a timecode value given as a 32 bit value, where the
//	individual bytes = HMSF. The offset is interpreted in the current timecode
//	format (MTCType) used by the generator.

#define	MTCGeneratorSetOffset_(izzyParamsPtr, inMTCGenerator, inHMSFValue) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mMTCGeneratorSetOffsetProc)((inMTCGenerator), (inHMSFValue))

typedef void			(*MTCGeneratorSetOffsetProc)(
							MTCGeneratorRef		inMTCGenerator,
							TimecodeHMSF		inHMSFValue);

// -----------------------------------------------------------------------------
// MTCConvert
// -----------------------------------------------------------------------------
//	Converts a MIDI Timecode value to frames or seconds

#define	MTCConvert_(izzyParamsPtr, inMTCType, inHMSF, outIzTimecode, outFrames, outSeconds) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mMTCConvertProc)((inMTCType), (inHMSF), (outIzTimecode), (outFrames), (outSeconds))

typedef Boolean			(*MTCConvertProc)(
							MTCType				inTCType,
							TimecodeHMSF		inHMSFValue,
							IzValue_Timecode*	outIzTimecode,		// CAN BE NULL
							UInt32*				outFrames,			// CAN BE NULL
							double*				outSeconds);		// CAN BE NULL

// -----------------------------------------------------------------------------
// GetDefaultTimecodeRateForCurrentRegion
// -----------------------------------------------------------------------------

#define	GetDefaultTimecodeRateForCurrentRegion_(izzyParamsPtr) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetDefaultTimecodeRateForCurrentRegionProc)()

typedef TimecodeType	(*GetDefaultTimecodeRateForCurrentRegionProc)();

// -----------------------------------------------------------------------------
// GetShowDefaultTimecodeRate
// -----------------------------------------------------------------------------

#define	ActorGetShowDefaultTimecodeRate_(izzyParamsPtr, inActorInfo, inMTCRatesOnly) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetShowDefaultTimecodeRateProc)((inActorInfo), NULL, (inMTCRatesOnly))

#define	ControlGetShowDefaultTimecodeRate_(izzyParamsPtr, inControlInfo) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetShowDefaultTimecodeRateProc)(NULL, inControlInfo, (inMTCRatesOnly))

typedef TimecodeType	(*GetShowDefaultTimecodeRateProc)(
							const ActorInfo*	inActorInfo,
							const ControlInfo*	inControlInfo,
							bool				inMTCRatesOnly);

// -----------------------------------------------------------------------------
// IzTimecodeConvert
// -----------------------------------------------------------------------------
//	Converts an Isadora Timecode value to frames or seconds

#define	IzActorTimecodeConvert_(izzyParamsPtr, inActorInfo, inIzTimecode, outMTCType, outMTCTimecode, outFrames, outSeconds) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIzTimecodeConvertProc)((inActorInfo), NULL, (inIzTimecode), (outMTCType), (outMTCTimecode), (outFrames), (outSeconds))

#define	IzControlTimecodeConvert_(izzyParamsPtr, inControlInfo, inIzTimecode, outMTCType, outMTCTimecode, outFrames, outSeconds) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIzTimecodeConvertProc)(NULL, inControlInfo, (inIzTimecode), (outMTCType), (outMTCTimecode), (outFrames), (outSeconds))

typedef Boolean			(*IzTimecodeConvertProc)(
							const ActorInfo*	inActorInfo,
							const ControlInfo*	inControlInfo,
							IzValue_Timecode	inTimecode,
							MTCType*			outMTCType,			// CAN BE NULL
							TimecodeHMSF*		outHMSFValue,		// CAN BE NULL
							UInt32*				outFrames,			// CAN BE NULL
							TimecodeSeconds*	outSeconds);		// CAN BE NULL

// -----------------------------------------------------------------------------
// IzTimecodeConvertToString
// -----------------------------------------------------------------------------
//	Converts an Isadora Timecode value to a string. Returns true if the
//	the conversion was successful.

// IS-763: Timecode Control Should Use Default Frame Rate Unless Otherwise Specified
//
// added TCStringMode enum. Previously this parameter was a Boolean (UInt8) so:
//
// kTCStringMode_DontShowRate equates to FALSE
// kTCStringMode_ShowRateAndReplaceDefWithCurrentRate equates to TRUE
//
// which offers backward compatibility

typedef UInt8 TCStringMode;

enum {
	kTCStringMode_DontShowRate = 0,
	kTCStringMode_ShowRateAndReplaceDefWithCurrentRate = 1,
	kTCStringMode_ShowActualRate = 2
};

#define	IzActorTimecodeConvertToString_(izzyParamsPtr, inActorInfo, inIzTimecode, inShowFrameRate, outBufferSize, outBuffer) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIzTimecodeConvertToStringProc)((inActorInfo), NULL, (inIzTimecode), (inShowFrameRate), (outBufferSize), (outBuffer))

#define	IzControlTimecodeConvertToString_(izzyParamsPtr, inControlInfo, inIzTimecode, inShowFrameRate, outBufferSize, outBuffer) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIzTimecodeConvertToStringProc)(NULL, inControlInfo, (inIzTimecode), (inShowFrameRate), (outBufferSize), (outBuffer))

typedef Boolean			(*IzTimecodeConvertToStringProc)(
							const ActorInfo*	inActorInfo,
							const ControlInfo*	inControlInfo,
							IzValue_Timecode	inTimecode,
							TCStringMode		inShowFrameRate,
							UInt32				outBufferSizeBytes,
							char*				outBuffer);

// -----------------------------------------------------------------------------
// ConvertSecondsToIzTimecode
// -----------------------------------------------------------------------------
//	Converts an seconds to either a MidiTimeCode value or an Isadora Timecode Value

#define	ActorConvertSecondsToIzTimecode_(izzyParamsPtr, inActorInfo, inTimecodeType, inSeconds, outIzTimecode) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mConvertSecondsToIzTimecodeProc)((inActorInfo), NULL, (inTimecodeType), (inSeconds), (outIzTimecode))

#define	ControlConvertSecondsToIzTimecode_(izzyParamsPtr, inControlInfo, inTimecodeType, inSeconds, outIzTimecode) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mConvertSecondsToIzTimecodeProc)(NULL, (inControlInfo), (inTimecodeType), (inSeconds), (outIzTimecode))

typedef Boolean			(*ConvertSecondsToIzTimecodeProc)(
							const ActorInfo*	inActorInfo,
							const ControlInfo*	inControlInfo,
							TimecodeType		inTimecodeType,
							double				inSeconds,
							IzValue_Timecode*	outIzTimecode);

// -----------------------------------------------------------------------------
// ConvertFramesToIzTimecode
// -----------------------------------------------------------------------------
//	Converts an seconds to either a MidiTimeCode value or an Isadora Timecode Value

#define	ActorConvertFramesToIzTimecode_(izzyParamsPtr, inActorInfo, inTimecodeType, inFrames, outIzTimecode) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mConvertFramesToIzTimecodeProc)((inActorInfo), NULL, (inTimecodeType), (inFrames), (outIzTimecode))

#define	ControlConvertFramesToIzTimecode_(izzyParamsPtr, inControlInfo, inTimecodeType, inFrames, outIzTimecode) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mConvertFramesToIzTimecodeProc)(NULL, (inControlInfo), (inTimecodeType), (inFrames), (outIzTimecode))

typedef Boolean			(*ConvertFramesToIzTimecodeProc)(
							const ActorInfo*	inActorInfo,
							const ControlInfo*	inControlInfo,
							TimecodeType		inTimecodeType,
							TimecodeFrames		inFrames,
							IzValue_Timecode*	outIzTimecode);		// CAN BE NULL

// -----------------------------------------------------------------------------
// Time Conversion (Percentage/Seconds In/Out)
// -----------------------------------------------------------------------------
// Converts a time percentage for a specific media to a timecode value or,
// or vice-versa. If the media indicated by the inMediaType and inMediaIndex
// is valid, the result will be TRUE, outValue value will be set to a valid
// value, and outTimecodeType will be set to the timecode type for the specified
// media. Otherwise, the result will be FALSE, outValue will be set to 0.0, and
// outTimecodeType will be set to kTimecodeType_Default.

#define	IzActorConvertTimeForMedia_(izzyParamsPtr, inActorInfo, inMediaType, inMediaIndex, inSrcType, inSrcValue, inDstType, outDstValue, outDstTCType) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIzConvertTimeForMedia)((inActorInfo), NULL, (inMediaType), (inMediaIndex), (inSrcType), (inSrcValue), (inDstType), (outDstValue), (outDstTCType))

#define	IzControlConvertTimeForMedia_(izzyParamsPtr, inControlInfo, inMediaType, inMediaIndex, inSrcType, inSrcValue, inDstType, outDstValue, outDstTCType) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIzConvertTimeForMedia)(NULL, (inControlInfo), (inMediaType), (inMediaIndex), (inSrcType), (inSrcValue), (inDstType), (outDstValue), (outDstTCType))

typedef Boolean			(*IzConvertTimeForMedia)(
							const ActorInfo*	inActorInfo,
							const ControlInfo*	inControlInfo,
							MediaCode			inMediaCode,
							MediaIndexT			inMediaIndex1,
							ValueType			inSrcType,
							double				inValue,			// e.g., either float from 0 to 100% or a timecode value
							ValueType			inDstType,			// destination type: kFloat or kTimecode
							double*				outValue,			// the converted value, with type as specified by inDestType
							TimecodeType*		outTimecodeType);	// used when inDestType is kTimecode; can be NULL if result is not required
							
// -----------------------------------------------------------------------------
// ConvertValueToValue
// -----------------------------------------------------------------------------
//	Converts an a Value struct to another type, e.g., from kInteger to kFloat or
//	kFloat to kString, etc. If the conversion is successful, the type member of
//	the returned Value struct will not be kInvalid; if it is kInvalid the
//	conversion failed and the value cannot be used.
//
//	IMPORTANT: if you are converting to a kString, the returned Value will have
//	an a string allocated by AllocateValueString_(). It is up to the caller to
//	release this value or a memory leak will occur!

#define	ActorConvertValueToValue_(izzyParamsPtr, inActorInfo, inValue, inOutputType) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mConvertValueToValueProc)((inActorInfo), NULL, (inValue), (inOutputType))

#define	ControlConvertValueToValue_(izzyParamsPtr, inControlInfo, inValue, inOutputType) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mConvertValueToValueProc)(NULL, (inControlInfo), (inValue), (inOutputType))

typedef Value			(*ConvertValueToValueProc)(
							const ActorInfo*	inActorInfo,
							const ControlInfo*	inControlInfo,
							const Value*		inValue,
							ValueType			inOutputType);

// -----------------------------------------------------------------------------
// Value Time Conversion (ValuePtr In/Out)
// -----------------------------------------------------------------------------
// With a ValuePtr as both a source and destination, converts a time percentage
// for a specific media to seconds, or vice-versa. If the media indicated by the
// inMediaType and inMediaIndex is valid, the result will be TRUE and the
// converted value will be set. Otherwise, the result will be FALSE and the
// outValue will be set a value of 0.0 with the requested type

#define	IzActorConvertTimeValueForMedia_(izzyParamsPtr, inActorInfo, inMediaType, inMediaIndex, inValuePtr, inDestType, outValuePtr) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIzConvertTimeValueForMedia)((inActorInfo), NULL, (inMediaType), (inMediaIndex), (inValuePtr), (inDestType), (outValuePtr))

#define	IzControlConvertTimeValueForMedia_(izzyParamsPtr, inControlInfo, inMediaType, inMediaIndex, inValuePtr, inDestType, outValuePtr) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIzConvertTimeValueForMedia)(NULL, (inControlInfo), (inMediaType), (inMediaIndex), (inValuePtr), (inDestType), (outValuePtr))

typedef Boolean			(*IzConvertTimeValueForMedia)(
							const ActorInfo*	inActorInfo,
							const ControlInfo*	inControlInfo,
							MediaCode			inMediaCode,
							MediaIndexT			inMediaIndex1,
							const ValuePtr		inValue,		// e.g., either float from 0 to 100% or a timecode value
							ValueType			inDestType,		// destination type: kFloat or kTimecode
							ValuePtr			outValue);		// the converted value, with type as specified by inDestType
							
// -----------------------------------------------------------------------------
// DrawAlphaBitmap_ (v3.0)
// -----------------------------------------------------------------------------

#define	DrawAlphaBitmap_(izzyParamsPtr, inGraphics, inAlphaBitmap, dstRect) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mDrawAlphaBitmapProc)((inGraphics), (inAlphaBitmap), NULL, (dstRect), 1.0)

#define	DrawAlphaBitmapSection_(izzyParamsPtr, inGraphics, inAlphaBitmap, srcRect, dstRect) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mDrawAlphaBitmapProc)((inGraphics), (inAlphaBitmap), (srcRect), (dstRect), 1.0)

#define	DrawAlphaBitmapWithOpacity_(izzyParamsPtr, inGraphics, inAlphaBitmap, dstRect, opacity) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mDrawAlphaBitmapProc)((inGraphics), (inAlphaBitmap), NULL, (dstRect), (opacity))

typedef IzzyError	(*DrawAlphaBitmapProc)(
						UGraphics*		inGraphics,
						AlphaBitmap*	inAlphaBitmap,
						const IzRect*	inSrcRect,
						const IzRect*	inDstRect,
						float			inOpacity);


// -----------------------------------------------------------------------------
// GetBitmapDataForAlphaBitmap_ (v3.0)
// -----------------------------------------------------------------------------

#define	GetBitmapDataForAlphaBitmap_(izzyParamsPtr, inAlphaBitmap, outBitmapData, outWidth, outHeight, outRowBytes) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mGetBitmapDataForAlphaBitmapProc)((inAlphaBitmap), (outBitmapData), (outWidth), (outHeight), (outRowBytes))

typedef IzzyError	(*GetBitmapDataForAlphaBitmapProc)(
						AlphaBitmap*	inAlphaBitmap,
						void**			outData,
						SInt32*			outWidth,
						SInt32*			outHeight,
						SInt32*			outRowBytes);


// -----------------------------------------------------------------------------
// PrepareToDrawIntoImageBuffer_ (v3.0)
// UnprepareToDrawIntoImageBuffer_ (v3.0)
// -----------------------------------------------------------------------------
//	If an actor or control wishes to use the standard drawing commmands to
//	draw into a bitmap, it can create a UGraphics context using these calls.
//	Call PrepareToDrawIntoImageBuffer_() to create the graphics context, and
//	UnprepareToDrawIntoImageBuffer_() to release it and update the data change
//	count for the image buffer.

#define PrepareToDrawIntoImageBuffer_(izzyParamsPtr, inImageBufferPtr) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mPrepareToDrawIntoImageBufferProc)((inImageBufferPtr))

typedef UGraphics*	(*PrepareToDrawIntoImageBufferProc)(
						ImageBufferPtr		inImageBuffer);
					
#define UnprepareToDrawIntoImageBuffer_(izzyParamsPtr, inImageBuffer, inGraphicsContext) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mUnprepareToDrawIntoImageBufferProc)((inImageBuffer), (inGraphicsContext))

typedef void		(*UnprepareToDrawIntoImageBufferProc)(
						ImageBufferPtr		inImageBuffer,
						UGraphics*			inGraphics);
						
// -----------------------------------------------------------------------------
// AcquireGlobalDataRef_
// -----------------------------------------------------------------------------
//
// Returns a reference to named global data ref, potentially creating the data
// reference if it does not already exist.
//
// If the named data ref does not yet exist, and if the inOKToCrate flag is true,
// then a block of memory will be created with the specified length and version
// number. The pointer to the data, the version number, the length, for this new
// reference will be passed back to the caller. The return value is the retain
// count, which is always 1 for a newly created object.
//
// If the named data ref does not yet exist, and if the inOKToCrate flag is false,
// the the function returns 0 to indicate that the object does not exist.
//
// If the named data ref already eixsts, its retain count is incremented and the
// pointer to the data, the version number, the length will be passed back to the
// caller. The return value is the incremented retain count.
//
// Note that you should check the data length and version number and use those
// to ensure compatibility for the expected data. It is entirely possible that
// a named data reference created by another client may not have the data length
// and version number passed to AcquireGlobalDataRef.
//
// The convenience function RetainGlobalDataRef_() will retain a named data ref
// if it already exists; in this case, the return value will be > 0. Otherwise,
// if the data ref does not exist, the return value will be 0.
//
// Added for 3.0.9b05

typedef void (*GlobalDataRefReleaseCallback)(
	const char* inDataName,
	void*		inDataRef,
	UInt32		inDataLength,
	UInt32		inDataVersion);

#define AcquireGlobalDataRef_(izzyParamsPtr, inDataName, inDataLength, inDataVersion, inReleaseCallback, inOKToCreate, outDataRef, outDataLength, outDataVersion ) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mAcquireGlobalDataRefProc)((inDataName), (inDataLength), (inDataVersion), (inReleaseCallback), (inOKToCreate), (outDataRef), (outDataLength), (outDataVersion))

typedef UInt32	(*AcquireGlobalDataRefProc)(
						const char*						inDataName,
						UInt32							inDataLength,
						UInt32							inDataVersion,
						GlobalDataRefReleaseCallback	inReleaseCallback,	// CAN BE NULL
						Boolean							inOKToCreate,
						void**							outDataRef,
						UInt32*							outDataLength,
						UInt32*							outDataVersion);
	
#define RetainGlobalDataRef_(izzyParamsPtr, inDataName, outDataRef, outDataLength, outDataVersion ) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mAcquireGlobalDataRefProc)((inDataName), 0, 0, NULL, 0, (outDataRef), (outDataLength), (outDataVersion))


// -----------------------------------------------------------------------------
// ReleaseGlobalDataRef_
// -----------------------------------------------------------------------------
// If the named data reference exists and its retain count is greater than 1,
// then the retain count is simply decremented.
//
// If it's retain count is 1, then the following actions are taken:
//
// 1) If inDeferredDelete is false:
//     -- remove the global refeference from the list of global refeferences
//     -- call the release callback (if specified) and ree the memory allocated
//        for the data ref pointer
//
// 2) If inDeferredDelete is true:
//     -- remove the global refeference from the list of global refeferences
//     -- schedule the reference for deletion as a task on the main thread
//     -- when the main thread task executes, call the release callback (if
//        specified) and ree the memory allocated for the data ref pointer
//
// The return value is the retain count BEFORE releasing the object. If this
// value is greater than 1, then the object still exists; if the value is 
// 1, then the object was just released; if the value is zero, then the object
// did not exist

#define ReleaseGlobalDataRef_(izzyParamsPtr, inDataName, inDeleteOnMainThread) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mReleaseGlobalDataRefProc)((inDataName), (inDeleteOnMainThread))

typedef UInt32	(*ReleaseGlobalDataRefProc)(
					const char*		inDataName,
					Boolean			inDeferredDelete);

// -----------------------------------------------------------------------------
// ADAAudioControlsInfo
// -----------------------------------------------------------------------------
// This structure is used by the AudioWidget_AttemptToHandleClick_ function to
// keep track of the volume/pan widget and/or the audio matrix thumbnail widget
// during tracking.
//
// The caller should initialize this struct to all zeroes, and then fill in the
// mVolumePanWidgetHeight, mMatrixTopOffset, and mMatrixHeight members. If you
// don't want a volume/pan widget, or a matrix thumbnail, set the respective
// height members to zero.

typedef struct {

	IzCoord 		mVolumePanWidgetHeight;		// required input: the height of the volume pan widget, or 0 if none is to be shown
	IzCoord 		mMatrixTopOffset;			// required input: the gap between the volume/pan widget and the audio matrix thumbnail
	IzCoord 		mMatrixHeight;				// required input: the height of the audio matrix thumbnail if one is to be shown
	bool			mPlaying;					// required input: when true, draw the widgets in their active state; otherwise drawn inactive
	
// VOLUME/PAN
	PropertyIndex	mVolumeInputPropertyIndex1;	// required input: the input property index for the volume
	PropertyIndex	mPanInputPropertyIndex1;	// required input: the input property index for the pan
	float			mCurrentVolume;				// required input: the current volume, between 0 and mMaxVolume
	float			mMinVolume;					// required input: the minimum volume
	float			mMaxVolume;					// required input: the maximum volume
	float			mCurrentPan;				// required input: the current pan value
	float			mMinPan;					// required input: the minimum pan
	float			mMaxPan;					// required input: the maximum pan
	
	bool			mTrackingVolume;			// calculated: true while tracking the volume
	IzRect			mLastADAVolumeRect;			// calculated: last rectangle for the volume/pan widget

	IzCoord			mVolumePanSize;				// calculated: current volume/pan size in pixels
	IzCoord			mLastCircleSize;			// calculated: pervious volume/pan size in pixels
	IzCoord			mLastCirclePosition;		// calculated: pervious volume/pan position in pixels
	
	IzPoint			mMouseTrackStart;			// calculated: where mouse tracking started
	IzPoint			mLastMousePoint;			// calculated: the last mouse position during tracking
	Value			mInitialVolume;				// calculated: remembers the initial volume
	Value			mInitialPan;				// calculated: remembers the initial pan
	Value			mNewVolume;					// calculated: the current volume during tracking
	Value			mNewPan;					// calculated: the current pan during tracking
	bool			mFollowVolume;				// calculated: true if we will track the volume
	bool			mFollowPan;					// calculated: true if we will track the pan

// AUDIO MATRIX THUMBNAIL
	
	UInt32			mMatrixSourceChannels;		// required input for the matrix thumbnail: the number of source channels in the matrix
	UInt32			mMatrixDestinationChannels;	// required input for the matrix thumbnail: the number of destinations channels in the matrix
												// 0 means the current number of output channels for the current Isadora output device

	PropertyType	mRoutingPropertyType;		// required input for the matrix thumbnail track: property type for the routing string
	PropertyIndex	mRoutingPropertyIndex1;		// required input for the matrix thumbnail track: property index for the routing string

	const char*		mRoutingString;				// required input for the matrix thumbnail draw: current routing string

	bool			mTrackingMatrix;			// calculated: true while tracking the matrix
	IzRect			mLastADAMatrixRect;			// calculated: last rectangle for the audio matrix widget

} ADAAudioControlsInfo;

// -----------------------------------------------------------------------------
// ADAAudioControls_Draw_
// -----------------------------------------------------------------------------
// Draws the Audio Matrix Thumbnail when the actor's DrawActorDefinedArea()
// callback function is called. The thumbnail represents the current
// multichannel audio routing.
//
// The thumbnail will be drawn into the rectangle described by inDrawRect, offset
// vertically by inTopOffestPixels and with the height specified by inHeightPixels.
// If inPlaying is true, the thumbnail will be drawn in an active state; otherwise
// it will be drawn in an inactive state.
//
// The caller must give a valid value for the number of source channels. The
// number of destination channels can be set to 0, in which case the number
// of channels for the current Isadora audio output device will be used.
//
// Finally, the current audio routing string from the actor should be provided
// by the inRoutingString parameter.
//
// Added for 3.2.3b01

#define ADAAudioControls_Draw_(izzyParamsPtr, inActorInfo, inAudioCtlInfo, inDrawRect) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mADAAudioControls_DrawProc)((inActorInfo), (inAudioCtlInfo), (inDrawRect))

typedef void	(*ADAAudioControls_DrawProc)(
						struct ActorInfo*		inActorInfo,				// actor's ActorInfo struct
						ADAAudioControlsInfo*	inAudioCtlInfo,				// information used to draw and track the ADA audio controls
						IzRect					inDrawRect);				// current routing string

// -----------------------------------------------------------------------------
// ADACreateUndoRedoDataProc (CALLBACK FUNCTION)
// -----------------------------------------------------------------------------
// Passed as a parameter to ADAAudioControls_AttemptToHandleClick_, this callback
// handles the creation of an opaque block of data that will be passed to the
// actor's HandleActorMessage message function, allowing the implementation of
// of undo and redo required when values are changed by the handle click function.

typedef void* (*ADACreateUndoRedoData)(struct ActorInfo* inActorInfo, const ADAAudioControlsInfo* inAudioCtlInfo, void* inRefCon);

// -----------------------------------------------------------------------------
// ADAAudioControls_AttemptToHandleClick_
// -----------------------------------------------------------------------------
// Handles a click in an Audio Matrix Thumbnail drawn in an actor when the
// actor's MouseTrackInActorDefinedArea() callback is called.
//
// The caller should use the same values for inDrawRect, inTopOffestPixels, and
// inHeightPixels as were passed to ADAAudioControls_Draw_()
//
// The caller must give a valid value for the number of source channels. The
// number of destination channels can be set to 0, in which case the number
// of channels for the current Isadora audio output device will be used.
//
// inMouseMessage and inPoint should be the values passed to the actor by the
// MouseTrackInActorDefinedArea() function.
//
// inDestPropertyType and inDestPropertyIndex1 give identify the property that
// will receive the audio routing string when the user closes the Audio Matrix
// widget.
//
// Added for 3.2.3b01

#define ADAAudioControls_AttemptToHandleClick_(izzyParamsPtr, inActorInfo, inAudioCtlInfo, inDrawRect, inMouseMessage, inPoint, createUndoRedoProc, createUndoRedoProcRefCon) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mADAAudioControls_AttemptToHandleClickProc)((inActorInfo), (inAudioCtlInfo), (inDrawRect), (inMouseMessage), (inPoint), (createUndoRedoProc), (createUndoRedoProcRefCon))

typedef Boolean	(*ADAAudioControls_AttemptToHandleClickProc)(
						struct ActorInfo*		inActorInfo,				// actor's ActorInfo struct
						ADAAudioControlsInfo*	inAudioCtlInfo,				// information used to draw and track the ADA audio controls
						IzRect					inDrawRect,					// rect into which the matrix selector will be drawn
						IzzyMouseMessage		inMouseMessage,				// the IzzyMouseMessage received by the
						IzPoint					inPoint,					// destination property index for the routing string
						ADACreateUndoRedoData	inCreateUndoRedoProc,		// callback to create undo/redo data -- can be NULL for no undo redo functionality
						void*					inCreateUndoRedoRefCon);	// refcon passed to CreateUndoRedoProc

// -----------------------------------------------------------------------------
// IzzyAudio_GetOutputChannels_
// -----------------------------------------------------------------------------
// Returns the number of output channels for the current audio device selected
// in the Audio View of the Isadora document that owns the actor.
//
// Added for 3.2.3b01

#define IzzyAudio_GetOutputChannels_(izzyParamsPtr, inActorInfo) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIzzyAudio_GetOutputChannelsProc)((inActorInfo))

typedef UInt32	(*IzzyAudio_GetOutputChannelsProc)(
						struct ActorInfo*			inActorInfo);		// actor's ActorInfo struct

// -----------------------------------------------------------------------------
// IzzyAudio_GetSampleRate_
// -----------------------------------------------------------------------------
// Returns the sample rate for the current audio device selected in the Audio
// View of the Isadora document that owns the actor.
//
// Added for 3.2.3b01

#define IzzyAudio_GetSampleRate_(izzyParamsPtr, inActorInfo) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIzzyAudio_GetSampleRateProc)((inActorInfo))

typedef UInt32	(*IzzyAudio_GetSampleRateProc)(
						struct ActorInfo*			inActorInfo);		// actor's ActorInfo struct

// -----------------------------------------------------------------------------
// IzzyAudioCallbackProc (CALLBACK FUNCTION)
// -----------------------------------------------------------------------------
// Passed as a parameter to IzzyAudio_StartAudioStream_, this callback needs to
// provide audio samples to the audio stream using the number of channels and
// sample rate provided to IzzyAudio_StartAudioStream_() when the stream was
// created.
//
// The function should return the number of samples written to the buffer. If
// this number is less than the number of samples requested by inSamples, then
// it is assumed that all samples have been played and the stream will stop
// playing.
//
// If the client does not have enough samples to fill the entire buffer, it is
// the client's responsibility to set all remaining samples to 0.0

// the audio notification message enumeration allows the callback client
// to determine which struct within the IzzyAudioNotification to use
enum IzzyAudioNotificationMsg
{
	kIzzyAudio_GetSamples		// indicates the IzzyAudioNotification's mGetSamples member is valid
};

// when the mNotificationMsg member of the IzzyAudioNotification
// struct is kIzzyAudio_GetSamples, this member of the union u
// will be valid.
typedef struct {
	float*					mSampleBuf;			// input: the buffer that will receive the audio samples
	unsigned int			mSamplesRequested;	// input: the number of samples the caller would like to receive
	unsigned int			mChannelOffset;		// input: number of samples between channels in non-interleaved output
	unsigned int			mSamplesWritten;	// output: the client must fill in the number of samples written
} IzzyAudioGetSamples;

// for this version of the SDK, there is only version 1
// of the IzzyAudioNotification struct
enum {
	IzzyAudioNotificationVersion_Initial = 1,
	IzzyAudioNotificationVersion_Current = IzzyAudioNotificationVersion_Initial
};

// the IzzyAudioNotification struct allows a flexible way to send notifications
// to the IzzyAudioCallbackProc. The client should examine the mNotificationMsg
// member to determine which struct within the union u is valid for the callback.

typedef struct {
	unsigned int				mNotificationVersion;	// version number so clients can be sure of the the structure
	IzzyAudioNotificationMsg	mNotificationMsg;		// see IzzyAudioNotification above
	union {
		IzzyAudioGetSamples		mGetSamples;			// valid if mNotification = kIzzyAudio_GetSamples;
	};
} IzzyAudioNotification;

typedef void (*IzzyAudioCallbackProc)(
	struct ActorInfo*		inActorInfo,			// pointer to the actor that owns this reference
	void*					inCallbackRefCon,		// callback ref con passed to IzzyAudio_CreateAudioStreamProc
	IzzyAudioNotification*	inNotification);		// see IzzyAudioNotification above

// -----------------------------------------------------------------------------
// IzzyAudioStream [OPAQUE PTR]
// -----------------------------------------------------------------------------
// Used to refernece an audio stream created by IzzyAudio_CreateAudioStream_

typedef struct OpaqueIzzyAudioStreamRef* IzzyAudioStreamRef;

// -----------------------------------------------------------------------------
// IzzyAudio_CreateAudioStream_
// -----------------------------------------------------------------------------

#define IzzyAudio_CreateAudioStream_(izzyParamsPtr, inActorInfo, inGetSamplesCallback, inGetSamplesRefCon, inChannels, inSampleRate, inInitVolume, inInitPan, inStartPlaying) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIzzyAudio_CreateAudioStreamProc)((inActorInfo), (inGetSamplesCallback), (inGetSamplesRefCon), (inChannels), (inSampleRate), (inInitVolume), (inInitPan), (inStartPlaying))

typedef IzzyAudioStreamRef	(*IzzyAudio_CreateAudioStreamProc)(
								struct ActorInfo*		inActorInfo,			// actor's ActorInfo struct
								IzzyAudioCallbackProc	inGetSamplesCallback,	// callback provides audio samples
								void*					inGetSamplesRefCon,		// pointer passed back to IzzyAudioCallbackProc
								unsigned int			inChannels,				// number of channels in the audio stream
								float					inSampleRate,			// sample rate of the audio stream
								float					inInitialVolume,		// initial volume
								float					inInitialPan,			// initial pan
								Boolean					inStartPlaying);		// if true, start playing immediately
						
// -----------------------------------------------------------------------------
// IzzyAudio_DisposeAudioStream_
// -----------------------------------------------------------------------------

#define IzzyAudio_DisposeAudioStream_(izzyParamsPtr, inIzzyAudioStream) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIzzyAudio_DisposeAudioStreamProc)((inIzzyAudioStream))

typedef void				(*IzzyAudio_DisposeAudioStreamProc)(
								IzzyAudioStreamRef		inStream);			// isadora audio stream reference
						
// -----------------------------------------------------------------------------
// IzzyAudio_StartAudioStream
// -----------------------------------------------------------------------------

#define IzzyAudio_StartAudioStream_(izzyParamsPtr, inIzzyAudioStream) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIzzyAudio_StartAudioStreamProc)((inIzzyAudioStream))

typedef void				(*IzzyAudio_StartAudioStreamProc)(
								IzzyAudioStreamRef		inStream);			// isadora audio stream reference
						
						
// -----------------------------------------------------------------------------
// IzzyAudio_StopAudioStream
// -----------------------------------------------------------------------------

#define IzzyAudio_StopAudioStream_(izzyParamsPtr, inIzzyAudioStream) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIzzyAudio_StopAudioStreamProc)((inIzzyAudioStream))

typedef void				(*IzzyAudio_StopAudioStreamProc)(
								IzzyAudioStreamRef		inStream);			// isadora audio stream reference


// -----------------------------------------------------------------------------
// IzzyAudio_SetVolumeAndPan
// -----------------------------------------------------------------------------

#define IzzyAudio_SetVolumeAndPan_(izzyParamsPtr, inIzzyAudioStream, inVolume, inPan) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIzzyAudio_SetVolumeAndPanProc)((inIzzyAudioStream), (inVolume), (inPan))

typedef void				(*IzzyAudio_SetVolumeAndPanProc)(
								IzzyAudioStreamRef		inStream,			// isadora audio stream reference
								float					inVolume,			// 1.0 = unity, 0.0 = silent, 2.0 = twice as loud, etc.
								float					inPan);				// 0.5 = center, 0.0 = left, 1.0 = right

// -----------------------------------------------------------------------------
// IzzyAudio_SetRouting
// -----------------------------------------------------------------------------

#define IzzyAudio_SetRouting_(izzyParamsPtr, inIzzyAudioStream, inRoutingString) \
	(*(izzyParamsPtr)->kConditionalCallbacks->mIzzyAudio_SetRoutingProc)((inIzzyAudioStream), (inRoutingString))

typedef void				(*IzzyAudio_SetRoutingProc)(
								IzzyAudioStreamRef		inStream,			// isadora audio stream reference
								const char*				inRoutingString);	// routing string


////////////////////////////////////////////////////////////////////////////
// VERSION_NUMBER MACRO
////////////////////////////////////////////////////////////////////////////
//
// the VERSION_NUMBER macro builds a version number from the following
// components:
//
// major version (0 to 63, 6 bits)
// minor version (0 to 255, 8 bits)
// point release (0 to 255, 8 bits)
// release type (0 to 3, 2 bits -- kReleaseType_Development to kReleaseType_Final)
// build number (0 to 255, 8 bits)
//
// The result is a 32 bit unsigned number that allows comparison, ensuring that
// an alpha or beta version of a build is always less than the equivalently 
// numbered final build, e.g., 3.0.9b05 is less than 3.0.9f05

#define kReleaseType_Development	0
#define kReleaseType_Alpha			1
#define kReleaseType_Beta			2
#define kReleaseType_Final			3

#define VERSION_NUMBER(major, minor, point, rt, build) (\
	  ((UInt32) (major) << 26) \
	| ((UInt32) (minor) << 18) \
	| ((UInt32) (point) << 10) \
	| ((UInt32) (rt) << 8) \
	| ((UInt32) (build) << 0))

// -----------------------------------------------------------------------------
// callback function structure

enum {

	// for Isadora 3.0.8b38 and earlier, we did not store the Isadora Callbacks version
	kIsadoraCallbacksVersion_308b38andEarlier = 0,
	
	// the values for kIsadoraCallbacksVersion_308b39 through kIsadoraCallbacksVersion_309b01 
	// were calculated using a less robust method than provided by the VERSION_NUMBER macro;
	// nevertheless, these still work with later versions since they are less than the versions
	// calculated by the VERSION_NUMBER macro
	#define kIsadoraCallbackBetaFlag	0
	#define kIsadoraCallbackFinalFlag	128
	kIsadoraCallbacksVersion_308b39 = (0x03 << 24) | (0x08 << 16) | kIsadoraCallbackFinalFlag | (39),	// 0x030827
	kIsadoraCallbacksVersion_308b47 = (0x03 << 24) | (0x08 << 16) | kIsadoraCallbackBetaFlag  | (47),	// 0x03082F
	kIsadoraCallbacksVersion_308f12 = (0x03 << 24) | (0x08 << 16) | kIsadoraCallbackFinalFlag | (12),	// 0x03088C
	kIsadoraCallbacksVersion_309b01 = (0x03 << 24) | (0x09 << 16) | kIsadoraCallbackBetaFlag  | (1),	// 0x030901

	// starting with Isadora 3.0.9b05 we calculated the version number using the
	// VERSION_NUMBER macro, which is also used within Isadora itself when generating
	// 32-bit version numbers
	kIsadoraCallbacksVersion_309b05 = kActorCompatilibilityVersion_v309b05,		// 3.0.9b5

	// Added SetPropertyIsInvalidFlagProc
	kIsadoraCallbacksVersion_309b06 = kActorCompatilibilityVersion_v309b06,		// 3.0.9b6

	// Added GetUTF8PathToTroikaTronixFolderProc
	kIsadoraCallbacksVersion_312b10 = kActorCompatilibilityVersion_v312b10,		// 3.1.2b10

	// IS-1250 - Code Editor Syntax Coloring Is Now Language Aware
	kIsadoraCallbacksVersion_322b50 = kActorCompatilibilityVersion_v322b50,		// 3.2.2b50

	// Audio Matrix Router - Actor Defined Area Additions
	kIsadoraCallbacksVersion_323b01 = kActorCompatilibilityVersion_v323b01,		// 3.2.3b01

	kIsadoraCallbacksVersion_Current = kIsadoraCallbacksVersion_323b01
};

// CALLBACK VERSION CHECK MACRO: you can use this macro to ensure that
// the callback version number is at least a certain version, e.g.,
// if (IsAtLeastIsadoraCallbackVersion(ip, kIsadoraCallbacksVersion_309b06)) {
//     ... do something
// }
#define IsAtLeastIsadoraCallbackVersion(inIsadoraParameters, inCallbackVersion) ((ip)->mCallbacks->mCallbackVersionNumber >= (inCallbackVersion))

struct IsadoraCallbacks {

	ShowAlertProc								mShowAlertProc;
	
	CreateMessageReceiverProc					mCreateMessageReceiverProc;
	DisposeMessageReceiverProc					mDisposeMessageReceiverProc;
	UpdateMessageReceiverProc					mUpdateMessageReceiverProc;
	UpdateMessageReceiverMaskProc				mUpdateMessageReceiverMaskProc;

	ChangePropertyTypeProc						mChangePropertyTypeProc;
	GetInputPropertyValueProc					mGetInputPropertyValueProc;
	SetOutputPropertyValueProc					mSetOutputPropertyValueProc;
	GetOutputPropertyValueProc					mGetOutputPropertyValueProc;
	GetOutputPropertyMinimumValueProc			mGetOutputPropertyMinimumValueProc;
	GetOutputPropertyMaximumValueProc			mGetOutputPropertyMaximumValueProc;

	QueueMediaCommandProc						mQueueMediaCommandProc;
	ExecuteMediaCommandProc						mExecuteMediaCommandProc;
	SendMidiMessageProc							mSendMidiMessageProc;

	void*										REMOVED_mCopyImageBufferToStageProc;
	GetStageBoundsProc							mGetStageBoundsProc;
	
	// for Isadora v3.0.0 through v3.0.7 and prior the following member was
	// void* REMOVED_mInvalAreaOnStageProc. In those versions the value of
	// REMOVED_mInvalAreaOnStageProc would have been NULL. As of Isadora
	// 3.0.8b39 and later, this spot now contains the version number of the
	// Isadora callbacks structure
	uint64_t									mCallbackVersionNumber;

	IsIntermediaryImageBufferRequired			mIsIntermediaryImageBufferRequiredProc;
	CreateComplimentaryImageBufferProc			mCreateComplimentaryImageBufferProc;
	DisposeImageBufferProc						mDisposeImageBufferProc;
	CopyImageBufferProc							mCopyImageBufferProc;
	ProcPtr										mNotUsed1;
	
	JumpToSceneProc								mJumpToSceneProc;
	
	PropertyTypeAndIndexToHelpIndexProc			mPropertyTypeAndIndexToHelpIndexProc;
	
	AddPropertyProc								mAddPropertyProc;
	RemovePropertyProc							mRemovePropertyProc;
	GetPropertyCountProc						mGetPropertyCount;
	
	GetPropertyMinMaxProc						mGetPropertyMinMaxProc;
	SetPropertyMinMaxProc						mSetPropertyMinMaxProc;

	GetInputPropertyScaleToMinMaxProc_DEPRECATED	mGetInputPropertyScaleToMinMaxProc_DEPRECATED;
	SetInputPropertyScaleToMinMaxProc_DEPRECATED	mSetInputPropertyScaleToMinMaxProc_DEPRECATED;

	AllNotesOffProc								mAllNotesOffProc;
	
	GetSceneNameProc							mGetSceneNameProc;
	ActivateDeactivateSceneProc					mActivateDeactivateSceneProc;
	
	GetPluginResourcePICTProc					mGetPluginResourcePICTProc;
	
	IzzyMallocProc								mIzzyMallocProc;
	
	IzzyFreeProc								mIzzyFreeProc;
	
	LookupPartialStringInListProc				mLookupPartialStringInListProc;
	
	ModKeyStringToMaskProc						mModKeyStringToMaskProc;
	
	ModKeyMaskToStringProc						mModKeyMaskToStringProc;
	
	GetScreenBoundsProc							mGetScreenBoundsProc;
	
	SendMidiNoteOnOffProc						mSendMidiNoteOnOffProc;
	
	GetMidiDancerEnergyProc						mGetMidiDancerEnergyProc;
	
	GetTextFromDialogProc						mGetTextFromDialogProc;
	
	PluginAssertProc							mPluginAssertProc;
	
	PrepareActorDefinedAreaPictProc				mPrepareActorDefinedAreaPictProc;
	
	DrawActorDefinedAreaPictProc				mDrawActorDefinedAreaPictProc;

	WarnLowMemoryProc							mWarnLowMemoryProc;

	RecallSnapshotProc							mRecallSnapshotProc;
	
	CreateImageBufferProc						mCreateImageBufferProc;
	
	EnterVideoProcessingProc					mEnterVideoProcessingProc;
	
	ExitVideoProcessingProc						mExitVideoProcessingProc;
	
	BroadcastControlValueProc					mBroadcastControlValueProc;
	
	DrawTextInRectProc							mDrawTextInRectProc;

	GetMediaIndexListFromDialogProc				mGetMediaIndexListFromDialogProc;
	
	GetMediaObjectInfoProc						mGetMediaObjectInfoProc;
	
	BroadcastControlValueDirectProc				mBroadcastControlValueDirectProc;
	
	GetPropertyDisplayFormatsProc				mGetPropertyDisplayFormatsProc;

	SetInputPropertyValueProc					mSetInputPropertyValueProc;
	
	CancelMidiNoteOffProc						mCancelMidiNoteOffProc;
	
	GetRandomSeedProc							mGetRandomSeed;
	
	SetCaptureModeProc							mSetCaptureModeProc;
	
	void*										REMOVED_mCanRenderToStageProc;
	
	EncodeStringProc							mEncodeStringProc;

	BroadcastValueProc							mBroadcastValueProc;
	
	void*										REMOVED_mCopyImageBufferToRecordedOutputProc;

	PluginLogInfoProc							mPluginLogInfoProc;
	
	StartRecordCaptureInputToDiskProc			mStartRecordCaptureInputToDiskProc;
	
	StopRecordCaptureInputToDiskProc			mStopRecordCaptureInputToDiskProc;
	
	GetMidiDancerEnergyPlusProc					mGetMidiDancerEnergyPlusProc;
	
	EnableDisableMIDIProc						mEnableDisableMIDIProc;
	
	ShowNumericValueEditBoxProc					mShowNumericValueEditBoxProc;
	
	void*										REMOVED_mDrawPascalStringProc;
	void*										REMOVED_mDrawCStringProc;
	void*										REMOVED_mDrawTextProc;
	
	void*										REMOVED_mTextWidthPascalStringProc;
	void*										REMOVED_mTextWidthCStringProc;
	void*										REMOVED_mTextWidthProc;
	
	GetBinNameAndTypeFromUniqueIDProc			mGetBinNameAndTypeFromUniqueIDProc;
	CountMediaObjectsInBinProc					mCountMediaObjectsInBinProc;
	GetMediaObjectInfoFromBinProc				mGetMediaObjectInfoFromBinProc;
	GetBinNameAndUniqueIDFromIndexProc			mGetBinNameAndUniqueIDFromIndexProc;

	void*										REMOVED_mValidAreaOnStageProc;
	
	SmartSetPropertyMinMaxProc					mSmartSetPropertyMinMaxProc;
	
	EnableDisableActorsProc						mEnableDisableActorsProc;
	
	GetInfoForIndexedBinProc					mGetInfoForIndexedBinProc;
	
	TrackPopupMenuProc							mTrackPopupMenuProc;
	
	GetAggregateBufferFormatProc				mGetAggregateBufferFormatProc;
	
	CreateAndRetainColorImageBufferProc			mCreateAndRetainColorImageBufferProc;
	
	IPAddressStringToInetAddressProc			mIPAddressStringToInetAddressProc;
	SendOSCValuesProc							mSendOSCValuesProc;
	
	SetSceneIntensityProc						mSetSceneIntensityProc;

	GetSceneIntensityProc						mGetSceneIntensityProc;
	
	OutputTextFileProc							mOutputTextFileProc;
	
	GetMediaIndexForItemInBinProc				mGetMediaIndexForItemInBinProc;
	
	WriteToSerialPortProc						mWriteToSerialPortProc;
	
	ReadFromSerialPortProc						mReadFromSerialPortProc;
	
	SetLocalTranslationRotationAndFOVProc		mSetLocalTranslationRotationAndFOVProc;
	
	AllocateValueStringProc						mAllocateValueStringProc;

	ReleaseValueStringProc						mReleaseValueStringProc;
	
	ControlAcquireImageProc						mControlAcquireImageProc;
	ControlReleaseImageProc						mControlReleaseImageProc;
	
	GetUserNameProc								mGetUserNameProc;
	
	GetActorDefaultPixelFormatProc				mGetActorDefaultPixelFormatProc;
	
	GetActorGraphicsProc						mGetActorGraphicsProc;
	
	GetControlGraphicsProc						mGetControlGraphicsProc;
	
	SetForegroundColorRefProc					mSetForegroundColorRefProc;
	DrawLineProc								mDrawLineProc;
	PaintRectProc								mPaintRectProc;
	DrawShadedBoxPercentProc					mDrawShadedBoxPercentProc;
	InvalidateControlProc						mInvalidateControlProc;
	Draw3DRectProc								mDraw3DRectProc;
	DrawTextAtProc								mDrawTextAtProc;
	GetTextWidthProc							mGetTextWidthProc;
	GetFontInfoProc								mGetFontInfoProc;
	SetupFontProc								mSetupFontProc;
	
	TestMTCTimecodeTimeProc						mTestMTCTimecodeTimeProc;
	
	SetOutputPropertyLimitMinMaxProc			mSetOutputPropertyLimitMinMaxProc;
	
	SetLocalLightingProc						mSetLocalLightingProc;
	
	void*										mDummy1;

	SendLanBoxProc								mSendLanBoxProc;
	
	EditLanBoxRectListDialog					mEditLanBoxRectListDialogProc;

	GetPropertyIDProc							mGetPropertyIDProc;

	ConvertTimeToPercentForMediaObjectProc		mConvertTimeToPercentForMediaObject;
	
	ConvertPercentToTimeForMediaObjectProc		mConvertPercentToTimeForMediaObject;
	
	EraseRectProc								mEraseRectProc;
	
	GetFPSProc									mGetFPSProc;
	
	MarkControlDocumentModifiedProc				mMarkControlDocumentModifiedProc;
	MarkActorDocumentModifiedProc				mMarkActorDocumentModifiedProc;

	BroadcastValueFromControlProc				mBroadcastValueFromControlProc;
	
	SetActorUndoActionProc						mSetActorUndoActionProc;
	SetControlUndoActionProc					mSetControlUndoActionProc;
	
	ShowHideStageProc							mShowHideStageProc;
	
	PaintFramePolyProc							mPaintFramePolyProc;
	
	DrawNativeBitmapProc						mDrawNativeBitmapProc;
	
	DrawPictureProc								mDrawPictureProc;
	
	PushPopSetClipProc							mPushPopSetClipProc;
	
	SetLocalScalingOptionsProc					mSetLocalScalingOptionsProc;
	
	void*										REMOVED_mSetLocalStageBackgroundOptionsProc;
	
	GetPerformanceDataProc						mGetPerformanceDataProc;

	FileSavedOnBigEndianMachineProc				mFileSavedOnBigEndianMachineProc;
	
	PostEditActorPrivateDataActionProc			mPostEditActorPrivateDataActionProc;

	ChangePropertyTypeAndNameProc				mChangePropertyTypeAndNameProc;
	
	GetNativeBitmapBoundsProc					mGetNativeBitmapBoundsProc;
	
	DrawImageBufferProc							mDrawImageBufferProc;
	
	GetPropertyConnectionInfoProc				mGetPropertyConnectionInfoProc;
	
	LinkPropertiesProc							mLinkPropertiesProc;
	
	TakeSnapshotProc							mTakeSnapshotProc;
	
	UpdateSnapshotProc							mUpdateSnapshotProc;
	
	SetActorFlagsProc							mSetActorFlagsProc;
	
	WriteToLogFileProc							mWriteToLogFileProc;
	
	GetUTF8PathToCurrentDocumentProc			mGetUTF8PathToCurrentDocumentProc;

	SetTCPStreamInfoProc						mSetTCPStreamInfoProc;
	
	EnableDisableTCPStreamProc					mEnableDisableTCPStreamProc;
	
	WriteToTCPStreamProc						mWriteToTCPStreamProc;
	
	ReadFromTCPStreamProc						mReadFromTCPStreamProc;
	
	void*										REMOVED_mGetStagePreviewFromControlProc;
	
	ActorControlEnableStagePreview				mActorControlEnableStagePreview;
	
	StartStopRecordingLanBoxDMXInputProc		mStartStopRecordingLanBoxDMXInputProc;
	
	GetSceneIndexProc							mGetSceneIndexProc;
	
	BroadcastStreamingDataProc					mBroadcastStreamingDataProc;

	GetChannelMapDialogProc						mGetChannelMapDialogProc;
	
	SetPropertyNameProc							mSetPropertyNameProc;
	
	CopyPropDefValueSourceProc					mCopyPropDefValueSourceProc;
	
	IsControlVisibleProc						mIsControlVisibleProc;
	
	PreloadSceneProc							mPreloadSceneProc;
	
	GetAudioDeviceListProc						mGetAudioDeviceListProc;
	
	AudioDeviceNameToAudioDeviceIDProc			mAudioDeviceNameToAudioDeviceIDProc;
	
	void*										REMOVED_mGetFamilyAndStyleFromFullFontNameProc;
	
	SetGlobalMoviePlaybackOptionsProc			mSetGlobalMoviePlaybackOptionsProc;
	
	SetLocalBackgroundLightingProc				mSetLocalBackgroundLightingProc;
	
	FillImageBufferProc							mFillImageBufferProc;
	
	ReleaseColorImageBufferProc					mReleaseColorImageBufferProc;
	
	MediaObjectIndexToUniqueIDProc				mMediaObjectIndexToUniqueIDProc;
	
	UniqueIDToMediaObjectIndexProc				mUniqueIDToMediaObjectIndexProc;
	
	SendArbitraryOSCDataProc					mSendArbitraryOSCDataProc;
	
	GetLocalTranslationRotationAndFOVProc		mGetLocalTranslationRotationAndFOVProc;
	
	ActorDefinedAreaSizeChangedProc				mActorDefinedAreaSizeChangedProc;
	
	CopyImageBufferInvertColorsProc				mCopyImageBufferInvertColorsProc;
	
	SetForceCursorVisibleOnStageProc			mSetForceCursorVisibleOnStageProc;
	
	GetCurrentMTCProc							mGetCurrentMTCProc;
	
	SendLongMidiMessageProc						mSendLongMidiMessageProc;
	
	PrintMessageToMonitorWindowProc				mPrintMessageToMonitorWindowProc;
	
	UpdateTextureBufferWithOwnedTextureProc		mUpdateTextureBufferWithOwnedTextureProc;
	
	DisposeTextureBufferWithOwnedTextureProc	mDisposeTextureBufferWithOwnedTextureProc;
	
	GetNativeGLContextForStageProc				mGetNativeGLContextForStageProc;
	
	StartRecordStageToDiskProc					mStartRecordStageToDiskProc;
	
	StopRecordStageToDiskProc					mStopRecordStageToDiskProc;
	
	GetMediaObjectPropertiesProc				mGetMediaObjectPropertiesProc;
	
	GetTCPStreamStatusProc						mGetTCPStreamStatusProc;
	
	AddRemoveMediaFromLoadedMediaListProc		mAddRemoveMediaFromLoadedMediaListProc;
	
	AcquireIzzyMoviePlayerTaskProc				mAcquireIzzyMoviePlayerTaskProc;
	
	RelinquishIzzyMoviePlayerTaskProc			mRelinquishIzzyMoviePlayerTaskProc;
	
	MutatePropertyProc							mMutatePropertyProc;
	
	GetOutputPropertyLimitMinMaxProc			mGetOutputPropertyLimitMinMaxProc;
	
	SetPropertyNotApplicableFlagProc			mSetPropertyNotApplicableFlagProc;
	
	GetColorRefsForCurrentThemeProc				mGetColorRefsForCurrentThemeProc;
	
	GetTextHeightProc							mGetTextHeightProc;

	IsModernGUIActiveProc						mIsModernGUIActiveProc;
	
	DrawColorRefFeedbackInRectProc				mDrawColorRefFeedbackInRectProc;
	
	CreateIzzyFrameBufferProc					mCreateIzzyFrameBufferProc;
	DisposeIzzyFrameBufferProc					mDisposeIzzyFrameBufferProc;
	UpdateIzzyFrameBufferProc					mUpdateIzzyFrameBufferProc;
	BindIzzyFrameBufferProc						mBindIzzyFrameBufferProc;
	UnbindIzzyFrameBufferProc					mUnbindIzzyFrameBufferProc;

	MovePropertyProc							mMovePropertyProc;

	GetPropertyNameProc							mGetPropertyNameProc;
	
	SimulateActorDoubleClickProc				mSimulateActorDoubleClickProc;
	
	GetTextFromDialogAdvancedProc				mGetTextFromDialogAdvancedProc;
	
	void*										REMOVED_mBitmapDrawFramedPolygonProc;
	
	SetGlobalVariableValueProc					mSetGlobalVariableValueProc;

	GetGlobalVariableValueProc					mGetGlobalVariableValueProc;
	
	RetainValueStringProc						mRetainValueStringProc;
	
	PrepareUnprepareToDrawActorInUserInterfaceProc		mPrepareUnprepareToDrawActorInUserInterfaceProc;
	
	PrepareUnprepareToDrawControlInUserInterfaceProc	mPrepareUnprepareToDrawControlInUserInterfaceProc;
	
	SetPropertyAllowedMutableFormatsProc		mSetPropertyAllowedMutableFormatsProc;
	
	GetPropertyAllowedMutableFormatsProc		mGetPropertyAllowedMutableFormatsProc;
	
	ConvertVideoStreamToBitmapProc				mConvertStreamVideoToBitmapProc;
	
	DisposeIzzyBitmapConverterProc				mDisposeIzzyBitmapConverterProc;
	
	ConvertBitmapToVideoStreamProc				mConvertBitmapToVideoStreamProc;
	
	GetActorGraphicScalingProc					mGetActorGraphicScalingProc;
	
	GetControlGraphicScalingProc				mGetControlGraphicScalingProc;
	
	DrawActorDefinedAreaPropertyIndicatorProc	mDrawActorDefinedAreaPropertyIndicatorProc;
	
	GetWidthForActorDefinedAreaPropertyIndicatorProc	mGetWidthForActorDefinedAreaPropertyIndicatorProc;

// v2.3
	IzzyNewHandleProc								mIzzyNewHandleProc;
	IzzyDisposeHandleProc							mIzzyDisposeHandleProc;
	IzzyGetHandleSizeProc							mIzzyGetHandleSizeProc;
	IzzyLockHandleProc								mIzzyLockHandleProc;
	IzzyUnlockHandleProc							mIzzyUnlockHandleProc;
	IzzySetHandleSizeProc							mIzzySetHandleSizeProc;

	SetPropertyTimecodeFPSProc_DEPRECATED			DEPRECATED_mSetPropertyTimecodeFPSProc;
	
	CreateIzzyUnicodeStringWithUTF8TextProc			mCreateIzzyUnicodeStringWithUTF8TextProc;
	DisposeIzzyUnicodeStringProc					mDisposeIzzyUnicodeStringProc;
	IzzyUnicodeStringLengthProc						mIzzyUnicodeStringLengthProc;
	IzzyUnicodeStringLineCountProc					mIzzyUnicodeStringLineCountProc;
	CreateUTF8StringWithIzzyUnicodeStringProc		mCreateUTF8StringWithIzzyUnicodeStringProc;
	IzzyUnicodeStringCompareProc					mIzzyUnicodeStringCompareProc;
	IzzyUnicodeStringGetSubstringProc				mIzzyUnicodeStringGetSubstringProc;

// v2.5
	MTCGeneratorCreateProc							mMTCGeneratorCreateProc;
	MTCGeneratorDisposeProc							mMTCGeneratorDisposeProc;
	MTCGeneratorControlProc							mMTCGeneratorControlProc;
	MTCGeneratorSetTimeProc							mMTCGeneratorSetTimeProc;
	MTCGeneratorGetCurrentTimeProc					mMTCGeneratorGetCurrentTimeProc;
	MTCGeneratorSetOffsetProc						mMTCGeneratorSetOffsetProc;
	
// v2.6
	GetSceneInfoProc								mGetSceneInfoProc;
	JumpToSceneWithBuiltInFadeProc					mJumpToSceneWithBuiltInFadeProc;
	CaptureStageToPictureFileProc					mCaptureStageToPictureFileProc;
	JumpToCueProc									mJumpToCueProc;
	GetOSCChanelAddressProc							mGetOSCChanelAddressProc;
	LogErrorToSystemConsoleProc						mLogErrorToSystemConsoleProc;

// v3.0
	DrawAlphaBitmapProc								mDrawAlphaBitmapProc;
	GetBitmapDataForAlphaBitmapProc					mGetBitmapDataForAlphaBitmapProc;
	PrepareToDrawIntoImageBufferProc				mPrepareToDrawIntoImageBufferProc;
	UnprepareToDrawIntoImageBufferProc				mUnprepareToDrawIntoImageBufferProc;
	GetNativeContextInfoProc						mGetNativeContextInfoProc;
	
	MTCConvertProc									mMTCConvertProc;
	IzTimecodeConvertProc							mIzTimecodeConvertProc;
	ConvertSecondsToIzTimecodeProc					mConvertSecondsToIzTimecodeProc;
	ConvertFramesToIzTimecodeProc					mConvertFramesToIzTimecodeProc;
	GetShowDefaultTimecodeRateProc					mGetShowDefaultTimecodeRateProc;
	GetDefaultTimecodeRateForCurrentRegionProc		mGetDefaultTimecodeRateForCurrentRegionProc;

	IzConvertTimeForMedia							mIzConvertTimeForMedia;
	IzConvertTimeValueForMedia						mIzConvertTimeValueForMedia;

	GetPropertyScaleToMinMaxProc					mGetInputPropertyScaleToMinMaxProc;
	SetPropertyScaleToMinMaxProc					mSetInputPropertyScaleToMinMaxProc;
	SetPropertyTimecodeTypeProc						mSetPropertyTimecodeTypeProc;
	
	InvalidateActorDefinedAreaProc					mInvalidateActorDefinedAreaProc;
	
	GetStageNameListProc							mGetStageNameListProc;
	GetIndexedStageNameProc							mGetIndexedStageNameProc;
	HandleStageValueToStringProc					mHandleStageValueToStringProc;
	HandleStageStringToValueProc					mHandleStageStringToValueProc;
	
	InvalidateActorProc								mInvalidateActorProc;
	
	GetUTF8PathToPluginResourcesProc				mGetUTF8PathToPluginResourcesProc;
	
	ActorControlCreateVideoPreviewProc				mActorControlCreateVideoPreviewProc;
	UpdateVideoPreviewProc							mUpdateVideoPreviewProc;
	DisposeVideoPreviewProc							mDisposeVideoPreviewProc;
	DrawVideoIntoPreviewProc						mDrawVideoIntoPreviewProc;

	GetGraphicsScalingProc							mGetGraphicsScalingProc;
	SetPropertyValueToStringCallbackProc			mSetPropertyValueToStringCallbackProc;
	
	MarkOutputPropertyAsNewForNextCycleProc			mMarkOutputPropertyAsNewForNextCycleProc;
	
	GetPropertyCustomDataTypeProc					mGetPropertyCustomDataTypeProc;
	SetPropertyCustomDataTypeProc					mSetPropertyCustomDataTypeProc;
	
	IzTimecodeConvertToStringProc					mIzTimecodeConvertToStringProc;
	ConvertValueToValueProc							mConvertValueToValueProc;
	BroadcastControlTimecodeProc					mBroadcastControlTimecodeProc;
	
	// 3.0.8b9
	SetCursorProc									mSetCursorProc;
	PenModeProc										mPenModeProc;
	PaintRgnProc									mPaintRgnProc;
	
	// 3.0.8b18
	GetSceneCountProc								mGetSceneCountProc;
	IsActiveAsPrimarySceneProc						mIsActiveAsPrimarySceneProc;
	IsActiveAsSecondarySceneProc					mIsActiveAsSecondarySceneProc;
	
	// 3.0.8b39
	SetPropertyFilePathModeProc						mSetPropertyFilePathModeProc;
	
	// 3.0.8b47
	SetPropertyDisallowWhitespaceFlagProc			mSetPropertyDisallowWhitespaceFlagProc;
	SetPropertyDisallowCarriageReturnsFlagProc		mSetPropertyDisallowCarriageReturnsFlagProc;
	
	// 3.0.9b01
	CreateOSCAddressReceiverProc					mCreateOSCAddressReceiverProc;
	DisposeOSCAddressReceiverProc					mDisposeOSCAddressReceiverProc;

	// 3.0.9b05 (kIsadoraCallbacksVersion_309b05)
	EnableDisableBitmapConverterAsyncReadback		mEnableDisableBitmapConverterAsyncReadbackProc;
	
	AcquireGlobalDataRefProc						mAcquireGlobalDataRefProc;
	ReleaseGlobalDataRefProc						mReleaseGlobalDataRefProc;

	SetPropertyIsPasswordFlagProc					mSetPropertyIsPasswordFlagProc;
	
	// 3.0.9b06 (kIsadoraCallbacksVersion_309b06)
	SetPropertyIsInvalidFlagProc					mSetPropertyIsInvalidFlagProc;
	
	// 3.1.2b10 (kIsadoraCallbacksVersion_312b10)
	GetUTF8PathToTroikaTronixFolderProc				mGetUTF8PathToTroikaTronixFolderProc;
	
	// 3.2.3b01 (kIsadoraCallbacksVersion_323b01)
	ADAAudioControls_DrawProc						mADAAudioControls_DrawProc;
	ADAAudioControls_AttemptToHandleClickProc		mADAAudioControls_AttemptToHandleClickProc;
	
	IzzyAudio_GetOutputChannelsProc					mIzzyAudio_GetOutputChannelsProc;
	IzzyAudio_GetSampleRateProc						mIzzyAudio_GetSampleRateProc;
	
	IzzyAudio_CreateAudioStreamProc					mIzzyAudio_CreateAudioStreamProc;
	IzzyAudio_DisposeAudioStreamProc				mIzzyAudio_DisposeAudioStreamProc;
	IzzyAudio_StartAudioStreamProc					mIzzyAudio_StartAudioStreamProc;
	IzzyAudio_StopAudioStreamProc					mIzzyAudio_StopAudioStreamProc;
	IzzyAudio_SetVolumeAndPanProc					mIzzyAudio_SetVolumeAndPanProc;
	IzzyAudio_SetRoutingProc						mIzzyAudio_SetRoutingProc;
};

typedef struct IsadoraCallbacks IsadoraCallbacks;

// ---------------------------------------------------------------
// Isadora parameters

#define	kConditionalCallbacks	mCallbacks

struct IsadoraParameters {

	UInt32						mIsadoraMagic;
	
	ImageProcessingPrefs		mImgProcPrefs;				// v1.0: image processing preferences
	
	IsadoraCallbacks*			mCallbacks;					// v1.0: isadora callbacks
	
	UInt32						mMillisecondCount;			// v1.0: current time in milliseconds

	Boolean						mHasFastMath;				// v1.0: true for Altivec (PowerPC) or MMX/SSE processor (Intel)
	
	Boolean						mNewPullModeActive;			// v1.3: true if the v2.0 actor processing model (new pull mode)
															// otherwise, revert to the v1.x actor processing model
	
	Boolean						mGroupLayersByScene;		// v2.0: if true, group layers by scenes
															// otherwise, revert to the 1.x ungrouped layering
	
	RenderGroupT				mRenderGroup;				// v2.0: group ID for rendering - for more info, see definition in IsadoraTypes.h
	
	Boolean						mSharedGLContexts;			// v2.0: when true, all OpenGL contexts created by Isadora share the same
															// resources. That means that textures can be shared between
															// stages.
	
	UInt32						mOSVersion;					// v3.0 - operating system version: MMmmRRbb (Major, Minor, Revision, Build)

	UInt32						mMaxDecimalDigits;			// v3.0 - number of digits to allow after the decimal point
};

typedef struct IsadoraParameters IsadoraParameters;

// ---------------------------------------------------------------
// Private Functions -- do not call from plugins

void	SetImageProcessingPreferences(const ImageProcessingPrefs* inImageProcPrefs);
Boolean	GetProcessVideoAtFixedSizeRect(IzRect* outRect);
Boolean	GetProcessVideoAtFixedSize(SInt32* outWidth, SInt32* outHeight);
void	InitIsadoraCallbacks();
void	DisposeIsadoraCallbacks();
double	GetVideoProcessingOverhead();

#ifdef __cplusplus
}
#endif

extern	IsadoraParameters	gIzzyParams;

#endif
